#ifndef lint
static char rcsid[] =
	"$Id: babid.c,v 4.0 1993/09/17 00:34:39 srp Exp srp $";
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <errno.h>
#include <fcntl.h>

#include <signal.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/time.h>

#include "config.h"
#include "batch.h"
#include "pathnames.h"
#include "protocol.h"

#include H_STRING

static char *use[] = {
	"babid -h [-Q queue] [-H host]",
	"\"Babid\" returns the current bid for a queue.",
	"",
	"Options:",
#ifdef USE_PORT_SELECTION
	"-L [string]	Local Unix socket to connect",
	"-R [int]	Remote port number to call",
#endif
	"-Q [string]	Specify queue to examine (or \"all\")",
	"-H [string]	Specify remote host to query (or \"all\")",
	"",
	"$Revision: 4.0 $",
	NULL
};


int s;

static SIGRTN	cleanup P((int));

extern char *	optarg;			/* Ptr to option argument */
extern char	optopt;			/* Current option */ 
extern int	optind;			/* Argument idx */
extern int	opterr;			/* Error handling by getopt? */


main(argc, argv)
int argc;
char *argv[];
{

	char *	cp;
	char *	host = NULL;
	char *	defhost = NULL;
	char *	queue = NULL;
	int	option;


	progname = argv[0];

	if (signal(SIGHUP, SIG_IGN) != SIG_IGN)
		(void) signal(SIGHUP, cleanup);

	if (signal(SIGINT, SIG_IGN) != SIG_IGN)
		(void) signal(SIGINT, cleanup);

	if (signal(SIGTERM, SIG_IGN) != SIG_IGN)
		(void) signal(SIGTERM, cleanup);

	progname = argv[0];			/* Assign the program name */

	if (gethostname(lhost, MAXHOSTNAMELEN) < 0) {
		(void) fprintf(stderr, "gethostname fails\n");
		exit(1);
	}

	/*
	 * Check first for a "BATCH" environment variable -
	 * backwards compatable, but clashes with Splus driver envvar.
	 */
	if ((cp = getenv("BATCH")) != NULL)
		queue = estrdup(cp);
	else
		queue = estrdup(_DEF_QUEUE);

	/*
	 * Check second for a "BATCHQUEUE" environment variable.
	 * If there is one, use it.
	 */
	if ((cp = getenv("BATCHQUEUE")) != NULL) {
		free(queue);
		queue = estrdup(cp);
	}

	/*
	 * Always start with the default host.
	 */
	defhost = _DEF_HOST;				

	/*
	 * ... modifiy if the default host is also the localhost.
	 */
	if (islocalhost(defhost) == 0)
		defhost = lhost;

	/*
	 * Parse options
	 */
	while ((option = getopt(argc, argv, "L:R:H:Q:")) != EOF)
		switch((char)option) {
		case 'h':			/* Help */
			usage(use);
			exit(1);
#ifdef USE_PORT_SELECTION
		case 'L':			/* Local port or socket */
			socketname = optarg;
			break;
		case 'R':			/* Remote port */
			remote_port = atoi(optarg);
			break;
#endif
		case 'Q':			/* Queue name */
			queue = optarg;
			break;
		case 'H':			/* Host name */
			host = optarg;
			break;
		case '?':			/* Unknown option */
			if (opterr == 0)
				blog(BLOG_WARNING,
				     "internal: bad flag \"-%c\"", optopt);
			usage(use);
			exit(1);
			break;
		default:			/* Unhandled option */
			blog(BLOG_WARNING, 
			     "internal: no action for \"-%c\" option",
			     (char) option);
			usage(use);
			exit(1);
			break;

		}


	/*
	 * Queue may be queue@host - if so, break it apart
	 */
	if ((cp = strchr(queue, '@')) != NULL) {
		*cp++ = '\0';
		host = cp;
	}

	/* If host is not set, use the default host */
	if (!host)
		host = defhost;

	if (host_query(host, QC_BID, queue) != 0)
		exit(1);

	exit(0);

}


/*
 * Client fuction for the host_query call:
 * If the return code is QR_OK, read the rest of the line for the bid
 * information, and store the hostname if the bid is the lowest so far.
 * This is designed for one round of bid requests only.  You would have to
 * assign lowbid to 9999.0 to allow lowbid calculation at another time.
 */
int
reply_ok(fd, fmt, to_queue)
FILE *	fd;
char *	fmt;
char *	to_queue;
{

	char	recvbuf[BUFSIZ];
	double	one, five, fifteen, bid;
	time_t	stamp;
	int	rtn;

	if (fgets(recvbuf, sizeof(recvbuf), fd) != NULL) {
		rtn = sscanf(recvbuf, fmt, &one, &five, &fifteen,
			      &bid, &stamp);
		printf("bid: %.2lf [%.2lf %.2lf %.2lf] %s",
		       bid, one, five, fifteen, ctime(&stamp));

	}

	return 0;
}


int
reply_mt(fd, fmt, to_queue)
FILE *	fd;
char *	fmt;
char *	to_queue;
{

}


/*
 * Clean up on interruption.
 */
static SIGRTN
cleanup(sig)
int	sig;
{
	extern int s;
	(void) close(s);
	exit(1);
}
