#ifndef lint
static char rcsid[] =
	"$Id: baview.c,v 4.0 1993/09/17 00:35:38 srp Exp srp $";
#endif

/*
 *
 * Author: Scott R. Presnell
 *
 * Copyright (C) 1989, 1990, The Regents of the University of California
 *                     All Rights Reserved
 *
 * THE REGENTS OF THE UNIVERSITY OF CALIFORNIA MAKE NO REPRESENTATION OR
 * WARRANTIES WITH RESPECT TO THE CONTENTS HEREOF AND SPECIFICALLY DISCLAIM
 * ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
 * PURPOSE.
 *
 * Further, the Regents of the University of California reserve the right to
 * revise this software and/or documentation and to make changes from time
 * to time in the content hereof without obligation of the Regents of the
 * University of California to notify any person of such revision or change.
 *
 */

/*
 * Baview: Graphical query and removal of batch system jobs.
 */

/*
 * $Log: baview.c,v $
 * Revision 4.0  1993/09/17  00:35:38  srp
 * Re-munged code for control by Configure script.  Added setsid() calls,
 * changed fork() calls for error handling, finer nice() priority control,
 * write functionality.  Cleaned up code definitions and declarations for
 * version 4.0 release.  Updated documentation, man pages, test scripts.
 *
 * Revision 3.5  1993/09/13  00:51:06  srp
 * Added broadcast, multicast, interface selection, new logging facility,
 * host <-> address cache, job renice, user settable soft limits.  Changed
 * all structures to bidirectional linked list.  Job attachment and queue
 * "drain" implemented.  Queues are locked during critical modification
 * (scanning, job removal).  Jobs are sorted as specified by batchcap -
 * user information dumped periodically.  Reworked prototype macros. SIGTERM
 * will terminate running jobs.  "Kick" message does not cause fork().
 * Console usage and console user detection decoupled.  Machine specific code
 * broken out into separate file.
 *
 * Revision 1.1  1993/03/15  15:32:44  srp
 * Initial revision
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <errno.h>
#include <fcntl.h>

#include <signal.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/time.h>

#include <grp.h>
#include <pwd.h>

#include <gl/device.h>

#include "config.h"

#include H_STRING

#include "batch.h"
#include "protocol.h"
#include "reply.h"
#include "pathnames.h"

#include "forms.h"
#include "baviewform.h"

int s;						/* Socket handle */

int current_browser_line = 0;			/* Current insert line */
int max_info_line = 0;				/* Maximum insert line */
int current_info_line = -1;			/* Current selection */

static struct	passwd pw;

struct jobinfo_t {
	int	pid;				/* Process ID */
	int	jid;				/* Job ID */
	long	job_state;			/* Job State */
	long	queue_state;			/* Queue State */
	int	b_line;				/* Browser line */
	char	user[LINE];			/* Full Mail addr */
	char	name[LINE];			/* Short name */
	char	delay[LINE];			/* Delay time/file/job */
	int	queueno;			/* Queue browser line */
	int	hostno;				/* Host browser line */
	time_t	st_ctime;			/* Create time */
	time_t	st_stime;			/* Start time */
	time_t	cpu_time;			/* Current CPU time */
};


/*
 * Maximum number of jobs from a given host+queue reading.
 */
#define MAX_JOBINFO	20

struct jobinfo_t Info[MAX_JOBINFO];

static char Title[80];

static int show_help = FALSE;
static int show_details = FALSE;

Notify	mail_opts;

UIDTYPE	myuid;
GIDTYPE	mygid;

char	bbuf[BUFFER];
char	pbuf[BUFFER];
struct passwd *	pass;			/* Password structure data */

char *	bp;

char *	submit_queue = NULL;

void	handle_submit_queue();

char	sfile[MAXPATHLEN];
char	ofile[MAXPATHLEN];
char	efile[MAXPATHLEN];
char	jname[MAXPATHLEN];

extern char *	optarg;			/* Ptr to option argument */
extern char	optopt;			/* Current option */ 
extern int	optind;			/* Argument idx */
extern int	opterr;			/* Error handling by getopt? */


main(argc, argv)
int	argc;
char **	argv;
{

	char *	cp;
	char *	host = NULL;			/* Host of current interest */
	char *	defhost = NULL;			/* Host of current interest */
	char *	queue = NULL;			/* Queue of current interest */
	int	status;

	progname = argv[0];			/* Assign the program name */

	myuid = getuid();
	mygid = getgid();

	if (gethostname(lhost, MAXHOSTNAMELEN) < 0) {
		fprintf(stderr, "gethostname fails\n");
		exit(1);
	}

	/*
	 * Check first for a "BATCH" environment variable -
	 * backwards compatable, but clashes with Splus driver envvar.
	 */
	if ((cp = getenv("BATCH")) != NULL)
		queue = estrdup(cp);
	else
		queue = estrdup(_DEF_QUEUE);

	/*
	 * Check second for a "BATCHQUEUE" environment variable.
	 * If there is one, use it.
	 */
	if ((cp = getenv("BATCHQUEUE")) != NULL) {
		free(queue);
		queue = estrdup(cp);
	}

	/*
	 * Always start with the default host.
	 */
	defhost = _DEF_HOST;

	/*
	 * ... modifiy if the default host is also the localhost.
	 */
	if (islocalhost(defhost) == 0)
		defhost = lhost;
	
	/*
	 * We're running as suid root, and maybe sgid daemon, so we need
	 * to get the *real* uid to handle security on the request.
	 */
	if ((pass = getpwuid(myuid)) == NULL) {
		blog(BLOG_ERR, epid, myuid);
	} else {
		pw.pw_name = estrdup(pass->pw_name);
		pw.pw_uid = pass->pw_uid;
		pw.pw_gid = pass->pw_gid;
	}

	/*
	 * Entry into the forms library - create all the forms that we might
	 * use in this program (no display done).
	 */
	create_the_forms();

	/*
	 * Constant width font for the display browser.
	 */
	fl_set_browser_fontstyle(main_display_browser, FL_FIXED_STYLE);

	/*
	 * Initialize the host and queue of the browsers.
	 */
	init_main_form(defhost, queue);
	init_submit_form(defhost, queue);

	(void) strcpy(Title, "BatchView@");
	(void) strcat(Title, lhost);

	/* Show the main form, placed by mouse position, with std. border. */
	fl_show_form(main_form, FL_PLACE_SIZE, TRUE, Title);

	/* Main loop. */
	doit();

	/* Later dudes. */
	gexit();
	exit(0);
}


/*
 * Initialize the main form before display.
 */
init_main_form(host, queue)
char *	host, *queue;
{

	char *	ho, *qu, *ma;
	int	h_line = 1;
	int	q_line = 1;

	/*
	 * Load the host browser.
	 */
	while ((ho = getbhent()) != NULL) {
		fl_add_browser_line(main_host_browser, ho);
		if (strcmp(ho, host) == 0)
			fl_select_browser_line(main_host_browser, h_line);
		h_line++;
	}
	/* endbhent(); */

	/*
	 * Load the queue browser.
	 */
	while ((qu = getquent()) != NULL) {
		fl_add_browser_line(main_queue_browser, qu);
		if (strcmp(qu, queue) == 0)
			fl_select_browser_line(main_queue_browser, q_line);
		q_line++;
	}
	endquent();

}


/*
 * Initialize the submit form before display.
 */
init_submit_form(host, queue)
char *	host, *queue;
{

	char *	ho, *qu, *ma;
	int	h_line = 1;
	int	q_line = 1;

	/*
	 * Load the host browser.
	 */
	while ((ho = getbhent()) != NULL) {
		fl_add_browser_line(submit_host_browser, ho);
		if (strcmp(ho, host) == 0)
			fl_select_browser_line(submit_host_browser, h_line);
		h_line++;
	}
	/* endbhent(); */

	/*
	 * Load the queue browser.
	 */
	while ((qu = getquent()) != NULL) {
		fl_add_browser_line(submit_queue_browser, qu);
		if (strcmp(qu, queue) == 0)
			fl_select_browser_line(submit_queue_browser, q_line);
		q_line++;
	}
	endquent();

	handle_submit_queue(submit_queue_browser, 1);
	fl_set_input(submit_output_input, "<daemon>");
	fl_set_input(submit_error_input, "<daemon>");
}


/*
 * Main loop for device/form/object processing
 */
doit()
{

	char *	ho, *qu;
	char	job[LINE];
	int	hostno, queueno, lineno;
	int	done = FALSE;
	short	val, rval;
	FL_OBJECT *	obj;
	
	/*
	 * Queue up some devices:
	 * WINQUIT is the quit selection from the frame menu.
	 * WINSHUT is the lightning bolt and "close" from the frame menu.
	 */
 	fl_qdevice(WINQUIT);
	fl_qdevice(WINSHUT);

	/*
	 * Main loop
	 */
	do {

		/* Get an event */
		obj = fl_do_forms();

		/*
		 * If it is a user queued event, take care of it here.
		 */
		if (obj == FL_EVENT) {
			switch(fl_qread(&val)) {
			case WINQUIT:
			case WINSHUT:
				done = TRUE;
				break;
			}
		
		/*
		 * Else if the query_button has been selected, evaluate
		 * the current browser states and query a queue@host
		 */
		} else if (obj == query_button) {

			if ((hostno = fl_get_browser(main_host_browser)) <= 0)
				continue;
			ho = fl_get_browser_line(main_host_browser, hostno);

			if ((queueno = fl_get_browser(main_queue_browser)) <= 0)
				continue;
			qu = fl_get_browser_line(main_queue_browser, queueno);

			if (host_query(ho, QC_DISP, qu) < 0)
				fl_show_message("Cannot make query:", "", "");

		/*
		 * Else if the display browser has been used, fix 
		 * the details window (always update even if window is
		 * closed).
		 */
		} else if (obj == main_display_browser) {
			if ((lineno = fl_get_browser(main_display_browser)) <= 0)
				continue;
			display_details(lineno);

		/*
		 * Else if the delete button has been selected, get the
		 * host and the queue info from the structre, and try
		 * to remove the job.
		 */
		} else if (obj == remove_button) {

			if (current_info_line < 0) {
				fl_show_message("No selection current!",
						"", "");
				continue;
			}
				
			ho = fl_get_browser_line(main_host_browser,
					      Info[current_info_line].hostno);

			qu = fl_get_browser_line(main_queue_browser,
					      Info[current_info_line].queueno);
			
			/* Check group access to this queue */
			if (check_group(pw.pw_name, myuid, mygid, qu) < 0) {
				(void) sprintf(line, "to the %s queue.", qu);
				fl_show_message("You do not have access",
					     line, "");
				continue;
			}

			(void) sprintf(line, "Remove job %d in",
				       Info[current_info_line].jid);

			(void) sprintf(line2, "queue %s on %s?", qu, ho);
			
			if (!fl_show_question(line, line2, ""))
				continue;

			(void) sprintf(job, "%d", Info[current_info_line].jid);
			if ((rval = queue_remove_job(qu, ho, job)) < 0)
				fl_show_message("Cannot make query:", "", "");
			else if (rval == 0) {
				/*
				 * If the call to remove the job was
				 * "sucessful, clear the detail form and
				 * "refresh" the display_browser with
				 * another call to for queue display.
				 */
				clear_detail_form();
				if (host_query(ho, QC_DISP, qu) < 0)
					fl_show_message("Refresh error",
						     "", "");
			}
		} else if (obj == main_submit_button) {
#ifdef ndef
			toggle_submit_form(main_submit_button, 0);
#else
			fl_show_message("Job Setup not currently implemented",
					"", "");
#endif

		}

	/*
	 * Until exit_button or WINQUIT or WINSHUT has been called.
	 */
	} while (obj != exit_button && !done);

	fl_hide_form(main_form);
	if (show_details)
	    fl_hide_form(detail_form);
	if (show_help)
		fl_hide_form(help_form);

	return 0;
}

/*
 * Attempt to remove the job from the queue on the (possibly remote) host.
 * through the host_query() function.
 */
int
queue_remove_job(tqueue, thost, job)
char *	tqueue;
char *	thost;
char *	job;
{

	char *	cp;
	char	list[BUFFER];
	struct passwd *	jpw;


	cp = list;

	/* First copy in the queue name, */
	(void) strcpy(list, tqueue);	

	/* ... then the current user name... */
	cp += strlen(cp);
	*cp++ = ' ';
	(void) strcpy(cp, pw.pw_name);

	cp += strlen(cp);
	*cp++ = ' ';
	(void) strcpy(cp, job);
	cp += strlen(cp);
	*cp++ = '\0';

	/*
	 * Try to remove the job
	 */
	return host_query(thost, QC_REM, list);

}


/*
 * Client function for host_query(): If the QC_DISP comes back with OK,
 * Then get the first line (status), and then parse and print the folowing
 * lines as batch jobs in the queue.
 */
int
reply_ok(fp, fmt, queue)
FILE *	fp;
char *	fmt;
char *	queue;
{

	char	recvbuf[BUFSIZ];


	if (fgets(recvbuf, sizeof(recvbuf), fp) != NULL) {
		recvbuf[strlen(recvbuf)-2] = '\0';
		fl_clear_browser(main_status_browser);
		fl_add_browser_line(main_status_browser, recvbuf);
	} else {
		blog(BLOG_WARNING, "reply_ok(): bad read from daemon");
	}

	bzero((void *) Info, sizeof(struct jobinfo_t) * MAX_JOBINFO);
	max_info_line = 0;
	current_info_line = -1;

	if (fmt != NULL) {
		fl_freeze_object(main_display_browser);
		fl_clear_browser(main_display_browser);
		current_browser_line = 0;
		(void) queue_print(fp, fmt, queue);
		fl_unfreeze_object(main_display_browser);
	} else {
		char recvbuf[BUFSIZ];

		while (fgets(recvbuf, sizeof(recvbuf), fp) != NULL) {
			fl_show_message(recvbuf, "", "");
		}
	}

	return 0;
}


/*
 * Client function for host_query(): If the QC_DISP comes back with MT,
 * We still report the status line.
 */
int
reply_mt(fp, fmt, queue)
FILE *	fp;
char *	fmt;
char *	queue;
{

	char	recvbuf[BUFSIZ];

	extern int clear_display;


	if (fgets(recvbuf, sizeof(recvbuf), fp) != NULL) {
		recvbuf[strlen(recvbuf)-2] = '\0';
		fl_clear_browser(main_status_browser);
		fl_add_browser_line(main_status_browser, recvbuf);
	} else {
		blog(BLOG_WARNING, "reply_mt(): bad read from daemon");
	}

	fl_freeze_object(main_display_browser);
	fl_clear_browser(main_display_browser);
	current_browser_line = 0;
	fl_add_browser_line(main_display_browser, "Empty"); 
	current_browser_line++;
	fl_unfreeze_object(main_display_browser);

	bzero((void *) Info, sizeof(struct jobinfo_t) * MAX_JOBINFO);
	max_info_line = 0;
	current_info_line = -1;

	clear_detail_form();

	return 0;
}


/*
 * reply_info(): print off the format with the arument (usually the queue
 * name).  If the format is NULL, don't print anything; just return.
 */
int
reply_info(fp, fmt, arg)
FILE *	fp;
char *	fmt;
char *	arg;
{

	if (fmt == NULL)
		return 0;

	/* Nuke the '\n' (new line) that is also passed along */

	fmt[strlen(fmt)-2] = '\0';

	(void) sprintf(line, fmt, arg);
	(void) fl_show_message(line, "", "");
	return 0;
}



#define SHORT_SFORMAT	"%-9.9s %-9.9s %-8.8s %-8.8s %-8.8s"
#define SHORT_NFORMAT	"%s%-8d %-9.9s %-8.8s %-8ld %-8.8s"


char *states[] = {
	"queued",  "running",
	"queued",  "suspended",
	"queued",  "interrupted",
	"queued",  "timeout",
	"queued",  "stopped",
	"held",    "attached",
	"pending", "sending",
	"delayed", "running"
};


/*
 * Retreive and print in human readable format the information on the
 * jobs in the queue.
 */
int
queue_print(fp, fmt, queue)
FILE *	fp;
char *	fmt;
char *	queue;
{

	char *	uptr, *status;
	char	full_name[LINE], name[LINE], user[LINE], delay[LINE];
	char	recvbuf[BUFSIZ];
	char	buf[BUFSIZ];
	long	queue_state, job_state;
	int	pid, restart, jid;
	int	hostno, queueno;
	time_t	st_ctime, st_stime, st_dtime, cpu_time;

	/*
	 * Translate and print individual job information from the daemon
	 */
	while (fgets(recvbuf, sizeof(recvbuf), fp) != NULL) {
		(void) sscanf(recvbuf, fmt,
			      &jid, name, user, &job_state,
			      &st_ctime, &st_stime, delay, &cpu_time,
			      &pid, &queue_state);

		/*
		 * Cut off the address part of the name for the short disp
		 */
		(void) strcpy(full_name, user);
		if ((uptr = strchr(user, '@')) != NULL)
			*uptr = '\0';

		if (queue_state == 0 || !(job_state & JOB_RUNNING)) {
			if (job_state & (JOB_ATTACH | JOB_ATTACHED))
				queue_state = 5;
			else if (job_state & (JOB_SEND | JOB_AGAIN))
				queue_state = 6;
			else if (job_state & JOB_DELAY)
				queue_state = 7;
			/* else retain old meaning */
		}

		if (job_state & JOB_RUNNING)
			status = states[(queue_state * 2)+1];
		else
			status = states[(queue_state * 2)];

		(void) sprintf(buf, SHORT_NFORMAT,
			       ((job_state & JOB_TOPQ) ? "*" : " "),
			       jid, name, user, cpu_time, status);

		/*
		 * Display this line of information in the borwser.
		 */
		fl_add_browser_line(main_display_browser, buf);
		current_browser_line++;

		/*
		 * Load the information into the structure.
		 */
		Info[max_info_line].pid = pid;
		Info[max_info_line].jid = jid;
		Info[max_info_line].job_state = job_state;
		Info[max_info_line].queue_state = queue_state;
		Info[max_info_line].b_line = current_browser_line;

		(void) strcpy(Info[max_info_line].user, full_name);
		(void) strcpy(Info[max_info_line].name, name);
		(void) strcpy(Info[max_info_line].delay, delay);

		Info[max_info_line].queueno = fl_get_browser(main_queue_browser);
		Info[max_info_line].hostno = fl_get_browser(main_host_browser);

		Info[max_info_line].st_ctime = st_ctime;
		Info[max_info_line].st_stime = st_stime;
		Info[max_info_line].cpu_time = cpu_time;

		max_info_line++;
	}

	if (current_browser_line == 1)
		display_details(current_browser_line);

	return 0;
	
}

display_details(line)
int line;
{

	int	i;
	char	full_name[LINE];
	char	jid_string[LINE];
	char	pid_string[LINE];
	char	time_string[LINE];
	char	queue_string[LINE];
	char	start_string[LINE];
	char	delay_string[LINE];
	time_t	st_dtime;


	/*
	 * Find the info structure for the line selected.
	 */
	for (i = 0; i < MAX_JOBINFO ; i++)
		if (Info[i].b_line == line)
			break;

	/*
	 * If we skipped over all the strucures, just return.
	 */
	if (i == MAX_JOBINFO) {
		current_info_line = -1;
		return;
	}

	current_info_line = i;

	/*
	 * Copy the ctimes into temporary buffers so that we can
	 * nuke the returned string at pos 16.
	 */
	if (Info[i].st_ctime > 0) {
		(void) strcpy(queue_string, ctime(&Info[i].st_ctime));
		queue_string[strlen(queue_string)-1] = '\0';
	} else
		(void) strcpy(queue_string, "before epoch?");
	
	if (Info[i].job_state & JOB_RUNNING) {
		(void) strcpy(start_string, ctime(&Info[i].st_stime));
		start_string[strlen(start_string)-1] = '\0';
	} else {
		(void) strcpy(start_string, "not yet");
	}

	if (Info[i].delay && strchr(Info[i].delay, '/') != NULL) {
		/* got a file */
		strcpy(delay_string, Info[i].delay);
	} else if (Info[i].delay && strchr(Info[i].delay, '%') != NULL) {
		/* got a job */
		strcpy(delay_string, Info[i].delay);
	} else if (Info[i].delay && (st_dtime = atol(Info[i].delay)) > 0) {
		(void) strcpy(delay_string, ctime(&st_dtime));
		delay_string[strlen(delay_string)-1] = '\0';
	} else {
		(void) strcpy(delay_string, "no delay");
	}


	/*
	 * Write the data to the browsers in the details form
	 */
	clear_detail_form();

	fl_freeze_form(detail_form);

	(void) sprintf(jid_string, "%d", Info[i].jid);
	fl_add_browser_line(detail_jid_browser, jid_string);

	(void) sprintf(pid_string, "%d", Info[i].pid);
	fl_add_browser_line(detail_pid_browser, pid_string);

	(void) sprintf(time_string, "%d", Info[i].cpu_time);
	fl_add_browser_line(detail_time_browser, time_string);

	fl_add_browser_line(detail_user_browser, Info[i].user);

	fl_add_browser_line(detail_queued_browser, queue_string);

	fl_add_browser_line(detail_started_browser, start_string);
	fl_add_browser_line(detail_delayed_browser, delay_string);

	fl_add_browser_line(detail_state_browser, 
			 states[ ((Info[i].job_state & JOB_RUNNING) ? 1 : 0) + (Info[i].queue_state * 2)]);

	fl_add_browser_line(detail_job_browser, Info[i].name);
	fl_unfreeze_form(detail_form);

}


/*
 * Clear the all the browsers in the detail form
 */
clear_detail_form()
{

	fl_freeze_form(detail_form);
	fl_clear_browser(detail_jid_browser);
	fl_clear_browser(detail_pid_browser);
	fl_clear_browser(detail_time_browser);
	fl_clear_browser(detail_user_browser);
	fl_clear_browser(detail_queued_browser);
	fl_clear_browser(detail_started_browser);
	fl_clear_browser(detail_delayed_browser);
	fl_clear_browser(detail_state_browser);
	fl_clear_browser(detail_job_browser);
	fl_unfreeze_form(detail_form);
}


/*
 * Popup for adding a new hostname
 */
void
add_host_name(obj, arg)
FL_OBJECT *	obj;
long	arg;
{
	char *	name;

	if ((name = fl_show_input("Enter host name", "")) == NULL)
		return;

	fl_add_browser_line(main_host_browser, name);

	return;
}


/*
 * Toggle the display of the details form/window.
 */
void
toggle_detail_form(obj, arg)
FL_OBJECT *	obj;
long	arg;
{

	if (show_details == TRUE) {
		fl_hide_form(detail_form);
		show_details = FALSE;
	} else {
		fl_show_form(detail_form, FL_PLACE_SIZE, FALSE, NULL);
		show_details = TRUE;
	}

	return;
}


/*
 * Show the help form.
 */
void
toggle_help_form(obj, arg)
FL_OBJECT *	obj;
long	arg;
{
	if (show_help == TRUE) {
		fl_hide_form(help_form);
		show_help = FALSE;
	} else {
		fl_show_form(help_form, FL_PLACE_SIZE, FALSE, NULL);
		show_help = TRUE;
	}
	return;
}

void
toggle_submit_form(obj, arg)
FL_OBJECT *	obj;
long	arg;
{

	if (obj == main_submit_button) {
		fl_show_form(submit_form, FL_PLACE_SIZE, FALSE, NULL);
	} else if (obj == submit_submit_button) {
		/* Submit job, pop up dialog showing job number */
		fl_hide_form(submit_form);
	} else  if (obj == submit_cancel_button) {
		/* Don't submit job */
		fl_hide_form(submit_form);
	}

	return ;

}

void
handle_submit_queue(obj, arg)
{
	char *	qu;
	char *	ma;
	int	queueno;
	int	status;

	if ((queueno = fl_get_browser(submit_queue_browser)) <= 0)
		return ;
	submit_queue = fl_get_browser_line(submit_queue_browser, queueno);

	/*
	 * Get the named queue into a buffer
	 */
	if ((status = bgetent(bbuf, submit_queue)) < 0) {
		blog(BLOG_FATAL, "cannot open queue description file");
		exit(1);
	} else if (status == 0) {
		blog(BLOG_ERR, enotq, submit_queue);
		exit(1);
	}

	/*
	 * Reset the queue name to the canonical (first) name
	 */
	submit_queue = bnamget(bbuf);

	bp = pbuf;

	/*
	 * Get the default mailing information...
	 */
	ma = bgetstr("ma", &bp);

	/*
	 * ...and use it if the user has not specified otherwise.
	 */
	mail_opts = setnotify(ma);
	display_mailopts();

	return;

}

void
submit_option(obj, arg)
FL_OBJECT *	obj;
long	arg;
{

}

void
set_submit_file(obj, arg)
FL_OBJECT *	obj;
long	arg;
{

	char *	cp, *jp;

	if (obj == set_script_button) {
		cp = fl_show_file_selector("Script To Execute",
					       "", "", "");
		(void) strcpy(sfile, cp);
		fl_set_input(submit_script_input, sfile);
		if ((jp = strrchr(sfile, '/')) != NULL) {
			jp++;
			strcpy(jname, jp);
		} else {
			strcpy(jname, cp);
		}
		fl_set_input(submit_jobname_input, jname);

	} else if (obj == set_output_button) {
		cp = fl_show_file_selector("Standard Output File",
					       "", "", "");
		(void) strcpy(ofile, cp);
		fl_set_input(submit_output_input, ofile);
	} else if (obj == set_error_button) {
		cp = fl_show_file_selector("Standard Error File",
					       "", "", "");
		(void) strcpy(efile, cp);
		fl_set_input(submit_error_input, efile);
	}

		
}

void
set_submit_mail(obj, arg)
FL_OBJECT *	obj;
long	arg;
{

	if (obj == mail_start_button) {
		if (fl_get_button(mail_start_button)) {
			mail_opts |= NOTIFY_START;
			fl_set_button(mail_none_button, 0);
		} else
			mail_opts &= ~NOTIFY_START;
	} else if (obj == mail_end_button) {
		if (fl_get_button(mail_end_button)) {
			mail_opts |= NOTIFY_END;
			fl_set_button(mail_none_button, 0);
		} else
			mail_opts &= ~NOTIFY_END;
	} else if (obj == mail_crash_button) {
		if (fl_get_button(mail_crash_button)) {
			mail_opts |= NOTIFY_CRASH;
			fl_set_button(mail_none_button, 0);
		} else
			mail_opts &= ~NOTIFY_CRASH;
	} else if (obj == mail_transfer_button) {
		if (fl_get_button(mail_transfer_button)) {
			mail_opts |= NOTIFY_TRANSFER;
			fl_set_button(mail_none_button, 0);
		} else
			mail_opts &= ~NOTIFY_TRANSFER;
	} else if (obj == mail_pause_button) {
		if (fl_get_button(mail_pause_button)) {
			mail_opts |= NOTIFY_PAUSE;
			fl_set_button(mail_none_button, 0);
		} else
			mail_opts &= ~NOTIFY_PAUSE;
	} else if (obj == mail_resource_button) {
		if (fl_get_button(mail_resource_button)) {
			mail_opts |= NOTIFY_RESOURCE;
			fl_set_button(mail_none_button, 0);
		} else
			mail_opts &= ~NOTIFY_RESOURCE;
	} else if (obj == mail_none_button) {
		if (fl_get_button(mail_none_button)) {
			mail_opts |= NOTIFY_NONE;
			fl_set_button(mail_start_button, 0);
			fl_set_button(mail_end_button, 0);
			fl_set_button(mail_crash_button, 0);
			fl_set_button(mail_transfer_button, 0);
			fl_set_button(mail_pause_button, 0);
			fl_set_button(mail_resource_button, 0);
		} else {
			mail_opts &= ~NOTIFY_NONE;
		}
	}

				
}

int
display_mailopts()
{

	if (mail_opts & NOTIFY_START)
		fl_set_button(mail_start_button, 1);
	else
		fl_set_button(mail_start_button, 0);

	if (mail_opts & NOTIFY_END)
		fl_set_button(mail_end_button, 1);
	else
		fl_set_button(mail_end_button, 0);

	if (mail_opts & NOTIFY_CRASH)
		fl_set_button(mail_crash_button, 1);
	else
		fl_set_button(mail_crash_button, 0);

	if (mail_opts & NOTIFY_TRANSFER)
		fl_set_button(mail_transfer_button, 1);
	else
		fl_set_button(mail_transfer_button, 0);

	if (mail_opts & NOTIFY_PAUSE)
		fl_set_button(mail_pause_button, 1);
	else
		fl_set_button(mail_pause_button, 0);

	if (mail_opts & NOTIFY_RESOURCE)
		fl_set_button(mail_resource_button, 1);
	else
		fl_set_button(mail_resource_button, 0);

	if (mail_opts & NOTIFY_NONE) {
		fl_set_button(mail_none_button, 1);
		fl_set_button(mail_start_button, 0);
		fl_set_button(mail_end_button, 0);
		fl_set_button(mail_crash_button, 0);
		fl_set_button(mail_transfer_button, 0);
		fl_set_button(mail_pause_button, 0);
		fl_set_button(mail_resource_button, 0);
	} else {
		fl_set_button(mail_none_button, 0);
	}
}
