#ifndef lint
static char rcsid[] =
	"$Id: aix3.c,v 4.0 1993/09/17 00:36:23 srp Exp srp $";
#endif

/*
 *
 * Author: Scott R. Presnell
 *
 * Copyright (C) 1989, 1990, 1991, 1992, 1993
 * The Regents of the University of California, All Rights Reserved.
 *
 * THE REGENTS OF THE UNIVERSITY OF CALIFORNIA MAKE NO REPRESENTATION OR
 * WARRANTIES WITH RESPECT TO THE CONTENTS HEREOF AND SPECIFICALLY DISCLAIM
 * ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
 * PURPOSE.
 *
 * Further, the Regents of the University of California reserve the right to
 * revise this software and/or documentation and to make changes from time
 * to time in the content hereof without obligation of the Regents of the
 * University of California to notify any person of such revision or change.
 *
 */

/*
 * aix3.c - determine the load on the machine, check the process
 *	variables and keep track of the time a process uses.
 */

#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <errno.h>

#if defined(MALLOC_TRACE) || defined(MALLOC_LEAK)
#include "dmalloc.h"
#else
#include <malloc.h>
#endif

#include <nlist.h>
#include <time.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>

#include <sys/file.h>

#include <sys/sysinfo.h>

#include "config.h"

#include H_STRING

#include "batch.h"
#include "batchd.h"
#include "pathnames.h"
#include "protocol.h"

#if defined(USE_CUMULTIME)
#define READ_PROC
#endif

#ifdef READ_PROC
#include <sys/var.h>
#include <sys/sysconfig.h>

#include <procinfo.h>

struct procinfo *	pbase;
struct userinfo		ups;

/* # processes */
int	nproc;

#endif /* READ_PROC */


#define svirtophys(x)	((uint)(x) & ~0xE0000000) /* Virtual to physical */
#define pagetok(x) (((x) << BPCSHIFT)/1024)	/* Pages to K */

#define	X_SYSINFO	0

struct	nlist nl[] = {
	{ "sysinfo" },
	{ 0 },
};


static char *	Unix = _PATH_VMUNIX;
static char *	Swap = _PATH_SWAP;
static char *	Kmem = _PATH_KMEM;
static char *	Mem  = _PATH_MEM;

static int	kmem = -1;
static int	swap = -1;
static int	mem  = -1;

static time_t	granule = SECPERMIN;		/* granule of avenrun check */

static int	kernel_init P((void));
static int	kernel_get_value P((off_t, char *, int, char *));

#ifdef READ_PROC
static int	kernel_get_user P((struct procinfo *));
#endif /* READ_PROC */


static int
kernel_init()
{

	int	i;

	knlist(nl, 1, sizeof(struct nlist));

	for (i = 0; i < sizeof(nl)/sizeof(struct nlist) - 1; i++) {
		if (!nl[i].n_value) {
			blog(BLOG_ERR, "%s: no offset for %s (%d)",
			    Unix, nl[i].n_name, i);
			return (-1);
		}
	}

	if ((kmem = open(Kmem, 0)) == -1) {
		blog(BLOG_ERR, eopen, Kmem);
		return (-1);
	}

#ifdef ndef
	/* Not used */
	if ((swap = open(Swap, 0)) == -1) {
		blog(BLOG_ERR, eopen, Swap);
		return (-1);
	}


	if ((mem = open(Mem, 0)) == -1) {
		blog(BLOG_ERR, eopen, Mem);
		return (-1);
	}
#endif

	return (0);
}


time_t
boottime_get()
{

	time_t	btime;

#if ndef
	if (kmem < 0)
		if (kernel_init() < 0)
			return (-1);

	if (kernel_get_value(nl[X_BOOTTIME].n_value,
			     (char *) &btime, sizeof(btime),
			     nl[X_BOOTTIME].n_name) < 0)
		return (-1);
#else
	btime = time((long *) 0);
#endif
	return btime;

}


#define INTERVAL1 60			/* average interval 1 (in seconds) */
#define INTERVAL2 (5*60)		/* average interval 2 (in seconds) */
#define INTERVAL3 (15*60)		/* average interval 3 (in seconds) */

int
load_get(aven)
double	aven[3];
{

	int	i;
	long	n, c;
	struct sysinfo si;
	unsigned long runocc, runque;
	static unsigned long orunocc, orunque;
	unsigned long swpocc, swpque;
	static unsigned long oswpocc, oswpque;
	double	exp1, exp2, exp3;
	double	newrun = 0;
	double	newswp = 0;

	static char first = 1;
	static double avenrun[3];

	static time_t	otime = 0;
	time_t	current, elapsed;


	if (kmem < 0)
		if (kernel_init() < 0)
			return (-1);

	current = time((long *)0);
	
	elapsed = current - otime;

	/*
	 * If it's been less than a minute, return the current values.
	 */
	if (elapsed < granule) {
		dbg((BLOG_RESOURCE, "elapsed %ld < granule %ld",
		     elapsed, granule));
		goto done;
	} else
		otime = current;

	exp1 = 0.0 - ((double) elapsed) / INTERVAL1;
	exp1 = exp(exp1);

	exp2 = 0.0 - ((double) elapsed) / INTERVAL2;
	exp2 = exp(exp2);

	exp3 = 0.0 - ((double) elapsed) / INTERVAL3;
	exp3 = exp(exp3);

	if (kernel_get_value((off_t) nl[X_SYSINFO].n_value, 
			     (char *) &si, sizeof(si),
			     nl[X_SYSINFO].n_name) < 0)
		return (-1);

	runque = si.runque;
	runocc = si.runocc;

	swpque = si.swpque;
	swpocc = si.swpocc;


	if (first) {

		newrun = ((double) runque) / runocc;
		newswp = ((double) swpque) / swpocc;

		avenrun[0] = avenrun[1] = avenrun[2] = newrun + newswp;

		first = 0;

	} else {

		/* get number of times updated */
		n = runocc - orunocc;
		if ( n <= 0 ) goto done;

		/* get number of runners */
		c = runque - orunque;
		if ( c < 0 ) c = 0;	/* mumble! */

		/* new instantaneous avg */
		newrun = ((double) c ) / ((double) n);

		/* get number of times updated */
		n = swpocc - oswpocc;
		if ( n <= 0 ) goto done;

		/* get number of runners */
		c = swpque - oswpque;
		if ( c < 0 ) c = 0;	/* mumble! */

		/* new instantaneous avg */
		newswp = ((double) c ) / ((double) n);

		/************************************************/
		/*   The following formwla is used to achieve   */
		/*   exponential decay of old measurements:	*/
		/*	avgN = avgN * expN  +  new * (1 - expN)	*/
		/*						*/
		/*   However, the factorized forms below	*/
		/*   require fewer floating point operations.	*/
		/************************************************/

		avenrun[0] = ((avenrun[0] - newrun - newswp) * exp1)
			+ newrun + newswp;
		avenrun[1] = ((avenrun[1] - newrun - newswp) * exp2)
			+ newrun + newswp;
		avenrun[2] = ((avenrun[2] - newrun - newswp) * exp3)
			+ newrun + newswp;
	}


	dbg((BLOG_RESOURCE, "runque %d orunque %d runocc %d orunocc %d",
	     runque, orunque, runocc, orunocc));

	dbg((BLOG_RESOURCE, "swpque %d oswpque %d swpocc %d oswpocc %d",
	     swpque, oswpque, swpocc, oswpocc));

	orunocc = runocc;
	orunque = runque;

	oswpocc = swpocc;
	oswpque = swpque;

done:

	for (i = 0; i < 3; i++)
		aven[i] = avenrun[i];


	return (0);
}


/*
 * Read kernel values
 */
static int
kernel_get_value(offset, ptr, size, refstr)
off_t	offset;
char *	ptr;
int	size;
char *	refstr;
{

	/*
	 * Seek to correct place in kmem...
	 */
	if (lseek(kmem, svirtophys(offset), 0) == -1)  {
		blog(BLOG_ERR, "%s: lseek to %s", Kmem, refstr);
		return (-1);
	}

	/*
	 * ...and read "size" bytes.
	 */
	if (read(kmem, ptr, (unsigned) size) != size)  {
		blog(BLOG_ERR, "%s: reading %s", Kmem, refstr);
		return (-1);
	}

	return (0);

}

#ifdef READ_PROC


/*
 * Now, set up everything we need to write a GetProc() routine.
 *
 * IMPORTANT NOTE TO PEOPLE SNARFING THIS CODE:
 *
 *	The AIX getproc() and getuser() routines are unsupported (not to
 *	mention, undocumented).  They are liable to be changed, or replaced
 *	in future versions of AIX.  Please include this comment in any
 *	code that uses these routines.  Thanks.
 *
 *	With that said, here's a brief (incomplete) description of each:
 *
 *		#include <procinfo.h>
 *
 *		struct procinfo pinfo[N];
 *		getproc(pinfo, N, sizeof(struct procinfo))
 *		RETURN: number of procs read (must be all), or -1 on error.
 *		[If N is too small, -1 is returned with errno == ENOSPC]
 *
 *		struct userinfo uinfo;
 *		struct procinfo *aproc;
 *		getuser(aproc, sizeof(struct procinfo), &uinfo, sizeof(uinfo))
 *		RETURN: 0 on success, -1 on error.
 */



/*
 * Snatch up the proc structure for examination.
 */
int
job_init_usage()
{

	int	pisize;
	static int p_bytes = -1;
	struct var v;


	if (kmem < 0)
		if (kernel_init() < 0)
			return (-1);

	if (p_bytes < 0) {

		if (sysconfig(SYS_GETPARMS, &v, sizeof(struct var)) < 0)
			nproc = 1024;	/* very big */
		else
			nproc = v.v_proc;

		/*
		 * Allocate space for proc structure array and
		 * array of pointers
		 */
		p_bytes = nproc * sizeof(struct procinfo);
		pbase = (struct procinfo *) ecalloc(nproc, sizeof(struct procinfo));
	}

	/*
	 * Read all the proc structures in one fell swoop
	 */
	pisize = sizeof(struct procinfo);

	if (getproc(pbase, (u_int) nproc, pisize) < 0)
		return (-1);

	return (0);
}



/*
 * Determine the resource use of a particular job.
 */
int
job_get_usage(jp)
Job *	jp;
{

	char *	mess = "job: %d; pid: %d; cannot get user area";
	int	i = 0;
	Bool	first = FALSE;				/* first child */
	time_t	newtime = 0;
	time_t	parent_time, child_time;
	struct procinfo * pp;


	/*
	 * Scan all the processes...
	 */
	for (pp = pbase, i = 0; i < nproc; pp++, i++) {

		/*
		 * If this process is in the process group we're
		 * interested in then get the user pages.
		 */
		if (pp->pi_pgrp != jp->j_pid)
			continue;


		/* 
		 * Don't look at "dead," zombie, idle, exiting processes.
		 */
		if (pp->pi_stat == 0
		    || pp->pi_stat == SZOMB || pp->pi_stat == SIDL
#ifdef SEXIT
		    || pp->pi_flag & SEXIT 
#endif
		    )
			continue;


		/*
		 * Read the user pages
		 */
		if (kernel_get_user(pp) < 0) {
			dbg((BLOG_RESOURCE, mess, jp->j_id, pp->pi_pid));
			continue;
		}

		if (first == FALSE) {
			/* re*-calculate the CPU time */
			/* jp->j_cputime = 0; */
			first = TRUE;
		}

		/*
		 * Sum in the times of this process...
		 */
		parent_time = ups.ui_ru.ru_utime.tv_sec +
			ups.ui_ru.ru_stime.tv_sec;

		/*
		 * ...and it's children.
		 */
		child_time = ups.ui_cru.ru_utime.tv_sec +
			ups.ui_cru.ru_stime.tv_sec;

		newtime += parent_time + child_time;

		dbg((BLOG_RESOURCE, "args: %s", ups.ui_comm));

		dbg((BLOG_RESOURCE,
		     "job: %d pid: %d ptime: %ld ctime: %ld",
		     jp->j_id, pp->pi_pid, parent_time, child_time));

		if (jp->j_pid == pp->pi_pid && jp->j_jname == NULL)
			jp->j_jname = estrdup(ups.ui_comm);

	}


	/*
	 * This is a sanity check on timing, if we gain more CPU time than
	 * real time in seconds, there is probably a time accounting problem.
	 * Give them some lee-way: we cannot gain more than twice the real
	 * time.
	 */
	if (jp->j_cputime && ((newtime - jp->j_cputime) / ptimer) > 2) {
		blog(BLOG_WARNING,
		     "job: %d; pid: %d; timing faulty: %d -> %d",
		     jp->j_id, jp->j_pid, jp->j_cputime, newtime);

		return (RTN_ERROR);
	} else {
		jp->j_cputime = newtime;
	}
		

	if (first == FALSE) {
		dbg((BLOG_RESOURCE,
		     "job: %d; pid: %d; proc structure not found",
		     jp->j_id, jp->j_pid));

		return (RTN_ERROR);

	} else  {


		dbg((BLOG_RESOURCE, "job: %d; times: %ld",
		     jp->j_id, jp->j_cputime));

		return (RTN_OK);
	}

}

/*
 *  Get the user structure for the process who's proc structure is pointed
 *  to by pp.  The user structure is returned in ups.
 */
static int
kernel_get_user(pp)
struct procinfo * pp;
{


	if (getuser(pp, sizeof(struct procinfo),
		    &ups, sizeof(struct userinfo)) < 0)
		return (-1);

	return(0);

}

#else

/*
 * Snatch up the proc structure for examination.
 */
int
job_init_usage()
{

	return (0);
}

/*
 * Determine the resource use of a particular job.
 */
int
job_get_usage(jp)
Job *	jp;
{

	return (RTN_OK);

}

#endif /* READ_PROC */

time_t
graphactive(current)
time_t	current;
{

	time_t	graphidle;			/* Time differences */

	graphidle = SECPERDAY;

	return graphidle;

}

