#ifndef lint
static char rcsid[] =
	"$Id: irix4.c,v 4.0 1993/09/17 00:36:23 srp Exp srp $";
#endif

/*
 *
 * Author: Scott R. Presnell
 *
 * Copyright (C) 1989, 1990, 1991, 1992, 1993
 * The Regents of the University of California, All Rights Reserved.
 *
 * THE REGENTS OF THE UNIVERSITY OF CALIFORNIA MAKE NO REPRESENTATION OR
 * WARRANTIES WITH RESPECT TO THE CONTENTS HEREOF AND SPECIFICALLY DISCLAIM
 * ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
 * PURPOSE.
 *
 * Further, the Regents of the University of California reserve the right to
 * revise this software and/or documentation and to make changes from time
 * to time in the content hereof without obligation of the Regents of the
 * University of California to notify any person of such revision or change.
 *
 */

/*
 * irix4.c - determine the load on the machine, check the process
 *	variables and keep track of the time a process uses.
 */


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <ctype.h>

#include <errno.h>
#include <fcntl.h>

#if defined(MALLOC_TRACE) || defined(MALLOC_LEAK)
#include "dmalloc.h"
#else
#include <malloc.h>
#endif

#include <nlist.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>

#include <sys/file.h>

#include <sys/statfs.h>
#include <sys/sysmp.h>
#include <sys/sysinfo.h>
#include <sys/sysmacros.h>

#include "config.h"

#include H_STRING

#include "batch.h"
#include "batchd.h"
#include "pathnames.h"

#if defined(USE_CUMULTIME)
#define READ_PROC
#endif

#ifdef READ_PROC
#include <sys/user.h>
#include <sys/proc.h>

#define	BPB	512				/* Bytes per Block */

struct user	ups;
struct proc *	pbase;

/* # processes */
int		nproc;

#endif /* READ_PROC */

#define svirtophys(x)	((uint)(x) & ~0xE0000000) /* Virtual to physical */
#define pagetok(x) (((x) << BPCSHIFT)/1024)	/* Pages to K */

#define	X_AVENRUN	0
#define X_BOOTTIME	1
#define X_PROC		2
#define X_VAR		3

struct	nlist nl[] = {
	{ "avenrun" },
        { "boottime" },
        { "proc" },
        { "v" },
	{ 0 },
};


static char *	Unix = _PATH_VMUNIX;
static char *	Swap = _PATH_SWAP;
static char *	Kmem = _PATH_KMEM;
static char *	Mem  = _PATH_MEM;

static int	kmem = -1;
static int	swap = -1;
static int	mem  = -1;

static time_t	granule = SECPERMIN;		/* granule of avenrun check */

static int	kernel_init P((void));
static int	kernel_get_value P((off_t, char *, int, char *));

#ifdef READ_PROC
static int	kernel_get_user P((struct proc *));
#endif /* READ_PROC */


static int
kernel_init()
{


	if ((nl[X_AVENRUN].n_value = sysmp(MP_KERNADDR, MPKA_AVENRUN)) == -1)
		blog(BLOG_ERR, "%s: no offset", nl[X_AVENRUN].n_name);

	if ((nl[X_PROC].n_value = sysmp(MP_KERNADDR, MPKA_PROC)) == -1)
		blog(BLOG_ERR, "%s: no offset", nl[X_PROC].n_name);

	if ((nproc = sysmp(MP_KERNADDR, MPKA_PROCSIZE)) == -1)
		blog(BLOG_ERR, "%s: cannot find", "nproc");

	if ((kmem = open(Kmem, 0)) == -1) {
		blog(BLOG_ERR, eopen, Kmem);
		return (-1);
	}

	if ((swap = open(Swap, 0)) == -1) {
		blog(BLOG_ERR, eopen, Swap);
		return (-1);
	}

	if ((mem = open(Mem, 0)) == -1) {
		blog(BLOG_ERR, eopen, Mem);
		return (-1);
	}

	return (0);
}


time_t
boottime_get()
{

	time_t	btime;

#if ndef
	if (kmem < 0)
		if (kernel_init() < 0)
			return (-1);

	if (kernel_get_value(nl[X_BOOTTIME].n_value,
			     (char *) &btime, sizeof(btime),
			     nl[X_BOOTTIME].n_name) < 0)
		return (-1);
#else
	btime = time((long *) 0);
#endif
	return btime;

}


int
load_get(aven)
double	aven[3];
{

	int	i;
	static long	avenrun[3];
	static time_t	otime = 0, elapsed = 0;
	time_t	current;

	current = time((long *)0);
	
	elapsed = current - otime;

	/*
	 * If it's been less than a minute, return the current values.
	 */
	if (elapsed < granule)
		goto done;
	else
		otime = current;

	if (kmem < 0)
		if (kernel_init() < 0)
			return (-1);

	if (kernel_get_value(nl[X_AVENRUN].n_value,
			     (char *) avenrun, sizeof(avenrun),
			     nl[X_AVENRUN].n_name) < 0)
		return (-1);
 done:

#define FSCALE 1024.0
	for (i = 0; i < 3; i++)
		aven[i] = (double) (avenrun[i] / FSCALE);

	return (0);
}


/*
 * Read kernel values
 */
static int
kernel_get_value(offset, ptr, size, refstr)
off_t	offset;
char *	ptr;
int	size;
char *	refstr;
{

	/*
	 * Seek to correct place in kmem...
	 */
	if (lseek(kmem, svirtophys(offset), 0) == -1)  {
		blog(BLOG_ERR, "%s: lseek to %s", Kmem, refstr);
		return (-1);
	}

	/*
	 * ...and read "size" bytes.
	 */
	if (read(kmem, ptr, (unsigned) size) != size)  {
		blog(BLOG_ERR, "%s: reading %s", Kmem, refstr);
		return (-1);
	}

	return (0);

}

#ifdef READ_PROC

/*
 * Snatch up the proc structure for examination.
 */
int
job_init_usage()
{

	static int p_bytes = -1;


	if (kmem < 0)
		if (kernel_init() < 0)
			return (-1);

	if (p_bytes < 0) {
		/*
		 * Allocate space for proc structure array and
		 * array of pointers
		 */
		p_bytes = nproc * sizeof(struct proc);
		pbase = (struct proc *) ecalloc(nproc, sizeof(struct proc));
	}


	/*
	 * Read all the proc structures in one fell swoop
	 */
	if (kernel_get_value(nl[X_PROC].n_value,
			     (char *) pbase, p_bytes,
			     nl[X_PROC].n_name) < 0)
		return (-1);

	return (0);
}



/*
 * Determine the resource use of a particular job.
 */
int
job_get_usage(jp)
Job *	jp;
{

	char *	mess = "job: %d; pid: %d; cannot get user area";
	int	i = 0;
	Bool	first = FALSE;				/* first child */
	time_t	newtime = 0;
	time_t	parent_time, child_time;
	struct proc * pp;


	/*
	 * Scan all the processes...
	 */
	for (pp = pbase, i = 0; i < nproc; pp++, i++) {

		/*
		 * If this process is in the process group we're
		 * interested in then get the user pages.
		 */
		if (pp->p_pgrp != jp->j_pid)
			continue;

		/* 
		 * Don't look at "dead," zombie, idle or exiting processes.
		 */
		if (pp->p_stat == 0
		    || pp->p_stat == SZOMB || pp->p_stat == SIDL
#ifdef SEXIT
		    || pp->p_flag & SEXIT 
#endif
		    )
			continue;


		/*
		 * Read the user pages
		 */
		if (kernel_get_user(pp) < 0) {
			dbg((BLOG_RESOURCE, mess, jp->j_id, pp->p_pid));
			continue;
		}

		if (first == FALSE) {
			/* re*-calculate the CPU time */
			/* jp->j_cputime = 0; */
			first = TRUE;
		}

		/*
		 * Sum in the times of this process...
		 */
		parent_time = ups.u_utime + ups.u_stime;

		/*
		 * ...and it's children.
		 */
		child_time = ups.u_cutime + ups.u_cstime;

		newtime += parent_time + child_time;

		dbg((BLOG_RESOURCE, "args: %s", ups.u_psargs));

		dbg((BLOG_RESOURCE,
		     "job: %d pid: %d ptime: %ld ctime: %ld",
		     jp->j_id, pp->p_pid, parent_time, child_time));

		if (jp->j_pid == pp->p_pid && jp->j_jname == NULL)
			jp->j_jname = estrdup(ups.u_comm);

	}

	/* convert time from clicks to seconds */
	if (newtime)
		newtime /= HZ;

	/*
	 * This is a sanity check on timing, if we gain more CPU time than
	 * real time in seconds, there is probably a time accounting problem.
	 * Give them some lee-way: we cannot gain more than twice the real
	 * time.
	 */
	if (jp->j_cputime && ((newtime - jp->j_cputime) / ptimer) > 2) {
		blog(BLOG_WARNING,
		     "job: %d; pid: %d; timing faulty: %d -> %d",
		     jp->j_id, jp->j_pid, jp->j_cputime, newtime);

		return (RTN_ERROR);
	} else {
		jp->j_cputime = newtime;
	}
		

	if (first == FALSE) {
		dbg((BLOG_RESOURCE,
		     "job: %d; pid: %d; proc structure not found",
		     jp->j_id, jp->j_pid));

		return (RTN_ERROR);

	} else  {

		dbg((BLOG_RESOURCE, "job: %d; times: %ld",
		     jp->j_id, jp->j_cputime));

		return (RTN_OK);
	}

}

/*
 *  Get the user structure for the process who's proc structure is pointed
 *  to by pp.  The user structure is returned in ups.
 */
static int
kernel_get_user(pp)
struct proc * pp;
{

	int	nbytes, i;
	unsigned int n;
	caddr_t	upage = (caddr_t) &ups;


#ifdef SULOAD
	if ((pp->p_flag & SULOAD) == 0) {
#else
	if ((pp->p_flag & SLOAD) == 0) {
#endif
		/*
		 * Read the swap file for the user pages.
		 */
		dbg((BLOG_RESOURCE, "process %d user pages are on swap", pp->p_pid));

		for (nbytes = sizeof(ups), i = 0; nbytes > 0;
		     nbytes -= NBPP, i++) {
			if (pp->p_upgs[i].pgm.dbd.dbd_type != DBD_SWAP)
				return (-1);
			if (lseek(swap, (long) pp->p_upgs[i].pgm.dbd.dbd_pgno * BPB, 0) < 0)
				return (-1);
			n = MIN(nbytes, NBPP);
			if (read(swap, upage, n) != n)
				return (-1);
			upage += n;
		}

	} else {

		/*
		 * Read the real memory location of the user pages.
		 */
		for (nbytes = sizeof(ups), i = 0; nbytes > 0;
		     nbytes -= NBPP, i++) {
			if (lseek(mem, (long) (pp->p_upgs[i].pgm.pg_pfn * NBPP), 0) < 0)
				return (-1);
			n = MIN(nbytes, NBPP);
			if (read(mem, upage, n) != n)
				return (-1);
			upage += n;
		}
	}

	return(0);

}

#else

/*
 * Snatch up the proc structure for examination.
 */
int
job_init_usage()
{

	return (0);
}

/*
 * Determine the resource use of a particular job.
 */
int
job_get_usage(jp)
Job *	jp;
{

	return (RTN_OK);

}

#endif /* READ_PROC */


long
freemem_get()
{

	struct rminfo ri;


	if (sysmp(MP_SAGET, MPSA_RMINFO, &ri, sizeof(ri)) < 0) {
		blog(BLOG_ERR, "sysmp");
		return (-1);
	}

	dbg((BLOG_RESOURCE,
	     "freemem %ld, freereal %ld freeswap %ld, physmem %ld",
	     pagetok(ri.availsmem), pagetok(ri.freemem),
	     pagetok(ri.availsmem - ri.freemem),
	     pagetok(ri.physmem)));

	return pagetok(ri.availsmem);
}


long
freedisk_get(path)
char *	path;
{

	struct statfs  bi;


	if (statfs(path, &bi, sizeof(bi), 0) < 0) {
		blog(BLOG_ERR, "statfs");
		return (-1);
	}

	dbg((BLOG_RESOURCE,
	     "totalK %ld freeK %ld totalI %ld freeI %ld",
	     (bi.f_blocks/2), (bi.f_bfree/2),
	     bi.f_files, bi.f_ffree));

	return (bi.f_blocks/2);
	
}


extern int	Gswitches;		/* factor for Graphics con. switches */

time_t
graphactive(current)
time_t	current;
{

	time_t	graphidle;			/* Time differences */

	static time_t	lasttime = 0L;		/* Last time console checked */
	static time_t	graphused;		/* Last time console in use */

	static u_long	lastgs = 0L;		/* Old graphics c-switches */
	struct sysinfo	si;			/* Sysinfo structure */


	if (!lasttime)
		lasttime = current - ptimer; /* fudge */

	graphused = 0;

	graphidle = 86400;

	if (sysmp(MP_SAGET, MPSA_SINFO, &si, sizeof(si)) < 0) {
		blog(BLOG_ERR, "graphactive(): problem with sysinfo");
		return graphidle;
	}

	if (!lastgs)
		lastgs = si.gswitch;

	if ((si.gswitch - lastgs)/(current - lasttime) >= Gswitches) {
		graphused = current;
	}

	/*
	 * Graphics idle time is the time between now, and when we last knew
	 * the graphics pipleine to be working up a sweat.
	 */
	if (graphused > 0) {
		graphidle = current - graphused;
		if (graphidle < 0)
			graphidle = 0L;
	}

	/*
	 * Remember the current settings for next time.
	 */
	lasttime = current;
	lastgs = si.gswitch;

	return graphidle;

}
