#ifndef lint
static char rcsid[] =
	"$Id: ultrix4.c,v 4.0 1993/09/17 00:38:17 srp Exp srp $";
#endif

/*
 *
 * Author: Scott R. Presnell
 *
 * Copyright (C) 1989, 1990, 1991, 1992, 1993
 * The Regents of the University of California, All Rights Reserved.
 *
 * THE REGENTS OF THE UNIVERSITY OF CALIFORNIA MAKE NO REPRESENTATION OR
 * WARRANTIES WITH RESPECT TO THE CONTENTS HEREOF AND SPECIFICALLY DISCLAIM
 * ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
 * PURPOSE.
 *
 * Further, the Regents of the University of California reserve the right to
 * revise this software and/or documentation and to make changes from time
 * to time in the content hereof without obligation of the Regents of the
 * University of California to notify any person of such revision or change.
 *
 */

/*
 * ultrix4.c - determine the load on the machine, check the process
 *	variables and keep track of the time a process uses.
 */

#include <stdio.h>
#include <ctype.h>

#include <errno.h>
#include <fcntl.h>

#if defined(MALLOC_TRACE) || defined(MALLOC_LEAK)
#include "dmalloc.h"
#else
#include <malloc.h>
#endif

#include <nlist.h>
#include <time.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/fixpoint.h>

#include <sys/file.h>

#include "config.h"

#include H_STRING

#include "batch.h"
#include "batchd.h"
#include "pathnames.h"
#include "protocol.h"

#if defined(USE_CUMULTIME)
#define READ_PROC
#endif

#ifdef READ_PROC
#include <machine/pte.h>
#include <sys/vm.h>
#include <sys/user.h>
#include <sys/proc.h>

struct user	ups;
struct proc *	proc;
struct proc *	pbase;

/* # processes */
int		nproc;

#endif /* READ_PROC */

#define	X_AVENRUN	0
#define X_BOOTTIME	1
#define X_PROC		2
#define X_PROC_BITMAP	3
#define X_NPROC		4

struct	nlist nl[] = {
	{ "_avenrun" },
        { "_boottime" },
        { "_proc" },
        { "_proc_bitmap" },
        { "_nproc" },
	{ 0 },
};


static char *	Unix = _PATH_VMUNIX;
static char *	Swap = _PATH_SWAP;
static char *	Kmem = _PATH_KMEM;
static char *	Mem  = _PATH_MEM;

static int	kmem = -1;
static int	swap = -1;
static int	mem  = -1;

static time_t	granule = SECPERMIN;		/* granule of avenrun check */

static int	kernel_init P((void));
static int	kernel_get_value P((off_t, char *, int, char *));

#ifdef READ_PROC
static int	kernel_get_user P((struct proc *));
#endif /* READ_PROC */


static int
kernel_init()
{

	int	i;

	nlist(Unix, nl);
	
	for (i = 0; i < sizeof(nl)/sizeof(struct nlist) - 1; i++) {
		if (nl[i].n_type == 0) {
			blog(BLOG_ERR, "%s: no offset for %s (%d)",
			    Unix, nl[i].n_name, i);
			return (-1);
		}
	}

	if ((kmem = open(Kmem, 0)) == -1) {
		blog(BLOG_ERR, eopen, Kmem);
		return (-1);
	}

	if ((swap = open(Swap, 0)) == -1) {
		blog(BLOG_ERR, eopen, Swap);
		return (-1);
	}

	if ((mem = open(Mem, 0)) == -1) {
		blog(BLOG_ERR, eopen, Mem);
		return (-1);
	}

	return (0);
}



time_t
boottime_get()
{

	struct timeval btime;

#if ndef
	if (kmem < 0)
		if (kernel_init() < 0)
			return (-1);

	if (kernel_get_value(nl[X_BOOTTIME].n_value,
			     (char *) &btime, sizeof(btime),
			     nl[X_BOOTTIME].n_name) < 0)
		return (-1);
#else
	btime.tv_sec = time((time_t *) 0);
#endif
	return btime.tv_sec;

}


int
load_get(aven)
double	aven[3];
{

	int	i;
	static fix	avenrun[3];
	static time_t	otime = 0, elapsed = 0;
	time_t	current;


	current = time((long *)0);
	
	elapsed = current - otime;

	/*
	 * If it's been less than a minute, return the current values.
	 */
	if (elapsed < granule)
		goto done;
	else
		otime = current;

	if (kmem < 0)
		if (kernel_init() < 0)
			return (-1);

	if (kernel_get_value(nl[X_AVENRUN].n_value,
			     (char *) avenrun, sizeof(avenrun),
			     nl[X_AVENRUN].n_name) < 0)
		return (-1);
 done:

	for (i = 0; i < 3; i++)
		aven[i] = (double) FIX_TO_DBL(avenrun[i]);

	return (0);
}

/*
 * Read kernel values
 */
static int
kernel_get_value(offset, ptr, size, refstr)
off_t	offset;
char *	ptr;
int	size;
char *	refstr;
{

	/*
	 * Seek to correct place in kmem...
	 */
	if (lseek(kmem, offset, 0) == -1)  {
		blog(BLOG_ERR, "%s: lseek to %s", Kmem, refstr);
		return (-1);
	}

	/*
	 * ...and read "size" bytes.
	 */
	if (read(kmem, ptr, (unsigned) size) != size)  {
		blog(BLOG_ERR, "%s: reading %s", Kmem, refstr);
		return (-1);
	}

	return (0);

}

#ifdef READ_PROC

/*
 * Snatch up the proc structure for examination.
 */
int
job_init_usage()
{

	static int p_bytes = -1;


	if (kmem < 0)
		if (kernel_init() < 0)
			return (-1);

	if (p_bytes < 0) {
		/*
		 * Allocate space for proc structure array and
		 * array of pointers
		 */
		if (kernel_get_value(nl[X_NPROC].n_value,
				     (char *) &nproc, sizeof(nproc),
				     nl[X_NPROC].n_name) < 0)
			return (-1);

		p_bytes = nproc * sizeof(struct proc);
		pbase = (struct proc *) ecalloc(nproc, sizeof(struct proc));

		if (kernel_get_value(nl[X_PROC].n_value,
				     (char *) &proc, sizeof(int),
				     nl[X_PROC].n_name) < 0)
			return(-1);

	}

	/*
	 * Read all the proc structures in one fell swoop
	 */
	if (kernel_get_value((off_t) proc,
			     (char *) pbase, p_bytes,
			     nl[X_PROC].n_name) < 0)
		return (-1);

	return (0);
}



/*
 * Determine the resource use of a particular job.
 */
int
job_get_usage(jp)
Job *	jp;
{

	char *	mess = "job: %d; pid: %d; cannot get user area";
	int	i = 0;
	unsigned int bitmap;
	Bool	first = FALSE;				/* first child */
	time_t	newtime = 0;
	time_t	parent_time, child_time;
	struct proc * pp;


	/*
	 * Scan all the processes...
	 */
	for (pp = pbase, i = 0; i < nproc; pp++, i++) {

#ifdef ndef
		/* 
		 * Ultrix 4.x uses a bitmap to tell if a proc structure is
		 * valid.  We read and check the bitmap.  The code here is
		 * structured around 32 bits/entry, and 4 bytes/entry.
		 * But for some reason - some of the valid procs are marked
		 * invalid - so scan the entire table for now.
		 */
		if ((i % 32) == 0) {
			if (kernel_get_value((off_t) nl[X_PROC_BITMAP].n_value,
					     (char *) &bitmap, sizeof(bitmap),
					     nl[X_PROC_BITMAP].n_name) < 0)
				return (-1);

			bitmap += (unsigned int) ((i / 32) * 4);
		}

		if ((bitmap & (1 << (i % 32))) == 0) {
			dbg((BLOG_RESOURCE, "proc %d (abs proc %d) inactive",
			    (int) pp->p_pid, i));
			continue;
		} else {
			dbg((BLOG_RESOURCE, "proc %d pgrp %d (abs proc %d) active",
			    (int) pp->p_pid, (int) pp->p_pgrp, i));
		}
#endif

		/*
		 * If this process is in the process group we're
		 * interested in then get the user pages.
		 */
		if (pp->p_pgrp != jp->j_pid)
			continue;

		/* 
		 * Don't look at "dead," zombie, idle or exiting processes.
		 */
		if (pp->p_stat == 0
		    || pp->p_stat == SZOMB || pp->p_stat == SIDL
#ifdef SWEXIT
		    || pp->p_type & SWEXIT
#endif
		    )
			continue;


		/*
		 * Read the user pages
		 */
		if (kernel_get_user(pp) < 0) {
			dbg((BLOG_RESOURCE, mess, jp->j_id, (int) pp->p_pid));
			continue;
		}

		if (first == FALSE) {
			/* re*-calculate the CPU time */
			/* jp->j_cputime = 0; */
			first = TRUE;
		}

		/*
		 * Sum in the times of this process...
		 */
		parent_time = ups.u_ru.ru_utime.tv_sec +
			ups.u_ru.ru_stime.tv_sec;

		/*
		 * ...and it's children.
		 */
		child_time = ups.u_cru.ru_utime.tv_sec +
			ups.u_cru.ru_stime.tv_sec;

		newtime += parent_time + child_time;

		dbg((BLOG_RESOURCE, "args: %s", ups.u_comm));

		dbg((BLOG_RESOURCE,
		     "job: %d pid: %d ptime: %ld ctime: %ld",
		     jp->j_id, pp->p_pid, parent_time, child_time));

		if (jp->j_pid == pp->p_pid && jp->j_jname == NULL)
			jp->j_jname = estrdup(ups.u_comm);
	}

	/*
	 * This is a sanity check on timing, if we gain more CPU time than
	 * real time in seconds, there is probably a time accounting problem.
	 * Give them some lee-way: we cannot gain more than twice the real
	 * time.
	 */
	if (jp->j_cputime && ((newtime - jp->j_cputime) / ptimer) > 2) {
		blog(BLOG_WARNING,
		     "job: %d; pid: %d; timing faulty: %d -> %d",
		     jp->j_id, jp->j_pid, jp->j_cputime, newtime);

		return (RTN_ERROR);
	} else {
		jp->j_cputime = newtime;
	}
		

	if (first == FALSE) {
		dbg((BLOG_RESOURCE,
		     "job: %d; pid: %d: proc structure not found",
		     jp->j_id, jp->j_pid));

		return (RTN_ERROR);

	} else  {

		dbg((BLOG_RESOURCE, "job: %d; times: %ld",
		     jp->j_id, jp->j_cputime));

		return (RTN_OK);
	}

}

#define	SW_UBYTES	sizeof(struct user)

/*
 *  Get the user structure for the process who's proc structure is pointed
 *  to by pp.  The user structure is returned in ups.
 */
static int
kernel_get_user(pp)
struct proc * pp;
{

	int	i;
	unsigned int n;
	int	nbytes, u_addr;

	static union {
		struct user user;
		char upgs[UPAGES][NBPG];
	} u;

	static struct dmap dmap;
	static struct pte uptes[UPAGES];


	/*
	 * If process is not in core, we simply snarf it's user struct
	 * from the swap device.
	 */
	if ((pp->p_sched & SLOAD) == 0) {
		if (kernel_get_value((off_t) pp->p_smap,
				     (char *) &dmap, sizeof(dmap),
				     "pp->p_smap") < 0)
			return (-1);
		if (kernel_get_value((off_t) dmap.dm_ptdaddr,
				     (char *) &u_addr, sizeof(u_addr),
				     "u_addr") < 0)
			return (-1);
		if (lseek(swap, (off_t) dtob(u_addr), 0) == -1)
			return (-1);
		if (read(swap, (char *) &u.user, SW_UBYTES) != SW_UBYTES)
			return (-1);
	} else {

		/*
		 * Process is in core.  Follow p_addr to read in the page
		 * table entries that map the u-area and then read in the
		 * physical pages that comprise the u-area.
		 */
		if (kernel_get_value((off_t) pp->p_addr,
				     (char *) &uptes, sizeof(uptes),
				     "uptes") < 0)
			return (-1);

		nbytes = sizeof(struct user);

		for (i = 0; i < UPAGES && nbytes > 0; i++) {
			
			if (lseek(mem, ptob(uptes[i].pg_pfnum), 0) == -1)
				return (-1);
			if (read(mem, (char *) u.upgs[i], NBPG) == -1)
				return (-1);
		}
		nbytes -= NBPG;
	}

	bcopy((voidp) &u.user, (voidp) &ups, SW_UBYTES);
	return (0);

}

#else

/*
 * Snatch up the proc structure for examination.
 */
int
job_init_usage()
{

	return (0);
}

/*
 * Determine the resource use of a particular job.
 */
int
job_get_usage(jp)
Job *	jp;
{

	return (RTN_OK);

}


#endif /* READ_PROC */

time_t
graphactive(current)
time_t	current;
{

	time_t	graphidle;			/* Time differences */

	graphidle = SECPERDAY;

	return graphidle;

}
