case $CONFIG in
'')
	if test -f config.sh; then TOP=.;
	elif test -f ../config.sh; then TOP=..;
	elif test -f ../../config.sh; then TOP=../..;
	elif test -f ../../../config.sh; then TOP=../../..;
	elif test -f ../../../../config.sh; then TOP=../../../..;
	else
		echo "Can't find config.sh."; exit 1
	fi
	. $TOP/config.sh
	;;
esac
: This forces SH files to create target in same directory as SH file.
: This is so that make depend always knows where to find SH derivatives.
case "$0" in
*/*) cd `expr X$0 : 'X\(.*\)/'` ;;
esac
echo "Extracting src/Makefile (with variable substitutions)"
: This section of the file will have variable substitutions done on it.
: Move anything that needs config subs from !NO!SUBS! section to !GROK!THIS!.
: Protect any dollar signs and backticks that you do not want interpreted
: by putting a backslash in front.  You may delete these comments.
$spitshell >Makefile <<!GROK!THIS!
#
# \$Id: Makefile.SH,v 4.0 1993/09/17 00:42:25 srp Exp srp $
#
# Makefile for batch daemon, and clients.
#
# Shell that make uses to run scripts
SHELL	=	/bin/sh

.SUFFIXES: .s .i .c

.c.s:
	$(CC) $(CFLAGS) -S $<

.c.i:
	$(CC) $(CFLAGS) -P $<
#
#
#
OSTYPE	=	$os

# Final destination for the daemon and the control programs
ETCDIR	=	$installetc

# Final destination for the batch, barm and baq programs
BINDIR	=	$installbin

# C compiler (maybe gcc?)
CC	=	$cc

# Yacc parser (maybe bison?)
YACC	=	$yacc

# Make depend (maybe makedepend?)
MKDEP	=	$mkdep

#
# If not using reserve ports, the client modes don't have to have setuid,
# setgid is for writing/managing the spooling directories.	
# The daemon must allways be started as root.
#
# Name of the owner (for chown)
OWNER =		root

# Name of the daemon group (for chown)
GROUP =		daemon

INSTALL	=	../install -c
INSTALL_PROG =	\$(INSTALL) -o \$(OWNER) -g \$(GROUP) -m 4711

# cc options
OPTS	=	$optimize
#DEFS	=	-D__STDC__ -DDEBUG -DUCSF -DR4000 -DMALLOC_DEBUG
DEFS	=	$ccflags
INCS	=	-I..
CFLAGS	=	\$(OPTS) \$(DEFS) \$(INCS)

DPFLAGS =	\$(DEFS) \$(INCS)

# ld options
LDFLAGS	=	\$(OPTS) $ldflags #-Wl,-m

# libraries
#LIBS	=	-lsun -lbsd -lmalloc # /usr/src/local/libmalloc_d/libmalloc_d.a
LIBS	=	$libs

# Lint options
LOPT	=	-abxyz \$(DEFS)
LFLAGS	=	\$(LOPT) \$(DEFS) \$(INCS) 

# yacc options - none needed
# YFLAGS	=

# end of configurables
!GROK!THIS!

: In the following dollars and backticks do not need the extra backslash.
$spitshell >>Makefile <<'!NO!SUBS!'

ERRORSRC=	error.c
MALLOCSRC=	emalloc.c
REQUESTSRC=	request.c reply.c
COMMONSRC=	batchcap.c global.c getsocket.c util.c vprintf.c $(MALLOCSRC)

BAQSRC	=	baq.c $(COMMONSRC) $(REQUESTSRC) $(ERRORSRC)

BAISRC	=	bainfo.c $(COMMONSRC) $(REQUESTSRC) $(ERRORSRC)

BACSRC	=	bac.c cmds.c cmdtab.c $(COMMONSRC) $(REQUESTSRC) \
		$(ERRORSRC)

BARMSRC	=	barm.c groups.c $(COMMONSRC) $(REQUESTSRC) $(ERRORSRC) 

BATCHSRC=	batch.c getdate.y groups.c getpment.c getusershell.c \
		$(COMMONSRC) $(REQUESTSRC) $(ERRORSRC)

BATCHDSRC=	batchd.c queuejob.c sendjob.c runjob.c rmjob.c bid.c \
		readcap.c resource.c getproc.c recvjob.c displayjob.c \
		faults.c lock.c status.c event.c xlog.c notify.c console.c \
		$(COMMONSRC)

ALLSRC	=	batch.c getdate.c getpment.c barm.c baq.c \
		groups.c bac.c cmds.c cmdtab.c \
		batchd.c queuejob.c sendjob.c runjob.c rmjob.c bid.c \
		readcap.c resource.c getproc.c recvjob.c notify.c console.c \
		displayjob.c faults.c lock.c status.c event.c xlog.c \
		$(COMMONSRC) $(REQUESTSRC) $(ERRORSRC)

ERROR	=	error.o 
MALLOC	=	emalloc.o
REQUEST	=	request.o reply.o
COMMON	=	batchcap.o global.o getsocket.o util.o vprintf.o $(MALLOC)

BAQ	=	baq.o $(COMMON) $(REQUEST) $(ERROR)

BAI	=	bainfo.o $(COMMON) $(REQUEST) $(ERROR)

BAC	=	bac.o cmds.o cmdtab.o $(COMMON) $(REQUEST) $(ERROR)

BARM	=	barm.o groups.o $(COMMON) $(REQUEST) $(ERROR)

BATCH	=	batch.o getdate.o groups.o getpment.o getusershell.o \
		$(COMMON) $(REQUEST) $(ERROR)

BATCHD	=	batchd.o queuejob.o sendjob.o runjob.o rmjob.o bid.o \
		readcap.o resource.o getproc.o recvjob.o displayjob.o \
		faults.o lock.o status.o event.o xlog.o notify.o console.o \
		$(COMMON) 

ALLOBJ	=	batch.o getdate.o getpment.o barm.o baq.o \
		bac.o cmds.o cmdtab.o \
		batchd.o queuejob.o sendjob.o runjob.o rmjob.o bid.o \
		readcap.o resource.o getproc.o recvjob.o notify.o console.o \
		displayjob.o faults.o lock.o status.o event.o xlog.o \
		$(COMMON) $(REQUEST) $(ERROR)

DAEMON	=	batchd
CONTROL	=	bac
CLIENTS	=	batch baq barm bainfo

ALL	=	$(DAEMON) $(CONTROL) $(CLIENTS)

all:	$(DAEMON) $(CONTROL) $(CLIENTS)

$(DAEMON): $(BATCHD)
	$(CC) $(LDFLAGS) -o $@ $(BATCHD) $(LIBS)

$(CONTROL): $(BAC)
	$(CC) $(LDFLAGS) -o $@ $(BAC) $(LIBS)

batch:	$(BATCH)
	$(CC) $(LDFLAGS) -o $@ $(BATCH) $(LIBS)

baq:	$(BAQ) 
	$(CC) $(LDFLAGS) -o $@ $(BAQ) $(LIBS)

barm:	$(BARM)
	$(CC) $(LDFLAGS) -o $@ $(BARM) $(LIBS)

bainfo:	$(BAI) 
	$(CC) $(LDFLAGS) -o $@ $(BAI) $(LIBS)

getproc.c: ../machdep/$(OSTYPE).c
	rm -f getproc.c getproc.o
	ln -s ../machdep/$(OSTYPE).c getproc.c
	
install:  install.daemon install.control install.clients

install.batchd install.daemon: $(ETCDIR)/$(DAEMON)

$(ETCDIR)/$(DAEMON): $(DAEMON)
	$(INSTALL_PROG) $? $@

install.bac install.control: $(ETCDIR)/$(CONTROL)

$(ETCDIR)/$(CONTROL): $(CONTROL)
	$(INSTALL_PROG) $? $@

install.clients: install.batch install.baq install.barm install.bainfo

install.batch: $(BINDIR)/batch

$(BINDIR)/batch: batch
	$(INSTALL_PROG) $? $@

install.baq: $(BINDIR)/baq

$(BINDIR)/baq:  baq
	$(INSTALL_PROG) $? $@

install.barm: $(BINDIR)/barm

$(BINDIR)/barm: barm
	$(INSTALL_PROG) $? $@

install.bainfo: $(BINDIR)/bainfo

$(BINDIR)/bainfo: bainfo
	$(INSTALL_PROG) $? $@

lint.daemon: 
	-lint $(LFLAGS) $(BATCHDSRC) $(LFILTER)

lint.batch: 
	-lint $(LFLAGS) $(BATCHSRC) $(LFILTER)

lint.baq:
	-lint $(LFLAGS) $(BAQSRC) $(LFILTER)

lint.barm:
	-lint $(LFLAGS) $(BARMSRC) $(LFILTER)

lint.bac:
	-lint $(LFLAGS) $(BACSRC) $(LFILTER)

depend: getdate.c getproc.c
	(sed '/^# DO NOT DELETE/q' Makefile && \
		$(MKDEP) $(DPFLAGS) -- $(ALLSRC) ) > Makefile.new
	mv Makefile Makefile.bak
	mv Makefile.new Makefile
	rm -f Makefile.new

clean:
	/bin/rm -f a.out *.o *.s *.i core

clobber: clean
	/bin/rm -f $(ALL) getdate.c getproc.c

.PHONY:		all install depend clean clobber

batchd.o bid.o emalloc.o event.o queuejob.o readcap.o: Makefile

# Dependencies generated by make depend
# DO NOT DELETE THIS LINE -- make depend relies on it

!NO!SUBS!
chmod 644 Makefile
$eunicefix Makefile
