#ifndef lint
static char rcsid[] =
	"$Id: bac.c,v 4.0 1993/09/17 00:41:04 srp Exp srp $";
#endif

/*
 * Copyright (c) 1983 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Berkeley.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef lint
char copyright[] =
"@(#) Copyright (c) 1983 Regents of the University of California.\n\
 All rights reserved.\n";
#endif /* not lint */

/*
 * static char sccsid[] = "@(#)lpc.c	5.7 (Berkeley) 11/20/88";
 */

/*
 * This code is derivative of lpc.c for lpc in the line printer
 * software.  The copyrights and identifiers for that code have been
 * retained, they are still in effect.
 */

/*
 * bac -- Batch queue control program.  Control various attributes and the
 *	  jobs within the batch queues.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <ctype.h>

#include <errno.h>
#include <fcntl.h>

#if defined(MALLOC_TRACE) || defined(MALLOC_LEAK)
#include "dmalloc.h"
#else
#include <malloc.h>
#endif

#include <pwd.h>
#include <setjmp.h>
#include <signal.h>

#include <sys/types.h>
#include <sys/param.h>

#include "config.h"

#include H_STRING

#include "batch.h"
#include "bac.h"

/*
 * If the user is in the "operator" group, then they are allowed to 
 * use the restricted functions.  The define may be changed to afford
 * these privileges to a different group name.
 * See batch.h:
 * #define OPERATOR	"operator"
 */

Bool	trusted = FALSE;			/* User an operator? */

int	fromatty;				/* Connection from a tty? */

char	cmdline[200];
int	margc;
char *	margv[20];
int	top;

UIDTYPE	myuid;
GIDTYPE	mygid;

jmp_buf	toplevel;				/* Where to jump on intr */

extern struct cmd cmdtab[];

static int	cmdscanner P((int toplev));
static struct cmd * getcmd P((char *name));
static int	makeargv P((void));
static SIGRTN	intr P((int sig));

extern int	ingroup P((char *, GIDTYPE, char *));

main(argc, argv)
int	argc;
char *	argv[];
{

	int		mygroup = -1;
	struct cmd *	c;
	struct passwd *	pw;


	progname = argv[0];

	myuid = getuid();
	mygid = getgid();

	if (gethostname(lhost, MAXHOSTNAMELEN) < 0) {
		(void) fprintf(stderr, "gethostname fails\n");
		exit(1);
	}

	/*
	 * Check for valid password entry
	 */
	if ((pw = getpwuid(myuid)) == NULL) {
		(void) printf("?Who are you?\n");
		exit(1);
	}


	/*
	 * If the user is an operator, set that variable to TRUE
	 */
	if ((mygroup = ingroup(pw->pw_name, mygid, OPERATOR)) < 0) {
		(void) printf("?Invalid group: %s\n", OPERATOR);
		exit(1);
	} else if (mygroup == 0) {
		trusted = TRUE;
	} else {
		trusted = FALSE;
	}

	argv++;

	/*
	 * Parse the arguments - if there is a leading dash, then the argument
	 * is an option.  Options must be separated from thier arguments by
	 * white space.
	 */
	while (--argc > 0) {
		if (argv[0][0] == '-') {
			switch (argv[0][1]) {
#ifdef USE_PORT_SELECTION
			case 'L':
				socketname = argv[1];
				argv++;
				argc--;
				break;

			case 'R':
				remote_port = atoi(argv[1]);
				argv++;
				argc--;
				break;
#endif
			case 'V':
				dump_version(printsf, 0);
				exit(0);
				break;
			}

		} else {
			/*
			 * Scan the command line for commands and
			 * their arguments to execute.
			 */
			c = getcmd(*argv);
			if (c == (struct cmd *)-1) {
				(void) printf("?Ambiguous command\n");
				exit(1);
			}
			if (c == 0) {
				(void) printf("?Invalid command\n");
				exit(1);
			}
			if (c->c_priv && getuid() && trusted == FALSE) {
				(void) printf("?Privileged command\n");
				exit(1);
			}
			/*
			 * Call the function that matched with the 
			 * rest of the command line argmuents.
			 */
			(*c->c_handler)(argc, argv);
			exit(0);
		}
	}

	/*
	 * Scan commands and arguments from stdin
	 */
	fromatty = isatty(fileno(stdin));

	top = setjmp(toplevel) == 0;

	if (top)
		(void) signal(SIGINT, intr);

	for (;;) {
		(void) cmdscanner(top);
		top = 1;
	}

	/*NOTREACHED*/
}


/*
 * Command parser.
 */
static int
cmdscanner(toplev)
int	toplev;
{

	struct cmd *	c;


	if (!toplev)
		putchar('\n');
	for (;;) {
		if (fromatty) {
			(void) printf("bac> ");
			(void) fflush(stdout);
		}

		if (fgets(cmdline, sizeof(cmdline), stdin) == NULL)
			quit(0, NULL);

		if (cmdline[0] == '\0')
			break;

		if (cmdline[0] == '\n')
			continue;

		makeargv();

		c = getcmd(margv[0]);
		if (c == (struct cmd *)-1) {
			(void) printf("?Ambiguous command\n");
			continue;
		}

		if (c == 0) {
			(void) printf("?Invalid command\n");
			continue;
		}

		if (c->c_priv && getuid() && trusted == FALSE) {
			(void) printf("?Privileged command\n");
			continue;
		}

		(*c->c_handler)(margc, margv);
	}

	longjmp(toplevel, 0);
}


static struct cmd *
getcmd(name)
char *	name;
{

	char *	p, *q;
	struct cmd *	c, *found;
	int	nmatches, longest;


	longest = 0;
	nmatches = 0;
	found = 0;
	for (c = cmdtab; p = c->c_name; c++) {
		for (q = name; *q == *p++; q++)
			if (*q == 0)		/* exact match? */
				return(c);
		if (!*q) {			/* the name was a prefix */
			if (q - name > longest) {
				longest = q - name;
				nmatches = 1;
				found = c;
			} else if (q - name == longest)
				nmatches++;
		}
	}
	if (nmatches > 1)
		return((struct cmd *)-1);

	return(found);
}

/*
 * Slice a string up into argc/argv.
 */
static int
makeargv()
{

	char *	cp;
	char **	argp = margv;


	margc = 0;
	for (cp = cmdline; *cp;) {
		while (isspace((int) *cp))
			cp++;
		if (*cp == '\0')
			break;
		*argp++ = cp;
		margc += 1;
		while (*cp != '\0' && !isspace((int) *cp))
			cp++;
		if (*cp == '\0')
			break;
		*cp++ = '\0';
	}
	*argp++ = 0;
}

#define HELPINDENT (sizeof ("directory"))

/*
 * Help command.
 */
int
help(argc, argv)
int	argc;
char *	argv[];
{
	
	struct cmd *	c;


	if (argc == 1) {
		int i, j, w;
		int columns, width = 0, lines;
		extern int NCMDS;

		(void) printf("Commands may be abbreviated.  Commands are:\n\n");
		for (c = cmdtab; c < &cmdtab[NCMDS - 1]; c++) {
			int len = strlen(c->c_name);

			if (len > width)
				width = len;
		}
		width = (width + 8) &~ 7;
		columns = 80 / width;
		if (columns == 0)
			columns = 1;
		lines = (NCMDS + columns - 1) / columns;
		for (i = 0; i < lines; i++) {
			for (j = 0; j < columns; j++) {
				c = cmdtab + j * lines + i;
				(void) printf("%s", c->c_name);
				if (c + lines >= &cmdtab[NCMDS]) {
					(void) printf("\n");
					break;
				}
				w = strlen(c->c_name);
				while (w < width) {
					w = (w + 8) &~ 7;
					putchar('\t');
				}
			}
		}
		return;
	}
	while (--argc > 0) {
		char *arg;
		arg = *++argv;
		c = getcmd(arg);
		if (c == (struct cmd *)-1)
			(void) printf("?Ambiguous help command %s\n", arg);
		else if (c == (struct cmd *)0)
			(void) printf("?Invalid help command %s\n", arg);
		else
			(void) printf("%-*s\t%s\n", HELPINDENT,
				c->c_name, c->c_help);
	}
}


/*
 * Interrupt handler
 */
static SIGRTN
intr(sig)
int	sig;
{

	if (!fromatty)
		exit(0);

	longjmp(toplevel, 1);

	/*NOTREACHED*/
	return;
}
