#ifndef lint
static char rcsid[] =
	"$Id: bainfo.c,v 4.0 1993/09/17 00:41:04 srp Exp srp $";
#endif

/*
 *
 * Author: Scott R. Presnell
 *
 * Copyright (C) 1989, 1990, The Regents of the University of California
 *                     All Rights Reserved
 *
 * THE REGENTS OF THE UNIVERSITY OF CALIFORNIA MAKE NO REPRESENTATION OR
 * WARRANTIES WITH RESPECT TO THE CONTENTS HEREOF AND SPECIFICALLY DISCLAIM
 * ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
 * PURPOSE.
 *
 * Further, the Regents of the University of California reserve the right to
 * revise this software and/or documentation and to make changes from time
 * to time in the content hereof without obligation of the Regents of the
 * University of California to notify any person of such revision or change.
 *
 */

/*
 * Bainfo: batch queue information program.
 *
 * bainfo [-H host] [-Q queue[@host] ]
 *
 * -H specify host (all - specify all hosts known to batch)
 * -Q specify queue (all - specify all queues known to batch)
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <errno.h>
#include <signal.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/time.h>

#include "config.h"

#include H_STRING

#include "batch.h"
#include "pathnames.h"
#include "protocol.h"
#include "reply.h"


static char *	use[] = {
	"bainfo -h [-Q queue] [-H host]",
	"\"Bainfo\" returns the information for a batch queue on a specified host.",
	"",
	"Options:",
	"-h		This help message",
#ifdef USE_PORT_SELECTION
	"-L [string]	Local Unix socket to connect",
	"-R [int]	Remote port number to call",
#endif
	"-Q [string]	Specify queue to examine (or \"all\")",
	"-H [string]	Specify remote host to query (or \"all\")",
	"-V		Print version information and exit",
	"",
	"$Revision: 4.0 $",
	NULL
};

Bool	hflag = FALSE;				/* User specified a host */
int	s;

UIDTYPE	myuid;
GIDTYPE	mygid;

extern char *	optarg;				/* Ptr to option argument */
extern int	optopt;				/* Current option */ 
extern int	optind;				/* Argument idx */
extern int	opterr;				/* Error handling by getopt */


static SIGRTN	cleanup P((int));

int
main(argc, argv)
int	argc;
char **	argv;
{

	char *	cp;
	char *	host = NULL;			/* Host of current interest */
	char *	defhost = NULL;			/* Host of current interest */
	char *	queue = NULL;			/* Queue of current interest */
	char *	qu;				/* Canonical queue name */
	char	bbuf[BUFFER];
	int	status;				/* Return status */
	int	option;				/* Option character */


	if (signal(SIGHUP, SIG_IGN) != SIG_IGN)
		(void) signal(SIGHUP, cleanup);

	if (signal(SIGINT, SIG_IGN) != SIG_IGN)
		(void) signal(SIGINT, cleanup);

	if (signal(SIGTERM, SIG_IGN) != SIG_IGN)
		(void) signal(SIGTERM, cleanup);

	progname = argv[0];			/* Assign the program name */

	myuid = getuid();
	mygid = getgid();

	if (gethostname(lhost, MAXHOSTNAMELEN) < 0) {
		(void) fprintf(stderr, "gethostname fails\n");
		exit(1);
	}

	/*
	 * Check first for a "BATCH" environment variable -
	 * backwards compatable, but clashes with Splus driver envvar.
	 */
	if ((cp = getenv("BATCH")) != NULL)
		queue = estrdup(cp);
	else
		queue = estrdup(_DEF_QUEUE);

	/*
	 * Check second for a "BATCHQUEUE" environment variable.
	 * If there is one, use it.
	 */
	if ((cp = getenv("BATCHQUEUE")) != NULL) {
		free(queue);
		queue = estrdup(cp);
	}

	/*
	 * Always start with the default host.
	 */
	defhost = _DEF_HOST;				

	/*
	 * ... modifiy if the default host is also the localhost.
	 */
	if (islocalhost(defhost) == 0)
		defhost = lhost;
	
	/*
	 * Parse options
	 */
	while ((option = getopt(argc, argv, "hL:R:H:Q:V")) != EOF)
		switch((char)option) {
		case 'h':			/* Help */
			usage(use);
			exit(1);
#ifdef USE_PORT_SELECTION
		case 'L':			/* Local port or socket */
			socketname = optarg;
			break;
		case 'R':			/* Remote port */
			remote_port = atoi(optarg);
			break;
#endif
		case 'Q':			/* Queue name */
			queue = optarg;
			break;
		case 'H':			/* Host name */
			host = optarg;
			break;
		case 'V':
			dump_version(printsf, 0);
			exit(0);
			break;
		case '?':			/* Unknown option */
			if (opterr == 0)
				blog(BLOG_WARNING,
				     "internal: bad flag \"-%c\"", optopt);
			usage(use);
			exit(1);
			break;
		default:			/* Unhandled option */
			blog(BLOG_WARNING,
			     "internal: no action for \"-%c\" option",
			     (char) option);
			usage(use);
			exit(1);
			break;

		}


	/*
	 * Queue may be queue@host - if so, break it apart
	 */
	if ((cp = strchr(queue, '@')) != NULL) {
		*cp++ = '\0';
		host = cp;
	}

	/* If host is not set, use the default host */
	if (!host)
		host = defhost;

	/*
	 * Take care of using queue name aliases...
	 * Look for the specified queue name
	 */
	if ((status = bgetent(bbuf, queue)) < 0) {
		blog(BLOG_FATAL, "cannot open queue description file");
		cleanup(0);
	} else if (status == 0) {
		blog(BLOG_WARNING, enotq, queue);
		return (-1);
	}

	/*
	 * Reset the queue name to the canonical (first) name
	 */
	qu = bnamget(bbuf);

	/*
	 * Parse the given queue and host
	 */
	(void) host_query(host, QC_INFO, qu);

	exit(0);
}

/*
 * Client function for host_query(): If the QC_INFO comes back with OK,
 * then parse and print the folowing lines as queue information.
 */
int
reply_ok(fp, fmt, rqueue)
FILE *	fp;
char *	fmt;
char *	rqueue;
{

	char	recvbuf[BUFSIZ];

	while (fgets(recvbuf, sizeof(recvbuf), fp) != NULL) {
		(void) fputs(recvbuf, stdout);
	}

	return 0;
}


int
reply_mt(fp, fmt, rqueue)
FILE *	fp;
char *	fmt;
char *	rqueue;
{
	return 0;
}



/*
 * Clean up on interruption.
 */
static SIGRTN
cleanup(sig)
int	sig;
{


	(void) close(s);
	exit(1);
}


