#ifndef lint
static char rcsid[] =
	"$Id: baq.c,v 4.0 1993/09/17 00:41:04 srp Exp srp $";
#endif

/*
 *
 * Author: Scott R. Presnell
 *
 * Copyright (C) 1989, 1990, The Regents of the University of California
 *                     All Rights Reserved
 *
 * THE REGENTS OF THE UNIVERSITY OF CALIFORNIA MAKE NO REPRESENTATION OR
 * WARRANTIES WITH RESPECT TO THE CONTENTS HEREOF AND SPECIFICALLY DISCLAIM
 * ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
 * PURPOSE.
 *
 * Further, the Regents of the University of California reserve the right to
 * revise this software and/or documentation and to make changes from time
 * to time in the content hereof without obligation of the Regents of the
 * University of California to notify any person of such revision or change.
 *
 */

/*
 * Baq: batch queue examination program.
 *
 * baq [-l] [-H host] [-Q queue[@host] ] <not now  [user...] [job...]>
 *
 * -l long output
 * -H specify host (all - specify all hosts known to batch)
 * -Q specify queue (all - specify all queues known to batch)
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <errno.h>
#include <fcntl.h>

#include <signal.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/time.h>

#include "config.h"

#include H_STRING

#include "batch.h"
#include "pathnames.h"
#include "protocol.h"
#include "reply.h"

static char *	use[] = {
	"baq -hlv [-Q queue] [-H host]",
	"\"Baq\" returns the state of a batch job queue on a specified host.",
	"",
	"Options:",
	"-l		localhost only",
	"-v		Verbose output",
	"-h		This help message",
#ifdef USE_PORT_SELECTION
	"-L [string]	Local Unix socket to connect",
	"-R [int]	Remote port number to call",
#endif
	"-Q [string]	Specify queue to examine (or \"all\")",
	"-H [string]	Specify remote host to query (or \"all\")",
	"-V		Print version information and exit",
	"",
	"$Revision: 4.0 $",
	NULL
};

Bool	dflag = FALSE;
Bool	lflag = FALSE;				/* Localhost only option */
Bool	vflag = FALSE;				/* Verbose output option */
int	s;

UIDTYPE	myuid;
GIDTYPE	mygid;

static int	request_parse P((char *queue, char *host, Bool isuserhost));
static int	queue_print P((FILE *fp, char *fmt, char *queue));

extern char    *optarg;				/* Ptr to option argument */
extern int	optopt;				/* Current option */ 
extern int	optind;				/* Argument idx */
extern int	opterr;				/* Error handling by getopt */


static SIGRTN	cleanup P((int));


int
main(argc, argv)
int	argc;
char **	argv;
{

	char *	cp;
	char *	host = NULL;			/* Host of current interest */
	char *	defhost = NULL;			/* Default host to query */
	char *	queue = NULL;			/* Queue of current interest */
	int	option;				/* Option character */


	if (signal(SIGHUP, SIG_IGN) != SIG_IGN)
		(void) signal(SIGHUP, cleanup);

	if (signal(SIGINT, SIG_IGN) != SIG_IGN)
		(void) signal(SIGINT, cleanup);

	if (signal(SIGTERM, SIG_IGN) != SIG_IGN)
		(void) signal(SIGTERM, cleanup);

	progname = argv[0];			/* Assign the program name */

	myuid = getuid();
	mygid = getgid();

	if (gethostname(lhost, MAXHOSTNAMELEN) < 0) {
		(void) fprintf(stderr, "gethostname fails\n");
		exit(1);
	}

	/*
	 * Check first for a "BATCH" environment variable -
	 * backwards compatable, but clashes with Splus driver envvar.
	 */
	if ((cp = getenv("BATCH")) != NULL)
		queue = estrdup(cp);
	else
		queue = estrdup(_DEF_QUEUE);

	/*
	 * Check second for a "BATCHQUEUE" environment variable.
	 * If there is one, use it.
	 */
	if ((cp = getenv("BATCHQUEUE")) != NULL) {
		free(queue);
		queue = estrdup(cp);
	}

	/*
	 * Always start with the default host.
	 */
	defhost = _DEF_HOST;				

	/*
	 * ... modifiy if the default host is also the localhost.
	 */
	if (islocalhost(defhost) == 0)
		defhost = lhost;
	
	/*
	 * Parse options
	 */
	while ((option = getopt(argc, argv, "dhlvL:R:H:Q:V")) != EOF)
		switch((char) option) {
#ifdef DEBUG
		case 'd':
			dflag = TRUE;
			break;
#endif
		case 'h':			/* Help */
			usage(use);
			exit(1);
		case 'l':
			lflag = TRUE;
			break;
		case 'v':			/* verbose output */
			vflag = TRUE;
			break;
#ifdef USE_PORT_SELECTION
		case 'L':			/* Local port or socket */
			socketname = optarg;
			break;
		case 'R':			/* Remote port */
			remote_port = atoi(optarg);
			break;
#endif
		case 'Q':			/* Queue name */
			queue = optarg;
			break;
		case 'H':			/* Host name */
			host = optarg;
			break;
		case 'V':
			dump_version(printsf, 0);
			exit(0);
			break;
		case '?':			/* Unknown option */
			if (opterr == 0)
				blog(BLOG_WARNING,
				     "internal: bad flag \"-%c\"", optopt);
			usage(use);
			exit(1);
			break;
		default:			/* Unhandled option */
			blog(BLOG_WARNING,
			     "internal: no action for \"-%c\" option",
			     (char) option);
			usage(use);
			exit(1);
			break;

		}


	/*
	 * Queue may be queue@host - if so, break it apart
	 */
	if ((cp = strchr(queue, '@')) != NULL) {
		*cp++ = '\0';
		host = cp;
	}

	/* If host is not set, use the default host */
	if (!host)
		host = defhost;

	/*
	 * Parse the given queue and host
	 */
	(void) request_parse(queue, host, (host == defhost) ? FALSE : TRUE);

	exit(0);

	/*NOTREACHED*/
}


/*
 * Check the queue and/or host requests for "all" and other special
 * circumstances (load balancing).  Break into single requests and 
 * pass to the function host_query(), which< calls the (possibly remote) host.
 */
static int
request_parse(queue, host, isuserhost)
char *	queue;
char *	host;
Bool	isuserhost;
{

	char *	ho;				/* Next host */
	char *	qu;				/* Next queue */
	char *	rm, *rq;			/* Remote host and queue */
	char	bbuf[BUFFER];
	char	pbuf[BUFFER];
	char *	bp;
	int	status;				/* Return status */
	Bool	loadb;				/* Load balance? */


	if (strcasecmp(queue, all) == 0) {		/* All queues */
		if (strcasecmp(host, all) == 0 && !lflag) {	/* All hosts */
			while ((ho = getbhent()) != NULL) {
				(void) printf("%s:\n", ho); /* Print host */
				while ((qu = getquent()) != NULL) {
					(void) host_query(ho, QC_DISP, qu);
					(void) fputc('\n', stdout);
				}
				endbaent();
				(void) fputc('\n', stdout);
			}
			/* endbhent(); */
		} else {				/* Just all queues */
			while ((qu = getquent()) != NULL) {
				(void) host_query(host, QC_DISP, qu);
				(void) fputc('\n', stdout);
			}
			endbaent();
		}

	} else {					/*  Single queue */

		/*
		 * Take care of using queue name aliases...
		 * Look for the specified queue name
		 */
		if ((status = bgetent(bbuf, queue)) < 0) {
			blog(BLOG_FATAL, "cannot open queue description file");
			cleanup(0);
		} else if (status == 0) {
			blog(BLOG_WARNING, enotq, queue);
			return (-1);
		}

		/*
		 * Reset the queue name to the canonical (first) name
		 */
		qu = bnamget(bbuf);

		loadb = bgetflag("lb");

		bp = pbuf;

		rm = bgetstr("rm", &bp);	/* Get remote host */
		rq = bgetstr("rq", &bp);	/* Get remtoe queue */

		/*
		 * User specified a single host
		 */
		if (isuserhost == TRUE && strcasecmp(host, all)) {

			(void) host_query(host, QC_DISP, qu);

		/*
		 * If we are to examine all hosts, or we load balance
		 */
		} else if ((strcasecmp(host, all) == 0 || loadb == TRUE) &&
			   !lflag) {

			while ((ho = getbhent()) != NULL) {
				(void) printf("%s:\n", ho);
				(void) host_query(ho, QC_DISP, qu);
				(void) fputc('\n', stdout);
			}
			/* endbhent(); */

		/*
		 * else, if we have a remote queue...
		 */
		} else if (rm != NULL && strcasecmp(rm, lhost) != 0) {

			(void) printf("%s:\n", lhost);
			(void) host_query(lhost, QC_DISP, qu);

			(void) printf("\n%s:\n", rm);
			(void) host_query(rm, QC_DISP, rq);

		/*
		 * else, just our default host...
		 */
	        } else {
			(void) host_query(host, QC_DISP, qu);
		}
	}

	return 0;

}


/*
 * Client function for host_query(): If the QC_DISP comes back with OK,
 * Then get the first line (status), and then parse and print the folowing
 * lines as batch jobs in the queue.
 */
int
reply_ok(fp, fmt, rqueue)
FILE *	fp;
char *	fmt;
char *	rqueue;
{

	char	recvbuf[BUFSIZ];


	if (fgets(recvbuf, sizeof(recvbuf), fp) != NULL) {
#ifdef printqueue
		recvbuf[strlen(recvbuf) - 1] = '\0';
		(void) printf("queue %s: %s\n", queue, recvbuf);
#else
		(void) fputs(recvbuf, stdout);
#endif
	}

	(void) queue_print(fp, fmt, rqueue);
	return 0;

}


/*
 * Client function for host_query(): If the QC_DISP comes back with MT,
 * We still report the status line.
 */
int
reply_mt(fp, fmt, rqueue)
FILE *	fp;
char *	fmt;
char *	rqueue;
{

	char	recvbuf[BUFSIZ];


	if (fgets(recvbuf, sizeof(recvbuf), fp) != NULL) {
#ifdef printqueue
		recvbuf[strlen(recvbuf) - 1] = '\0';
		(void) printf("queue %s: %s\n", queue, recvbuf);
#else
		(void) fputs(recvbuf, stdout);
#endif
	}


	if (fmt != NULL)
		(void) printf(fmt, rqueue);

	return 0;
}


#define SHORT_SFORMAT	"%-9.9s %-20.20s %-12.12s %-8.8s %-12.12s\n"
#define SHORT_NFORMAT	"%s%-8d %-20.20s %-12.12s %-8ld %-12.12s\n"
#define LONG_NFORMAT	"ID:  %-08d%s\tJob: %-12s\tFor: %s\nPid: %-8d\tCPU Time: %-8ld\tState: %s\nQueued: %s\nStart:  %s\nDelay:  %s %s\n\n"



char *states[] = {
	"queued",  "running",
	"queued",  "suspended",
	"queued",  "interrupted",
	"queued",  "timeout",
	"queued",  "stopped",
	"held",    "attached",
	"pending", "sending",
	"delayed", "running"
};


/*
 * Retreive and print in human readable format the information on the
 * jobs in the queue.
 */
static int
queue_print(fp, fmt, queue)
FILE *	fp;
char *	fmt;
char *	queue;
{

	char *	uptr;
	char *	status;
	char	name[80], user[80], delayer[80];
	char	queue_string[80];
	char	start_string[80];
	char	delay_string[80];
	char	recvbuf[BUFSIZ];
	int	queue_state;
	int	pid, restart, jid;
	long	job_state;
	time_t	st_ctime, st_stime, st_dtime, cpu_time;

	/*
	 * Print out the header information
	 */
	if (vflag == FALSE)
		(void) printf(SHORT_SFORMAT,
			      "ID", "Job", "Owner", "Time", "State");
	
	/*
	 * Translate and print individual job information from the daemon
	 */
	while (fgets(recvbuf, sizeof(recvbuf), fp) != NULL) {
		(void) sscanf(recvbuf, fmt,
			      &jid, name, user, &job_state,
			      &st_ctime, &st_stime, delayer, &cpu_time,
			      &pid, &queue_state);

#ifdef DEBUG
		if (dflag) {
			printf("JobID:    %d\n", jid);
			printf("JobName:  \"%s\"\n", name);
			printf("UserName: \"%s\"\n", user);
			printf("JobState: %lo\n", job_state);
			printf("QueueState:%d\n", queue_state);
			printf("ctime:    %ld\n", st_ctime);
			printf("stime:    %ld\n", st_stime);
			printf("dtime:    %s\n",  delayer);
			printf("CPUtime:  %ld\n", cpu_time);
			printf("Pid:      %d\n", pid);
		}
#endif

		/*
		 * Cut off the address part of the name for the short disp
		 */
		if (!vflag)
			if ((uptr = strchr(user, '@')) != NULL)
				*uptr = '\0';

		/*
		 * Copy the ctimes into temporary buffers so that we can
		 * nuke the returned string at pos 16.
		 */
		if (st_ctime > 0) {
			(void) strcpy(queue_string, ctime(&st_ctime));
			queue_string[strlen(queue_string)-1] = '\0';
		} else
			(void) strcpy(queue_string, "before epoch?");

		if (st_stime > 0) {
			if (!(job_state & JOB_RUNNING)) {
				(void) strcpy(start_string, "not yet");
			} else {
				(void) strcpy(start_string, ctime(&st_stime));
				start_string[strlen(start_string)-1] = '\0';
			}
		} else {
			(void) strcpy(start_string, "not yet");
		}

		if (delayer && strchr(delayer, '/') != NULL) {
			/* got a file */
			strcpy(delay_string, delayer);
		} else if (delayer && strchr(delayer, '%') != NULL) {
			/* got a job */
			strcpy(delay_string, delayer);
		} else if (delayer && (st_dtime = atol(delayer)) > 0) {
			(void) strcpy(delay_string, ctime(&st_dtime));
			delay_string[strlen(delay_string)-1] = '\0';
		} else {
			(void) strcpy(delay_string, "no delay");
		}

		if (queue_state == 0 || !(job_state & JOB_RUNNING)) {
			if (job_state & (JOB_ATTACH | JOB_ATTACHED))
				queue_state = 5;
			else if (job_state & (JOB_SEND | JOB_AGAIN))
				queue_state = 6;
			else if (job_state & JOB_DELAY)
				queue_state = 7;
			/* else retain old meaning */
		}

		if (job_state & JOB_RUNNING)
			status = states[(queue_state * 2)+1];
		else
			status = states[(queue_state * 2)];

		/* Mark the topqed jobs with a asterisk */
		if (vflag) {
			(void) printf(LONG_NFORMAT,
				      jid, 
				      ((job_state & JOB_TOPQ) ? "*" : " "),
				      name, user,
				      pid,
				      cpu_time,
				      status,
				      queue_string,
				      start_string,
				      delay_string,
				      ((job_state & JOB_RESTART) ? "\n\t\t This job is the result of a restart" : ""));
		} else {
			(void) printf(SHORT_NFORMAT,
				      ((job_state & JOB_TOPQ) ? "*" : " "),
				      jid, name, user, cpu_time, status);
		}
	}

	return 0;
	
}


/*
 * Clean up on interruption.
 */
static SIGRTN
cleanup(sig)
int	sig;
{


	(void) close(s);
	exit(1);
}


