#ifndef lint
static char rcsid[] =
	"$Id: batch.c,v 4.0 1993/09/17 00:41:04 srp Exp srp $";
#endif

/*
 *
 * Author: Scott R. Presnell
 *
 * Copyright (C) 1989, 1990, The Regents of the University of California
 *                     All Rights Reserved
 *
 * THE REGENTS OF THE UNIVERSITY OF CALIFORNIA MAKE NO REPRESENTATION OR
 * WARRANTIES WITH RESPECT TO THE CONTENTS HEREOF AND SPECIFICALLY DISCLAIM
 * ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
 * PURPOSE.
 *
 * Further, the Regents of the University of California reserve the right to
 * revise this software and/or documentation and to make changes from time
 * to time in the content hereof without obligation of the Regents of the
 * University of California to notify any person of such revision or change.
 *
 */


/*
 *	Batch - submit a command or commands to the batch processor by 
 * creating an apropriately named file in the batch queue spool area.
 * This command must run set uid root and set gid daemon to write the
 * appropriate directories.
 *
 */

/*
 * $Log: batch.c,v $
 * Revision 4.0  1993/09/17  00:41:04  srp
 * Re-munged code for control by Configure script.  Added setsid() calls,
 * changed fork() calls for error handling, finer nice() priority control,
 * write functionality.  Cleaned up code definitions and declarations for
 * version 4.0 release.  Updated documentation, man pages, test scripts.
 *
 * Revision 3.5  1993/05/31  23:20:33  srp
 * Added broadcast, multicast, interface selection, new logging facility,
 * host <-> address cache, job renice, user settable soft limits.  Changed
 * all structures to bidirectional linked list.  Job attachment and queue
 * "drain" implemented.  Queues are locked during critical modification
 * (scanning, job removal).  Jobs are sorted as specified by batchcap -
 * user information dumped periodically.  Reworked prototype macros. SIGTERM
 * will terminate running jobs.  "Kick" message does not cause fork().
 * Console usage and console user detection decoupled.  Machine specific code
 * broken out into separate file.
 *
 * Revision 3.4  1992/01/25  18:22:40  srp
 * Intermediate checkin: added forward function declarations, major
 * changes for an event scheduler, ability to dump state file,
 * moved lock and status file manipulations to new files, inline
 * allocation of system structures.  Clean up code some more.
 *
 * Revision 3.3.1.2  1991/04/06  20:35:24  srp
 * patch2: Addition of working directory translation (for systems with auto-
 * patch2: mounters). Buffer renames. User name in Mail subject line.
 *
 * Revision 3.3.1.1  1991/04/06  16:23:31  srp
 * patch1: Fix local port assignment problems, and missing ioctl.h include.
 *
 * Revision 3.3  90/11/19  20:50:45  srp
 * The package is now configurable based on a determination of the system
 * capabilities.  Signal handlers now execute minimal code.  Fixed some
 * race conditions in rmjob.  Switched over to incremental logging levels.
 * Added pathanmes header file.  Removed extraneous ifdefs.
 * 
 * Revision 3.2  90/09/14  16:50:13  srp
 * 1) Fixed up some of the last bugs in sending and receiving jobs.
 * 2) Added the code to handle the new timing options from batch.
 * 3) Did the rough documentation in the man pages.
 * 4) Basically a running, Beta test version of the code.
 * 
 * Revision 3.1  90/07/20  21:36:32  srp
 * Checkin of patches 1 through 4 for batch3.0. 
 * Patch 1: 1 -> 2. Priority: HIGH.  Fixes: 1) Memory leak in recovery
 * of structures after sending a job.  2) Bad interaction between
 * malloc of job data and stdio - resolved by buffer allocation for stdio.
 * Patch 2: 2 -> 3. Priority: MEDIUM. Fixes: 1) Reread the queue names
 * if the batchcap file has changed.  2) Better timing of getwd calls.
 * Patch 3: 3 -> 4. Priority: MEDIUM.  Fixes: 1) batch daemon reads swap.
 * Patch 4: 4 -> 5. Priority: HIGH. Fixes: 1) file descriptors left open.
 * 
 * Revision 3.0  90/05/21  22:29:48  srp
 * More debugging in resource checking and load balancing.  Added
 * patchlevel.h.  Final checkin before second major release to the world.
 * 
 * Revision 2.8  90/05/12  16:33:58  srp
 * Finished implementing remote queueing, load balancing, and system
 * limit checks.  Performed as much testing of these functions as
 * reasonably possible.  Added defines for some functionalities.
 * Rewrote select() calls to use fd_set variables.  Fixed nickname
 * problem for baq and barm.  Jobs are now referred to by number.
 * 
 * Revision 2.7  90/04/02  08:07:27  srp
 * Added load balancing and cleaned up job transfer some.
 * This is a checkpoint version.
 * 
 * Revision 2.6  90/03/31  16:12:31  srp
 * Added capability to transfer jobs between machines, and to have
 * remote queues.  Also added data files (df) files to hold the shell
 * script.  Control files hold information about the job (who, whereform
 * return mailaddress, etc.)  Also reinstated status files and their
 * examination in baq.  This is a preliminary version.
 * 
 * Revision 2.5  90/03/16  20:18:21  srp
 * Made man pages and header files configurable from Makefile.
 * Added linked lists to structures, malloc'ed space as needed.
 * Added "correct" varargs code, and standardized error messages
 * in the client programs.
 * Enhanced the return mail service for finished jobs.
 * Added ancillary files README, Installation, TODO for release.
 * 
 * Revision 1.1  89/04/07  07:25:56  srp
 * Initial revision
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <ctype.h>

#include <errno.h>
#include <fcntl.h>

#if defined(MALLOC_TRACE) || defined(MALLOC_LEAK)
#include "dmalloc.h"
#else
#include <malloc.h>
#endif

#include <grp.h>
#include <pwd.h>
#include <signal.h>
#include <time.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/dir.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <sys/resource.h>

#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>

#include "config.h"

#include H_STRING

#if defined(USE_ATTACH) && defined(PROCFS)
#include <sys/syscall.h>
#include <sys/procfs.h>
#endif

#include "batch.h"
#include "pathnames.h"
#include "shells.h"
#include "protocol.h"
#include "reply.h"

#include "pm.h"

static char *	use[] = {
	"batch -chqlkrv [-a date] [-H host] [-Q queue] [command|stdin]",
	"\"Batch\" submits a command to be executed at a lower priority.",
	"",
	"Options:",
	"-a [string]	Start the job after the following time",
	"-c		exec the command that follows.",
	"-h		This help message.",
	"-k		Combine stderr and stdout of the job",
	"-l		Job stays on localhost (even under load balancing)",
	"-q		Don\'t print out the job id.",
	"-f [int]	Force to run a specific CPU",
#ifdef USE_CLASSES
	"-C [string]	Specify class of job",
#endif
	"-H [string]	Specify host on which to run the job.",
	"-Q [string]	Specify queue under which to run the job.",
	"-J [string]	Symbolic job name.",
	"-o [string]	File for stdout",
	"-e [string]	File for stderr",
	"-m (string)	Send mail under the follwing conditions:",
	"	s	Start of the job",
	"	t	Transfer of the job to another machine",
	"	e	End of the job",
	"	c	Crash of the job",
	"	p	Pause of job (graphics|console use, etc)",
	"	r	Resource overrun of the job",
	"	n	None, override batchd defaults, don't send any mail",
	"-M [string]	Send mail to this address (or \"no\" for no mail)",
	"-n		What if...(ala make)",
#ifdef USE_ATTACH
	"-p [pid ...]	Attach process <pid> to the batch daemon", 
#endif
	"-r (string)	User settable soft resource limits",
	"	p	CPU time",
	"	f	File size",
	"	d	Data size",
	"	s	Stack size",
	"	c	Coredump size",
	"	r	Resident set size",
	"	o	Open files",
	"-S		This job is to be restarted after a crash",
	"-s [shell]	Specify a shell to use (must be from /etc/shells)",
	"-v		Verbose - tell about the job submission", 
	"-V		Print version information and exit",
	"-w (string)	Write the user under the follwing conditions:",
	"	s	Start of the job",
	"	t	Transfer of the job to another machine",
	"	e	End of the job",
	"	c	Crash of the job",
	"	p	Pause of job (graphics|console use, etc)",
	"	r	Resource overrun of the job",
	"	n	None, override batchd defaults, don't send any mail",
	"-W [string]	Write the user at this address (or \"no\" for no write)",
	"-x		Export the environment",
	"-X		Do NOT export the environment",
	"",
	"$Revision: 4.0 $",
	NULL
};

#define OPTIONS		"chklnqvxDXSKa:e:f:J:m:o:p:r:s:w:C:L:R:H:Q:M:W:V"

/*
 * Syntax notes:
 *
 * The output and error file names have the syntax:
 *
 * [machine:][[/]path/]filename
 *
 * If machine is not specified then it is the local machine.
 * If path is not specified it is the orignal working directory,
 * or if the root is not specified it is from the working directory.
 *
 */

static char *	cfname;			/* Names of files to be created */
static char *	dfname;
static char *	tfname;

static char *	jname;			/* JobName of the job */
static char *	cname;			/* Class of the job */

static char *	maddress;		/* Mail address */
static char *	waddress;		/* Write address */
static char *	efile;			/* File for stderr */
static char *	ofile;			/* File for stdout */
static char *	delayer;		/* Keep the start date */
static char *	fdelay;			/* Name of file to track for delay. */
static char *	jdelay;			/* Job spec to track for delay. */
static char *	shell;			/* Name of shell */
static int	processor = -1;		/* The CPU to run on */

char *	sd, *rm, *rq, *wd, *ma, *wa, *lf;
char	 upath[MAXPATHLEN];		/* User space path */

char *	bhost;				/* The host we're making a bid to */
int	jid = 0;			/* Numerical job identifier */
int	pid = 0;			/* Process identifier (attach) */
int	s;				/* Socket */

static time_t	delaysec = -1;		/* Delay time in seconds */

Bool	cflag = FALSE,			/* exec the commands in the shell */
	kflag = FALSE,			/* Combine stderr and stdout of job */
	lflag = FALSE,			/* user want's job local */
	nflag = FALSE,			/* Just tell us */
	maddflag = FALSE,		/* add to mail opts if true */
	waddflag = FALSE,		/* add to write opts if true */
	qflag = FALSE,			/* Shaad-up! */
	Sflag = FALSE,			/* Job is to be restarted after crash */
	Kflag = FALSE,			/* Keep cf job file as kf file */
	xflag = TRUE,			/* Export environment */
	Dflag = FALSE,			/* Debug the script - not advertized */
	vflag = FALSE;			/* Verbose */

static Bool	loadb;

static Notify	mail_opts = -1;
static Notify	write_opts = -1;

static Bool	user_resources = TRUE;

static struct passwd pw;
static struct group gr;

UIDTYPE	myuid;
GIDTYPE	mygid;

#ifdef HAVE_RLIMIT
static struct rlimit	rlimits[RLIM_NLIMITS];
static char *	rlimidx = "pfdscrom";
#endif

static int	file_check_access P((char *));
static int	mkbatchtemps P((char *));
static char *	splicetmp P((char *, char *, int, int));
static int	card P((FILE *, int, char *));
static int	make_control P((char *, char *, char *, Bool));
static int	make_data P((char *, int, char **));
static void	dump_sh_env P((FILE *, char *, char *));
static void	dump_csh_env P((FILE *, char *, char *));
static int	file_check_access P((char *));
static int	assign_credentials P((int, int));

extern int	check_group P((char *name, UIDTYPE, GIDTYPE, char *queue));

extern time_t 	parse_date P((char *, time_t *));

#ifdef HAVE_RLIMIT
extern int 	atorlimit P((struct rlimit [], char *));
extern int	rlimittoc P((int));
#endif

static void	what_if P((char *, char *, Bool));

extern char *	optarg;				/* Ptr to option argument */
extern int	optopt;				/* Current option */ 
extern int	optind;				/* Argument idx */
extern int	opterr;				/* Error handling by getopt */

extern char **	environ;

static SIGRTN	cleanup P((int));


main(argc, argv)
int	argc;
char **	argv;
{

	char *	cp;			
	char *	host = NULL;			/* Host of current interest */
	char *	defhost = NULL;			/* Default host to query */
	char *	queue = NULL;			/* Queue of current interest */
	char	bbuf[BUFFER];
	char	pbuf[BUFFER];
	char *	bp;
	int	option;				/* Option character */
	int	status;				/* Return status */
	int	flag;				/* Flag status (could be -1) */
	int	i;

	struct hostent *	hp;


	if (signal(SIGHUP, SIG_IGN) != SIG_IGN)
		(void) signal(SIGHUP, cleanup);

	if (signal(SIGINT, SIG_IGN) != SIG_IGN)
		(void) signal(SIGINT, cleanup);

	if (signal(SIGTERM, SIG_IGN) != SIG_IGN)
		(void) signal(SIGTERM, cleanup);

	progname = argv[0];			/* Assign the program name */

	myuid = getuid();
	mygid = getgid();

	if (gethostname(lhost, MAXHOSTNAMELEN) < 0) {
		(void) fprintf(stderr, "gethostname fails\n");
		exit(1);
	}

	/*
	 * Check first for a "BATCH" environment variable -
	 * backwards compatable, but clashes with Splus driver envvar.
	 */
	if ((cp = getenv("BATCH")) != NULL)
		queue = estrdup(cp);
	else
		queue = estrdup(_DEF_QUEUE);

	/*
	 * Check second for a "BATCHQUEUE" environment variable.
	 * If there is one, use it.
	 */
	if ((cp = getenv("BATCHQUEUE")) != NULL) {
		free(queue);
		queue = estrdup(cp);
	}

	/*
	 * Always start with the default host...
	 */
	defhost = _DEF_HOST;

	/*
	 * ... modifiy if the default host is also the localhost.
	 */
	if (islocalhost(defhost) == 0)
		defhost = lhost;
	
	while ((option = getopt(argc, argv, OPTIONS)) != EOF)
		switch ((char) option) {
		/*
		 * If / in the argument - then waiting on a file.
		 * If % in the argument - then waiting on another process.
		 * <jobid>%queue@host.
		 * 	else waiting on a date.
		 */
		case 'a':
			if (*optarg == '/') {
				fdelay = optarg;
			} else if (strchr(optarg, '%') != NULL) {
				jdelay = optarg;
			} else {
				delayer = optarg;
				delaysec = parse_date(optarg, (time_t *) NULL);
				if (delaysec < 0) {
					blog(BLOG_WARNING,
					     "could not parse \"%s\"", optarg);
					delayer = NULL;
				}
			}
			break;
		case 'D':
			Dflag = TRUE;
			break;
		case 'c':
			cflag = TRUE;
			break;
		case 'e':			/* Set up an error file */
			efile = optarg;

#ifndef USE_SEND_REMOTE
			if (strchr(efile, ':') != NULL)
				blog(BLOG_WARNING, "output must remain local");
#endif
			break;
		case 'f':
			if ((processor = atoi(optarg)) < 0) {
				blog(BLOG_WARNING, "Bad CPU: %s", optarg);
			}
			break;
		case 'h':			/* Help */
			usage(use);
			exit(1);
			break;
		case 'C':
			cname = optarg;
			break;
		case 'J':			/* Give the job a name */
			jname = optarg;
			break;
		case 'l':			/* Job will stay local */
			lflag = TRUE;
			break;
		case 'k':
			kflag = TRUE;
			break;
		case 'm':
			mail_opts = setnotify(optarg);
			if (*optarg == '+')
				maddflag = TRUE;
			break;
		case 'w':
			write_opts = setnotify(optarg);
			if (*optarg == '+')
				waddflag = TRUE;
			break;
		case 'n':			/* What if... */
			nflag = TRUE;
			break;
		case 'o':			/* Set up an output file */
			ofile = optarg;
#ifndef USE_SEND_REMOTE
			if (strchr(ofile, ':') != NULL)
				blog(BLOG_WARNING, "output must remain local");
#endif

			break;

		case 'p':
#ifdef USE_ATTACH
			pid = atoi(optarg);
#else
			blog(BLOG_WARNING, "attach option not available");
#endif
			break;
		case 'q' :			/* Shut up */
			qflag = TRUE;
			break;
#ifdef HAVE_RLIMIT
		case 'r':
			(void) atorlimit(rlimits, optarg);
			break;
#endif
		case 's':
			shell = optarg;
			break;
		case 'x':
			xflag = TRUE;
			break;
		case 'X':
			xflag = FALSE;
			break;
		case 'K':			/* Keep job cf as kf */
			Kflag = TRUE;
			break;
		case 'S':			/* Job restarts on crash */
			Sflag = TRUE;
			break;
		case 'v':			/* Be verbose */
			vflag = TRUE;
			break;
#ifdef USE_PORT_SELECTION
		case 'L':			/* Local port or socket */
			socketname = optarg;
			break;
		case 'R':			/* Remote port */
			remote_port = atoi(optarg);
			break;
#endif
		case 'M':			/* Mail address */
			maddress = optarg;
			break;
		case 'W':			/* Write address */
			waddress = optarg;
			break;
		case 'H':			/* Host name */
			host = optarg;
			break;
		case 'Q' :			/* Queue name */
			queue = optarg;
			break;
		case 'V':
			dump_version(printsf, 0);
			exit(0);
			break;
		case '?':			/* Unknown option */
			if (opterr == 0)
				blog(BLOG_WARNING,
				     "internal: bad flag \"-%c\"", optopt);
			usage(use);
			exit(1);
			break;
		default:			/* Unhandled option */
			blog(BLOG_WARNING,
			     "internal: no action for \"-%c\" option",
			     (char) option);
			usage(use);
			exit(1);
			break;
		}

	/*
	 * Queue may be queue@host - if so, break it apart
	 */
	if ((cp = strchr(queue, '@')) != NULL) {
		*cp++ = '\0';
		host = cp;
	}

	/* If host is not set, use the default host */
	if (!host)
		host = defhost;

	if (!pid && ofile != NULL && file_check_access(ofile) < 0) {
		blog(BLOG_FATAL, eopen, ofile);
		if (!nflag)
			cleanup(0);
	}

	if (!pid && efile != NULL && file_check_access(efile) < 0) {
		blog(BLOG_FATAL, eopen, efile);
		if (!nflag)
			cleanup(0);
	}

	/*
	 * If the user didn't specify a job name, and there is no pid for
	 * attachment, get it from the command line.
	 */
	if (!pid && jname == NULL) {
		if (argv[optind] == NULL) {
			jname = "stdin";
		} else {
			/*
			 * Remove the leading path component.
			 */
			if ((cp = strrchr(argv[optind], '/')) != NULL)
				jname = ++cp;
			else
				jname = argv[optind];
		}
	}


	/*
	 * Get the named queue into a buffer
	 */
	if ((status = bgetent(bbuf, queue)) < 0) {
		blog(BLOG_FATAL, "cannot open queue description file");
		cleanup(0);
	} else if (status == 0) {
		blog(BLOG_ERR, enotq, queue);
		exit(1);
	}

	/*
	 * Reset the queue name to the canonical (first) name
	 */
	queue = bnamget(bbuf);

	/*
	 * The following bits of code check what will happen to the
	 * ob after it is submitted (printed out if the user so specifies).
	 *
	 * ALL JOBS ARE FIRST SUBMITTED TO THE LOCAL QUEUE
	 */
	loadb = bgetflag("lb");

	bp = pbuf;

	/* 
	 * Check to see wether we shuld consider using the users
	 * requested soft resource limits.
	 * Some administrators may want to turn this off.
	 */
	if ((flag = bgetflag("us")) < 0)
		user_resources = TRUE;			/* default is TRUE */
	else
		user_resources = flag;			/* set to result */

#ifdef HAVE_RLIMIT
	for (i = 0; i < RLIM_NLIMITS; i++) {
		if (rlimits[i].rlim_cur != 0 && user_resources == FALSE) {
			blog(BLOG_WARNING, ques, queue,
			    "user resources are disabled");
			break;
		}
	}
#endif

	/*
	 * Get the default mailing information...
	 */
	ma = bgetstr("ma", &bp);

	/*
	 * ...and use it if the user has not specified otherwise.
	 *
	 * If maddflag was set the user specified to add mail options rather
	 * than set them absolutely.
	 */
	if (maddflag == TRUE && ma != NULL) {
		mail_opts |= setnotify(ma);
	} else if (mail_opts == -1) {
		if (ma != NULL) {
			mail_opts = setnotify(ma);
		} else {
			mail_opts = 0;
		}
	}


	/*
	 * Get the default write information...
	 */
	wa = bgetstr("wa", &bp);

	/*
	 * ...and use it if the user has not specified otherwise.
	 *
	 * If waddflag was set the user specified to add write options rather
	 * than set them absolutely.
	 */
	if (waddflag == TRUE && wa != NULL) {
		write_opts |= setnotify(wa);
	} else if (write_opts == -1) {
		if (wa != NULL) {
			write_opts = setnotify(wa);
		} else {
			write_opts = 0;
		}
	}

	/*
	 * We must get the spooling directory so we can create a path to lf
	 */
	if ((sd = bgetstr("sd", &bp)) == NULL)
		sd = _DEF_SPOOL;

	/*
	 * Lock file
	 */
	if ((lf = bgetstr("lf", &bp)) == NULL)
		lf = _DEF_LOCK;

	/*
	 * Map a machine space path to a user space path.
	 * using pathmap and the configuration file in wd.
	 * If wd == NULL, use the default.
	 */
	if ((wd = bgetstr("wd", &bp)) != NULL)
		setpmfile(wd);

	if (getcwd(home, MAXPATHLEN) == NULL) {
		blog(BLOG_FATAL, "getcwd fails: %m");
		if (!nflag) cleanup(0);
	} else if (vflag) {
		(void) printf("current dir is : \"%s\"\n", home);
	}

	(void) pathmap(upath, home);
	if (vflag)
		(void) printf("cwd: \"%s\" maps to \"%s\"\n", home, upath);

#ifdef USE_ATTACH

	/*
	 * We're running as suid root, and maybe sgid daemon, so we need
	 * to get the *real* uid to handle security on the request.
	 */
	if (pid && islocalhost(host) != 0) {
		blog(BLOG_FATAL, "you can only attach processes on the localhost");
		if (!nflag) cleanup(0);
	}

	if (pid && pid != getpgrp(pid)) {
		blog(BLOG_FATAL, "pid %d is not a process group leader", pid);
		if (!nflag) cleanup(0);
	}

#ifdef PROCFS
	if (pid) {

		int	fd;
		prcred_t	pcred;

		(void) sprintf(line, "%s/%05d", PROCFS, pid);
		if ((fd = open(line, O_RDONLY)) < 0) {
			blog(BLOG_FATAL,
			    "cannot open pid %d for read \"%s\"\n",
			    pid, line);
			if (!nflag) cleanup(0);
		}

		if (ioctl(fd, PIOCCRED, &pcred) < 0) {
			blog(BLOG_FATAL, "cannot get credentials for pid: %d",
			    pid);
			if (!nflag) cleanup(0);
		}
	
		if ((myuid == 0 && pcred.pr_ruid != 0) ||
		    (myuid == pcred.pr_ruid)) {
			(void) assign_credentials(pcred.pr_ruid, pcred.pr_rgid);
		} else {
			blog(BLOG_FATAL, "permission denied for pid: %d",
			    pid);
			if (!nflag) cleanup(0);
		}

		(void) close (fd);
	} else
#endif	/* PROCFS */
#endif	/* USE_ATTACH */

	(void) assign_credentials(myuid, mygid);


	if ((hp = gethostbyname(host)) == NULL) {
		blog(BLOG_FATAL, eaddr, host);
		if (!nflag) cleanup(0);
	}

	/*
	 * If the user does not have the group access - we're outta here
	 */
	if (check_group(pw.pw_name, myuid, mygid, queue) < 0) {
		(void) printf(ques, queue,
			      "not a member of the appropriate group(s)\n");
		exit(0);
	}

	/*
	 * Check to make sure queuing is enabled if useruid is not root
	 */
	(void) sprintf(line, "%s/%s", sd, lf);
	if (stat(line, &stbuf) < 0) {
		blog(BLOG_FATAL, estat, line);
		cleanup(0);
	}

	/* Mail address for results */
	if (maddress == NULL) {
#ifdef UCSF
		if ((cp = getenv("REPLY_TO")) != NULL)
			(void) strcpy(line, cp);
		else
#endif
		(void) sprintf(line, "%s@%s", pw.pw_name, lhost);
		maddress = estrdup(line);
	}

	/* Maybe this user does not want mail. */
	if (maddress && (strcasecmp(maddress, no) == 0 ||
			 strcasecmp(maddress, "none") == 0))
		maddress = NULL;

	/* Write address for results */
	if (waddress == NULL) {
		(void) sprintf(line, "%s@%s", pw.pw_name, lhost);
		waddress = estrdup(line);
	}

	/* Maybe this user does not want writes. */
	if (waddress &&
	    (strcasecmp(waddress, no) == 0 ||
	     strcasecmp(waddress, "none") == 0))
		waddress = NULL;
	    

	if (nflag == FALSE && pw.pw_uid != 0 && (stbuf.st_mode & 0001)) {
		(void) printf("queue %s: is draining, try again later.\n",
		      queue);
		exit (0);
	}

	if (nflag == FALSE && pw.pw_uid != 0 && (stbuf.st_mode & 0010)) {
		(void) printf("queue %s: is disabled, try again later.\n",
		      queue);
		exit (0);
	}

	/*
	 * Initialize the job files
	 */
	jid = mkbatchtemps(sd);

	/*
	 * Set up the control file
	 */
	(void) make_control(tfname, host, queue,
			    (host != defhost ||
			     (host == defhost && defhost != lhost))
			      ? TRUE : FALSE);

	/*
	 * Set up the command script in the data file
	 */
	(void) make_data(dfname, argc, argv);

	if (nflag == FALSE && chmod(dfname, DFMOD) < 0) {
		blog(BLOG_FATAL, echmo, dfname);
		cleanup(0);
	}

	/*
	 * Copy the temp command file to the real command file
	 */
	if (nflag == FALSE && link(tfname, cfname) < 0) {
		blog(BLOG_FATAL, "couldn't rename %s", tfname);
		cleanup(0);
	}

	if (nflag == FALSE && chmod(cfname, CFMOD) < 0) {
		blog(BLOG_FATAL, echmo, cfname);
		cleanup(0);
	}

	
	(void) unlink(tfname);

	if (nflag == FALSE) {
		if (host_query(lhost, QC_KICK, queue) != 0) {
			(void) printf("Job cannot be submitted (files removed).\n");
			cleanup(0);
		}
	} else {
		what_if(host, queue,
			(host != defhost ||
			 (host == defhost && defhost != lhost))
			? TRUE : FALSE);
	}

	exit(0);

	/*NOTREACHED*/
}


/*
 * Make the temporary file names.  Return the numerical job identifier.
 */
static int
mkbatchtemps(spooldir)
char *	spooldir;
{

	int	n, oldn, len;

#ifdef USE_SEQFILE

	int	c, fd;
	char *	cp;

	(void) sprintf(line, "%s/.seq", sd);

	if ((fd = open(line, (O_RDWR|O_CREAT), 00640)) < 0) {
		blog(BLOG_FATAL, eopen, line);
		cleanup(0);
	}

#ifdef HAVE_LOCKF
	if (lockf(fd, F_LOCK, 0) < 0) {
#else
#ifdef HAVE_FLOCK
	if (flock(fd, LOCK_EX) < 0) {
#endif /* HAVE_FLOCK */
#endif /* HAVE_LOCKF */

		blog(BLOG_FATAL, "cannot lock %s", line);
		cleanup(0);
	}

	n = 0;

	if ((len = read(fd, line, sizeof(line))) > 0) {
		for (cp = line; len--; ) {
			if (*cp < '0' || *cp > '9')
				break;
			n = n * 10 + (*cp++ - '0');
		}
	}

#else
	/* Just use the process ID */
	n = getpid();

#endif

	if (vflag == TRUE || nflag == TRUE)
		(void) printf("The job id %s be: %d\n",
			      nflag == TRUE ? "would" : "will", n);
	
	len = strlen(spooldir) + strlen(lhost) + 20;
	cfname = splicetmp(spooldir, "cf", n, len);
	dfname = splicetmp(spooldir, "df", n, len);
	tfname = splicetmp(spooldir, "tf", n, len);

	oldn = n;

#ifdef USE_SEQFILE

	/* Write out the new sequence number. */
	n = (n + 1) % 1000;
	(void) lseek(fd, 0L, 0);
	(void) sprintf(line, "%03d\n", n);
	(void) write(fd, line, strlen(line));
	(void) close(fd);	/* unlocks as well */

#endif

	return (oldn);
}


/*
 * Make a single temporary file name.
 */
static char *
splicetmp(spooldir, id, num, len)
char *	spooldir;
char *	id;
int	num, len;
{

	char *	filename, *thp;
	char	host[MAXHOSTNAMELEN];

	if ((filename = emalloc(len)) == NULL)
		cleanup(0);

	(void) strcpy(host, lhost);

	/* Use just the short name of the host */
	if ((thp = strchr(host, '.')) != NULL)
		*thp = '\0';

/*	(void) sprintf(s, "%s/%sA%03d%s", SD, id, num, host);*/
	(void) sprintf(filename, "%s/%s%d%s", spooldir, id, num, host);

	return(filename);
}


/*
 * Make the control file fo the job.
 */
static int
make_control(file_name, rhost, rqueue, isremotehost)
char *	file_name;
char *	rhost;
char *	rqueue;
Bool	isremotehost;
{
	
	FILE *	tfp;
	int	i;


	if (nflag == FALSE && (tfp = fopen(file_name, "w+")) == NULL) {
		blog(BLOG_FATAL, eopen, file_name);
		cleanup(0);
	}

	if (vflag == TRUE)
		(void) printf("Command file:\n\n");

	/*
	 * a =	Start *a*t this time, or after this file, or after this job.
	 * 	if leading / then file.
	 *	if num%queue@host then when this job completes.
	 *	if num then after this time.
	 * c =	*c*reate time of the original job.
	 * D =	Starting (current) working *D*irectory 
	 * f =	*f*orce to run on a specific CPU
	 * R =	Send the job to the *R*emote host.
	 * H =	This job started on *H*ost.
	 * Q =	This job is destined for *Q*ueue.
	 * F =	Finished with these return values - written in batchd.
	 * G =	Users *G*roup.
	 * g =	Users numerical *g*roup id.
	 * K =	Job was *K*ill()'ed with this signal - written in batchd.
	 * U =	User *u*ser name submitted this job.
	 * u =	Users numerical *u*ser id.
	 * m =	*m*ail flags/arguments.
	 * M =	*M*ail address.
	 * O =	*O*utput (stdout) file.
	 * E =	*E*rror (stderr) file.
	 * C =	*C*ombine stdout and stderr
	 * L =	*L*ocal job - don't use load balancing.
	 * N =	*N*ew nice value of the job - written in batchd.
	 * J =	*J*ob name.
	 * P =	Job *P*id assigned - written in batchd.
	 * T =	*T*ransfer event - written in batchd. 
	 * S =	re*S*tart this job should it crash.
	 * r =	*R*esource Limits desired by the user.
	 *	c = CPU time
	 *	f = file size
	 *	d = data size
	 *	s = stack size
	 *	c = core size
	 *	r = resident image size
	 *	o = # open files
	 *	m = memory mapped
	 * j =	*j*ob id.
	 * p =	*p*rocess id of job to attach.
	 * w =  *w*rite flags/arguments.
	 * W =  *W*rite address.
	 * Z =	Keep cf file as kf file after job has finished.
	 */

	/* Current time */
	(void) sprintf(line, "%ld", (time_t) time((time_t *) 0));
	(void) card(tfp, 'c', line);

	/* Current directory */
	(void) card(tfp, 'D', upath);

	/* Required Cards: */

	/* Initial host */
	(void) card(tfp, 'H', lhost);

	/* Queue */
	(void) card(tfp, 'Q', rqueue);

	/* Person queued the job */
	(void) card(tfp, 'U', pw.pw_name);

	/* Uid of that person */
	(void) sprintf(line, "%u", pw.pw_uid);
	(void) card(tfp, 'u', line);

	/* Group membership at the time of queuing */
	(void) card(tfp, 'G', gr.gr_name);

	/* Gid at the time of queueing */
	(void) sprintf(line, "%u", gr.gr_gid);
	(void) card(tfp, 'g', line);

	/* Job id of this job */
	(void) sprintf(line, "%d", jid);
	(void) card(tfp, 'j', line);


	/* Optional Cards: */


	/* Delayed start time */
	if (fdelay) {
		/* Delayed start on file */
		(void) card(tfp, 'a', fdelay);
	} else if (jdelay) {
		/* Delayed start on job */
		(void) card(tfp, 'a', jdelay);
	} else if (delaysec >= 0) {
		(void) sprintf(line, "%ld", (time_t) delaysec);
		(void) card(tfp, 'a', line);
	}

	/* Processor to run on */
	if (processor >= 0) {
		(void) sprintf(line, "%d", processor);
		(void) card(tfp, 'f', line);
	}

	/* Job name */
	if (jname != NULL)
		(void) card(tfp, 'J', jname);

	/* Mail options and address */
	if (maddress != NULL) {
		(void) sprintf(line, "%d", (int) mail_opts);
		(void) card(tfp, 'm', line);
		(void) card(tfp, 'M', maddress);
	}

	/* Write options and address */
	if (waddress != NULL) {
		(void) sprintf(line, "%d", (int) write_opts);
		(void) card(tfp, 'w', line);
		(void) card(tfp, 'W', waddress);
	}

	/*
	 * Remote host for job:
	 * If the user specifies a host, or the user doesn't specify a job and 
	 * defhost is not the localhost, we submit an R record.
	 *
	 * On the daemon:
	 * If there is an R record:
	 *	Then the job goes to that host.
	 *	If R is the localhost, then the job stays on the localhost.
	 *
	 * If there is not an R record:
	 *	If the queue is load balancing then the job goes out,
	 *	Otherwise it stays on the localhost.
	 *
	 */
	if (isremotehost == TRUE || lflag == TRUE)
		(void) card(tfp, 'R', rhost);
	
	/* File for stdout */
	if (ofile != NULL)
		(void) card(tfp, 'O', ofile);

	/* File for stderr */
	if (efile != NULL)
		(void) card(tfp, 'E', efile);

	/* If one is set and kflag is given, signal it to the daemon. */
	if ((ofile || efile) && kflag)
		(void) card(tfp, 'C', yes);

	/* Process id of this job */
	if (pid != 0) {
		(void) sprintf(line, "%d", pid);
		(void) card(tfp, 'p', line);
	}

#ifdef HAVE_RLIMIT
	for (i = 0; i < RLIM_NLIMITS; i++) {
		if (user_resources && rlimits[i].rlim_cur != 0) {
			(void) sprintf(line, "%c%d",
				       rlimittoc(i), rlimits[i].rlim_cur);
			(void) card(tfp, 'r', line);
		}
	}
#endif

	/* Restart the job upon crash? */
	if (Sflag == TRUE)
		(void) card(tfp, 'S', yes);

	/* Keep cf file as kf file? */
	if (Kflag == TRUE)
		(void) card(tfp, 'Z', yes);

	if (nflag == FALSE)
		(void) fclose(tfp);

	return 0;
}


/*
 * Write a "card" to the control file.
 */
static int
card(fp, c, val)
FILE *	fp;
int	c;
char *	val;
{

	if (nflag == FALSE)
		(void) fprintf(fp, "%c%s\n", c, val);

	if (vflag == TRUE) 
		(void) printf("Ident:%c\tValue:%s\n", c, val);

	return 0;
}


/*
 * Make the data file for the job.
 */
static int
make_data(file_name, ac, av)
char *	file_name;
int	ac;
char **	av;
{

	int	c;			/* Character reader */
	int	scanned;		/* Scanned vars */
	char **	envp;			/* Environment pointer */
	char *	s1;			/* Ptr used in inserting qoute */
	char *	s2;			/* in environment variable names */
	char	name[BUFFER];		/* Environment variable name  */
	char	value[BUFFER];		/* Environment variable value */
	char	value2[BUFFER];		/* Environment value after escape */
	char *	ptr;			/* Ptr to tail of shell */
	char *	valid;			/* Valid shell pointer */
	Bool	sh;			/* Shell type - TRUE == sh like */
	FILE *	dfp;


	if (nflag == FALSE && (dfp = fopen(file_name, "w+")) == NULL) {
		blog(BLOG_FATAL, eopen, file_name);
		cleanup(0);
	}

	if (shell == NULL) {
		if ((shell = getenv("SHELL")) == NULL) {
			blog(BLOG_FATAL, "no shell!?");
			cleanup(0);
		}
	}

	/*
	 * Get environment variables for command file.  Scanf is used to
	 * seperate values from names in order to place quotes around the
	 * values so that those containing spaces will be handled
	 * correctly by 'sh'
	 */
	if ((ptr = strrchr(shell, '/')) == NULL || *shell != '/') {
		blog(BLOG_FATAL, "shell must be specified as a full path (%s)",
		    shell);
		cleanup(0);
	} else {
		ptr++;
	}

	do {
		if ((valid = getusershell()) == NULL) {
			blog(BLOG_FATAL, "%s is not sanctioned as a shell",
			    shell);
			cleanup(0);
		}
	} while (strcmp(shell, valid) != 0);

	endusershell();

	if (strcmp(ptr, "sh") == 0 || strcmp(ptr, "ksh") == 0 ||
	    strcmp(ptr, "bash") == 0) {
		sh = TRUE;
	} else if (strcmp(ptr, "csh") == 0 || strcmp(ptr, "tcsh") == 0) {
		sh = FALSE;
	} else {
		blog(BLOG_FATAL, "unknown shell: %s", shell);
		cleanup(0);
	}

	if (vflag == TRUE)
		(void) printf("\nData file:\n\n");

	if (vflag == TRUE) {
		(void) printf("Comand Shell:\t\t%s\n", shell);
		(void) printf("Umask setting:\t\t%o\n", umask(0));
		(void) printf("Working directory:\t%s\n", upath);
	}

	if (nflag == FALSE) {
		(void) fprintf(dfp, "#!%s%s\n", shell, Dflag ? " -xv" : "");
		(void) fprintf(dfp, "umask %o\n", umask(0));
		(void) fprintf(dfp, "cd %s\n", upath);
	}


	if (vflag == TRUE && xflag == TRUE)
		(void) printf("\nEnvironment:\n\n");

	if (xflag == TRUE) {
		for (envp = environ; *envp != NULL; envp++) {
			if ((scanned = sscanf(*envp, "%[^=]=%[^\n]", name, value)) == 0) {
				blog(BLOG_FATAL, "bad sscanf %s", *envp);
				cleanup(0);
			}

			/*
			 * Check for occurence of qoute (') in environment
			 * values and replace them with '\'' in order that
			 * sh does not get confused when executing job.
			 */
			if (scanned == 2) {
				s1 = value;
				s2 = value2;
				while (*s2++ = *s1) {	/* Assignment */
					if (*s1++ == '\'') {
						*s2++ = '\\';
						*s2++ = '\'';
						*s2++ = '\'';
					}
				}
			} else {
				value2[0] = NULL;
			}

			/*
			 * This should be more flexible - to allow arbitrary
			 * shells but we've got to know how to set up the
			 * environment. sh, ksh, csh, tcsh are probably
			 * the most common shells.
			 */
			if (sh == TRUE) {
				if (vflag == TRUE)
					dump_sh_env(stdout, name, value2);
				if (nflag == FALSE)
					dump_sh_env(dfp, name, value2);
			} else {
				if (vflag == TRUE)
					dump_csh_env(stdout, name, value2);
				if (nflag == FALSE)
					dump_csh_env(dfp, name, value2);
			}
		}
	}

	/*
	 * Place user's commands in command file
	 */
	if (!pid && av[optind] == NULL && nflag == FALSE) {

		if (vflag == TRUE)
			(void) printf("\nCommands:\n\n");

		while ((c = getchar()) != EOF) {
			if (vflag == TRUE)
				(void) putc(c, stdout);
			(void) putc(c, dfp);
		}

	} else if (av[optind] != NULL) {

		if (vflag == TRUE)
			(void) printf("\nCommands:\n\n");

		if (vflag == TRUE && cflag == TRUE)
			(void) printf("exec ");

		if (nflag == FALSE && cflag == TRUE)
			(void) fprintf(dfp, "exec ");

		for (; av[optind]; ++optind) {
			if (vflag == TRUE)
				(void) printf("%s ", av[optind]);
			if (nflag == FALSE)
				(void) fprintf(dfp, "%s ", av[optind]);
		}
		if (vflag == TRUE)
			(void) printf("\n\n");
		if (nflag == FALSE)
			(void) fprintf(dfp, "\n");
	}

	/*
	 * Make the control shell wait for all its jobs
	 */
	if (nflag == FALSE) {
		(void) fprintf(dfp, "wait\n");

		(void) fclose(dfp);
	}

	return 0;
}

static void
dump_sh_env(fp, u, t)
FILE *	fp;
char *	u, *t;
{

	if (t == NULL)
		(void) fprintf(fp, "%s\n", u);
	else
		(void) fprintf(fp, "%s='%s'\n", u, t);
			
	(void) fprintf(fp, "export %s\n", u);
}

static void
dump_csh_env(fp, u, t)
FILE *	fp;
char *	u, *t;
{

	if (t == NULL)
		(void) fprintf(fp, "setenv %s\n", u);
	else
		(void) fprintf(fp, "setenv %s '%s'\n", u, t);
}


static SIGRTN
cleanup(sig)
int	sig;
{

	extern int errno;
	int	orig = errno;		/* This is to preserve the errno */
					/* during the unlink calls */

	if (cfname)
		(void) unlink(cfname);
	if (tfname)
		(void) unlink(tfname);
	if (dfname)
		(void) unlink(dfname);

	errno = orig;

	(void) close(s);

	exit(1);
}


/*
 * Print what would happen if the job was submitted right now.
 */
static void
what_if(rhost, rqueue, isremotehost)
char *	rhost;
char *	rqueue;
Bool	isremotehost;
{

	(void) printf("The job would be submitted to the %s queue on %s...\n",
		      rqueue, lhost);

	if (rm != NULL) {
		(void) printf("\t...where it would have been sent to %s to execute.\n", rm);
	} else if (isremotehost == FALSE && lflag == FALSE && loadb == TRUE) {
		(void) printf("\t...which is a load balancing queue: the job would have been sent\n");
		(void) printf("\tto the lowest loaded machine from there.\n");
	} else if (isremotehost == TRUE) {
		(void) printf("\t...where it would be sent to %s per your command line request.\n", rhost);
	} else {
		(void) printf("\t...where it would be executed in turn.\n");
	}

	/* Delayed start time */
	if (fdelay) {
		/* Delayed start on file */
		(void) printf("The job will start after file \"%s\" is created\n",
			      delayer);
	} else if (jdelay) {
		/* Delayed start on job */
		(void) printf("The job will start after job \"%s\" is finishes\n",
			      delayer);
	} else if (delaysec >= 0) {
		(void) strcpy(line, ctime(&delaysec));
		line[19] = '\0';
		(void) printf("The job will start at or after \"%s\" (%s)\n",
			      delayer, line);
	}

	if (jname != NULL)
		(void) printf("The name of the job would be \"%s\".\n", jname);

	if (maddress != NULL) {
		(void) printf("Mail will be sent to %s.\t", maddress);

		if (mail_opts) {
			(void) printf("(mail flags set: \"%s%s%s%s%s%s\")\n",
				      (mail_opts & NOTIFY_START) ? "s" : "",
				      (mail_opts & NOTIFY_END) ?   "e" : "",
				      (mail_opts & NOTIFY_CRASH) ? "c" : "",
				      (mail_opts & NOTIFY_DEBUG) ? "d" : "",
				      (mail_opts & NOTIFY_PAUSE) ? "p" : "",
				      (mail_opts & NOTIFY_TRANSFER) ? "t" : "",
				      (mail_opts & NOTIFY_RESOURCE) ? "r" : "");
		} else {
			(void) printf("(no mail flags set)\n");
		}
	} else {
		(void) printf("Mail would not be sent to the user.\n");
	}


	if (waddress != NULL) {
		(void) printf("Tty writes will be made to %s.\t", waddress);

		if (write_opts) {
			(void) printf("(write flags set:\"%s%s%s%s%s%s\")\n",
				      (write_opts & NOTIFY_START) ?  "s" : "",
				      (write_opts & NOTIFY_END) ?    "e" : "",
				      (write_opts & NOTIFY_CRASH) ?  "c" : "",
				      (write_opts & NOTIFY_DEBUG) ?  "d" : "",
				      (write_opts & NOTIFY_PAUSE) ?  "p" : "",
				      (write_opts & NOTIFY_TRANSFER) ?  "t" : "",
				      (write_opts & NOTIFY_RESOURCE) ?  "r" : "");
		} else {
			(void) printf("(no write flags set)\n");
		}
	} else  {
		(void) printf("Tty writes would not be made to the user.\n");
	}


	if (ofile != NULL) 
		(void) printf("Output would be placed in %s %s.\n",
			      ofile,
			      (strchr(ofile, ':') == NULL) ?
			      "in real time" :
			      "after the job is done");

	if (efile != NULL) 
		(void) printf("Output would be placed in %s %s.\n",
			      efile,
			      (strchr(efile, ':') == NULL) ?
			      "in real time" :
			      "after the job is done");

	(void) printf("The job %s restart after a machine crash%s.\n",
		      Sflag == TRUE ? "would" : "would not", 
		      Sflag == TRUE ? " if the queue allows" : "");

	

	return ;
}

/* 
 * Client function for the host_query call:
 * If QC_KICK returns OK, we report that the job has been queued.
 */
int
reply_ok(fp, fmt, rqueue)
FILE *	fp;
char *	fmt;
char *	rqueue;
{


	if (!qflag)
		(void) printf("Job %d has been submitted to the %s queue.\n",
			      jid, rqueue);

	return 0;
}


/*
 * Client function for the host_query call:
 * If QC_KICK returns MT, we weren't able to queue the job.
 */
int
reply_mt(fp, fmt, rqueue)
FILE *	fp;
char *	fmt;
char *	rqueue;
{


	if (!qflag) {
		(void) printf("Job %d has been submitted to the %s queue\n",
			      jid, rqueue);
		(void) printf("but cannot be entered into the queue.\n");
	}

	return (-1);
}


/*
 * Return the ability to write the file specified.
 * Access uses uid instead of euid - Great!
 */
static int
file_check_access(thisfile)
char *	thisfile;
{

	char *	fptr, temp_file[LINE];
	extern int errno;


	errno = 0;

	if (stat(thisfile, &stbuf) >= 0) {

		/*
		 * File is a directory
		 */
		if ((stbuf.st_mode & S_IFMT) == S_IFDIR) {
			errno = EISDIR;
			return (-1);
		}
	}

	(void) strcpy(temp_file, thisfile);

	/*
	 * File does not exist - scan the path before it.
	 */
	if (errno == ENOENT) {
		if ((fptr = strrchr(temp_file, '/')) == NULL) {
			(void) strcpy(temp_file, ".");
		} else {
			*fptr = '\0';
		}
	}

	/*
	 * Check the file, or the path before it.
	 */
	if (access(temp_file, 02) < 0) {
		return (-1);
	} else {
		return 0;
	}

}


static int
assign_credentials(uid, gid)
int	uid;
int	gid;
{

	struct passwd *	pass;			/* User password structure */
	struct group *	grp;			/* User group structure */

	if ((pass = getpwuid(uid)) == NULL) {
		blog(BLOG_FATAL, epid, uid);
		cleanup(0);
	} else {
		pw.pw_name = estrdup(pass->pw_name);
		pw.pw_uid = pass->pw_uid;
		pw.pw_gid = pass->pw_gid;
	}

	if ((grp = getgrgid(gid)) == NULL) {
		blog(BLOG_FATAL, egid, gid);
		cleanup(0);
	} else {
		gr.gr_name = estrdup(grp->gr_name);
		gr.gr_gid = grp->gr_gid;
	}

	return 0;
}
