/*
 * $Id: batch.h,v 4.0 1993/09/17 00:41:04 srp Exp srp $
 */

/*
 *
 * Author: Scott R. Presnell
 *
 * Copyright (C) 1989, 1990, The Regents of the University of California
 *                     All Rights Reserved
 *
 * THE REGENTS OF THE UNIVERSITY OF CALIFORNIA MAKE NO REPRESENTATION OR
 * WARRANTIES WITH RESPECT TO THE CONTENTS HEREOF AND SPECIFICALLY DISCLAIM
 * ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
 * PURPOSE.
 *
 * Further, the Regents of the University of California reserve the right to
 * revise this software and/or documentation and to make changes from time
 * to time in the content hereof without obligation of the Regents of the
 * University of California to notify any person of such revision or change.
 *
 */

/*
 * Global definitions for the batch job control system.
 * The following conditional compilation options are rarely used, 
 * and should probably not be altered except for debugging purposes.
 */

/*
 * USE_REMOTE_DELETE:
 * If this is defined, barm, the job removal program will look to
 * all hosts considered in load balancing to try to remove a jobs
 * specified by a user name.  Similiarly, for "remote" queues barm
 * will look to the localhost *and* the remote host.  This makes it work
 * similarly to lprm.
 *
 * This could be deemed a mis-feature so I've made it configurable. 
 * If this is not defined, the user must specify a host other than the
 * default host to remove remote jobs.  This makes it work *not* like lprm. 
 */
/*#undef	USE_REMOTE_DELETE	/**/
 
/* 
 * USE_RRESVPORT:
 * This should be defined if you want to use rresvport() to obtain
 * reserved ports when calling a foriegn host. This is used as part
 * of the host authentication procedure.  Recommended.
 */
#define USE_RRESVPORT		/**/

/*
 * USE_ATTACH should be defined if you want to attach already running jobs
 * to the daemon - this is experimental, and only supported under irix4,5
 * with PROCFS defined.  The code in the daemon  is always there,
 * this define enables or disables the code in the batch client.
 */
#undef	USE_ATTACH		/**/

/*
 * USE_LOCALBUF:
 * SGI/SYSV stdio seems to grab more space than it ever needs, giving
 * the program have a voracious appetite for memory.This sets up local
 * buffers for stdio.  
 */
#define USE_LOCALBUF		/**/

/*
 * USE_LIBMALLOC:
 * Use the malloc library so that we can examine the memory stats.
 * Please check to see that -lmalloc is specified in the make file
 * libraries macro if you set this define.  And check the man page
 * to see that the mallinfo structure actually exists in malloc.h.
 */
#undef USE_LIBMALLOC		/**/

/*
 * HAVE_FLAT_NAMES:
 * If your name space is flat, that is, all the uid's line up across
 * the different machines on which you will run batch, you define this
 * conditional.  Numerical User and group information will be taken as
 * authoritative: j_uid and j_gid are accepted from the control file.
 * If the same uid maps to different users on different machines, this
 * conditional should not be defined: j_uid and j_gid are calculated
 * from the string/ascii user name and group in the control file.
 */
#define HAVE_FLAT_NAMES		/**/

/*
 * Group with permissions to run bac.
 */
#define OPERATOR	"operator"

/*
 * Macro for determining Unix domain socket length.
 */
#ifndef SUN_LEN
#define SUN_LEN(su) \
	(sizeof(*(su)) - sizeof((su)->sun_path) + strlen((su)->sun_path))
#endif

/* Absolute value */
#ifndef ABS
# define ABS(x)		((x) < 0.0 ? -(x) : (x))
#endif

#define TRUE	1
#define FALSE	0

/*
 * When files are created in the spooling area, they are normally
 *   readable only by their owner and the spooling group.  If you
 *   want otherwise, change this mode.  In the particular case of the
 *   data files (with a "df" prefix), they must also be executable.
 *
 * Leave the execute bit on the CF files off, as this is how batchd
 *   identifies jobs that are to be sent to the top of the queue.
 */
#define CFMOD		00640
#define DFMOD		00740
#define OFMOD		00640

#define MAXPWNAMLEN	80	/* Maximum length name for a login name
				   and mail address!!! */

#define MAXQUEUENAMELEN	16	/* maximum queue name length */

#define	BUFFER		1024	/* Multipurpose buffer size */
#define	LINE		128	/* Multipurpose line size */

/*
 * These constants contain the following number of seconds, minutes, hours.
 */
#define SECPERMIN	60
#define	SECPERHOUR	3600
#define SECPERDAY	86400

#define MINPERHOUR	60
#define	MINPERDAY	1440

#define HOURPERDAY	24

typedef unsigned int	Bool;
typedef int		Notify;

#define	BLOG_FATAL	0000001	/* critical conditions */
#define	BLOG_ERR	0000002	/* error conditions */
#define	BLOG_WARNING	0000004	/* warning conditions */
#define	BLOG_USER	0000010	/* debug-level messages */
#define	BLOG_INFO	0000020	/* informational */
#define	BLOG_STATS	0000040	/* normal but significant condition */
#define BLOG_TRACE	0000100	/* trace protocol messages */
#define BLOG_DEBUG	0000200	/* debugging messages */
#define BLOG_ALL	0000177 /* All error messages */

/* Under what conditions do we want to send off some mail */
#define NOTIFY_NONE	0000
#define	NOTIFY_START	0001
#define	NOTIFY_END	0002
#define	NOTIFY_CRASH	0004
#define NOTIFY_TRANSFER	0010
#define NOTIFY_RESOURCE	0020
#define NOTIFY_PAUSE	0040
#define NOTIFY_DEBUG	1000

/* Job state [Run|Send|Delay|Again],Restart */
#define	JOB_NULL	0000000			/* No status */
#define JOB_RUN		0000001			/* Job is to be run */
#define JOB_SEND	0000002			/* Job is to be sent */
#define JOB_DELAY	0000004			/* Job is a delayed start */
#define	JOB_RESTART	0000010			/* Job is to be restarted */
#define JOB_LOCAL	0000020			/* Job is to remain local */
#define JOB_AGAIN	0000040			/* Job is to be sent again */
#define	JOB_ATTACH	0000100			/* Job is to be attached */
#define	JOB_TOPQ	0000200			/* Job has been rescheduled */
#define	JOB_RUNNING	0001000			/* Job is running */
#define	JOB_ATTACHED	0002000			/* Job has been atttached */
#define	JOB_DELAYER	0010000			/* Job is waiting for me */
#define	JOB_KEEP	0020000			/* Keep cf file as kf file */

typedef struct	rc_t	Rc;

struct rc_t {
	char *	r_user;
	char *	r_name;
	time_t	r_mtime;
};

/*
 * These should be defined and set in each "main" program
 */
extern char *	progname;		/* Program name */

extern int	remote_port;		/* Remote port in host byte order */
extern int	local_port;		/* Local port in host byte order */
extern Bool	iamdaemon;		/* Identifier */

/*
 * Socketname is used in getport and batchd - it is defined in getport
 * because the batch, baq, bac, and barm programs use getport too.
 */
extern char *	socketname;		/* Socket connection name */

extern char	lhost[MAXHOSTNAMELEN];	/* Local hostname */
extern char	shost[MAXHOSTNAMELEN];	/* Short hostname */

extern char	line[LINE];		/* Multipurpose buffers */
extern char	line2[LINE];
extern char	home[MAXPATHLEN];

#ifdef USE_LOCALBUF
extern char	iobuf[BUFSIZ];
extern char	rciobuf[BUFSIZ];
#endif

extern struct stat	stbuf;

extern char *	localhost;		/* _DEF_HOST */
extern char *	batchhosts;		/* BATCHHOSTS */


extern char *	service;
extern char *	tcp;
extern char *	udp;

extern char *	yes;
extern char *	no;
extern char *	all;
extern char *	unk;

extern int errno, sys_nerr;	/* Current error, max error */
extern char *sys_errlist[];	/* Array of error strings corresp. to errno */

extern int	h_errno;	/* Some Unixes do not extern this for us. */


/*
 * Global error messages associated with port reading and writing
 */
extern char esock[];		/* "cannot create socket" */
extern char eserv[];		/* "cannot find service %s" */

/*
 * Global protocol error messages
 */
extern char elost[];		/* "lost connection" */
extern char ewrit[];		/* "bad write on %d for %s" */
extern char erecv[];		/* "recieved unknown message" */
extern char enotq[];		/* "queue %s: unknown" */

/*
 * System error messages
 */
extern char elock[];		/* "cannot lock %s" */
extern char efork[];		/* "cannot fork %s" */
extern char eopen[];		/* "cannot open/create %s" */
extern char efdop[];		/* "cannot fdopen on %s" */
extern char estat[];		/* "cannot stat %s" */
extern char eunlk[];		/* "cannot unlink %s" */
extern char echmo[];		/* "cannot chmod %s" */
extern char echow[];		/* "cannot chown %s" */
extern char egetd[];		/* "cannot get current working directory" */
extern char echad[];		/* "cannot chdir to %s" */
extern char epent[];		/* "cannot find passwd entry for %s" */
extern char epid[];
extern char egent[];		/* "cannot find groups entry for %s" */
extern char egid[];
extern char edup[];		/* "cannot dup to descriptor %d" */
extern char ekill[];		/* "cannot kill queue %s: pid:%d, kill:%d" */
extern char eaddr[];		/* "cannot obtain address for host %s" */

extern char ques[];		/* "%s: %s" */
extern char quesr[];		/* "%s: %s (%m) */

extern char quej[];		/* "%s: job %d: %s" */
extern char quejr[];		/* "%s: job %d: %s (%m) */

/*
 * These functions that take variable arguments.
 */
extern void 	blog V((int, char *, ...));
extern void	printsf V((int, char *, ...));

#if !defined(MALLOC_TRACE) && !defined(MALLOC_LEAK)
extern voidp 	emalloc P((unsigned int));
extern voidp 	ecalloc P((unsigned int, unsigned int));
/*extern voidp	erealloc P((voidp, unsigned int));*/
extern char *	estrdup P((char *));
extern void 	efree P((voidp));
#endif

extern char *	bnamget P((char *));
extern int 	bgetflag P((char *));
extern int 	bgetnum P((char *));
extern char *	bgetstr P((char *, char **));

extern int 	islocalhost P((char *));

extern int 	getbaent P((char *));
extern int 	endbaent P((void));
extern int 	bgetent P((char *, char *));
extern int 	getsocket P((char *, char *, char *, u_long *, u_short *));

extern char *	getbhent P((void));
extern void 	endbhent P((void));
extern char *	getrcent P((char *));
extern char *	scanrcline P((char *, char *, char *, char *));
extern void 	endrcent P((void));
extern char *	getquent P((void));
extern void 	endquent P((void));
extern char *	prttime P((time_t));
extern time_t	getmtime P((char *));
extern Notify 	setnotify P((char *));
extern int 	getrand P((int, int));

extern int 	ewrite P((int, voidp, size_t, char *));
extern int 	ewriteone P((int, int));

extern void 	usage P((char *[]));

extern void 	dump_version P((void (*)V((int, char *,...)), int));

extern int 	gethostname P((char *, int));

extern int 	host_query P((char *, int, char *));

