#ifndef lint
static char rcsid[] =
 	"$Id: batchd.c,v 4.0 1993/09/17 00:41:04 srp Exp srp $";
#endif

/*
 *
 * Author: Scott R. Presnell
 *
 * Copyright (C) 1989, 1990, The Regents of the University of California
 *                     All Rights Reserved
 *
 * THE REGENTS OF THE UNIVERSITY OF CALIFORNIA MAKE NO REPRESENTATION OR
 * WARRANTIES WITH RESPECT TO THE CONTENTS HEREOF AND SPECIFICALLY DISCLAIM
 * ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
 * PURPOSE.
 *
 * Further, the Regents of the University of California reserve the right to
 * revise this software and/or documentation and to make changes from time
 * to time in the content hereof without obligation of the Regents of the
 * University of California to notify any person of such revision or change.
 *
 */

/* 
 * Batch job control daemon.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>

#include <errno.h>
#include <fcntl.h>

#if defined(MALLOC_TRACE) || defined(MALLOC_LEAK)
#include "dmalloc.h"
#else
#include <malloc.h>
#endif

#include <signal.h>
#include <syslog.h>
#include <time.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/file.h>
#include <sys/dir.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/resource.h>

#include <sys/socket.h>
#include <sys/un.h>

#include <netinet/in.h>
#include <net/if.h>

#if defined(sgi) || defined(__sgi)
#include <net/soioctl.h>
#endif

#include <arpa/inet.h>

#include <netdb.h>

#include "config.h"

#if (defined(I_STDARG) && defined(CAN_VAPROTO)) || defined(__STDC__)
#include <stdarg.h>
#else
#include <varargs.h>
#endif

#include H_STRING

#include "batch.h"
#include "batchd.h"
#include "bid.h"
#include "event.h"
#include "flags.h"
#include "pathnames.h"
#include "protocol.h"
#include "patchlevel.h"

#ifdef I_SELECT
#include H_SELECT
#endif /* I_SELECT */

#ifndef BSEC
#define	BSEC	(60*60)				/* Min. bid ttl */
#endif

#ifndef CSEC
#define	CSEC	(60)				/* Console checks */
#endif

#ifndef PSEC
#define	PSEC	(5*60)				/* Running job checks */
#endif

#ifndef	CPUOFF
#define CPUOFF	(1.00)				/* Relative CPU power */
#endif

#ifndef	GSWITCH
#define	GSWITCH	(2)				/* graphics gw per second */
#endif

#ifndef	DEFAULT_TTL
#define	DEFAULT_TTL	(32)			/* Multicast ttl */
#endif

#ifdef MASTLOCK					/* Lock/pid file for daemon */
char *	masterlock = MASTLOCK;
# else
char *	masterlock = _PATH_MASTLOCK;
#endif

#ifdef EQUIVHOSTS				/* hosts.equiv location */
char *	equivhosts = EQUIVHOSTS;
#else
char *	equivhosts = _PATH_EQUIV;
#endif

#ifndef HAVE_FD_MACROS
#define NFDBITS         32
#define FD_SETSIZE      32
#define FD_SET(n, p)    ((p)->fds_bits[(n)/NFDBITS] |= (1 << ((n) % NFDBITS)))
#define FD_CLR(n, p)    ((p)->fds_bits[(n)/NFDBITS] &= ~(1 << ((n) % NFDBITS)))
#define FD_ISSET(n, p)  ((p)->fds_bits[(n)/NFDBITS] & (1 << ((n) % NFDBITS)))
#endif /* HAVE_FD_MACROS */

#ifdef FD_ZERO
#undef FD_ZERO
#endif
#define FD_ZERO(p)      (void) memset((voidp)(p), 0, sizeof(*(p)))

int nkids = 0;					/* Number of children */
int njobs = 0;		       			/* Number of jobs */

static int alarmed = 1;				/* True if alarmed() */
static int forced  = 1;				/* Force rescan of the queue */
static int returned = 0;			/* Child has returned. */ 
static int doreload = 1;			/* Reload capabilities db */
static int dostats = 0;				/* Dump the statistics */
static int dodb = 0;				/* Dump the db */

qelem	queues	= { &queues, &queues };		/* Queue linked list */
qelem	runs	= { &runs,   &runs   };		/* Running job linked list */
qelem	hosts	= { &hosts,  &hosts  };		/* Host info linked list */

Info	qinfo;					/* Machine information */

State log_lvl = 				/* Loging level */
	(BLOG_FATAL|BLOG_ERR|BLOG_WARNING|BLOG_INFO|BLOG_STATS);

Bool logging = TRUE;				/* Logging state */
Bool debugging = FALSE;				/* Debugging state */


int aflag = 0;					/* Keep accounting records */
int batchd_pid = -1;				/* Pid of parent batchd */

static int flocal;				/* socket descriptors */
static int finet;    

static char remotehost[MAXHOSTNAMELEN];		/* derived remote hostname */

int fudp;					/* UDP socket descriptor */
int udp_port;					/* and port number */

time_t now;					/* Current time */
time_t starttime = -1;				/* Time daemon was started */
time_t boottime = -1;				/* Time machine was booted */

double loadave[3];				/* Load averages */

struct balarm_t {
	int	alarmed;
	char *	queue;
} balarm ;					/* Specific queue alarm */

#if defined(MALLOC_TRACE)
FILE *	mtfp = NULL;				/* Malloc trace file */
#endif

#if defined(MALLOC_LEAK)
FILE *	mlfp = NULL;				/* Malloc leak file */
#endif

int btimer = BSEC;				/* Generate bids this often */
int stimer = CSEC;				/* Check console this often */
int ptimer = PSEC;				/* Check jobs this often */

Bool use_rhosts	= TRUE;				/* Use hosts.equiv file */
Bool use_cache	= TRUE;				/* Use addr <-> hosts cache */

double	CPUfactor	= CPUOFF;		/* Relative CPU power */
int	Gswitches	= GSWITCH;		/* Graphics pipeline use */

char *	interface	= NULL;			/* Name of interface */

int	Bflag		= 0;			/* Broadcast in use */
int	Mflag		= 0;			/* Multicast in use */

/*
 * Set the time-to-live for the packet:
 *  0 == same host,  1 == same subnet 
 * 32 == same site, 64 == same region
 */
static u_char	ttl	= DEFAULT_TTL;		/* Multicast ttl */

struct in_addr bcastaddr;			/* Broadcast address */

/*
 * Signal handlers
 */
static SIGRTN	reload P((int));
static SIGRTN	cleanup P((int));
static SIGRTN	stats P((int));
static SIGRTN	db P((int));
static SIGRTN	togglelog P((int));
static SIGRTN	toggledebug P((int));
static SIGRTN	reap P((int));
static SIGRTN	terminate P((int));

/*
 * Internal functions.
 */
static int	init_sockets P((char *));
static int	init_lock P((void));
static int	init_interface P((struct in_addr *, struct in_addr *,
				  int, char *));

static void	main_loop P((void));

static void	queue_scan P((char *, int));
static int	host_check P((struct sockaddr_in *, int, char []));

static void	process_reap P((void));


static int	schedule_bid P((Queue *));
static int	schedule_periodic P((Queue *));
static int	schedule_timeout P((Queue *));
static int	schedule_usage P((Queue *));

static void	addrcache_flush P((void));

static int 	ipc_request P((int, char *));
static int	queue_kick P((int, char *));

static int	provide_info P((char *));
static int	provide_bid P((char *queue));
static int	provide_stats P((char *queue));

/*
 * Event functions
 */
void		do_bid P((Queue *));
void		do_timein P((Queue *));
void		do_timeout P((Queue *));
void		do_periodic P((Queue *));
void		do_scan P((Queue *));
void		do_delay P((Job *));
void		do_clear P((Queue *));

/*
 * External fucntions.
 */
extern void	do_consolecheck P((void));	/* in console.c */

extern int 	load_get P((double []));	/* in getproc.c */
extern time_t	boottime_get P((void));		/* in getproc.c */

extern int 	queue_display P((char *));	/* in displayjob.c */

extern void	dump_stats P((void));		/* in xlog.c */
extern void	dump_db P((void));		/* in xlog.c */

extern State	decode_loglvl P((char *));	/* in xlog.c */
extern State	decode_dbglvl P((char *));	/* in xlog.c */

extern int 	switch_to_logfile P((char *));	/* in xlog.c */

/* in batchd.c but used before defined */			
extern void	process_handle P((Run *, wait_t *, rusage_t *));

extern void 	cap_reread P((void));		/* in readcap.c */
extern char *	statetoa P((Queue *));		/* in readcap.c */

extern Hostinfo * hostinfo_alloc P((void));	/* in bid.c */


extern char *	optarg;				/* Ptr to option argument */
extern int	optopt;				/* Current option */ 
extern int	optind;				/* Argument idx */
extern int	opterr;				/* Error handling by getopt */

extern int	cevent;				/* Console check event */

void		send_results P((Queue *, Job *, wait_t *, Notify, char *));

#define OPTSTRING	"ab:i:s:r:l:x:BFC:D:E:G:L:R:M:ST:X:V"

char *logfile = "stderr";

static	char *sel = "select: %s %lds";

static char *mend =
	"Queue %s: job %d (%s) finished for %s; exit: %d/%d; time: %ld.";
static char *aend = "stop  %-6.6s uid:%-5d job:%-5d exit:%d/%d time:%-8lu";


int
main(argc, argv)
int	argc;
char *	argv[];
{

	char *	cp;
	int	f;
	int	option;


	iamdaemon = TRUE;			/* We are the deamon */

	if ((progname = strrchr(argv[0], '/')) != NULL)
		progname++;
	else
		progname = argv[0];

	/* Get the real local hostname */
	if (gethostname(lhost, MAXHOSTNAMELEN) < 0) {
		blog(BLOG_ERR, "%s: gethostname fails", progname);
		exit(1);
	}

	/* Get the short hostname */
	(void) strcpy(shost, lhost);
	if ((cp = strchr(shost, '.')) != NULL)
		*cp = '\0';
		
	/*
	 * Read the command line arguments.
	 */
	while ((option = getopt(argc, argv, OPTSTRING)) != EOF) {
		switch (option) {
		case 'a':			/* accounting */
			aflag = 1;
			break;
		case 'b':			/* *B*roadcast timeout */
			if ((btimer = atoi(optarg)) < 1) {
				blog(BLOG_WARNING, "broadcast time must be > 1");
				btimer = BSEC;
			}
			break;
		case 'l':			/* logfile to write */
			logfile = optarg;
			break;
		case 'i':			/* interface to use */
			interface = optarg;
			break;
		case 'c':			/* *C*onsole check timeout */
		case 's':			/* ...or  (*S*hort time) */
			if ((stimer = atoi(optarg)) < 1) {
				blog(BLOG_WARNING, "console time must be > 1");
				stimer = CSEC;
			}

			break;
		case 'r':			/* per. *R*esource timeout  */
			if ((ptimer = atoi(optarg)) < 1) {
				blog(BLOG_WARNING, "resource time must be > 1");
				ptimer = PSEC;
			}
			break;
		case 'x':			/* logging flags */
			log_lvl = decode_loglvl(optarg);
			logging = TRUE;
			break;
		case 'B':			/* Classic Broadcast */
			Bflag = 1;
			break;
		case 'C':			/* Console device */
			console = optarg;
			break;
		case 'X':			/* Console info file */
			consoleinfo = optarg;
			break;
		case 'D':			/* Debugging flags */
#ifdef DEBUG
			log_lvl |= decode_dbglvl(optarg);
			debugging = TRUE;
#else
			blog(BLOG_WARNING, "sorry - not compiled with DEBUG");
#endif

			break;
		case 'E':			/* CPU Efficency */
			if ((CPUfactor = atof(optarg)) < 1.0) {
				blog(BLOG_WARNING, "CPU factor must be > 1.0");
				CPUfactor = CPUOFF;
			}
			break;
		case 'F':
			use_cache = FALSE;
			break;
		case 'G':			/* Graphics Contect switches */
			if ((Gswitches = atoi(optarg)) < 1) {
				blog(BLOG_WARNING, "Gswitches must be > 1");
				Gswitches = GSWITCH;
			}
			break;
		case 'M':			/* Multicast */
#ifdef IFF_MULTICAST
			Mflag = 1;
			bcastaddr.s_addr = inet_addr(optarg);
			if (!IN_MULTICAST(bcastaddr.s_addr)) {
				blog(BLOG_WARNING,
				     " invalid multicast group: %s", optarg);
				exit(1);
			}
#else
			blog(BLOG_WARNING,
			     "Multicast not supported under %s", OS);
#endif	/* IFF_MULTICAST */
			break;
		case 'T':			/* Multicast Time-to-live */
#ifdef IFF_MULTICAST
			if ((ttl = (u_char) atoi(optarg)) < 1.0) {
				blog(BLOG_WARNING, "Multicast ttl must be > 1");
				ttl = DEFAULT_TTL;
			}
#else
			blog(BLOG_WARNING,
			     "Multicast not supported under %s", OS);
#endif	/* IFF_MULTICAST */
			break;
#ifdef USE_PORT_SELECTION
			/*
			 * local_port and remote_port stay in
			 * host byte order.
			 */
		case 'L':
			socketname = optarg;
			break;
		case 'R':
			remote_port = atoi(optarg);
			break;
#endif /* USE_PORT_SELECTION */
		case 'S':			/* less Security */
			use_rhosts = FALSE;
			break;
		case 'V':
			dump_version(printsf, 0);
			exit(0);
			break;
		case '?':			/* Unknown option */
			if (opterr == 0)
				blog(BLOG_WARNING,
				     "internal: bad flag \"-%c\"", optopt);
			exit(1);
			break;
		default:			/* Unhandled option */
			blog(BLOG_WARNING,
			     "internal: no action for \"-%c\" option",
			     (char) option);
			exit(1);
			break;
		}
	}


	if (!(log_lvl & BLOG_NOFORK)) {

		if (getuid() != 0) {
			blog(BLOG_ERR, "must be root to start %s", progname);
			exit(1);
		}

		/*
		 * Fork.  The parent exits; the child becomes the daemon.
		 */
		if (fork())
			exit(0);

		/*
		 * Why would any of these be open?  I don't know. 
		 * Close them all anyway.
		 */
		for (f = getdtablesize() - 1; f > 2; f--)
			(void) close(f);

		/*
		 * Reopen stdin, stdout and stderr to /dev/null,
		 * or / or something.
		 */
		if ((f = open(_PATH_DEVNULL, O_RDONLY)) < 0) {
			blog(BLOG_ERR, eopen, _PATH_DEVNULL);
			exit(1);
		} else {
			(void) dup2(f, 0);
			(void) dup2(f, 1);
			(void) dup2(f, 2);
			if (f > 2)
				(void) close(f);
		}
#ifdef HAVE_SETSID
		(void) setsid();
#else
		/*
		 * Disassociate ourselves from the controlling tty.
		 * N.B.: we don't disassociate if we don't have
		 * TIOCNOTTY but we do have HAVE_BSDPGRP (?)
		 */
#ifdef TIOCNOTTY
		if ((f = open(_PATH_TTY, O_RDONLY)) >= 0) {
			(void) ioctl(f, TIOCNOTTY, 0);
			(void) close(f);
		}
#ifdef HAVE_BSDPGRP
		(void) setpgrp(0, 0);
#else
		(void) setpgrp();
#endif
#else /* TIOCNOTTY */
#ifdef HAVE_BSDPGRP
		(void) setpgrp(0, 0);
#else
		(void) setpgrp();
#endif
#endif /* TIOCNOTTY */
#endif
	}

	/*
	 * Get the pid *after* we fork.
	 */
	batchd_pid = getpid();

	/*
	 * Open syslog() *after* we fork.
	 */
	(void) switch_to_logfile(logfile);

	/* Write version information to the system log */
	dump_version(blog, BLOG_INFO);

#ifdef MALLOC_TRACE
	mtfp = fopen("/usr/tmp/malloc.trace", "w+");
	mal_setstatsfile(mtfp);
	mal_trace(1);
#endif
#ifdef MALLOC_DEBUG
	mal_debug(3);
#endif

	/*
	 * Created files are not world readable.
	 */
	(void) umask(002);

	/*
	 * Get the current working directory.  Under normal operating
	 * situations this is "/".
	 */
#ifdef DEBUG
	(void) getcwd(home, MAXPATHLEN);
#else
	(void) strcpy(home, "/");
#endif

	/*
	 * Set lock file.
	 */
	if (init_lock() < 0)
		exit(2);

	/*
	 * Set up the sockets.
	 */
	if (init_sockets(socketname) < 0)
		exit(2);

	/* 
	 * Set the boottime.
	 */
	boottime = boottime_get();

	/*
	 * Signal traps.
	 */
	(void) signal(SIGHUP, reload);
	(void) signal(SIGINT, cleanup);		/* Exit the process */

#ifdef DEBUG
	/* Leave SIGQUIT available */
#else
	(void) signal(SIGQUIT, SIG_IGN);	/* Ignore */
#endif

	(void) signal(SIGIOT, stats);
	(void) signal(SIGEMT, db);

	(void) signal(SIGPIPE, SIG_IGN);	/* Ignore */

	(void) signal(SIGALRM, alarming);	/* Examine the queues. */
	(void) signal(SIGCHLD, reap);		/* Catch the children */
	(void) signal(SIGUSR1, togglelog);	/* Toggle writing logfile  */
	(void) signal(SIGUSR2, toggledebug);	/* Toggle writing debug */

	(void) signal(SIGTERM, terminate);	/* Kill jobs and exit */

	qinfo.i_start = starttime = time((long *) 0);

	/* start the random number generator */
	srand((u_int) qinfo.i_start);

	main_loop();

	return 0;
}


/*
 * Open the batchd master lock file.  Try to get a lock on it and then
 * write the pid of the parent (this) process into that file.  Leave
 * the file open (as it is locked).  This ensures that only one batchd
 * is running at any given time.
 */
static int
init_lock()
{

	int	lfd;
	unsigned int n;

	if ((lfd = open(masterlock, (O_WRONLY|O_CREAT), 00644)) < 0) {
		blog(BLOG_ERR, eopen, masterlock);
		return (-1);
	}

	/* Try to get a lock */
#ifdef HAVE_LOCKF
	if (lockf(lfd, F_TLOCK, 0) < 0) {
		if (errno == EACCES || errno == EAGAIN) {
#else
#ifdef HAVE_FLOCK
	if (flock(lfd, (LOCK_EX|LOCK_NB)) < 0) {
		if (errno == EWOULDBLOCK) {
#endif /* HAVE_FLOCK */
#endif /* HAVE_LOCKF */
			/* Deamon present */
			blog(BLOG_ERR, "active batchd present");
		} else {
			/* Other reason */
			blog(BLOG_ERR, elock, masterlock);
		}
		return (-1);
	}

	/* Write process id for others to know */
	(void) ftruncate(lfd, 0L);
	(void) sprintf(line, "%d\n", batchd_pid);
	n = strlen(line);
	if (write(lfd, line, n) != n) {
		blog(BLOG_ERR, "cannot write %s", masterlock);
		return (-1);
	}

	return 0;

}


/*
 * Initialize the local and remote sockets.
 */
static int
init_sockets(sname)
char *	sname;
{

	int		ifd;			/* interface test socket */
	int		on;			/* reuse address */
	struct servent *	sp;		/* /etc/services pointer */
	struct sockaddr_in sin;			/* Inet domain socket */
	struct sockaddr_un slo;			/* UNIX domain socket */

	struct in_addr	 ifaddr;		/* Interface addrress */
	struct sockaddr_in sud;			/* UDP socket */
#ifdef IFF_MULTICAST
	u_char		 loop;			/* Multicast loopback (off) */
	struct ip_mreq	 mreq;			/* Multicast service (recv) */
#endif /* IFF_MULTICAST */

	static char ebind[] = "cannot bind service %s (%m)";
	static char elstn[] = "cannot set listen for service %s (%m)";


	/*
	 * Get a socket, bind it to the local service, and start listening.
	 */
	(void) unlink(sname);

	(void) memset((voidp) &slo, 0, sizeof (slo));

	if ((flocal = socket(AF_UNIX, SOCK_STREAM, 0)) < 0) {
		blog(BLOG_ERR, esock);
		return (-1);
	}

	(void) strcpy(slo.sun_path, sname);

	slo.sun_family = AF_UNIX;

	if (bind(flocal, (struct sockaddr *) &slo, SUN_LEN(&slo)) < 0) {
		blog(BLOG_ERR, ebind, sname);
		return (-1);
	}

	if (listen(flocal, 5) < 0) {
		blog(BLOG_ERR, elstn, sname);
		return (-1);
	}

	/* Bind to the default interface */
	ifaddr.s_addr = htonl(INADDR_ANY);

	/*  Mflag - bcastaddr set at command line */
	if (Bflag) {
		/* Bind to the default broadcast address */
		bcastaddr.s_addr = htonl(INADDR_BROADCAST);
	}

	/*
	 * If we're asked about a certain interface, open up a socket
	 * and get the configuration information.
	 */
	if (interface) {

		/* DataGram... */
		if ((ifd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
			blog(BLOG_ERR, esock);
			return (-1);
		}

		/*
		 * Set the broadcast and interface addresses if the
		 * requested interface is found.
		 */
		if (init_interface(&bcastaddr, &ifaddr, ifd, interface) < 0)
			return (-1);
		close(ifd);
	}


	blog(BLOG_INFO, "interface address: %s", inet_ntoa(ifaddr));
	if (Bflag || Mflag)
		blog(BLOG_INFO, "broadcast address: %s", inet_ntoa(bcastaddr));


	/*
	 * Get a socket, bind it to the general service, and start listening.
	 */
	(void) memset((voidp) &sin, 0, sizeof (sin));

	if ((finet = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		blog(BLOG_ERR, esock);
		return (-1);
	}

	if (remote_port == 0) {
		if ((sp = getservbyname(service, tcp)) == NULL) {
			blog(BLOG_ERR, eserv, service);
			return (-1);
		}
		/* Already in network byte order */
		sin.sin_port = sp->s_port;
		/* Change to host byte order */ 
		remote_port = ntohs(sp->s_port);
	} else {
		/* Change from host byte order (back) to network byte order */
		sin.sin_port = htons(remote_port);
	}

	/* long network byte order */
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = ifaddr.s_addr;

	if (bind(finet, (struct sockaddr *) &sin, sizeof(sin)) < 0) {
		blog(BLOG_ERR, ebind, service);
		return (-1);
	}

	if (listen(finet, 5) < 0) {
		blog(BLOG_ERR, elstn, service);
		return (-1);
	}

	(void) memset((voidp) &sud, 0, sizeof (sud));

	/*
	 * Get a socket for UDP, and start listening.
	 */
	if ((fudp = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		blog(BLOG_ERR, esock);
		return (-1);
	}

	if (udp_port == 0) {
		if ((sp = getservbyname(service, udp)) == NULL) {
			blog(BLOG_ERR, eserv, service);
			return (-1);
		}
		/* Already in network byte order */
		sud.sin_port = sp->s_port;
		/* Change to host byte order */ 
		udp_port = ntohs(sp->s_port);
	} else {
		/* Change from host byte order (back) to network byte order */
		sud.sin_port = htons(udp_port);
	}

	on = 1;
	if (setsockopt(fudp, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) < 0) {
		blog(BLOG_ERR, "udp reuseaddr (%m)");
		return (-1);
	}

	/* For binding (recv) */
	sud.sin_family = AF_INET;
	sud.sin_addr.s_addr = ifaddr.s_addr;

#ifdef IFF_MULTICAST
	if (Mflag) {

		/* Don't send it to the loopback interface */
		loop = 0;
		if (setsockopt(fudp, IPPROTO_IP, IP_MULTICAST_LOOP,
			       &loop, sizeof(loop)) < 0) {
			blog(BLOG_ERR, "multicast loopback (%m)");
			return (-1);
		}


		if (setsockopt(fudp, IPPROTO_IP, IP_MULTICAST_TTL,
			       &ttl, sizeof(ttl)) < 0) {
			blog(BLOG_ERR, "multicast ttl (%m)");
			return (-1);
		}

		if (setsockopt(fudp, IPPROTO_IP, IP_MULTICAST_IF,
			       &ifaddr, sizeof(ifaddr)) < 0) {
			blog(BLOG_ERR, "multicast interface (%m)");
			return (-1);
		}

		/* Listen to multicast address on given interface */
		mreq.imr_multiaddr.s_addr = bcastaddr.s_addr;
		mreq.imr_interface.s_addr = ifaddr.s_addr;
		if (setsockopt(fudp, IPPROTO_IP, IP_ADD_MEMBERSHIP,
			       &mreq, sizeof(mreq)) < 0) {
			blog(BLOG_ERR, "multicast add member (%m)");
			return (-1);
		}
	} else
#endif	/* IFF_MULTICAST */

	if (Bflag) {
		/* ...CandyGram... */
		if (setsockopt(fudp, SOL_SOCKET, SO_BROADCAST, &on, sizeof(on)) < 0) {
			blog(BLOG_ERR, "set broadcast (%m)");
			return (-1);
		}
	}

	if (bind(fudp, (struct sockaddr *) &sud, sizeof(sud)) < 0) {
		blog(BLOG_ERR, ebind, service);
		return (-1);
	}

	return 0;
}


/*
 * Initialize data from an interface if we are asked.
 * This code does not handle broadcasting to multiple interfaces, or
 * point-to-point virtual broadcasting.
 * For now, select one interface to use for Inet sockets and broadcasts.
 */
static int
init_interface(bcaddr, ifaddr, fd, name)
struct in_addr *	bcaddr;
struct in_addr *	ifaddr;
int	fd;
char *	name;
{

	char		 buf[BUFSIZ];		/* interface information */
	int		 i = 0;
	int		 found = 0;		/* Found the interface? */
	struct ifconf	 ifc;			/* Interface configuration */
	struct ifreq *	ifr;			/* Interface flags */

	/* 
	 * Make sure the specified interface exists and is capable
	 * of doing multicasting if so asked.
	 */
	ifc.ifc_len = sizeof(buf);
	ifc.ifc_buf = buf;
	if (ioctl(fd, SIOCGIFCONF, (char *) &ifc) < 0) {
		blog(BLOG_ERR, "ioctl SIOCGIFCONF");
		return (-1);
	}

	ifr = ifc.ifc_req;
	for (i = ifc.ifc_len/sizeof(*ifr); --i >= 0; ifr++) {

		if (ifr->ifr_addr.sa_family != AF_INET)
			continue;
		
		/* Is this the interface we asked for? */
		if (strncmp(ifr->ifr_name, name, strlen(ifr->ifr_name)) != 0)
			continue;
			
		found = 1;

		/* Obtain the interface's network address */
		memcpy ((voidp) ifaddr, 
			(voidp) &((struct sockaddr_in *) &ifr->ifr_addr)->sin_addr,
			sizeof(*ifaddr));

		/* Obtain the interface's config. flags */
		if (ioctl(fudp, SIOCGIFFLAGS, (char *)ifr) < 0) {
			blog(BLOG_ERR, "ioctl SIOCGIFFLAGS");
			return (-1);
		}

		/* Interface is loopback */
		if (ifr->ifr_flags & IFF_LOOPBACK) {
			blog(BLOG_ERR, "%s: interface is loopback",
			    ifr->ifr_name);
			return (-1);
		}

		if (!(ifr->ifr_flags & IFF_UP)) {
			blog(BLOG_ERR, "%s: interface down", ifr->ifr_name);
			return (-1);
		}

#ifdef IFF_MULTICAST
		/* Interface does not multicast */
		if (Mflag && !(ifr->ifr_flags & IFF_MULTICAST)) {
			blog(BLOG_ERR, "%s: cannot multicast", ifr->ifr_name);
			return (-1);
		}
#endif /* IFF_MULTICAST */
				
		if (Bflag && !(ifr->ifr_flags & IFF_BROADCAST)) {
			blog(BLOG_ERR, "%s: cannot broadcast", ifr->ifr_name);
			return (-1);
		} else if (Bflag) {
			if (ioctl(fudp, SIOCGIFBRDADDR, (char *) ifr) < 0) {
				blog(BLOG_ERR, "%s: cannot get bcastaddr",
				    ifr->ifr_name);
				return (-1);
			}

			memcpy((voidp) bcaddr,
			       (voidp) &((struct sockaddr_in *)&ifr->ifr_broadaddr)->sin_addr,
			       sizeof(*bcaddr));
		}
	}

	if (!found) {
		blog(BLOG_ERR, "%s: interface not found", name);
		return -1;
	}

	return 0;
}

/*
 * Main Loop...
 *
 * 1) Note the time.
 * 2) Check for dead children.
 * 3) Check for alarm.
 * 4) While nothing else, wait for communications requests.
 *
 */
static void
main_loop()
{
	
	Bool	error = FALSE;
	Bool	ilocal;
	int 	fromlen;
	int	nfds, s;			/* Fds for select */
	int	cc;
	char *	queue;				/* queue name if alarm */
	char *	funcname = "main";
	fd_set	rfds;				/* Read the sockets */
	struct timeval
		tvs;				/* Select timeout */
	time_t	sstart = 0;			/* Select start */
	time_t	sdiff = 0;			/* Select difference */

	struct sockaddr_in frominet;
	struct sockaddr_un fromlocal;


#ifdef MALLOC_LEAK
	mlfp = fopen("/usr/tmp/malloc.leak", "w+");
	mal_leaktrace(1);
#endif /* MALLOC_LEAK */

	while (error == FALSE) {
		
		qinfo.i_main++;

		/* Note time of day */
		now = time((long *) 0);

		blog(BLOG_TRACE, ques, funcname, "top of loop");

		if (dostats) {
			dostats = 0;
			dump_stats();
		}
		
		if (dodb) {
			dodb = 0;
			dump_db();
		}
		
		if (doreload) {
			
			qinfo.i_reload++;
			
			doreload = 0;

			/*
			 * Delete pending console check event.  Reload the
			 * queue and then restart the console check event
			 * chain.
			 */
			if (cevent) {
				(void) event_remove(cevent);
				cevent = 0;
			}

			addrcache_flush();

			cap_reread();

			/* do_consolecheck() sets itself up after each call. */
			do_consolecheck();
		}
		
		if (returned) {
			returned = 0 ;
			process_reap();
		}
		
		if (alarmed || balarm.alarmed) {
			
			alarmed = 0;
			queue = NULL;

			/*
			 * If only one queue was signaled to be scanned
			 * do that, otherwise all queues are rescanned.
			 */
			if (balarm.alarmed) {
				balarm.alarmed = 0;
				queue = balarm.queue;
			}
			
			/*
			 * Scan every queue.
			 * We force a rescan on the first pass through
			 * the main loop and when a job comes back.
			 */
			queue_scan(queue, forced);
			forced = 0;
			
		}
		
		/*
		 * Ok. Time to set up for the select call...
		 * 1) Set the file descriptors to watch.
		 * 2) Determine the timeout length.
		 * 3) Check for signals received.
		 * 4) Start select().
		 */
		FD_ZERO(&rfds);
		
		FD_SET(flocal, &rfds);
		FD_SET(finet, &rfds);
		FD_SET(fudp, &rfds);
		
		ilocal = FALSE;

#ifdef MALLOC_LEAK
		mal_dumpleaktrace(mlfp);
#endif /* MALLOC_LEAK */
		

		/*
		 * tvs must be reset to the timeout each invocation.
		 * In some future *IX select() may return the time
		 * at which select was interrupted.
		 *
		 * next_event() has the side effect of executing any
		 * functions that need to be run from the event queue.
		 */
		tvs.tv_sec = event_next();
		do {
			switch (tvs.tv_sec) {
			case -1:		/* No events */
				tvs.tv_sec = ptimer;
				break;
			case 0:			/* Maybe more events */
				tvs.tv_sec = event_next();
				break;
			default:		/* Event in future */
				break;
			}
		/* call event_next() until rtn > 0 */
		} while (tvs.tv_sec <= 0);

		tvs.tv_usec = 0;

		if (njobs > 0)
			blog(BLOG_TRACE, "%s: %d job%s running",
			    funcname, njobs, (njobs > 1) ? "s" : "");

		/*
		 * If one of the signals should happen to have been set
		 * go back to the top of the loop.
		 */
		if (alarmed || returned || doreload || dostats || dodb)
			continue;

		blog(BLOG_TRACE, sel, "waits for", tvs.tv_sec);

		sstart = time((time_t *) 0);
		s = 0;
		errno = 0;

		/*
		 * Select call.
		 */
		nfds = select(FD_SETSIZE, &rfds, 0, 0, &tvs);
		
		now =   time((time_t *) 0);

		sdiff = now - sstart;
		
		if (nfds < 0) {
			if (errno == EINTR) {
				/* Signal Interrupt */
				blog(BLOG_TRACE, sel, "interrupted after", sdiff);
			} else {
				/* Error in select */
				blog(BLOG_ERR, sel, "error after", sdiff);
			}
			continue;

		} else if (nfds == 0) {
			/* Time expired */
			blog(BLOG_TRACE, sel, "expired after", sdiff);
			continue;

		} else if (FD_ISSET(flocal, &rfds)) {
			/* Local service */
			blog(BLOG_TRACE, sel, service, sdiff);

			fromlen = sizeof(fromlocal);
			s = accept(flocal, &fromlocal, &fromlen);
			ilocal = TRUE;
			qinfo.i_loc++;
			
		} else if (FD_ISSET(finet, &rfds)) {
			/* Internet tcp service */
			blog(BLOG_TRACE, sel, tcp, sdiff);
			
			fromlen = sizeof(frominet);
			s = accept(finet, &frominet, &fromlen);
			qinfo.i_inet++;

		} else if (FD_ISSET(fudp, &rfds)) {

			Bidpacket bid;

			/* Internet udp service */
			blog(BLOG_TRACE, sel, udp, sdiff);

			fromlen = sizeof(Bidpacket);
			cc = recvfrom(fudp, (voidp) &bid, fromlen, 0,
				      (struct sockaddr *) &frominet, &fromlen);

			if (cc <= 0) {
				if (cc < 0 && errno != EINTR) {
					blog(BLOG_ERR, "recvfrom failed");
				}
				continue;
			}

			/*
			 * Receive the bid and go back to the top of the loop
			 * for another select() call.
			 */
			if (host_check(&frominet, 0, remotehost) < 0) {
#ifdef LOG_UDP
				blog(BLOG_WARNING,
				     "access denied: %s (%s)",
				     remotehost, inet_ntoa(frominet.sin_addr));
#endif
			} else if (islocalhost(remotehost) == 0) {
				/*
				 * Ugh - drop broadcast to ourself, otherwise,
				 * we get into infinite telling ourselves how
				 * the daemon just started.
				 */
				  ;
			} else {
#ifdef LOG_UDP
				blog(BLOG_INFO, "recv udp: %s (%s)",
				     remotehost, inet_ntoa(frominet.sin_addr));
#endif
				(void) bid_receive(remotehost, &bid, now);
				qinfo.i_udp++;
			}
			continue;
			
		} else {

			blog(BLOG_ERR, "problem with select? (0x%x)", nfds);
			continue;
		}
		
		if (s < 0) {
			if (errno != EINTR) 
				blog(BLOG_ERR, "cannot accept socket");
			continue;
		}

		if (ilocal) {
			(void) strcpy(remotehost, lhost);
		} else {
			if (host_check(&frominet, 1, remotehost) < 0) {
				blog(BLOG_WARNING,
				     "access denied: %s", remotehost);
				(void) ewriteone(s, QR_EHPERM);
				(void) close(s);
				continue;

			}
		}

		(void) ipc_request(s, remotehost);
		(void) close(s);
	}

	return ;
}



static char *commands[] = {	/* strings that describe the requests */
	"",
	"kick",
	"receive",
	"display",
	"delete",
	"bid",
	"stats",
	"info",
	""
	};

static char	combuf[BUFFER];		/* command line buffer */


/*
 * Handle an interprocess communication request for a service of this daemon.
 * Possibly log the request with the hostname of the orignating machine and
 * the "command line" of the request.
 */
static int
ipc_request(fd, from_host)
int	fd;
char *	from_host;
{
	char * 	cp = combuf;
	char *	que;
	char *	thp;
 	int	n, pid, result;
	int	idx;
	char *	mess = "%s requests %s on %s";
	char	host[MAXHOSTNAMELEN];
	
	
	do {
		if (cp >= &combuf[sizeof(combuf) - 1]) {
			blog(BLOG_ERR, "Command line too long");
			return (RTN_ERROR);
		}
		if ((n = read(fd, cp, 1)) != 1) {
			if (n < 0)
				blog(BLOG_ERR, elost, from_host);
			return (RTN_ERROR);
		}
	} while (*cp++ != '\n');
	
	*--cp = '\0';
	
	cp = combuf;
	
	que = (cp + 1);
	
	idx = (int) (combuf[0] - '\0');
	
	while (*cp && *cp != ' ' && *cp != '\n')
		cp++;
	*cp++ = '\0';	
	
	if (from_host == NULL)
		return RTN_ERROR;
	
	(void) strncpy(host, from_host, MAXHOSTNAMELEN);
	
	/* Use just the short name of the host for identification */
	if ((thp = strchr(host, '.')) != NULL)
		*thp = '\0';
	
	/* Puts the message in the debugging file as well */
	blog(BLOG_INFO, mess, host, commands[idx], que);
	
	switch (combuf[0]) {
		
	case QC_KICK:	/* Reevaluate the runque */
		return queue_kick(fd, que);
		break;
		
	}
	
	/* if we get this far, it means we need to fork. */
	
	(void) sighold(SIGCHLD);		 /* until in parent table */
	
	switch (pid = fork()) {
	case -1:	/* ERROR */
		blog(BLOG_ERR, efork, "command request");
		break;
	case 0:		/* CHILD -> handle request  */
		/* The child won't have any SIGCHLD */
		/* (void) sigrelse(SIGCHLD); */
			
		(void) signal(SIGHUP, SIG_IGN);
		(void) signal(SIGINT, SIG_IGN);
		(void) signal(SIGIOT, SIG_IGN);
		(void) signal(SIGEMT, SIG_IGN);
		(void) signal(SIGALRM, SIG_IGN);
		(void) signal(SIGPIPE, SIG_IGN);
		(void) signal(SIGUSR1, SIG_IGN);
		(void) signal(SIGUSR2, SIG_IGN);
		(void) signal(SIGTERM, SIG_IGN);
		
		(void) close(flocal);
		(void) close(finet);
		(void) close(fudp);

		if (dup2(fd, 1) < 0)
			blog(BLOG_ERR, edup, fd);

		(void) close(fd);

		switch (combuf[0]) {
	
		case QC_RECV:	/* Receive files to be queued */
			result = job_receive(que, from_host);
			break;
		
		case QC_DISP:	/* Send queue info for display */
			result = queue_display(que);
			break;
		
		case QC_REM:	/* Remove job(s) from the queue */
			result = job_remove(que, cp);
			break;

		case QC_BID:	/* Provide a bid on a queue */
			result = provide_bid(que);
			break;

		case QC_STATS: /* Provide a stats report on a queue */
			result = provide_stats(que);
			break;

		case QC_INFO: /* Provide a stats report on a queue */
			result = provide_info(que);
			break;

		default:
			(void) ewriteone(1, QR_ENOIMP);
			blog(BLOG_WARNING, erecv, combuf[0]);
			result = RTN_ERROR;
			break;
		}

		exit(result);
		break;

	default:	/* PARENT -> add process, return */
		/* This is the parent - it returns for more requests */
		(void) process_add(pid, (Queue *) NULL, (Job *) NULL, "command request");

		/* Job is safely in table now */
		(void) sigrelse(SIGCHLD);
		break;
	}

	return 0;
}


/*
 * Alarm the batch daemon.
 */
static int
queue_kick(fd, queue)
int	fd;
char *	queue;
{

	Queue *	qp;


	if ((qp = getqueuebyname(queue)) == (Queue *) NULL) {
		(void) ewriteone(1, QR_ENOENT);
		return (RTN_ERROR); 
	}

	/* alarm the daemon to check on queues. */
	alarmed++;

	balarm.alarmed = 1;
	balarm.queue = queue;

	(void) ewriteone(fd, QR_OK);
	/* Normal return code */
	return (RTN_OK);

}


/*
 * Provide a bid.
 */
static int
provide_bid(queue)
char *	queue;
{

	Queue *	qp;


	if ((qp = getqueuebyname(queue)) == (Queue *) NULL) {
		(void) ewriteone(1, QR_ENOENT);
		return (RTN_ERROR); 
	}

	if (qp->q_loadb != TRUE) {
		(void) ewriteone(1, QR_MT);
		return (RTN_OK);
	}

	(void) ewriteone(1, QR_OK);

	/* Re-calcualte the load averages. */
	(void) load_get(&loadave[0]);

	(void) sprintf(line, BID_FMT, loadave[0], loadave[1], loadave[2],
		       bid_calculate(qp), time((time_t *) 0));

	(void) ewrite(1, line, strlen(line), "provide_bid");

	(void) ewriteone(1, '\n');

	return (RTN_OK);
}



/*
 * Provide a status.
 */
static int
provide_stats(queue)
char *	queue;
{

	Queue *	qp;


	if ((qp = getqueuebyname(queue)) == (Queue *) NULL) {
		(void) ewriteone(1, QR_ENOENT);
		return (RTN_ERROR);
	}

	(void) ewriteone(1, QR_OK);

	(void) sprintf(line, STATS_FMT, qp->q_state,
		       qp->q_maxexec, qp->q_nexec, qp->q_inactive);

	(void) ewrite(1, line, strlen(line), "provide_stats");

	(void) ewriteone(1, '\n');

	return (RTN_OK);
}

/*
 * Provide basic queue information.
 */
static int
provide_info(queue)
char *	queue;
{

	Queue *	qp;		/* Queue pointer */
	FILE *	tfp;		/* Temporary file pointer to hold old value */


	if ((qp = getqueuebyname(queue)) == (Queue *) NULL) {
		(void) ewriteone(1, QR_ENOENT);
		return (RTN_ERROR);
	}

	(void) ewriteone(1, QR_OK);

	tfp = sfp;
	sfp = stdout;
	dump_queue(printsf, 0, qp);
	sfp = tfp;

	return (RTN_OK);
}


/*
 * Find a queue by its name.
 */
Queue *
getqueuebyname(queue)
char *	queue;
{

	Queue *	qp;

	ITER(qp, Queue, &queues) {
		if (qp->q_name != NULL && strcmp(qp->q_name, queue) == 0)
			return qp;
	}			

	return ((Queue *) NULL);
}


/*
 * Gather up the children that have died and process each accordingly.
 */
static void
process_reap()
{

	int	pid = 0;
	wait_t	rtn;
	rusage_t rused;
	char *	funcname = "reap";
	Run *	rp;

	
	blog(BLOG_TRACE, ques, funcname, "reaping children");

	/*
	 * Will there be a problem under heavy load with signal interrupts?
	 */
	errno = 0;
	while ((pid = wait3(&rtn, WNOHANG, &rused)) > 0) {

		dbg((BLOG_PROC, "%s: pid: %d; sig: %d, exit: %d status %o",
		    funcname, pid, WTERMSIG(rtn), WEXITSTATUS(rtn),
		     rtn.w_status));

		qinfo.i_wait++;
	
	
		ITER(rp, Run, &runs)
			if (rp->r_pid == pid)
				break;
	
		/*
		 * This should never happen, provided SIGCHLD is properly
		 * held from before the fork until after the process is
		 * put in the run table.
		 */
		if (rp == HEAD(Run, &runs)) {
			blog(BLOG_ERR, "%s: process %d *not* in proc table?",
			    funcname, pid);
			continue;
		}

		nkids--;
	
		process_handle(rp, &rtn, &rused);

	}

	if (pid < 0 && errno != ECHILD)
		blog(BLOG_WARNING, "wait3 exited with -1 (%m)");

#ifndef HAVE_BSDSIGNALS
	(void) signal(SIGCHLD, reap);
#endif

	return ;
}


/*
 * Check the queues.
 *	1) Possibly reread the batchcap file.
 *	2) Re-determine the load.
 *	3) Check for someone on the console.
 *	4) Check each queue spooling directory for jobs.
 */
static void
queue_scan(queue, force)
char *	queue;
int	force;
{

	char *	sline;
	char	lhome[MAXPATHLEN];
	int	find_rval = 0;
	int	run_rval = 0;
	int	i, lfd = 0;
	int	retries = 3;
	Queue *	qp;


	blog(BLOG_TRACE, "%s:%sscanning",
	     (queue != NULL) ? queue : all,  (force > 0) ? " forced " : " ");

	/*
	 *  Check each potential queue...
	 */
	ITER(qp, Queue, &queues) {

		/* If this is an empty queue then skip. */
		if (qp->q_name == NULL)
			continue;

		/*
		 * If we asked for a specific queue, and this is not
		 * it, then skip.
		 */
		 if (queue && strcmp(qp->q_name, queue) != 0)
			continue;

		(void) queue_prt_flag(BLOG_QUEUES, qp);

		/*
		 * Save the current working directory.
		 * Restore at the end of the function.
		 */
		(void) getcwd(lhome, MAXPATHLEN);

		if (chdir(qp->q_spool) < 0) {
			blog(BLOG_ERR, echad, qp->q_spool);
			continue;
		}


		/* Grab lock */
		i = 1;
	lock_again:
		switch (lfd = queue_grab_lock(qp)) {
		case -1: /* error in opening the file */
			blog(BLOG_ERR, eopen, qp->q_lname);
			continue;
			break;

		case 0:	/* file is already locked */
			if (i < retries) {
				i++;
				(void) sleep((unsigned) i*i);
				goto lock_again;
			} else {
				blog(BLOG_WARNING, elock, qp->q_lname);
				continue;
			}
			break;

		default:
			dbg((BLOG_QUEUES, ques, qp->q_name, "locked"));
			/* All is cool, we've locked the file */
			break;
		}

		
		/*
		 * Return of (-1) means can't do queueing or running.
		 *			(STATUS_DISABLED)
		 * 	      (0) means no jobs affected.
		 *			(STATUS_NOCHANGE)
		 *	      (1) means queue was enabled or jobs queued.
		 *			(STATUS_ENABLED)
		 *	      (2) means message printed, don't overwrite.
		 *			(STATUS_PERSIST)
		 *
		 * The intent is to only update the status file
		 * when we need to, and only have that code called once
		 *
		 * Truth table: U = Update, N = No update.
		 *	
		 * Old		New return val
		 * Val -1	0	1	2
		 * -1	U	N	U	N
		 *  0	U	N	U	N
		 *  1	U	N	U	N
		 *  2	U	U	U	N
		 *
		 * ? should we leave the old value for people to see
		 * or clear it?
		 */
#undef KEEP_OLD
		
		find_rval = queue_find_jobs(qp, force);

		if ((find_rval <= STATUS_DISABLED) ||
#ifndef KEEP_OLD
		    (find_rval == STATUS_NOCHANGE &&
		     qp->q_return > STATUS_ENABLED) ||
#endif
		    (find_rval == STATUS_ENABLED)) {

			dbg((BLOG_QUEUES, ques, qp->q_name,
			    "queue: writing status file"));

			sline = statetoa(qp);
			queue_write_status(qp, sline);
		}

		run_rval = queue_run_jobs(qp, force);

		if ((run_rval <= STATUS_DISABLED) ||
#ifndef KEEP_OLD
		    (run_rval == STATUS_NOCHANGE &&
		     qp->q_return > STATUS_ENABLED) ||
#endif
		    (run_rval == STATUS_ENABLED)) {

			dbg((BLOG_QUEUES, ques, qp->q_name,
			    "run: writing status file"));

			sline = statetoa(qp);
			queue_write_status(qp, sline);
		}

		/*
		 * If everything passed queueing and running
		 * then reset the status file after the user
		 * has had a chance to look.
		 */
		dbg((BLOG_QUEUES, "%s: find: %d; run: %d; old: %d",
		    qp->q_name, find_rval, run_rval, qp->q_return));

		if (find_rval == STATUS_PERSIST ||
		    run_rval == STATUS_PERSIST) 
			qp->q_return = STATUS_PERSIST;

		else if (find_rval <= STATUS_DISABLED ||
			 run_rval <= STATUS_DISABLED)

			qp->q_return = STATUS_DISABLED;

		else if (find_rval == STATUS_ENABLED ||
			 run_rval == STATUS_ENABLED)

			qp->q_return = STATUS_ENABLED;
		else

			qp->q_return = STATUS_NOCHANGE;


		/*
		 * Reap attached jobs that have completed
		 */
		if (qp->q_nexec > 0 && qp->q_pevent == 0)
			(void) queue_attach_reap(qp);

		/*
		 * Check the limits of currently running jobs.
		 */
		if (qp->q_nexec > 0 && qp->q_pevent == 0)
			(void) queue_limit_jobs(qp);

		/*
		 * Reset the nice values of currently running jobs.
		 */
		if (qp->q_nexec > 0 && qp->q_pevent == 0)
			(void) queue_prioritize_jobs(qp);


		/* Release the lock when were done examining the queue. */
		if (queue_release_lock(qp, lfd) < 0) {
			blog(BLOG_ERR, eopen, qp->q_lname);
		} else {
			dbg((BLOG_QUEUES, ques, qp->q_name, "released"));
		}


#ifndef USE_TCP_BID
		/*
		 * If this is a load balancing queue, and
		 * 	1) the state of the queue has changed ||
		 * 	2) there are jobs in the queue ||
		 * 	3) the capabilites file has been reloaded
		 * broadcast a new bid message to all load balancing hosts.
		 */

		if (qp->q_loadb)
			(void) bid_send((char *) NULL, qp, now, force);

		if (qp->q_loadb && qp->q_bevent == 0)
			qp->q_bevent = schedule_bid(qp);

#endif

		/*
		 * Set up the next periodic scan of the queue.  This
		 * event manages checking the time on jobs and the
		 * return of attached jobs.
		 * Only one will be scheduled at a time.
		 */
		if (qp->q_nexec > 0 && qp->q_pevent == 0)
			qp->q_pevent = schedule_periodic(qp);

		/*
		 * Set up the next timeout for this queue.
		 * Only one timeout will be scheduled at a time.
		 */
		if (qp->q_tevent == 0)
			qp->q_tevent = schedule_timeout(qp);

		/*
		 * Set up event to dump+clear user CPU/job usage
		 * information
		 */
		if (qp->q_uevent == 0)
			qp->q_uevent = schedule_usage(qp);

		if (queue_get_flag(qp, RESET) == 0)
			queue_clr_flag(qp, RESET);

		(void) queue_prt_flag(BLOG_QUEUES, qp);

		if (qp->q_count)
			blog(BLOG_TRACE, "%s: jobs in the queue: %d\n",
			    qp->q_name, qp->q_count);

		/*
		 * Go back to where we came from
		 */
		if (chdir(lhome) < 0)
			blog(BLOG_ERR, echad, lhome);

	}

	return ;
}


/*
 * A process has returned, pick up the pieces, and take care of
 * job accounting.
 */
void
process_handle(rp, rtn, ruse)
Run *	rp;
wait_t *	rtn;
rusage_t *	ruse;
{
	
	Job *	jp;
	Queue *	qp;
	Usr *	up;
	int	pid;
	int	retcode = 0;
	int	termsig = 0;
	int	mtest;
	int	delayer = 0;
	char	lhome[MAXPATHLEN];
	char *	emess;
	static char *	mess2 =
		"%s: job %d: finished - exited %d/%d; time: %ds";
	static char *	mess3 =
		"Queue %s: job %d has transfered from %s\n\tto %s for %s (pid: %d).";
	static char *	mess5 =
		"Queue %s: transfer of job %d (%s -> %s\nfor %s) has been aborted (pid: %d).";
	static char *	mess6 =
		"Queue %s: error reported for transfer of job %d (%s -> %s\nfor %s), job terminated (pid: %d)";
	static char *	mess7 =
		"Queue %s: not available for transfer of job %d (%s -> %s\nfor %s), job terminated (pid: %d)";
	static char *	mess8 = "Queue %s: finished draining";

	now = time((time_t *) 0);

	if (rtn != NULL) {
		if (WIFEXITED(*rtn))
			retcode = WEXITSTATUS(*rtn);
		if (WIFSIGNALED(*rtn))
			termsig = WTERMSIG(*rtn);
	}

	/*
	 * If there is no pointer to a queue structure, this is the
	 * return of a mail process, an ipc_request()  or a resource
	 * checking fork.
	 */
	if ((qp = rp->r_qp) == NULL) {
		rp->r_pid = 0;
		return;
	} 

	/*
	 * All running jobs that point to a queue must also point
	 * to a job structure.  This should never happen...
	 */
	if ((jp = rp->r_jp) == NULL) {
		blog(BLOG_WARNING, "%s: jp for pid %d is NULL",
		     qp->q_name, rp->r_pid);
		rp->r_pid = 0;
		return;
	}

	/*
	 * ...and that job structure must be valid.  The file must *not*
	 * point to null. This should never happen.
	 */
	if (jp->j_file == NULL) {
		blog(BLOG_WARNING, "%s: jp->j_file for pid %d is NULL",
		     qp->q_name, rp->r_pid);
		rp->r_pid = 0;
		return;
	}

	pid = rp->r_pid;

	/*
	 * Save the current working directory.
	 * (Restore at the end of the function.)
	 */
	(void) getcwd(lhome, MAXPATHLEN);
	
	if (chdir(qp->q_spool) < 0) {
		blog(BLOG_ERR, echad, qp->q_spool);
		return ;
	}
	
	/*
	 * Handle transmission jobs specially...
	 */
	if (jp->j_state & JOB_SEND) {
		
		/*
		 * Subject line
		 */
		(void) sprintf(line, quej, qp->q_name, jp->j_id,
			       "transfered");
		
		/*
		 * Transmissions that are aborted might come back
		 * with a signal value.  Send SIGNAL message
		 * to the users in that case.
		 */
		if (termsig != RTN_OK)
			goto jabort;

		switch (retcode) {

		case RTN_OK:
			emess = mess3;
			/*
			 * This is a normal ending, send mail only if
			 * they ask for it specifically.
			 */
			mtest = NOTIFY_TRANSFER;

			qp->q_stats.s_received++;

			goto doit;
			break;

		case RTN_ACCESS:
			emess = mess7;

			/* This is the "end" of their job */
			mtest = (NOTIFY_TRANSFER | NOTIFY_END);
			goto doit;
			break;

		case RTN_ERROR:
			emess = mess6;
			/* This is the "end" of their job */
			mtest = (NOTIFY_TRANSFER | NOTIFY_END);
			goto doit;
			break;

		case RTN_SIGNAL:
		jabort:
			emess = mess5;
			/* This is the "end" of their job */
			mtest = (NOTIFY_TRANSFER | NOTIFY_END);

			qp->q_stats.s_aborted++;

		doit:
			/*
			 * Only if it was a normal exit do we nuke the
			 * job stuff...
			 */
			(void) sprintf(line2, quej, qp->q_name, jp->j_id,
				       "transfered");

			blog(BLOG_STATS, line2);
			queue_write_status(qp, line2);

			if (jp->j_usermail & mtest)
				send_mail(jp->j_umaddress, line,
					  emess, qp->q_name, jp->j_id,
					  lhost, jp->j_shost,
					  jp->j_user, pid);
			
			if (qp->q_supmail & mtest)
				send_mail(qp->q_smaddress, line,
					  emess, qp->q_name, jp->j_id,
					  lhost, jp->j_shost,
					  jp->j_user, pid);

			if (jp->j_userwrite & mtest)
				send_write(jp->j_uwaddress,
					   emess, qp->q_name, jp->j_id,
					   lhost, jp->j_shost,
					   jp->j_user, pid);
			
			if (qp->q_supwrite & mtest)
				send_write(qp->q_swaddress,
					   emess, qp->q_name, jp->j_id,
					   lhost, jp->j_shost,
					   jp->j_user, pid);

			goto term_remove;
			break;	

		case RTN_RETRY:
		case RTN_REPORT:

			/*
			 * ...otherwise it's still a valid job,
			 * to be sent at a later time.
			 */
			(void) sprintf(line2, quej, qp->q_name, jp->j_id,
				       "transfer failed");

			blog(BLOG_STATS, line2);
			queue_write_status(qp, line2);

			/*
			 * Job is to be sent again later.  Set up the event.
			 */
			jp->j_state = JOB_AGAIN;
			jp->j_shost = NULL;
			jp->j_pid = 0;
			jp->j_state &= ~(JOB_RUNNING);
			jp->j_stime = now + ptimer;

			if (jp->j_sevent != 0) {
				(void) event_remove(jp->j_sevent);
				jp->j_sevent = 0;
			}

			jp->j_sevent = event_add(ptimer + getrand(-5, 5),
						 do_scan, (voidp) qp,
						 "retransmit");

			goto term_done;
			break;	

		default:
			blog(BLOG_WARNING,
			    "process_handle(): bogus exit value (%d)",
			    retcode);
			goto term_remove;
			break;
		}
	} /* endif JOB_SEND */

	njobs--;

#ifdef DEBUG
	if (njobs < 0)
		blog(BLOG_WARNING, ques, qp->q_name, "njobs < 0!");
#endif

	qp->q_nexec--;

#ifdef DEBUG
	if (qp->q_nexec < 0)
		blog(BLOG_WARNING, ques, qp->q_name, "nexec < 0!");
#endif

	if (ruse != NULL)
		jp->j_cputime = ruse->ru_utime.tv_sec +
			ruse->ru_stime.tv_sec;

	if (qp->q_supmail & NOTIFY_END) {

		(void) sprintf(line, quej, qp->q_name, jp->j_id,
			       "done");

		send_mail(qp->q_smaddress, line,
			  mend, qp->q_name, jp->j_id, jp->j_jname,
			  jp->j_user, retcode, termsig, jp->j_cputime);
	}

	if (qp->q_supwrite & NOTIFY_END)
		send_write(qp->q_swaddress, 
			   mend, qp->q_name, jp->j_id, jp->j_jname,
			   jp->j_user, retcode, termsig, jp->j_cputime);


	qp->q_stats.s_returned++;


	/* Tell the user about his job */
	send_results(qp, jp, rtn, (Notify) (NOTIFY_END | NOTIFY_DEBUG), (char *)0);

	if (jp->j_userwrite & NOTIFY_END)
		send_write(jp->j_uwaddress,
			   mend, qp->q_name, jp->j_id, jp->j_jname,
			   jp->j_user, retcode, termsig, jp->j_cputime);


	/* Write a message to the syslog file */
	(void) sprintf(line, mess2,
		       qp->q_name, jp->j_id, retcode, termsig, jp->j_cputime);

	/* Write a message to the syslog file */
	blog(BLOG_STATS, line);

	/* Write a message to the status file */
	queue_write_status(qp, line);

	/* Log an accounting entry */
	if (aflag > 0)
		account(qp, aend, qp->q_name, jp->j_uid, jp->j_id,
			retcode, termsig, jp->j_cputime);

	/* Do our internal user CPU/job usage accounting */
	if ((up = getusrbyid(&(qp->q_usrs), qp->q_ftype, jp)) == NULL) {
		blog(BLOG_WARNING, "%s: cannot find uid: %u",
		    qp->q_name, jp->j_uid);
	} else {
		up->u_cputime += jp->j_cputime;

		/*
		 * Remove the CPU time from the running job CPU accounting.
		 *
		 * This may result in a slightly lower runtime count
		 * until the running jobs are tallied next.
		 */
		if ((up->u_runtime -= jp->j_cputime) < 0)
			up->u_runtime = 0;
			
		/* Another job completed */
		up->u_complete++;
	}

	/* Drained? */
	if (queue_get_flag(qp, DRAINING) == 0 && qp->q_nexec == 0) {
		blog(BLOG_INFO, ques, qp->q_name, "drained");
		(void) queue_set_flag(qp, DRAINED);

		if (qp->q_supmail & NOTIFY_RESOURCE)
			send_mail(qp->q_smaddress, "queue drained",
				  mess8, qp->q_name);

		if (qp->q_supwrite & NOTIFY_RESOURCE)
			send_write(qp->q_swaddress, mess8, qp->q_name);

	}

 term_remove:

	/* Job completed: decrement the count of jobs in the queue. */
	qp->q_count--;

#ifdef DEBUG
	if (qp->q_count < 0)
		blog(BLOG_WARNING, ques, qp->q_name, "q_count < 0!");
#endif

	/* Remove the job description files. */
	job_unlink(jp);

	delayer = (jp->j_state & JOB_DELAYER);

	/* Free/zero the job structures. */
	job_free(jp);

 term_done:

	/* "free" the structure just before we return. */
	rp->r_pid = 0;

	/* Go back to wherever we were */
	if (chdir(lhome) < 0)
		blog(BLOG_ERR, echad, lhome);

	if (delayer) {
		/* Someone was waiting on this job: re-scan all queues */
		alarming(0);
	} else if (!(jp->j_state & JOB_AGAIN)) {
		/*
		 * Unless it is a failed transmission, rescan the queue that
		 * just had a job return.
		 */
		queue_scan(qp->q_name, TRUE);
	}
	return;
}


/*
 *  Send a signal to all jobs in a particular queue.
 */
void
queue_signal_jobs(qp, sig, explain)
Queue *	qp;
int	sig;
char *	explain;
{

	Job *	jp;
	int	cfd;


	blog(BLOG_INFO, ques, qp->q_name, explain);

	if (qp->q_nexec == 0)
		return;

	qp->q_stats.s_signaled++;

	now = time((time_t *)0);

	ITER(jp, Job, &(qp->q_jobs)) {

		/* Job is not running, skip on ahead in the list */
		if (!(jp->j_state & JOB_RUNNING))
			continue;
#ifdef ndef
		/* nostopme code - not functional right now. */
		if (!(qp->q_uctl & STOPPED) && !cuid && jp->j_uid == cuid) {
			dbg((BLOG_PROC, "skipped killing job: %d (pid: %d)",
			     jp->j_id, jp->j_pid));
			continue;
		}
#endif
		dbg((BLOG_PROC, "killing job: %d (pid: %d) signal %d",
		     jp->j_id, jp->j_pid, sig));

		(void) job_killpg(jp, sig);

		/*
		 * Update the control file with an audit trail of sending
		 * the signal.
		 */
		if ((cfd = open(jp->j_file, (O_RDWR|O_APPEND))) < 0) {
			blog(BLOG_WARNING, eopen, jp->j_file);
		} else {
			(void) sprintf(line, "%c%s %d %d %ld\n",
				       'K', explain, jp->j_pid, sig, now);
			(void) ewrite(cfd, line, strlen(line), "signaljob()");
			(void) close(cfd);
		}

		/* Make a subject line */
		(void) sprintf(line, quej, qp->q_name, jp->j_id, explain);

		/* Send mail to the owner that the job has been signaled. */
		if (jp->j_usermail & NOTIFY_PAUSE)
			send_mail(jp->j_umaddress, line, (char *) NULL);

		if (qp->q_supmail & NOTIFY_PAUSE)
			send_mail(qp->q_smaddress, line, (char *) NULL);

		/* Write the owner that the job has been signaled. */
		if (jp->j_userwrite & NOTIFY_PAUSE)
			send_write(jp->j_uwaddress, line);

		if (qp->q_supwrite & NOTIFY_PAUSE)
			send_write(qp->q_swaddress, line);
		

	}
	return;
}


/*
 * Kill a process group - signal an error if kill fails.
 *
 * All OK	- return 0
 * No process	- return 1
 * Other error	- return -1
 */
int
job_killpg(jp, sig)
Job *	jp;
int	sig;
{



	/*
	 * Zap the entire process group for the job  (use a negative pid
	 * to indicate the group).
	 */
	errno = 0;
	if (kill(-(jp->j_pid), sig) < 0) {
		if (errno == ESRCH) {
			return 1;
		} else {
			blog(BLOG_ERR, ekill, jp->j_queue->q_name, jp->j_pid, sig);
			return (-1);
		}

	} else {
		blog(BLOG_TRACE, "kill: sent signal %d to job %d (pid: %d)",
		    sig, jp->j_id, jp->j_pid);

		return 0;
	}

}


/*
 * Add a forked process to the internal table of running processes.
 */
void
process_add(pid, rpq, rjp, class)
int	pid;
Queue *	rpq;
Job *	rjp;
char *	class;
{

	Run *	rp;
	Run *	nextrp = NULL;


	nkids++;

	ITER(rp, Run, &runs) {
		if (rp->r_pid == 0) {
			/* Found an empty slot */
			nextrp = rp;
			break;
		}
	}

	if (rp == HEAD(Run, &runs)) {

		/* Make space for a new queue. */
		dbg((BLOG_ALLOC, "allocating Run structure"));

		rp = ALLOC(Run);

		qinfo.i_runlist++;
		ins_tail_que(&rp->r_runs, &runs);

	} else {
		rp = nextrp;
	}
	
	rp->r_pid = pid;
	rp->r_qp = rpq;
	rp->r_jp = rjp;

	dbg((BLOG_PROC, "adding process pid:%d (%s)", pid, class));

	return;
}


static SIGRTN
reload(sig)
int	sig;
{

	doreload++;
	alarmed++;

#ifndef HAVE_BSDSIGNALS
	if (sig)
		(void) signal(sig, reload);
#endif

	return ;
}


SIGRTN
alarming(sig)
int	sig;
{

	alarmed++;

	if (sig < 0)
		forced++;

	dbg((BLOG_PROC, "%salarm signaled", (forced > 0) ? "forced " : ""));

#ifndef HAVE_BSDSIGNALS
	if (sig > 0)
		(void) signal(sig, alarming);
#endif

	return ;
}


static SIGRTN 
cleanup(sig)
int	sig;
{

	(void) close(flocal);

	(void) close(finet);

	(void) close(fudp);

	(void) unlink(socketname);
	(void) unlink(masterlock);

	blog(BLOG_INFO, "killed by signal %d", sig);

	(void) closelog();

	exit(sig);

}

static SIGRTN 
terminate(sig)
int	sig;
{

	char	lhome[MAXPATHLEN];
	Queue *	qp;

	/* Ignore dying children */
	signal(SIGCHLD, SIG_DFL);

	/*
	 * Save the current working directory.
	 * Restore at the end of the function.
	 */
	(void) getcwd(lhome, MAXPATHLEN);

	ITER(qp, Queue, &queues) {

		/* If this is an empty queue then skip it */
		if (qp->q_name == NULL)
			continue;

		if (chdir(qp->q_spool) < 0) {
			blog(BLOG_ERR, echad, qp->q_spool);
			continue;
		}

		(void) queue_signal_jobs(qp, SIGTERM, "terminated");
	}

	/*
	 * Go back to where we came from
	 */
	if (chdir(lhome) < 0)
		blog(BLOG_ERR, echad, lhome);

	cleanup(sig);

}


static SIGRTN
reap(sig)
int	sig;
{

	returned++;

	/*
	 * If you reset the signal handler here, the process appears to
	 * continually receive SIGCHLD.   The signal is lowered after the
	 * wait call has completed.   So, we reset the signal handler
	 * after gathering up all the children that have returned.
 	 * IRIX 3.2 manual page confirms.
	 */

	return ;
}


static SIGRTN
stats(sig)
int	sig;
{

	dostats++;

#ifndef HAVE_BSDSIGNALS
	if (sig)
		(void) signal(sig, stats);
#endif

	return;
}


static SIGRTN
db(sig)
int	sig;
{

	dodb++;

#ifndef HAVE_BSDSIGNALS
	if (sig)
		(void) signal(sig, db);
#endif

	return ;
}


static SIGRTN
togglelog(sig)
int	sig;
{

	if (logging == TRUE) {
		switch_to_logfile((char *) NULL);
		logging = FALSE;
	} else {
		switch_to_logfile(logfile);
		logging = TRUE;
	}
	

#ifndef HAVE_BSDSIGNALS
	if (sig)
		(void) signal(sig, togglelog);
#endif

	return ;
}


static SIGRTN
toggledebug(sig)
int	sig;
{

	if (debugging == TRUE)
		debugging = FALSE;
	else
		debugging = TRUE;
			

#ifndef HAVE_BSDSIGNALS
	if (sig)
		(void) signal(sig, toggledebug);
#endif

	return ;
}


/*VARARGS*/
#if (defined(I_STDARG) && defined(CAN_VAPROTO)) || defined(__STDC__)

void
account(Queue *qp, char *fmt, ...)

#else

void
account(qp, fmt, va_alist)
Queue *	qp;
char *	fmt;
va_dcl

#endif /* __STDC__ */
{


	char *	cp;
	char	priv[LINE];
	int	i = 0;
	int	retries = 3;
	int	afd = 0;
	va_list	args;


	/* Fire up the variable argument mechanism. */
	/*CONSTCOND*/
#if (defined(I_STDARG) && defined(CAN_VAPROTO)) || defined(__STDC__)
	va_start(args, fmt);
#else
	va_start(args);
#endif

	i = 1;

 account_again:
	if ((afd = open(qp->q_aname, (O_RDWR|O_CREAT|O_APPEND), 00640)) < 0) {
		blog(BLOG_ERR, eopen, qp->q_aname);
		return ;
	}

#ifdef HAVE_LOCKF
	if (lockf(afd, F_TLOCK, 0) < 0) {
#else
#ifdef HAVE_FLOCK
	if (flock(afd, (LOCK_EX|LOCK_NB)) < 0) {
#endif /* HAVE_FLOCK */
#endif /* HAVE_LOCKF */

		if (i < retries) {
			i++;
			(void) close(afd);
			(void) sleep((unsigned)(i*i));
			goto account_again;
		} else {
			blog(BLOG_ERR, elock, qp->q_aname);
			return ;
		}
	} 

	now = time((long *) 0);
	(void) strcpy(priv, ctime(&now));

	priv[24] = '\0';	/* Cut off after seconds */
	cp = &priv[4];		/* Start at the month */

	(void) ewrite(afd, cp, strlen(cp), (char *) NULL);
	(void) ewriteone(afd, ' ');

	(void) vsprintf(line, fmt, args);

	(void) ewrite(afd, line, strlen(line), "writeaccount()");
	(void) ewriteone(afd, '\n');

	(void) close(afd);

	/* Shut down the variable argument mechanism. */
	va_end(args);

	return;
}


/*
 * Check to see if the from host has access to the batch queue system.
 */
static int
host_check(f, reserved, fhost)
struct sockaddr_in *	f;
int	reserved;
char	fhost[];
{


	char *	cp;
	int	hit_cache = 0;
	Hostinfo *	hip;
	struct hostent *	hp;


	errno = 0;

	if (f->sin_family != AF_INET) {
		(void) strcpy(fhost, "<unknown>");
		blog(BLOG_WARNING,
		     "connection not AF_INET? (%d)", f->sin_family);
#ifdef ndef
		/* Odd problem this... just pretend it *is* INET */
		return (-1);
#endif
	}

#ifdef USE_RRESVPORT
	if (reserved && ntohs(f->sin_port) >= IPPORT_RESERVED) {
		(void) strcpy(fhost, inet_ntoa(f->sin_addr));
		blog(BLOG_WARNING,
		     "connection from %s not on a reserved port", fhost);
		return (-1);
	}
#endif /* USE_RRESVPORT */

	/*
	 * If calls to gethostbyaddr are expensive, scan the hosts
	 * structure till we find the address. Assumes address
	 * stays valid, and that the list is shorter/faster than the
	 * host list.
	 */
	ITER(hip, Hostinfo, &hosts) {
		/*
		 * If this is an empty entry then skip it
		 */
		if (hip->hi_name == NULL)
			continue;
		
		if (hip->hi_addr == f->sin_addr.s_addr) {
			/* If the address is the same, use this name */
			hit_cache = 1;
			break ;
		}
	}

	/*
	 * Check the hostname from the address...
	 */
	if (use_cache && hit_cache) {

#ifdef LOG_CACHE_STATS
		blog(BLOG_STATS, "addr hit: %lu -> \"%s\"",
		     hip->hi_addr, hip->hi_name);
#endif
		(void) strcpy(fhost, hip->hi_name);

	} else {
#ifdef LOG_CACHE_STATS
		blog(BLOG_STATS, "addr cache miss: %lu",
		     f->sin_addr.s_addr);
#endif

		/* Copy the string version of dotted quad into the hostname */
		(void) strcpy(fhost, inet_ntoa(f->sin_addr));

		h_errno = 0;
		if ((hp = gethostbyaddr(&f->sin_addr, sizeof(struct in_addr),
					f->sin_family)) == NULL) {
			blog(BLOG_WARNING,
			     "cannot obtain hostname for address %s (%h)", fhost);
			return (-1);
		} else if (hp->h_name == NULL) {
			blog(BLOG_WARNING,
			     "hostname is NULL for address %s", fhost);
			return (-1);
		} else {
			/* Copy the determined hostname into fhost */
			(void) strcpy(fhost, hp->h_name);
		}
	}

	/*
	 * If the "from" host appears to be the same as this host return
	 */
	if (strcasecmp(fhost, lhost) == 0)
		return (0);

	/*
	 * No fancy checks here, just see if the calling hostname is the
	 * same as the listed hostname.
	 */
	while ((cp = getbhent()) != NULL) {

		/*
		 * Check the hostname. Case is ignored.
		 */
		if (strcasecmp(fhost, cp) == 0) {
			if (use_cache && hit_cache == 0) {
				/*
				 * We're checking hosts we'll probably want
				 * to call in the future, so save the address.
				 */
				hip = hostinfo_alloc();
				hip->hi_name = estrdup(cp);
				hip->hi_addr = f->sin_addr.s_addr;
#ifdef LOG_CACHE_STATS
				blog(BLOG_STATS, "addto cache: %lu -> \"%s\" (%s)",
				     hip->hi_addr, cp, inet_ntoa(f->sin_addr));
#endif

			}
			endbhent();
			return (0);
		}
	}

	/* endbhent(); */

	/*
	 * We use ruserok() here so that yp/NIS will work correctly.
	 */
	if (use_rhosts && ruserok(fhost, 1, "root", "root") == 0)
		return (0);


	/*NOTREACHED*/
	return (-1);
}


void
queue_prt_flag(lvl, qp)
int	lvl;
Queue *	qp;
{

	dbg((lvl, "%s: flags are: %s (%lo)",
	     qp->q_name, statetoa(qp), qp->q_state));
	return;
}


/*
 * Execute a queue timeout
 */
void
do_timeout(qp)
Queue *	qp;
{

	qp->q_tevent = 0;
	qp->q_timein = FALSE;
	queue_scan(qp->q_name, FALSE);

	return;

}


/*
 * Execute a queue timein
 */
void
do_timein(qp)
Queue *	qp;
{

	qp->q_tevent = 0;
	qp->q_timein = TRUE;
	queue_scan(qp->q_name, FALSE);

	return;

}


/*
 * Dump and Clear the usr CPU usage.
 */
void
do_clear(qp)
Queue *	qp;
{

	Usr *	up;


	qp->q_uevent = 0;

	/* Dump to the system log before zeroing the array */
	dump_usr(blog, BLOG_STATS, qp);

	/*
	 * Clear the accumulated CPU time, but not the currently running
	 * job CPU accounting.
	 */
	ITER(up, Usr, &(qp->q_usrs))
		up->u_cputime = 0;
	
	return;

}


/*	
 * Generate a bid for this queue
 */
void
do_bid(qp)
Queue *	qp;
{

	qp->q_bevent = 0;

	if (qp->q_loadb)
		(void) bid_send((char *) NULL, qp, time((time_t *) NULL), 0);

	if (qp->q_loadb)
		qp->q_bevent = schedule_bid(qp);

	return;
}


/*	
 * Check process timing and job reprioritization.
 */
void
do_periodic(qp)
Queue *	qp;
{

	char	 lhome[MAXPATHLEN];


	qp->q_pevent = 0;

	/*
	 * Save the current working directory.
	 * Restore at the end of the function.
	 */
	(void) getcwd(lhome, MAXPATHLEN);

	if (chdir(qp->q_spool) < 0) {
		blog(BLOG_ERR, echad, qp->q_spool);
		return;
	}

	/*
	 * Reap attached jobs that have completed
	 */
	if (qp->q_nexec > 0)
		(void) queue_attach_reap(qp);

	/*
	 * Check the limits of currently running jobs.
	 */
	if (qp->q_nexec > 0)
		(void) queue_limit_jobs(qp);

	/*
	 * Reset the nice values of currently running jobs.
	 */
	if (qp->q_nexec > 0)
		(void) queue_prioritize_jobs(qp);

	/*
	 * Go back to where we came from
	 */
	if (chdir(lhome) < 0)
		blog(BLOG_ERR, echad, lhome);

	/*
	 * If there are still running jobs, schedule another timeout
	 */
	if (qp->q_nexec > 0)
		qp->q_pevent = schedule_periodic(qp);


	return;
}


/*
 * Do a scan of a queue.
 */
void
do_scan(qp)
Queue *	qp;
{

	queue_scan(qp->q_name, FALSE);
	return;
}


/*
 * Do a scan of a queue and zero the delay event for this queue.
 */
void
do_delay(jp)
Job *	jp;
{

	jp->j_devent = 0;
	queue_scan(jp->j_queue->q_name, FALSE);
	return;
}


/*
 * Schedule a timeout to generate bids.
 * Returns event id.
 */
static int
schedule_bid(qp)
Queue *	qp;
{

	int bid_time = MAX((btimer/2), stimer) + getrand(-10, 0);


	/* Create a description. */
	(void) sprintf(line, ques, qp->q_name, "bid generation");

	/* Set the bid generation event. */
	
	return event_add(bid_time, do_bid, (voidp) qp, line);
}


/*
 * Schedule a periodic scan of the proc table
 * Returns event id.
 */
static int
schedule_periodic(qp)
Queue *	qp;
{

	/* Create a description. */
	(void) sprintf(line, ques, qp->q_name, "periodic scan");

	/* Set the periodic scan event. */
	return event_add(ptimer + getrand(-5, 5),
			 do_periodic, (voidp) qp, line);
}


/*
 * Schedule the timein or timeout of a queue
 * Returns event id.
 */
static int
schedule_timeout(qp)
Queue *	qp;
{


	int		cday;
	int		day, i, j;
	int		stop;
	int		start;
	Bool		timein = FALSE;
	time_t		tout = 0;
	struct tm *	local_time;


	now = time((time_t *) 0);

	local_time = localtime(&now);
	cday = local_time->tm_wday;

	for (i = cday, j = 0; j < 8; i++, j++) {

		/* Wrap around at Sunday */
		day = i % 7;

		stop  = qp->q_days[day].d_stop;
		start = qp->q_days[day].d_start;
		
		/* If the same, then we don't do anything on this day. */
		if (stop == start)
			continue;

		/* If we are currently timed in... */
		if (qp->q_timein == TRUE) {
			if ((tout = event_time(j, stop)) < 0)
				continue;
			else
				timein = FALSE;
		/* Or timed out... */
		} else {
			if ((tout = event_time(j, start)) < 0)
				continue;
			else
				timein = TRUE;
		}

		break ;
	}


	if (tout <= 0)
		return (0);

	/* Create a description. */
	(void) sprintf(line, ques, qp->q_name, timein ? "timein" : "timeout");

	/* Set the timeout event. */
	return event_add(tout, timein ? do_timein : do_timeout, (voidp) qp, line);

}

/*
 * Schedule the dump and clear of user information.
 * Returns event id.
 */
static int
schedule_usage(qp)
Queue *	qp;
{

	int		cday;
	time_t		tout;
	struct tm *	local_time;


	now = time((time_t *) 0);

	local_time = localtime(&now);
	cday = local_time->tm_wday;

	if (qp->q_ftype & SORT_WEEKLY) {
		tout = event_time(7 - cday, 0); /* next sunday at midnight */
	} else if (qp->q_ftype & SORT_DAILY) {
		tout = event_time(1, 0); /* tomorrow at midnight */
	} else
		tout = 0;

	if (tout <= 0)
		return 0;

	/* Create a description. */
	(void) sprintf(line, ques, qp->q_name, "dump+clear usageo");

	/* Set the timeout event. */
	return event_add(tout, do_clear, (voidp) qp, line);

}

void
ins_head_que(elem, pred)
qelem *	elem;
qelem *	pred;
{
	qelem *p = pred->q_forw;
	elem->q_back = pred;
	elem->q_forw = p;
	pred->q_forw = elem;
	p->q_back = elem;
}

void
ins_tail_que(elem, succ)
qelem *	elem;
qelem *	succ;
{
	qelem *p = succ->q_back;
	elem->q_forw = succ;
	elem->q_back = p;
	succ->q_back = elem;
	p->q_forw = elem;
}

void
rem_que(elem)
qelem *	elem;
{
	qelem *p = elem->q_forw;
	qelem *p2 = elem->q_back;
	p2->q_forw = p;
	p->q_back = p2;
}

static void
addrcache_flush()
{

	Hostinfo *	hip;

	/* Flush the address cache. */
	ITER(hip, Hostinfo, &hosts) {
		/*
		 * If this is an empty entry then skip it
		 */
		if (hip->hi_name == NULL)
			continue;
		
		hip->hi_addr = 0;
	}

}
