/*
 * $Id: batchd.h,v 4.0 1993/09/17 00:41:04 srp Exp srp $
 */

/*
 *
 * Author: Scott R. Presnell
 *
 * Copyright (C) 1989, 1990, The Regents of the University of California
 *                     All Rights Reserved
 *
 * THE REGENTS OF THE UNIVERSITY OF CALIFORNIA MAKE NO REPRESENTATION OR
 * WARRANTIES WITH RESPECT TO THE CONTENTS HEREOF AND SPECIFICALLY DISCLAIM
 * ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
 * PURPOSE.
 *
 * Further, the Regents of the University of California reserve the right to
 * revise this software and/or documentation and to make changes from time
 * to time in the content hereof without obligation of the Regents of the
 * University of California to notify any person of such revision or change.
 *
 */

/*
 * Global definitions for the batchd job control system.
 */

/*
 * Return state from queuejob and runjob functions:
 *	We decide wether to update the status file based on these codes.
 */
#define STATUS_ERROR	-2
#define STATUS_DISABLED	-1
#define STATUS_NOCHANGE	 0
#define STATUS_ENABLED	 1
#define STATUS_PERSIST	 2

/* Debugging flags */
#define	BLOG_NOFORK	0001000
#define BLOG_LOAD	0002000
#define BLOG_XFER	0004000
#define BLOG_PROC	0010000
#define BLOG_JOBS	0020000
#define BLOG_QUEUES	0040000
#define BLOG_ALLOC	0100000
#define BLOG_EVENT	0200000
#define BLOG_RESOURCE	0400000
#define BLOG_FULL      	0777000

/* Send/receive return and exit codes */
#define	RTN_OK		0		/* Function return OK */
#define	RTN_RETRY	1		/* Expect to retry later */
#define	RTN_REPORT	2		/* Something happend */
#define RTN_ACCESS	3		/* Access problem */
#define RTN_ERROR	4		/* Error other error */
#define RTN_SIGNAL	5		/* Aborted by signal */

/* Sorting types */
#define SORT_UID	0000001
#define SORT_GID	0000002
#define SORT_ROUND	0000010
#define SORT_TIME	0000020
#define	SORT_USAGE	0000100
#define SORT_COMPLETE	0000200
#define	SORT_WEEKLY	0001000
#define	SORT_DAILY	0002000

/* Delay types */
#define	DELAY_NONE	0
#define	DELAY_TIME	1
#define	DELAY_FILE	2
#define	DELAY_JOB	3

/* Nice (and ndpriority) options */
#define NICE_HIGH	0
#define NICE_LOW	1

#define NICE_INTERP	0000001
#define NICE_STEP	0000002
#define NICE_JUMP	0000004
#define NICE_LOAD	0000010
#define NICE_GANG	0000100
#define NICE_SEPARATE	0000200

#define NICE_MAX	(-20)
#define NICE_MIN	(20)

#define LOAD_1MIN	0
#define LOAD_5MIN	1
#define LOAD_15MIN	2

/* Maximum number of sort functions */
#define MAXSORTFUNC	10

/* Safely large number of jobs (for array pointer storage) */
#define MAXJOBS		1000

/* Debugging is really just like logging */
#ifdef DEBUG
#define	dbg(x)	blog x
#else
#define	dbg(x)
#endif

/*
 * Linked list
 */
typedef struct qelem qelem;
struct qelem {
	qelem *	q_forw;
	qelem *	q_back;
};

#define	FIRST(ty, q)	((ty *) ((q)->q_forw))
#define	LAST(ty, q)	((ty *) ((q)->q_back))
#define	NEXT(ty, q)	((ty *) (((qelem *) q)->q_forw))
#define	PREV(ty, q)	((ty *) (((qelem *) q)->q_back))
#define	HEAD(ty, q)	((ty *) q)
#define	ITER(v, ty, q) \
	for ((v) = FIRST(ty,(q)); (v) != HEAD(ty,(q)); (v) = NEXT(ty,(v)))

#define	ALLOC(ty)	((ty *) ecalloc(1, sizeof(ty)))

/*
 * Typedefs need be specified before they are used
 */
typedef long 		State;

typedef struct	day_t	Day;

typedef struct	stat_t	Stats;

typedef struct	info_t	Info;

typedef struct	job_t	Job;
typedef struct	queue_t	Queue;

typedef struct	run_t	Run;

typedef struct	usr_t	Usr;

#ifdef HAVE_RLIMIT
typedef struct	rlimit	rlimit_t;
#endif
typedef struct	rusage	rusage_t;

struct stat_t {
	int	s_started;	/* Jobs started per queue */
	int	s_returned;	/* Jobs returned per queue */
	int	s_transfered;	/* Job transfers started per queue */
	int	s_received;	/* Job transfers succeeded per queue */
	int	s_aborted;	/* Job transfers aborted per queue */
	int	s_signaled;	/* Signals sent to jobs */
};

struct djob_t {
	int	d_id;
	char *	d_queue;
	char *	d_host;
};

struct info_t {
	time_t	i_start;	/* Start time of the daemon */
	int	i_fork;		/* Process forks - not used */
	int	i_wait;		/* Sucessful waits */
	int	i_main;		/* Main loop cycles */
	int	i_reload;	/* Capabilities file reloads */
	int	i_loc;		/* Local connections */
	int	i_udp;		/* Remote UDP information received. */
	int	i_inet;		/* Remote (internet) connections */
	int	i_queuelist;	/* Queuelist */
	int	i_joblist;	/* Joblist */
	int	i_fjoblist;	/* Free Joblist */
	int	i_usrlist;	/* Usrlist */		
	int	i_runlist;	/* Runlist */
	int	i_bidlist;	/* Bidlist */
	int	i_hostlist;	/* Hostlist */
	int	i_eventlist;	/* Event list */
	int	i_feventlist;	/* Free Event list */

};

struct day_t {
	int	d_start;	/* Hour to start this queue */
	int	d_stop;		/* Hour to stop this queue */
};

struct job_t {
	qelem	j_jobs;		/* Linked list of job entries */
	Queue *	j_queue;	/* Pointer back to it's queue */
	Usr *	j_usr;		/* Pointer back to it's usr structure */
	char *	j_rhost;	/* User specified host for this job */
	char *	j_wd;		/* Starting working directory */
	char *	j_file;		/* Control file name */
	char *	j_ofile;	/* User specified output file */
	char *	j_efile;	/* User specified error file */
	char *	j_user;		/* User name of job owner */
	char *	j_group;	/* Group name of job owner */
	char *	j_umaddress;	/* Send mail to this address */
	char *	j_uwaddress;	/* Send writes to this address */
	char *	j_jname;	/* Symbolic name for the job */
	char *	j_shost;	/* Job is being sent to this host */
	int	j_id;		/* Job id associated with the file */
	UIDTYPE	j_uid;		/* Uid of the job owner */
	GIDTYPE	j_gid;		/* Gid of the job owner */
	PIDTYPE	j_pid;		/* Process ID associated with file */

#ifdef HAVE_RLIMIT
				/* Job Resource hard/soft limits */
	rlimit_t j_rlimit[RLIM_NLIMITS];
#endif

        struct {	
		int	d_type;
		char *	d_string;
		union {
			time_t	d_time;
			char *	d_file;
			struct djob_t	d_djob;
		} delay_u ;
	} j_delay;	
	int	j_priority;	/* Job priority (nice) */
	int	j_rpriority;	/* Requested Job priority (nice) */
	int	j_ndpriority;	/* Job Non degrading priority */
	int	j_mustrun;	/* Processor associated with the job */
	Notify	j_usermail;	/* Mail args associated with the job */
	Notify	j_userwrite;	/* Write args associated with the job */
	State	j_state;	/* State of the job */
	time_t	j_ctime;	/* Creation time of the job */
	time_t	j_mtime;	/* Modification time of the job */
	time_t	j_stime;	/* Start time of the job / retransmit time */
	int	j_devent;	/* Delayed start event number */
	int	j_sevent;	/* retransmit event number */
	time_t	j_cputime;	/* accumulated CPU time in seconds */
	int	j_rsize;	/* current resident set size in bytes */
	int	j_dsize;	/* current process data size in bytes */
	int	j_ssize;	/* current process stack size in bytes  */
	Bool	j_restart:1;	/* This job was a restart */
	Bool	j_seen:1;	/* Tally existence */
};


struct queue_t {
	qelem	q_queues;	/* Linked list */
	qelem	q_jobs;		/* Linked list of jobs */
	qelem	q_usrs;		/* Linked list of users */
	char *	q_name;		/* Queue name */
	char *	q_rhost;	/* Queue host */
	char *	q_rname;	/* Queue remote name */
	char *	q_lname;	/* Lock file name */
	char *	q_sname;	/* Status file name */
	char *	q_aname;	/* Accounting file */
	char *	q_ename;	/* Error file */
	char *	q_spool;	/* Spooling area */
	char *	q_smaddress;	/* Queue supervisor mail address or file */
	char *	q_swaddress;	/* Queue supervisor write address */
	State	q_state;	/* State of the queue */
	time_t	q_mtime;	/* Mod. time of queue spooling directory */
	Notify	q_supmail;	/* Type of mail for supervisor */
	Notify	q_usermail;	/* Type of mail for user */
	Notify	q_supwrite;	/* Type of writes for supervisor */
	Notify	q_userwrite;	/* Type of writes for user */
	State	q_ftype;	/* Sort jobs with these flags */
	int	q_ulimit;	/* Sort jobs with cutoff of job usage */
	int	q_climit;	/* Sort jobs with cutoff of jobs completed */
	                        /* Pointers to sort functs */
	int	(*q_func[MAXSORTFUNC]) P((Job *, Job *));
	int	q_nfunc;	/* Number of sorting functions to consider */
	int	q_count;	/* Number of jobs in the queue */
	int	q_maxexec;	/* Max # of executing jobs */
	int	q_nexec;	/* Current # of executing jobs */
	int	q_loadsched;	/* Stop scheduling if over this load */
	int	q_loadstop;	/* Stop jobs if over this load */
	State	q_ptype;	/* Reschedule job nice from these flags */
	int	q_priority[2];	/* Job priority (nice) */
	int	q_ndpriority;	/* Job Non degrading priority */
	int	q_mustrun;	/* Processor for the job */

#ifdef HAVE_RLIMIT		
	rlimit_t		/* Queue default Resource hard/soft limits */
		q_rlimit[RLIM_NLIMITS];
#endif

	Stats	q_stats;	/* Job Statistics for each queue */
	Day	q_days[7];	/* Start and stop times for each days */
	int	q_idle;		/* Stop queue / Idle time ? */
	int	q_return;	/* Return code of queue manipulations */
	int	q_bevent;	/* Bid generation event number */
	int	q_pevent;	/* Periodic job scan event number */
	int	q_tevent;	/* Time[in|out] event number */
	int	q_uevent;	/* Usage clear event number */
	State	q_uctl;		/* User control of the queue */
	Bool	q_restart:1;	/* Restart on system crash? */
	Bool	q_loadb:1;	/* Load balanceing is ok in this queue */
	Bool	q_inactive:1;	/* Is the console considered active ? */
	Bool	q_timein:1;	/* Is the queue in timein ? */
	Bool	q_cumetime:1;	/* Use cumulative CPU time */
	Bool	q_userrc:1;	/* Use the users rcfile information */

};


struct run_t {
	qelem	r_runs;	/* Linked list of running jobs */
	Queue *	r_qp;		/* Pointer to queue table */
	Job *	r_jp;		/* Pointer to control file table */
	PIDTYPE	r_pid;		/* Process ID of running job */
};


struct usr_t {
	qelem	u_usrs;	/* Linked list of users/groups */
	UIDTYPE	u_uid;		/* numerical uid */
	GIDTYPE	u_gid;		/* numerical gid */
	time_t	u_cputime;	/* Completed job CPU time (seconds) */
	time_t	u_runtime;	/* Current running job CPU time (seconds) */
	int	u_queued;	/* Number of jobs currently queued */
	int	u_submit;	/* Number of jobs submitted */
	int	u_complete;	/* Number of jobs completed */
};

struct names_t {
	State	n_flag;
	char *	n_up;
	char *	n_down;
};


#ifdef HAVE_SIGBLOCK

/*
 * Emulate SYSV style signal control.
 */
#define	mask(s)		(1 << ((s) -1))
#define	sighold(s)	sigblock(mask(s))
#define	sigrelse(s)	sigsetmask(sigblock(0) & ~mask(s))

#endif

#if !defined(MIN)
#define MIN(a, b) ((a) < (b) ? (a) : (b))
#define MAX(a, b) ((a) > (b) ? (a) : (b))
#endif

extern qelem	queues;
extern qelem	runs;
extern qelem	hosts;

extern Info	qinfo;
extern Rc	rcfile;

extern int	batchd_pid;
extern int	njobs, nkids;
extern int	btimer, stimer, ptimer;
extern int	aflag;
extern Bool	use_cache;
extern State	log_lvl;
extern char **	oargv;
extern char *	console;
extern char *	consoleinfo;
extern double	loadave[3];
extern time_t	now;
extern time_t	boottime;
extern time_t	starttime;

extern FILE *	logfp;
extern FILE *	sfp;

extern int	fudp;
extern int	udp_port;
extern int	Mflag;
extern int	Bflag;

extern struct names_t namestab[10];


/*
 * Function declarations
 */
extern void	account V((Queue *, char *, ...));
extern void	send_mail V((char *, char *, char *, ...));
extern void	send_write V((char *, char *, ...));
extern void	queue_write_status V((Queue *, char *, ...));

extern Queue *	getqueuebyname P((char *));
extern int 	queue_find_jobs P((Queue *, int));
extern int 	queue_run_jobs P((Queue *, int));

/* Macros in flags.h */
/*extern int 	queue_get_flag P((Queue *, State));*/
/*extern void 	queue_set_flag P((Queue *, State));*/
/*extern void 	queue_clr_flag P((Queue *, State));*/

extern void 	queue_prt_flag P((int, Queue *));

extern void 	queue_set_lock P((Queue *, State));
extern int 	queue_get_lock P((Queue *, State));
extern int 	queue_grab_lock P((Queue *));
extern int 	queue_release_lock P((Queue *, int));

extern int 	queue_limit_jobs P((Queue *));
extern int 	queue_prioritize_jobs P((Queue *));

extern void 	queue_signal_jobs P((Queue *, int, char *));

extern int 	queue_attach_reap P((Queue *));

extern void 	dump_queue P((void (*)V((int, char *,...)), int, Queue *));
extern int 	queue_change_state P((Queue *, int, State, int));

extern SIGRTN	alarming P((int));

extern int	load_get P((double [3]));

extern int 	job_check_usage P((Job *));

extern int 	job_get_usage P((Job *));
extern int 	job_init_usage P((void));
extern int 	job_send P((Job *));
extern int 	job_receive P((char *, char *));
extern int 	job_remove P((char *, char *));

extern void 	job_kill P((Job *, int));
extern int 	job_killpg P((Job *, int));
extern void 	job_unlink P((Job *));
extern void 	job_free P((Job *));
extern void 	job_free_struct P((Job *));

extern void 	process_add P((int, Queue *, Job *, char *));

extern Usr * 	getusrbyid P((qelem *, int, Job *));
extern Usr *	usr_alloc P((qelem *));
extern void 	dump_usr P((void (*)V((int, char *,...)), int, Queue *));

extern void 	ins_tail_que P((qelem *, qelem *));
extern void 	ins_head_que P((qelem *, qelem *));
extern void 	rem_que P((qelem *));

