#ifndef lint
static char rcsid[] =
	"$Id: cmds.c,v 4.0 1993/09/17 00:41:04 srp Exp srp $";
#endif

/*
 * Copyright (c) 1983 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Berkeley.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef lint
static char sccsid[] = "@(#)cmds.c	5.4 (Berkeley) 6/30/88";
#endif /* not lint */

/*
 * cmds.c -- Batch job execution control program commands.
 * This code is derivative of the cmds.c for lpc.
 * The copyrights for that code have been retained, they are still in effect.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <ctype.h>

#include <errno.h>
#include <fcntl.h>
#include <utime.h>

#if defined(MALLOC_TRACE) || defined(MALLOC_LEAK)
#include "dmalloc.h"
#else
#include <malloc.h>
#endif

#include <sys/types.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/file.h>
#include <sys/dir.h>

#include "config.h"

#include H_STRING

#include "bac.h"

#include "batch.h"
#include "pathnames.h"
#include "protocol.h"
#include "reply.h"


static char	bbuf[BUFFER];
static char	pbuf[BUFFER];
static char *	bp;

char *	queue;				/* Queue of current interest */

int s;					/* Socket out of which to call */

int		abortq P((int));
int		cleanq P((void));
int		clobberq P((void));
int		enableq P((void));
int		disableq P((void));
int		drainq P((int, char **));
int		downq P((int, char **));
int		kickq P((void));
int		startq P((int));
int		statq P((void));
int		stopq P((void));
int		upq P((void));

static int	openqueue P((char *));
static int	checkarg P((struct direct *, char *	));
static int	unlinkf P((char *));

static int 	touch P((struct direct *d));

/*
 * Kill existing jobs and disable executions.
 */
int
doabort(argc, argv)
int	argc;
char **	argv;
{

	int	c;
	char *	cp1, *cp2;
	char	qbuf[100];


	if (argc == 1) {
		(void) printf("Usage: abort {all | queue ...}\n");
		return (-1);
	}

	if (argc == 2 && !strcmp(argv[1], all)) {
		queue = qbuf;
		while (getbaent(bbuf) > 0) {
			cp1 = qbuf;
			cp2 = bbuf;
			while ((c = *cp2++) && c != '|' && c != ':')
				*cp1++ = c;
			*cp1 = '\0';
			(void) abortq(1);
		}
		return 0;
	}

	while (--argc) {
		queue = *++argv;
		if (openqueue(queue) != 0)
			continue;
		(void) abortq(1);
	}

	return 0;
}


int
abortq(dis)
int	dis;
{

	int	fd;
	char *	lf, *sd;
	char	list[BUFFER];


	bp = pbuf;

	if ((lf = bgetstr("lf", &bp)) == NULL)
		lf = _DEF_LOCK;
	if (lf[0] != '/') {
		if ((sd = bgetstr("sd", &bp)) == NULL)
			sd = _DEF_SPOOL;
		(void) strcpy(line, sd);
		(void) strcat(line, "/");
		(void) strcat(line, lf);
	} else {
		(void) strcpy(line, lf);
	}

	(void) printf("%s:\n", queue);

	/*
	 * Turn on the owner execute bit of the lock file to
	 * disable executions.
	 */
	if (dis) {
		if (stat(line, &stbuf) >= 0) {
			if (chmod(line, (stbuf.st_mode & 06777) | 04100) < 0)
				(void) printf("\tcannot disable execution.\n");
			else
				(void) printf("\texecution is disabled.\n");
		} else if (errno == ENOENT) {
			if ((fd = open(line, (O_WRONLY|O_CREAT), 04766)) < 0)
				(void) printf("\tcannot create lock file.\n");
			else {
				(void) close(fd);
				(void) printf("\texecution is disabled.\n");
			}
			return (0);
		} else {
			(void) printf("\tcannot stat lock file.\n");
			return (-1);
		}
	}

	/*
	 * Send batchd a message to abort the queue.
	 */
	(void) strcpy(list, queue);
	(void) strcat(list, " root all");

	if (host_query(lhost, QC_REM, list) == 0)
		(void) printf("\tjobs aborted.\n");
	else
		(void) printf("\tcouldn't abort jobs.\n");

	return 0;
}


/*
 * Remove *incomplete* jobs from the spooling area.
 */
int
clean(argc, argv)
int	argc;
char **	argv;
{

	int	c;
	char *	cp1, *cp2;
	char	qbuf[100];


	if (argc == 1) {
		(void) printf("Usage: clean {all | queue ...}\n");
		return (-1);
	}

	if (argc == 2 && !strcmp(argv[1], all)) {
		queue = qbuf;
		while (getbaent(bbuf) > 0) {
			cp1 = qbuf;
			cp2 = bbuf;
			while ((c = *cp2++) && c != '|' && c != ':')
				*cp1++ = c;
			*cp1 = '\0';
			(void) cleanq();
		}
		return 0;
	}

	while (--argc) {
		queue = *++argv;
		if (openqueue(queue) != 0)
			continue;
		(void) cleanq();
	}

	return 0;
}


/*
 * Remove *all* jobs from the spooling area.
 */
int
clobber(argc, argv)
int	argc;
char **	argv;
{

	int	c;
	char *	cp1, *cp2;
	char	qbuf[100];


	if (argc == 1) {
		(void) printf("Usage: clobber {all | queue ...}\n");
		return (-1);
	}

	if (argc == 2 && !strcmp(argv[1], all)) {
		queue = qbuf;
		while (getbaent(bbuf) > 0) {
			cp1 = qbuf;
			cp2 = bbuf;
			while ((c = *cp2++) && c != '|' && c != ':')
				*cp1++ = c;
			*cp1 = '\0';
			(void) clobberq();
		}
		return 0;
	}

	while (--argc) {
		queue = *++argv;
		if (openqueue(queue) != 0)
			continue;
		(void) clobberq();
	}

	return 0;
}


int
file_select(d)
struct direct *	d;
{

	int 	c = d->d_name[0];

	/*
	 * of = output file.
	 * tf = temporary control file.
	 * cf = control file.
	 * df = data file.
	 * kf = kept control file.
	 */
	if ((c == 'o'|| c == 't' || c == 'c' || c == 'd' || c == 'k') &&
	    d->d_name[1] == 'f')
		return(1);
	return(0);
}


/*
 * Comparison routine for scandir. Sort by job number and machine, then
 * by `cf', `tf', or `df', then by the sequence letter A-Z, a-z.
 */
int
sortq(d1, d2)
struct direct **	d1, **	d2;
{

	int	c1, c2;


	if (c1 = strcmp((*d1)->d_name + 2, (*d2)->d_name + 2))
		return(c1);
	c1 = (*d1)->d_name[0];
	c2 = (*d2)->d_name[0];
	if (c1 == 'c')
		return(-1);
	if (c1 == 'd' || c1 == 'o' || c2 == 'c')
		return(1);
	return(-1);
}


int
cleanq()
{

	int	i, n;
	char *	cp, *cp1, *lp;
	char *	sd;
	struct direct ** queue_dir;
	int	nitems;


	bp = pbuf;

	if ((sd = bgetstr("sd", &bp)) == NULL)
		sd = _DEF_SPOOL;

	(void) printf("%s:\n", queue);

	for (lp = line, cp = sd; *lp++ = *cp++; )
		;
	lp[-1] = '/';

	nitems = scandir(sd, &queue_dir, file_select, sortq);

	if (nitems < 0) {
		(void) printf("\tcannot examine spool directory.\n");
		return (-1);
	}

	if (nitems == 0) {
		(void) printf("\tspool directory is clean.\n");
		return 0;
	}

	i = 0;
	do {
		cp = queue_dir[i]->d_name;
		if (*cp == 'c') {
			n = 0;
			while (i + 1 < nitems) {
				cp1 = queue_dir[i + 1]->d_name;
				if ((*cp1 != 'd' && *cp1 != 'o') ||
				    strcmp(cp + 2, cp1 + 2))
					break;
				i++;
				n++;
			}
			if (n == 0) {
				(void) strcpy(lp, cp);
				(void) unlinkf(line);
			}
		} else {
			/*
			 * Must be a df with no cf (otherwise, it would have
			 * been skipped above) or a tf file (which can always
			 * be removed).
			 */
			(void) strcpy(lp, cp);
			(void) unlinkf(line);

		}

     	} while (++i < nitems);

	return 0;
}


int
clobberq()
{

	int	i;
	char *	cp, *lp;
	char *	sd;
	struct direct ** queue_dir;
	int	nitems;


	bp = pbuf;

	if ((sd = bgetstr("sd", &bp)) == NULL)
		sd = _DEF_SPOOL;

	(void) printf("%s:\n", queue);

	for (lp = line, cp = sd; *lp++ = *cp++; )
		;
	lp[-1] = '/';

	nitems = scandir(sd, &queue_dir, file_select, sortq);
	if (nitems < 0) {
		(void) printf("\tcannot examine spool directory.\n");
		return (-1);
	}

	if (nitems == 0) {
		(void) printf("\tspool directory does not contain any jobs.\n");
		return 0;
	}

	i = 0;
	do {

		cp = queue_dir[i]->d_name;
		(void) strcpy(lp, cp);
		(void) unlinkf(line);

     	} while (++i < nitems);

	return 0;
}


static int 
unlinkf(name)
char *	name;
{
	if (unlink(name) < 0)
		(void) printf("\tcannot remove %s.\n", name);
	else
		(void) printf("\tremoved %s.\n", name);

	return 0;
}


/*
 * Enable submission to the queue.
 */
int
enable(argc, argv)
int	argc;
char **	argv;
{

	int	c;
	char *	cp1, *cp2;
	char	qbuf[100];


	if (argc == 1) {
		(void) printf("Usage: enable {all | queue ...}\n");
		return (-1);
	}

	if (argc == 2 && !strcmp(argv[1], all)) {
		queue = qbuf;
		while (getbaent(bbuf) > 0) {
			cp1 = qbuf;
			cp2 = bbuf;
			while ((c = *cp2++) && c != '|' && c != ':')
				*cp1++ = c;
			*cp1 = '\0';
			(void) enableq();
		}
		return 0;
	}

	while (--argc) {
		queue = *++argv;
		if (openqueue(queue) != 0)
			continue;
		(void) enableq();
	}

	return 0;
}


int
enableq()
{

	char *	sd, *lf;


	bp = pbuf;

	if ((lf = bgetstr("lf", &bp)) == NULL)
		lf = _DEF_LOCK;
	if (lf[0] != '/') {
		if ((sd = bgetstr("sd", &bp)) == NULL)
			sd = _DEF_SPOOL;
		(void) strcpy(line, sd);
		(void) strcat(line, "/");
		(void) strcat(line, lf);
	} else {
		(void) strcpy(line, lf);
	}

	(void) printf("%s:\n", queue);

	/*
	 * Turn off the group execute bit of the lock file to
	 * enable submission.
	 */
	if (stat(line, &stbuf) >= 0 && 
	    (stbuf.st_mode & 2000) &&
	    (stbuf.st_mode & 0010)) {
		if (chmod(line, (stbuf.st_mode & 02777) & ~02010) < 0)
			(void) printf("\tcannot enable submission.\n");
		else
			(void) printf("\tsubmission is enabled.\n");
	} else
		(void) printf("\tqueue was not disabled.\n");

	if (host_query(lhost, QC_KICK, queue) != 0)
		(void) printf("\tcouldn't enable submission in daemon.\n");

	return 0;
}


/*
 * Disable job submission to a queue.
 */
int
disable(argc, argv)
int	argc;
char **	argv;
{

	int	c;
	char *	cp1, *cp2;
	char	qbuf[100];


	if (argc == 1) {
		(void) printf("Usage: disable {all | queue ...}\n");
		return (-1);
	}
	if (argc == 2 && !strcmp(argv[1], all)) {
		queue = qbuf;
		while (getbaent(bbuf) > 0) {
			cp1 = qbuf;
			cp2 = bbuf;
			while ((c = *cp2++) && c != '|' && c != ':')
				*cp1++ = c;
			*cp1 = '\0';
			(void) disableq();
		}
		return 0;
	}

	while (--argc) {
		queue = *++argv;
		if (openqueue(queue) != 0)
			continue;
		(void) disableq();
	}

	return 0;
}


int
disableq()
{

	char *	sd, *lf;
	int	fd;


	bp = pbuf;

	if ((lf = bgetstr("lf", &bp)) == NULL)
		lf = _DEF_LOCK;
	if (lf[0] != '/') {
		if ((sd = bgetstr("sd", &bp)) == NULL)
			sd = _DEF_SPOOL;
		(void) strcpy(line, sd);
		(void) strcat(line, "/");
		(void) strcat(line, lf);
	} else {
		(void) strcpy(line, lf);
	}

	(void) printf("%s:\n", queue);

	/*
	 * Turn on the group execute bit of the lock file to
	 * disable submission.
	 */
	if (stat(line, &stbuf) >= 0) {
		if (chmod(line, (stbuf.st_mode & 06777) | 02010) < 0)
			(void) printf("\tcannot disable submission.\n");
		else
			(void) printf("\tsubmission is disabled.\n");
	} else if (errno == ENOENT) {
		if ((fd = open(line, (O_WRONLY|O_CREAT), 02676)) < 0)
			(void) printf("\tcannot create lock file.\n");
		else {
			(void) close(fd);
			(void) printf("\tsubmission is disabled.\n");
		}
	} else
		(void) printf("\tcannot stat lock file.\n");

	if (host_query(lhost, QC_KICK, queue) != 0)
		(void) printf("\tcouldn't disable submission in daemon.\n");

	return 0;
}


/*
 * Disable submissions and executions and put a message into the status file
 * (reason for being down).
 */
int
down(argc, argv)
int	argc;
char **	argv;
{
	int	c;
	char *	cp1, *cp2;
	char	qbuf[100];

	if (argc == 1) {
		(void) printf("Usage: down {all | queue} [message ...]\n");
		return (-1);
	}
	if (!strcmp(argv[1], all)) {
		queue = qbuf;
		while (getbaent(bbuf) > 0) {
			cp1 = qbuf;
			cp2 = bbuf;
			while ((c = *cp2++) && c != '|' && c != ':')
				*cp1++ = c;
			*cp1 = '\0';
			(void) downq(argc - 2, argv + 2);
		}
		return 0;
	}
	queue = argv[1];

	if (openqueue(queue) != 0)
		return (-1);
	(void) downq(argc - 2, argv + 2);

	return 0;
}


int
downq(argc, argv)
int	argc;
char **	argv;
{

	int	fd;
	char *	cp1, *cp2;
	char *	sd, *lf, *sf;


	bp = pbuf;

	(void) printf("%s:\n", queue);

	/*
	 * Turn on the group execute bit of lock file to disable
	 * submission and turn on the owner execute bit of lock file
	 * to disable executions.
	 */
	if ((lf = bgetstr("lf", &bp)) == NULL)
		lf = _DEF_LOCK;
	if (lf[0] != '/') {
		if ((sd = bgetstr("sd", &bp)) == NULL)
			sd = _DEF_SPOOL;
		(void) strcpy(line, sd);
		(void) strcat(line, "/");
		(void) strcat(line, lf);
	} else {
		(void) strcpy(line, lf);
	}

	if (stat(line, &stbuf) >= 0) {
		if (chmod(line, (stbuf.st_mode & 06777) | 06110) < 0)
			(void) printf("\tcannot disable submission and execution.\n");
		else
			(void) printf("\tsubmission and execution are disabled.\n");
	} else if (errno == ENOENT) {
		if ((fd = open(line, (O_WRONLY|O_CREAT), 06776)) < 0)
			(void) printf("\tcannot create lock file.\n");
		else {
			(void) close(fd);
			(void) printf("\tsubmission and execution are disabled.\n");
		}
	} else
		(void) printf("\tcannot stat lock file.\n");


	/*
	 * Write the message into the status file.
	 */
	if ((sf = bgetstr("sf", &bp)) == NULL)
		sf = _DEF_STAT;
	if (strchr(sf, '/') == NULL) {
		if ((sd = bgetstr("sd", &bp)) == NULL)
			sd = _DEF_SPOOL;
		(void) strcpy(line, sd);
		(void) strcat(line, "/");
		(void) strcat(line, sf);
	} else {
		(void) strcpy(line, sf);
	}

	fd = open(line, (O_WRONLY|O_CREAT), 00664);

#ifdef HAVE_LOCKF
	if (fd < 0 || lockf(fd, F_TLOCK, 0) < 0) {
#else
#ifdef HAVE_FLOCK
	if (fd < 0 || flock(fd, (LOCK_SH|LOCK_NB)) < 0) {
#endif /* HAVE_FLOCK */
#endif /* HAVE_LOCKF */
		(void) printf("\tcannot create status file.\n");
		return (-1);
	}
	(void) ftruncate(fd, 0);

	/*
	 * If they forgot to give a reason, just make the obvious statement.
	 */
	if (argc <= 0) {
		(void) write(fd, "is down\n", 8);
	} else {
		cp1 = line2;
		while (--argc >= 0) {
			cp2 = *argv++;
			while (*cp1++ = *cp2++)
				;
			cp1[-1] = ' ';
		}
		cp1[-1] = '\n';
		*cp1 = '\0';

		(void) write(fd, line2, strlen(line2));
	}
	(void) close(fd);

	if (host_query(lhost, QC_KICK, queue) != 0)
		(void) printf("\tcouldn't shut down queue in daemon.\n");
	else
		(void) printf("\tqueue is down.\n");

	return 0;

}


/*
 * Disable submissions and new executions but keep currently running jobs
 * going. Put a message into the status file (reason for draining).
 */
int
drain(argc, argv)
int	argc;
char **	argv;
{
	int	c;
	char *	cp1, *cp2;
	char	qbuf[100];

	if (argc == 1) {
		(void) printf("Usage: down {all | queue} [message ...]\n");
		return (-1);
	}
	if (!strcmp(argv[1], all)) {
		queue = qbuf;
		while (getbaent(bbuf) > 0) {
			cp1 = qbuf;
			cp2 = bbuf;
			while ((c = *cp2++) && c != '|' && c != ':')
				*cp1++ = c;
			*cp1 = '\0';
			(void) downq(argc - 2, argv + 2);
		}
		return 0;
	}
	queue = argv[1];

	if (openqueue(queue) != 0)
		return (-1);
	(void) drainq(argc - 2, argv + 2);

	return 0;
}


int
drainq(argc, argv)
int	argc;
char **	argv;
{

	int	fd;
	char *	cp1, *cp2;
	char *	sd, *lf, *sf;


	bp = pbuf;

	(void) printf("%s:\n", queue);

	/*
	 * Turn on the world execute bit of lock file to set draining.
	 */
	if ((lf = bgetstr("lf", &bp)) == NULL)
		lf = _DEF_LOCK;
	if (lf[0] != '/') {
		if ((sd = bgetstr("sd", &bp)) == NULL)
			sd = _DEF_SPOOL;
		(void) strcpy(line, sd);
		(void) strcat(line, "/");
		(void) strcat(line, lf);
	} else {
		(void) strcpy(line, lf);
	}

	if (stat(line, &stbuf) >= 0) {
		if (chmod(line, (stbuf.st_mode & 06777) | 02011) < 0)
			(void) printf("\tcannot drain queue.\n");
		else
			(void) printf("\tdraining queue.\n");
	} else if (errno == ENOENT) {
		if ((fd = open(line, (O_WRONLY|O_CREAT), 02671)) < 0)
			(void) printf("\tcannot create lock file.\n");
		else {
			(void) close(fd);
			(void) printf("\tdraining queue.\n");
		}
	} else
		(void) printf("\tcannot stat lock file.\n");


	/*
	 * Write the message into the status file.
	 */
	if ((sf = bgetstr("sf", &bp)) == NULL)
		sf = _DEF_STAT;
	if (strchr(sf, '/') == NULL) {
		if ((sd = bgetstr("sd", &bp)) == NULL)
			sd = _DEF_SPOOL;
		(void) strcpy(line, sd);
		(void) strcat(line, "/");
		(void) strcat(line, sf);
	} else {
		(void) strcpy(line, sf);
	}

	fd = open(line, (O_WRONLY|O_CREAT), 00664);

#ifdef HAVE_LOCKF
	if (fd < 0 || lockf(fd, F_TLOCK, 0) < 0) {
#else
#ifdef HAVE_FLOCK
	if (fd < 0 || flock(fd, (LOCK_SH|LOCK_NB)) < 0) {
#endif /* HAVE_FLOCK */
#endif /* HAVE_LOCKF */
		(void) printf("\tcannot create status file.\n");
		return (-1);
	}
	(void) ftruncate(fd, 0);

	/*
	 * If they forgot to give a reason, just make the obvious statement.
	 */
	if (argc <= 0) {
		(void) write(fd, "is down\n", 8);
	} else {
		cp1 = line2;
		while (--argc >= 0) {
			cp2 = *argv++;
			while (*cp1++ = *cp2++)
				;
			cp1[-1] = ' ';
		}
		cp1[-1] = '\n';
		*cp1 = '\0';

		(void) write(fd, line2, strlen(line2));
	}
	(void) close(fd);

	if (host_query(lhost, QC_KICK, queue) != 0)
		(void) printf("\tcouldn't ask daemon to drain queue.\n");
	else
		(void) printf("\tqueue will be drained.\n");

	return 0;

}


/*
 * Send SIGALRM to a queue.
 */
int
kick(argc, argv)
int	argc;
char **	argv;
{

	int	c;
	char *	cp1, *cp2;
	char	qbuf[100];


	if (argc == 1) {
		(void) printf("Usage: kick {all | queue ...}\n");
		return (-1);
	}

	if (argc == 2 && !strcmp(argv[1], all)) {
		queue = qbuf;
		while (getbaent(bbuf) > 0) {
			cp1 = qbuf;
			cp2 = bbuf;
			while ((c = *cp2++) && c != '|' && c != ':')
				*cp1++ = c;
			*cp1 = '\0';
			(void) kickq();
			sleep(2);
		}
		return 0;
	}

	while (--argc) {
		queue = *++argv;
		if (openqueue(queue) != 0)
			continue;
		(void) kickq();
	}

	return 0;
}

kickq()
{

	(void) printf("%s:\n", queue);

	if (host_query(lhost, QC_KICK, queue) != 0)
		(void) printf("\tcouldn't kick daemon.\n");
	else
		(void) printf("\tkicked.\n");
}


/*
 * Exit bac
 */
int
quit(argc, argv)
int	argc;
char **	argv;
{

	exit(0);
}


/*
 * Kill and restart the daemon and all current jobs.
 */
int
restart(argc, argv)
int	argc;
char **	argv;
{

	int	c;
	char *	cp1, *cp2;
	char	qbuf[100];


	if (argc == 1) {
		(void) printf("Usage: restart {all | queue ...}\n");
		return (-1);
	}

	if (argc == 2 && !strcmp(argv[1], all)) {
		queue = qbuf;
		while (getbaent(bbuf) > 0) {
			cp1 = qbuf;
			cp2 = bbuf;
			while ((c = *cp2++) && c != '|' && c != ':')
				*cp1++ = c;
			*cp1 = '\0';
			(void) abortq(0);
			(void) startq(0);
		}
		return 0;
	}

	while (--argc) {
		queue = *++argv;
		if (openqueue(queue) != 0)
			continue;
		(void) abortq(0);
		(void) sleep(5);
		(void) startq(0);
	}

	return 0;
}


/*
 * Enable executions on the specified queue and start jobs.
 */
int
start(argc, argv)
int	argc;
char **	argv;
{

	int	c;
	char *	cp1, *cp2;
	char	qbuf[100];


	if (argc == 1) {
		(void) printf("Usage: start {all | queue ...}\n");
		return (-1);
	}

	if (argc == 2 && !strcmp(argv[1], all)) {
		queue = qbuf;
		while (getbaent(bbuf) > 0) {
			cp1 = qbuf;
			cp2 = bbuf;
			while ((c = *cp2++) && c != '|' && c != ':')
				*cp1++ = c;
			*cp1 = '\0';
			(void) startq(1);
		}
		return 0;
	}

	while (--argc) {
		queue = *++argv;
		if (openqueue(queue) != 0)
			continue;
		(void) startq(1);
	}

	return 0;
}


int
startq(enable)
int	enable;
{

	char *	sd, *lf;


	bp = pbuf;

	if ((lf = bgetstr("lf", &bp)) == NULL)
		lf = _DEF_LOCK;
	if (lf[0] != '/') {
		if ((sd = bgetstr("sd", &bp)) == NULL)
			sd = _DEF_SPOOL;
		(void) strcpy(line, sd);
		(void) strcat(line, "/");
		(void) strcat(line, lf);
	} else {
		(void) strcpy(line, lf);
	}

	(void) printf("%s:\n", queue);

	/*
	 * Turn off the owner execute bit of the lock file to
	 * enable executions.
	 */
	if (enable && stat(line, &stbuf) >= 0) {
		if (enable == 3 && 
		    (stbuf.st_mode & 02000) && (stbuf.st_mode & 00010) &&
		    (stbuf.st_mode & 00001)) {
			if (chmod(line, (stbuf.st_mode & 06777) & ~02011) < 0)
				(void) printf("\tcannot reset draining.\n");
			else
				(void) printf("\tnolonger draining.\n");

		} else if (enable == 2 &&
		    ((stbuf.st_mode & 02000) && (stbuf.st_mode & 00010) &&
		     (stbuf.st_mode & 04000) && (stbuf.st_mode & 00100))) {
			if (chmod(line, (stbuf.st_mode & 06777) & ~06110) < 0)
				(void) printf("\tcannot enable submission and execution.\n");
			else
				(void) printf("\tsubmission and execution are enabled.\n");
		} else if ((stbuf.st_mode & 4000) && (stbuf.st_mode & 0100)) {
			if (chmod(line, (stbuf.st_mode & 06777) & ~04100) < 0)
				(void) printf("\tcannot enable execution.\n");
			else
				(void) printf("\texecution is enabled.\n");
		} else
			(void) printf("\texecution not disabled.\n");
	}

	if (host_query(lhost, QC_KICK, queue) != 0)
		(void) printf("\tcouldn't start job queue.\n");
	else
		(void) printf("\tjobs started.\n");

	return 0;
}


/*
 * Print the status of each queue listed or all the queues.
 */
int
status(argc, argv)
int	argc;
char **	argv;
{

	int	c;
	char *	cp1, *cp2;
	char	qbuf[100];


	if (argc == 1) {
		queue = qbuf;
		while (getbaent(bbuf) > 0) {
			cp1 = qbuf;
			cp2 = bbuf;
			while ((c = *cp2++) && c != '|' && c != ':')
				*cp1++ = c;
			*cp1 = '\0';
			(void) statq();
		}
		return 0;
	}
	while (--argc) {
		queue = *++argv;
		if (openqueue(queue) != 0)
			continue;
		(void) statq();
	}
	
	return 0;
}


/*
 * Print the status of the queue queue.
 */
int
statq()
{

	int	fd, i;
	char *	sd, *lf, *sf;
	struct direct * dp;
	DIR *	dirp;


	bp = pbuf;

	(void) printf("%s:\n", queue);

	if ((sd = bgetstr("sd", &bp)) == NULL)
		sd = _DEF_SPOOL;

	if ((lf = bgetstr("lf", &bp)) == NULL)
		lf = _DEF_LOCK;
	if (lf[0] != '/') {
		(void) strcpy(line, sd);
		(void) strcat(line, "/");
		(void) strcat(line, lf);
	} else {
		(void) strcpy(line, lf);
	}

	if (stat(line, &stbuf) >= 0) {
		(void) printf("\tsubmission is %s %s.\n",
		        (stbuf.st_mode & 02000) ? "externally" : "internally",
			(stbuf.st_mode & 00010) ? "disabled" : "enabled");
		if (stbuf.st_mode & 00001) 
			(void) printf("\texecution is draining.\n");
		else
			(void) printf("\texecution is %s %s.\n",
				      (stbuf.st_mode & 04000) ? "externally" : "internally",
				      (stbuf.st_mode & 00100) ? "disabled" : "enabled");
	} else {
		(void) printf("\tsubmission is enabled.\n");
		(void) printf("\texecution is enabled.\n");
	}

	if ((dirp = opendir(sd)) == NULL) {
		(void) printf("\tcannot examine spool directory.\n");
		return (-1);
	}

	i = 0;

	while ((dp = readdir(dirp)) != NULL) {
		if (*dp->d_name == 'c' && dp->d_name[1] == 'f')
			i++;
	}

	(void) closedir(dirp);
	if (i == 0)
		(void) printf("\tno entries.\n");
	else if (i == 1)
		(void) printf("\t1 entry in spool area.\n");
	else
		(void) printf("\t%d entries in spool area.\n", i);

	/*
	 * If modified by bac, show the message...
	 */
	if ((stbuf.st_mode & 02000) || (stbuf.st_mode & 04000)) {
		if ((sf = bgetstr("sf", &bp)) == NULL)
			sf = _DEF_STAT;
		if (strchr(sf, '/') == NULL) {
			(void) strcpy(line, sd);
			(void) strcat(line, "/");
			(void) strcat(line, sf);
		} else {
			(void) strcpy(line, sf);
		}

		fd = open(line, O_RDONLY);
		if (fd >= 0) {

#ifdef HAVE_LOCKF
			(void) lockf(fd, F_TLOCK, 0);
#else
#ifdef HAVE_FLOCK
			(void) flock(fd, LOCK_SH);
#endif /* HAVE_FLOCK */
#endif /* HAVE_LOCKF */
			putchar('\t');
			while ((i = read(fd, line, sizeof(line))) > 0)
				(void) fwrite(line, 1, i, stdout);
			(void) close(fd);	/* unlocks as well */
		}
	}

	return 0;

}


/*
 * Stop the specified daemon after completing the current job and disable
 * executions.
 */
int
stop(argc, argv)
int	argc;
char **	argv;
{

	int	c;
	char *	cp1, *cp2;
	char	qbuf[100];

	if (argc == 1) {
		(void) printf("Usage: stop {all | queue ...}\n");
		return (-1);
	}

	if (argc == 2 && !strcmp(argv[1], all)) {
		queue = qbuf;
		while (getbaent(bbuf) > 0) {
			cp1 = qbuf;
			cp2 = bbuf;
			while ((c = *cp2++) && c != '|' && c != ':')
				*cp1++ = c;
			*cp1 = '\0';
			(void) stopq();
		}
		return 0;
	}

	while (--argc) {
		queue = *++argv;
		if (openqueue(queue) != 0)
			continue;
		(void) stopq();
	}

	return 0;
}


int
stopq()
{
	
	char *	sd, *lf;
	int	fd;


	bp = pbuf;

	(void) printf("%s:\n", queue);

	if ((lf = bgetstr("lf", &bp)) == NULL)
		lf = _DEF_LOCK;
	if (lf[0] != '/') {
		if ((sd = bgetstr("sd", &bp)) == NULL)
			sd = _DEF_SPOOL;
		(void) strcpy(line, sd);
		(void) strcat(line, "/");
		(void) strcat(line, lf);
	} else {
		(void) strcpy(line, lf);
	}

	/*
	 * Turn on the owner execute bit of the lock file to
	 * disable executions.
	 */
	if (stat(line, &stbuf) >= 0) {
		if (chmod(line, (stbuf.st_mode & 06777) | 04100) < 0)
			(void) printf("\tcannot disable execution.\n");
		else 
			(void) printf("\texecution is disabled.\n");
	} else if (errno == ENOENT) {
		if ((fd = open(line, (O_WRONLY|O_CREAT), 04766)) < 0)
			(void) printf("\tcannot create lock file.\n");
		else {
			(void) close(fd);
			(void) printf("\texecution is disabled.\n");
		}
	} else
		(void) printf("\tcannot stat lock file.\n");

	if (host_query(lhost, QC_KICK, queue) != 0)
		(void) printf("\tcouldn't stop job queue.\n");
	else
		(void) printf("\tqueue stopped.\n");

	return 0;

}

time_t mtime = 0;

/*
 * Put the specified jobs at the top of the queue.
 */
topq(argc, argv)
int	argc;
char **	argv;
{

	char *	sd, *lf;
	int	i, changed = 0;
	struct direct * d;
	DIR *	qdir;
	time_t	cfmtime;


	mtime = time((long *) 0); /* set to now */

	if (argc < 3) {
		(void) printf("Usage: topq queue [jobnum ...] [user ...]\n");
		return (-1);
	}

	--argc;
	queue = *++argv;
	if (openqueue(queue) != 0)
		return (-1);

	bp = pbuf;
	if ((sd = bgetstr("sd", &bp)) == NULL)
		sd = _DEF_SPOOL;

	(void) printf("%s:\n", queue);

	if (chdir(sd) < 0) {
		(void) printf("\tcannot chdir to %s.\n", sd);
		return (-1);
	}

	if ((lf = bgetstr("lf", &bp)) == NULL)
		lf = _DEF_LOCK;

	(void) sprintf(line, "%s/%s", sd, lf);

	if ((qdir = opendir(".")) == NULL) {
		blog(BLOG_ERR, eopen, sd);
		return (-1);
	}


	while ((d = readdir(qdir)) != NULL) {

		/*
		 * We're only looking for "cf" files
		 */
		if (d->d_name[0] != 'c' || d->d_name[1] != 'f')
			continue;

		if (cfmtime = getmtime(d->d_name) < 0) {
			continue;
		}

		/*
		 * Find the earliest queued time 
		 */
		if (cfmtime < mtime)
			mtime = cfmtime;
	}

	(void) rewinddir(qdir);

	while ((d = readdir(qdir)) != NULL) {

		/*
		 * We're only looking for "cf" files
		 */
		if (d->d_name[0] != 'c' || d->d_name[1] != 'f')
			continue;

		for (i = argc; --i; ) {
			if (checkarg(d, argv[i]) == 0) {
				continue;
			} else
				changed++;
		}
	}

	(void) closedir(qdir);

	/*
	 * Set the modification time of the directory, and kick the daemon.
	 */
	if (changed) {

		struct utimbuf utb;
		utb.actime = utb.modtime = time((time_t *) 0);

		if (utime(".", &utb) < 0)
			(void) printf("\tcouldn't set directory utimes.\n");

		if (host_query(lhost, QC_KICK, queue) != 0)
			(void) printf("\tcouldn't notify the daemon.\n");
		else
			(void) printf("\tqueue order changed.\n");
	} else {
		(void) printf("\tqueue order unchanged.\n");

	}

	return 0;

} 



/*
 * Checks if specified job name is in the queue's queue.
 * Returns:  negative (-1) if argument name is not in the queue.
 */
static int
checkarg(d, job)
struct direct *	d;
char *	job;
{

	char *	cp, *machine;
	int	jobnum, n;
	int	cnt = 0;
	FILE *	fp;


	/*
	 * Look for a job item consisting of system name, colon, number 
	 * (example: ucbarpa:114)  
	 */
	if ((cp = strchr(job, ':')) != NULL) {
		machine = job;
		*cp++ = '\0';
		job = cp;
	} else
		machine = NULL;

	/*
	 * Check for job specified by number (example: 112 or 235ucbarpa)
	 */
	if (isdigit((int) *job)) {
		jobnum = 0;
		do
			jobnum = jobnum * 10 + (*job++ - '0');
		while (isdigit((int) *job));
		n = 0;
		for (cp = d->d_name+2; isdigit((int) *cp); )
			n = n * 10 + (*cp++ - '0');
		if (jobnum != n)
			return (cnt);
		if (*job && strcmp(job, cp) != 0)
			return (cnt);
		if (machine != NULL && strcmp(machine, cp) != 0)
			return (cnt);
		if (touch(d) == 0) {
			(void) printf("\tmoved %s.\n", d->d_name);
			cnt++;
		}
		return(cnt);
	}

	/*
	 * Process item consisting of owner's name (example: henry).
	 */
	if ((fp = fopen(d->d_name, "r")) == NULL)
		return (cnt);

	while (fgets(line, sizeof(line), fp) != NULL)
		if (line[0] == 'U')
			break;

	(void) fclose(fp);

	if (line[0] != 'U' || strcmp(job, line+1) != 0)
		return (cnt);

	if (touch(d) == 0) {
		(void) printf("\tmoved %s.\n", d->d_name);
		cnt++;
	}

	return(cnt);
}


/*
 * Reposition the job by changing the modification time of
 * the control file and setting the user execute bit.
 */
static int
touch(d)
struct direct * d;
{

	struct utimbuf	utb;

	/*
	 * Turn on the owner execute bit of the cf file to mark as topqueued
	 */
	if (stat(d->d_name, &stbuf) >= 0) {
		if (chmod(d->d_name, (stbuf.st_mode & 06777) | 00100) < 0)
			(void) printf("\tcannot set mode on \"%s\".\n",
				      d->d_name);
		else 
			(void) printf("\tmarked \"%s\".\n", d->d_name);
	} else
		(void) printf("\tcannot stat \"%s\".\n", d->d_name);

	utb.actime = utb.modtime = --mtime;

	return(utime(d->d_name, &utb));
}

/*
 * Enable everything and start queue (undo `down').
 */
int
up(argc, argv)
int	argc;
char **	argv;
{

	int	c;
	char *	cp1, *cp2;
	char	qbuf[100];


	if (argc == 1) {
		(void) printf("Usage: up {all | queue ...}\n");
		return 0;
	}

	if (argc == 2 && !strcmp(argv[1], all)) {
		queue = qbuf;
		while (getbaent(bbuf) > 0) {
			cp1 = qbuf;
			cp2 = bbuf;
			while ((c = *cp2++) && c != '|' && c != ':')
				*cp1++ = c;
			*cp1 = '\0';
			(void) startq(2);
		}
		return 0;
	}

	while (--argc) {
		queue = *++argv;
		if (openqueue(queue) != 0)
			continue;
		(void) startq(2);
	}

	return 0;
}


/*
 * Stop draining a queue (undo `drain').
 */
int
continu(argc, argv)
int	argc;
char **	argv;
{

	int	c;
	char *	cp1, *cp2;
	char	qbuf[100];


	if (argc == 1) {
		(void) printf("Usage: continue {all | queue ...}\n");
		return 0;
	}

	if (argc == 2 && !strcmp(argv[1], all)) {
		queue = qbuf;
		while (getbaent(bbuf) > 0) {
			cp1 = qbuf;
			cp2 = bbuf;
			while ((c = *cp2++) && c != '|' && c != ':')
				*cp1++ = c;
			*cp1 = '\0';
			(void) startq(3);
		}
		return 0;
	}

	while (--argc) {
		queue = *++argv;
		if (openqueue(queue) != 0)
			continue;
		(void) startq(3);
	}

	return 0;
}


static int
openqueue(q)
char *	q;
{

	int	rtn;


	if ((rtn = bgetent(bbuf, q)) < 0) {
		(void) printf("cannot open queue description file.\n");
		return (-1);
	} else if (rtn == 0) {
		(void) printf("%s: unknown queue.\n", q);
		return (-1);
	}
	
	return 0;

}


int
reply_ok(fp, fmt, rqueue)
FILE *	fp;
char *	fmt;
char *	rqueue;
{

	return 0;
}


int
reply_mt(fp, fmt, rqueue)
FILE *	fp;
char *	fmt;
char *	rqueue;
{

	return 0;
}


