#ifndef lint
static char rcsid[] =
	"$Id: console.c,v 4.0 1993/09/17 00:41:04 srp Exp srp $";
#endif

/*
 *
 * Author: Scott R. Presnell
 *
 * Copyright (C) 1989, 1990, The Regents of the University of California
 *                     All Rights Reserved
 *
 * THE REGENTS OF THE UNIVERSITY OF CALIFORNIA MAKE NO REPRESENTATION OR
 * WARRANTIES WITH RESPECT TO THE CONTENTS HEREOF AND SPECIFICALLY DISCLAIM
 * ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
 * PURPOSE.
 *
 * Further, the Regents of the University of California reserve the right to
 * revise this software and/or documentation and to make changes from time
 * to time in the content hereof without obligation of the Regents of the
 * University of California to notify any person of such revision or change.
 *
 */

/*
 * Sending mail back to the users...
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <ctype.h>

#include <limits.h>

#if defined(MALLOC_TRACE) || defined(MALLOC_LEAK)
#include "dmalloc.h"
#else
#include <malloc.h>
#endif

#include <pwd.h>
#include <grp.h>
#include <utmp.h>


#include <sys/types.h>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/resource.h>

#include "config.h"

#if (defined(I_STDARG) && defined(CAN_VAPROTO)) || defined(__STDC__)
#include <stdarg.h>
#else
#include <varargs.h>
#endif

#include H_STRING

#include "batch.h"
#include "batchd.h"
#include "event.h"
#include "flags.h"
#include "pathnames.h"

static struct utmp utrecord;

#ifdef USE_LOCALBUF
static char	utmpiobuf[BUFSIZ];
static char	myconbuf[LINE];
#endif	/* USE_LOCALBUF */

#ifndef UTMP_FILE
#define	UTMP_FILE	"/etc/utmp"
#endif

#ifdef CONSOLE
char *	console = CONSOLE;
#else
char *	console = _PATH_CONS;
#endif

char *	consoleinfo = NULL;

Rc	rcfile;				/* Resource file of console user */
int	cevent = 0;			/* Id of the console test event */

static int	cuid;			/* Uid of the console user */
static int	cgid;			/* Gid of the console user */
static int	last_user_cuid = 0;

static time_t	consactive P((char *, char *, char [], time_t));
static int	setrcopts P((char *));
static State	decodercopts P((State, char *));
static void	clear_uctl P((void));

void		do_consolecheck P((void));

extern time_t	graphactive P((time_t));

/*
 * Check the console device(s) for activity
 */
void
do_consolecheck()
{

	Queue *	qp;
	static int	last_cuid = 0;
	int	rcstate = 0;
	char	cuser[12];
	static char lcuser[12];
	char	rcfilename[80];
	time_t	idle = 0;
	time_t	mtime = 0;
	static double oload5min;
	static	int first = 1;
	struct passwd *	pwent;

	/* Clear the console user */
	cuser[0] = '\0';

	/* Clear the console event */
	cevent = 0;

	/* Get the current time */ 
	now = time((time_t *)0);

	/*
	 * Set to current time if never been set.
	 */	
	if (first) {
		lcuser[0] = '\0';
		last_cuid = -1;

		first = 0;
	}

	/* Get the current load averages... */
	(void) load_get(&loadave[0]);

	/* ...and print them to the log file. */
	dbg((BLOG_RESOURCE, "loadave 1min: %4.2lf 5min: %4.2lf 15min: %4.2lf",
	     loadave[0], loadave[1], loadave[2]));

	/*
	 * This function call not only returns the idle time,
	 * but also places  the console user in the variable "cuser."
	 * Seconds of idle time  returned.
	 */
	idle = consactive(console, consoleinfo, cuser, now);

	/* If there is a console user */
	if (cuser[0]) {

		/* If that user is not the old user */
		if (strcmp(cuser, lcuser)) {

			/* get the password entry and set struct rcfile  */
			if ((pwent = getpwnam(cuser)) != NULL)  {

				cuid = (int) pwent->pw_uid;
				cgid = (int) pwent->pw_gid;

				(void) strcpy(rcfilename, pwent->pw_dir);
				(void) strcat(rcfilename, "/.batchrc");

				rcfile.r_user = estrdup(cuser);
				rcfile.r_name = estrdup(rcfilename);
			} else {
				/* What ? no password entry? */
				dbg((BLOG_RESOURCE, "%s: no passwd entry",
				     cuser));
				rcstate = 0;
			}

			/* Save the console username, and uid */
			(void) strcpy(lcuser, cuser);
			last_cuid = cuid;

		}

		/* try to access the file */
		if (access(rcfile.r_name, F_OK) == 0)
			mtime = getmtime(rcfile.r_name);
		/* else mtime = 0; stays */

		/*
		 * If there is an mtime and it's more recent
		 * clear the user controls and (re)read the file.
		 */
		if (mtime > rcfile.r_mtime) {
			rcfile.r_mtime = mtime;
			clear_uctl();
			(void) setrcopts(rcfile.r_name);
			blog(BLOG_INFO, "read %s .batchrc", cuser);
			rcstate = 2;

		/* The file is there but unchanged or less time. */
		} else if (mtime > 0) {
			dbg((BLOG_RESOURCE, ".batchrc file unchanged"));
			rcstate = 1;

		/* There is no resource file. */
		} else /* mtime <= 0 */ {
			dbg((BLOG_RESOURCE, "no .batchrc file"));
			rcstate = 0;
		}

	} else {

		dbg((BLOG_RESOURCE, "no console user"));
		rcstate = 0;
		cuid = -1;
	}

	if (!rcstate)
		clear_uctl();

	/*
	 * Setting "forced" causes signals to be resent to jobs
	 * in queues that should be disabled.  This is how user
	 * jobs get stopped when they should after the user with
	 * a "nostopme" has logged out.
	 */
	if (cuid != last_cuid || rcstate > 1) {
		dbg((BLOG_RESOURCE, "change of user -> forced alarm"));
		alarming(-1);
	}

	/*
	 * We also want to check the loadave on a reasonable schedule.
	 * cause an alarm if the load has changed appreciably.
	 */
	if (ABS(oload5min - loadave[LOAD_5MIN]) > 0.5) {
		alarming(0);
		oload5min = loadave[LOAD_5MIN];
	}


	/* Save the uid of the last person to use the console. */
	if (cuid == -1 && last_cuid >= 0) {

		/* There is no console user - nuke the rcfile data */
		efree((voidp) rcfile.r_name);
		rcfile.r_name = NULL;
		efree((voidp) rcfile.r_user);
		rcfile.r_user = NULL;
		rcfile.r_mtime = 0;

		blog(BLOG_INFO, "user %s logged out", lcuser);
		last_user_cuid = last_cuid;
		last_cuid = -1;
		strcpy(lcuser, unk);
	}


	ITER(qp, Queue, &queues) {

		/* If this is an empty queue then skip. */
		if (qp->q_name == NULL)
			continue;

		/* A user is logged on and requires a dedicated console. */
		if (idle >= 0 && ((qp->q_uctl & ANYCONS) && qp->q_userrc)) {
			if (qp->q_inactive == TRUE) {
				alarming(0);
				dbg((BLOG_RESOURCE, ques, qp->q_name,
				     "console requested"));
			}
			qp->q_inactive = FALSE;

		/* there is no console checking done */
		} else if (qp->q_idle < 0) {
			if (qp->q_inactive == FALSE)
				dbg((BLOG_RESOURCE, ques, qp->q_name,
				     "console carefree"));
			qp->q_inactive = TRUE;		/* just to be safe */

		/* noone is logged onto the console */
		} else if (idle < 0) {
			if (qp->q_inactive == FALSE) {
				alarming(0);
				dbg((BLOG_RESOURCE, ques, qp->q_name,
				     "console is free"));
			}
			qp->q_inactive = TRUE;

		/* there is a console user, but they are not working */
		} else if (idle > qp->q_idle) {
			if (qp->q_inactive == FALSE) {
				alarming(0);
				dbg((BLOG_RESOURCE, ques, qp->q_name,
				     "console quiescent"));
			}
			qp->q_inactive = TRUE;

		/* the console user is working */
		} else /* if (idle <= qp->q_idle) */ {
			if (qp->q_inactive == TRUE) {
				alarming(0);
				dbg((BLOG_RESOURCE, ques, qp->q_name,
				     "console active"));
			}
			qp->q_inactive = FALSE;
		}



	}

	if (cevent == 0)
		cevent = event_add(stimer + getrand(-2, 2),
				   do_consolecheck, (voidp) NULL,
				   "console check");

	return;
}


static time_t
consactive(cons, icons, cuser, current)
char *	cons;
char *	icons;
char	cuser[];
time_t	current;
{

	char *	utmpfile = UTMP_FILE;
	char *	conline;
	char	cline[24];
	Bool	foundconsole = FALSE;	/* Console login located */
	FILE *	consf = NULL;
	static FILE *	ut = NULL;
	time_t	considle, graphidle;
	time_t	lowidle;	/* Time differences */
	static time_t	 lasttime = 0L;		/* Last time console checked */
	static time_t	 consused;		/* Last time console in use */


	if (!lasttime)
		lasttime = current - ptimer; /* fudge */

	consused =  0;

	lowidle  = SECPERDAY;
	graphidle = SECPERDAY;

	graphidle = graphactive(current);

	if (ut == NULL) {

		if ((ut = fopen(utmpfile, "r")) == NULL) {
			blog(BLOG_ERR, eopen, utmpfile);
			return 0;
		}

#ifdef USE_LOCALBUF
		(void) setbuf(ut, utmpiobuf);
#endif /* USE_LOCALBUF */

	}
	(void) rewind(ut);

	cuser[0] = '\0';
	cline[0] = '\0';

	/* Set the console line name if given as a path. */
	if ((conline = strrchr(cons, '/')) == NULL)
		conline = cons;
	else
		conline++;

	/* If there is a file to read for the console user, do it here. */
	if (icons) {

		dbg((BLOG_RESOURCE, "reading \"%s\"", icons));
		
		if ((consf = fopen(icons, "r")) != NULL) {
#ifdef USE_LOCALBUF
			(void) setbuf(consf, myconbuf);
#endif /* USE_LOCALBUF */

			(void) fgets(cuser, 10, consf);
			cuser[strlen(cuser) - 1] = '\0';

			(void) fclose(consf);

			if (!isalnum((int) cuser[0]))
				cuser[0] = '\0';
			else
				foundconsole = TRUE;

			dbg((BLOG_RESOURCE, "\"%s\": cuser set to %s",
			     icons, cuser));

		} else {
			dbg((BLOG_RESOURCE, eopen, icons));
			cuser[0] = '\0';
		}
	}

	/*
	 * First look for the console user if not already set.
	 */
	if (cuser[0] == '\0') {
		while (fread(&utrecord, sizeof(utrecord), 1, ut) != 0) {

			if (utrecord.ut_name[0] == '\0')
				continue;		/* That tty is free */

#ifdef USER_PROCESS
			if (utrecord.ut_type != USER_PROCESS)
				continue;
#endif

			dbg((BLOG_RESOURCE, "user: %s; line %s",
			     utrecord.ut_name, utrecord.ut_line));

			if (strncmp(conline, utrecord.ut_line,
				    strlen(utrecord.ut_line)) == 0) {
				(void) strcpy(cuser, utrecord.ut_name);
				foundconsole = TRUE;
			}
		}
	}
	rewind(ut);

	/*
	 * Now scan for tty's they might be on, record the most active.
	 */
	while (fread(&utrecord, sizeof(utrecord), 1, ut) != 0) {
	
		consused = 0;

		if (utrecord.ut_name[0] == '\0')
			continue;		/* That tty is free */

#ifdef USER_PROCESS
		if (utrecord.ut_type != USER_PROCESS)
			continue;
#endif

		/*
		 * Is this the console user? If so, get the mtime for
		 * that tty.
		 */
		if (cuser[0] != '\0' &&
		    strncmp(cuser, utrecord.ut_name,
			    strlen(utrecord.ut_name)) == 0) {

			(void) strcpy(line, "/dev/");
			(void) strcat(line, utrecord.ut_line);

			/*
			 * If a valid mtime and the idle time is the lowest,
			 * save away the idle time, and the line it was on.
			 */
			if ((consused = getmtime(line)) > 0) {

				considle = current - consused;

				dbg((BLOG_RESOURCE, "cuser: %s; idle: %d (line %s)",
				     cuser, considle, utrecord.ut_line));
				
				if (considle < 0)
					considle = 0L;

				if (considle < lowidle) {
					lowidle = considle;
					(void) strcpy(cline, utrecord.ut_line);
				}
			}
		}			
	}


	if (foundconsole) {

		dbg((BLOG_RESOURCE,
		     "cuser: %s; cidle: %ld (line %s); gidle: %d",
		     cuser, lowidle, cline, graphidle));

		return MIN(lowidle, graphidle);

	}

	dbg((BLOG_RESOURCE, "console not in use"));

	return (-1);				/* Logged out */

}

static int
setrcopts(filename)
char *	filename;
{
	char	cmd[40];
	char	var[40];
	char	state[80];
	char	bbuf[BUFFER];
	int	status;
	Bool	setall = FALSE; 
	Queue  *qp;


	while (scanrcline(filename, cmd, var, state) != NULL) {

		setall = FALSE; 

		if (strcasecmp(var, "all") == 0)
			setall = TRUE;
		/*
		 * Take care of using queue name aliases...
		 * Look for the specified queue name
		 */
		if ((status = bgetent(bbuf, var)) < 0) {
			blog(BLOG_WARNING,
			     "cannot open queue description file");
		} else if (status == 0) {
			blog(BLOG_WARNING, "status == 0");
			/* No such queue, scan next line */
			if (!setall)
				continue;
		}

		/* Reset the queue name to the canonical (first) name */
		(void) strcpy(var, bnamget(bbuf));

		blog(BLOG_TRACE, "resource: \"%s\" state: \"%s\"\n",
		    setall ? "all" : var, state);

		ITER(qp, Queue, &queues) {

			/* If this is an empty queue then skip. */
			if (qp->q_name == NULL)
				continue;

			if (strcasecmp(qp->q_name, var) == 0 || setall) {
				qp->q_uctl = decodercopts(qp->q_uctl, state);
				/* cannot use break; here if using setall */

			}
		}
	}

	endrcent();

	return 0;
}

	


static State
decodercopts(val, opts)
State	val;
char *	opts;
{


	if (strstr(opts, "noconsole") != NULL) 
		val &= ~(INTERRUPT);
	else if (strstr(opts, "anyconsole") != NULL)
		val |= ANYCONS;
	else if (strstr(opts, "console") != NULL)
		val |= INTERRUPT;

	if (strstr(opts, "noload") != NULL) 
		val &= ~(SUSPEND);
	else if (strstr(opts, "load") != NULL)
		val |= SUSPEND;

	if (strstr(opts, "notime") != NULL) 
		val &= ~(TIMEOUT);
	else if (strstr(opts, "time") != NULL)
		val |= TIMEOUT;

	if (strstr(opts, "nostopme") != NULL) 
		val &= ~(STOPPED);
	else if (strstr(opts, "stopme") != NULL)
		val |= STOPPED;

	return val;
}


/*
 * Clear the user originated queue control
 */
static void
clear_uctl()
{

	Queue *	qp;

	ITER(qp, Queue, &queues) {
		qp->q_uctl =  CLEAR;
		qp->q_uctl |= ANYRUN;
		qp->q_uctl |= ANYSUB;
	}

	return ;
}



