#ifndef lint
static char rcsid[] =
	"$Id: displayjob.c,v 4.0 1993/09/17 00:41:04 srp Exp srp $";
#endif

/*
 *
 * Author: Scott R. Presnell
 *
 * Copyright (C) 1989, 1990, The Regents of the University of California
 *                     All Rights Reserved
 *
 * THE REGENTS OF THE UNIVERSITY OF CALIFORNIA MAKE NO REPRESENTATION OR
 * WARRANTIES WITH RESPECT TO THE CONTENTS HEREOF AND SPECIFICALLY DISCLAIM
 * ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
 * PURPOSE.
 *
 * Further, the Regents of the University of California reserve the right to
 * revise this software and/or documentation and to make changes from time
 * to time in the content hereof without obligation of the Regents of the
 * University of California to notify any person of such revision or change.
 *
 */

/*
 * displayjob.c - display job queue
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <fcntl.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/file.h>
#include <sys/resource.h>

#include "config.h"

#include H_STRING

#include "batch.h"
#include "batchd.h"
#include "flags.h"
#include "protocol.h"
#include "reply.h"

/*
 * Report the current status of the queue and provide information on the
 * state of the jobs within.
 */
int
queue_display(queue)
char *queue;
{

	Job *jp;
	Queue *qp;
	int	first = 0;
	int	state = 0;
	int	sfd = 0;
	int	n;

	
	if ((qp = getqueuebyname(queue)) == (Queue *) NULL) {
		(void) fputc(QR_ENOENT, stdout);
		return (RTN_ERROR);
	}

	(void) strcpy(line, qp->q_spool);
	(void) strcat(line, "/");
	(void) strcat(line, qp->q_sname);

	if ((sfd = open(line, O_RDONLY)) < 0) {
		blog(BLOG_WARNING, eopen, line);
		(void) sprintf(line,
			       "%s: status file could not be opened\n",
			       lhost);
	} else {

		if ((n = read(sfd, line, sizeof(line))) <= 0) {
			blog(BLOG_WARNING,
			     "%s: could not get status line", qp->q_name);
			(void) sprintf(line, "%s: could not get status line\n",
				       lhost);
		}
		line[n] = '\0';
		(void) close(sfd);
	}
	
	ITER(jp, Job, &(qp->q_jobs))
		jp->j_seen = FALSE;

	/*
	 * For each active, job in the queue...
	 */
	ITER(jp, Job, &(qp->q_jobs)) {

		if (jp->j_file == NULL)
			continue ;

		if (!first) {
			ack(); /* OK signal */
			(void) ewrite(1, line, strlen(line), (char *) NULL);
		}

		first++;

		/*
		 * Set the state of the queue for that job.
		 */
		if (queue_get_flag(qp, STOPPED) == 0) {
			state = 4;		/* Stopped or queued */
		} else if (queue_get_flag(qp, TIMEOUT) == 0) {
			state = 3;		/* Timeout or queued */
		} else if (queue_get_flag(qp, INTERRUPT) == 0) {
			state = 2;		/* Intrupt or queued */
		} else if (queue_get_flag(qp, SUSPEND) == 0) {
			state = 1;		/* Suspend or queued */
		} else {
			state = 0;		/* Running or queued */
		}

		(void) sprintf(line2, DISP_FMT,
			      jp->j_id,
			      (jp->j_jname) ? jp->j_jname : unk,
			      (jp->j_umaddress) ? jp->j_umaddress : unk,
			      jp->j_state,
			      jp->j_ctime,
			      jp->j_stime,
			      (jp->j_delay.d_string) ?
			       jp->j_delay.d_string : "0",
			      jp->j_cputime,
			      jp->j_pid,
			      state);

		(void) ewrite(1, line2, strlen(line2), "displayjob()");
		(void) ewriteone(1, '\n');

	}

	if (!first) {
		mack();
		(void) ewrite(1, line, strlen(line), "displayjob()");
	}

	return (RTN_OK);
}

