#ifndef lint
static char rcsid[] =
	"$Id: getpment.c,v 4.0 1993/09/17 00:41:39 srp Exp srp $";
#endif

/*
 *
 *
 * Copyright (C) 1989, 1990, The Regents of the University of California
 *                     All Rights Reserved
 *
 * THE REGENTS OF THE UNIVERSITY OF CALIFORNIA MAKE NO REPRESENTATION OR
 * WARRANTIES WITH RESPECT TO THE CONTENTS HEREOF AND SPECIFICALLY DISCLAIM
 * ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
 * PURPOSE.
 *
 * Further, the Regents of the University of California reserve the right to
 * revise this software and/or documentation and to make changes from time
 * to time in the content hereof without obligation of the Regents of the
 * University of California to notify any person of such revision or change.
 *
 */

/* 
 * Path map functions:
 * 	void setpment(void)		- rewind the file
 * 	void endpment(void)		- close the file.
 * 	struct pm * getpment(void)	- get next entry.
 * 	struct pm * getpmnam(char *)	- get entry by name.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>

#include <sys/types.h>
#include <sys/param.h>

#define INIT		char *	sp = instring;
#define GETC()		(*sp++)
#define PEEKC()		(*sp)
#define UNGETC(c)	(--sp)
#define RETURN(c)	return;
#define ERROR(c)	pmerror(c);

#include "config.h"

#include H_STRING

#include "batch.h"
#include "pathnames.h"
#include "pm.h"

extern void	pmerror P((int));

#include <regexp.h>


#define MAXEXP		256

int _pm_stayopen = 0;

static char *	_pm_file =  _PATH_PATHMAP;

static FILE *	pmf = NULL;
static char _pm_buf[BUFSIZ+1];
static struct pathm pm;


void
setpmfile(newfile)
char *	newfile;
{
	_pm_file = newfile;

	return;
}

void
setpment()
{

	if (pmf == NULL)
		pmf = fopen(_pm_file, "r");
	else
		rewind(pmf);

	return;
}


void
endpment()
{

	if (pmf != NULL) {
		(void) fclose(pmf);
		pmf = NULL;
	}

	return;
}


struct pathm *
getpment()
{

	char *	cp, *ep, *np;
	static char	_pm_expr[MAXPATHLEN];
	static char	_pm_cname[MAXPATHLEN];
	static char	_pm_comp[MAXEXP];

	if (pmf == NULL) {
		if ((pmf = fopen(_pm_file, "r")) == NULL)
			return 0;
	}

 again:
	if (fgets(_pm_buf, BUFSIZ, pmf) == NULL)
		return 0;
	
	/*
	 * If the first character is a comment, skip the line
	 */
	if (_pm_buf[0] == '#' || _pm_buf[0] == '\n')
		goto again;

	/*
	 * Chop off the newline character
	 */
	_pm_buf[strlen(_pm_buf) - 1] = '\0';

	cp = _pm_buf;
	ep = _pm_expr;
	np = _pm_cname;

	/* Skip leading white space */
	while (cp && isspace(*cp))
		cp++;
			
	/* get first string */
	while (cp && !isspace(*cp))
		*ep++ = *cp++;

	*ep = '\0';

	/* Skip intermediate white space */
	while (cp && isspace(*cp))
		cp++;

	/* get second string */
	while ((*np++ = *cp++) != '\0')
		;

	compile (_pm_expr, _pm_comp, &_pm_comp[MAXEXP], '\0');

	pm.pm_circf = circf;
	pm.pm_expr  = _pm_expr;
	pm.pm_cname = _pm_cname;
	pm.pm_comp  = _pm_comp;


	return (&pm);
}
	
struct pathm *
getpmnam(path)
char *	path;
{

	struct pathm *	pmp;

	setpment();
	while (pmp = getpment()) {
		circf = pmp->pm_circf;
		if (step(path, pmp->pm_comp)) {
			if (!_pm_stayopen)
				endpment();
			return (pmp);
		}
	}

	return (struct pathm *) 0;
}


/*
 * pathmap -  recognize a path by a regex and transmute into a 
 * canonical name. Handle \(xx\) -> \1 transformations.
 * 	called like strcpy(to, from);
 * Make sure there is space in the "to" buffer.
 */
char *
pathmap(to, from)
char *	to;
char *	from;
{

	char *	cp, *bp, *tp;
	int	i;
		
	struct pathm *pmp;

	tp = to;

	if (to == NULL)
		return NULL;

	if ((pmp = getpmnam(from)) == NULL) 
		return strcpy(to, from);

	/*
	 * Copy up to the first match point 
	 */
	for (cp = from; cp < loc1; cp++)
		*to++ = *cp;

	for (cp = pmp->pm_cname; *cp; cp++) {
		/*
		 * This section of the code  will do a regular expression
		 * replacement from \(xxx\) in the expr to \1 in the cname
		 */
		if (*cp == '\\') {	/* could be a \# */
			cp++;
			/* find the matching string */
			if (*cp >= '1' && *cp <= '9') {
				i = *cp - '0';
				for (bp = braslist[i-1];
				     bp < braelist[i-1] && *bp; bp++)
					*to++ = *bp;
			/* 
			 * Nope, it was a literal
			 */
			} else
				*to++ = *cp;
		/*
		 * Normal char
		 */
		} else
			*to++ = *cp;
	}

	/*
	 * Copy the rest of the string into place
	 */
	(void) strcpy(to, loc2);
	
	/*
	 * Return the beginning of to string 
	 */
	return tp;
}
		
void
pmerror(c)
int	c;
{
#ifdef TEST	
	(void) fprintf(stderr, "pm regular epression error #%d", c);
#else
	blog(BLOG_WARNING, "pm regular epression error #%d", c);
#endif

	return ;
}

#ifdef TEST

main (argc, argv)
int	argc;
char **	argv;
{

	char	npath[MAXPATHLEN];


	if (argc != 2) {
		(void) fprintf(stderr, "provide an argument\n");
		exit(1);
	}
  
	setpmfile ("/usr/local/lib/wd.cf");

	(void) pathmap (npath, argv[1]);
	(void) printf ("%s -> %s\n", argv[1], npath);
	exit (0);
}

#endif
