#ifndef lint
static char rcsid[] =
	"$Id: getsocket.c,v 4.0 1993/09/17 00:41:39 srp Exp srp $";
#endif

/*
 *
 * Author: Scott R. Presnell
 *
 * Copyright (C) 1989, 1990, The Regents of the University of California
 *                     All Rights Reserved
 *
 * THE REGENTS OF THE UNIVERSITY OF CALIFORNIA MAKE NO REPRESENTATION OR
 * WARRANTIES WITH RESPECT TO THE CONTENTS HEREOF AND SPECIFICALLY DISCLAIM
 * ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
 * PURPOSE.
 *
 * Further, the Regents of the University of California reserve the right to
 * revise this software and/or documentation and to make changes from time
 * to time in the content hereof without obligation of the Regents of the
 * University of California to notify any person of such revision or change.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <ctype.h>

#include <errno.h>
#include <fcntl.h>

#include <sys/types.h>
#include <sys/param.h>

#include <sys/socket.h>
#include <sys/uio.h>
#include <sys/un.h>

#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>

#include "config.h"

#include H_STRING

#include "batch.h"
#include "pathnames.h"


#ifdef SOCKETNAME				/* local socket name */
char *	socketname = SOCKETNAME;
#else
char *	socketname = _PATH_SOCKET;
#endif

#ifdef USE_RRESVPORT
extern int 	rresvport P((int *));
#endif

extern long	gethostid P((void));

static char econn[] = "cannot connect to %s";

/*
 * Get a socket connected to host via the port for the batch service.
 * If address and/or port point to non NULL space, insert the address and/or
 * port number used for the connection.
 */
int
getsocket(host, serv, proto, address, port)
char *	host;
char *	serv;
char *	proto;
u_long *	address;
u_short *	port;
{
	int	s;
#ifdef USE_RRESVPORT
	int	lport = IPPORT_RESERVED - 1;
	int	timo = 1;
	int	error;
#endif
	static char *	mess  = "the batch daemon is unavailable";
	struct sockaddr_un	slo;
	struct sockaddr_in	sin;
	struct servent *	sp;
	struct hostent *	hp;

	h_errno = 0;

	/*
	 * If this is the localhost, and we desire the "batch" service,
	 * do it through a UNIX domain socket.
	 */
	if (islocalhost(host) == 0 && strcmp(serv, service) == 0) {

		if ((s = socket(AF_UNIX, SOCK_STREAM, 0)) < 0) {
			blog(BLOG_ERR, esock);
			return(-1);
		}		

		(void) strcpy(slo.sun_path, socketname);

		if (address != NULL)
			*address = (u_long) gethostid();

		if (port != NULL)
			*port = -1;

		slo.sun_family = AF_UNIX;

		if (connect(s, (struct sockaddr *) &slo, SUN_LEN(&slo)) < 0) {
			(void) close(s);
			blog(BLOG_ERR, mess);
			return (-1);
		}
		return (s);

	} else {

		(void) memset((voidp) &sin, 0, sizeof(sin));
		sin.sin_family = AF_INET;

		/*
		 * If the address is given, it overrides the host name
		 */
		if (address != NULL && *address != 0) {
			(void) memcpy((voidp) &sin.sin_addr.s_addr,
				      (voidp) address,
				      sizeof(address));
		} else {
			h_errno = 0;

			/* Use loopback device if localhost */
			if (islocalhost(host) == 0) {
				sin.sin_addr.s_addr = (u_long) gethostid();
			} else if ((hp = gethostbyname(host)) != NULL) {
				(void) memcpy((voidp) &sin.sin_addr,
					      (voidp) hp->h_addr,
					      hp->h_length);
			} else {
				blog(BLOG_ERR, eaddr, host);
				return(-1);
			}

			if (address != NULL) {
				*address = sin.sin_addr.s_addr;
			}

		}

		/*
		 * Save the remote port only for the batch service,
		 * if we're looking up something else, don't use or
		 * save the remote_port.
		 */
		if (strcmp(serv, service) == 0) {
			if (remote_port == 0) {
				if ((sp = getservbyname(serv, proto)) == NULL) {
					blog(BLOG_ERR, eserv, serv);
					return(-1);
				}
				sin.sin_port = sp->s_port;
				remote_port = ntohs(sp->s_port);
			} else {
				sin.sin_port = htons(remote_port);
			}
		} else {
			if ((sp = getservbyname(serv, proto)) == NULL) {
				blog(BLOG_ERR, eserv, serv);
				return(-1);
			}
			sin.sin_port = sp->s_port;
		}

		if (port != NULL)
			*port = sin.sin_port;

#ifdef USE_RRESVPORT

retry:

		if ((s = rresvport(&lport)) < 0) {
			blog(BLOG_ERR, esock);
			return(-1);
		}

		if (connect(s, (struct sockaddr *) &sin, sizeof(sin)) < 0) {
			error = errno;
			(void) close(s);
			errno = error;
			if (errno == EADDRINUSE) {
				lport--;
				goto retry;
			}

			if (iamdaemon == TRUE)
				blog(BLOG_ERR, econn, host);
			else
				blog(BLOG_ERR, mess);

			return(-1);
		}

		return(s);
#else

		if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
			blog(BLOG_ERR, esock);
			return(-1);
		}

		if (connect(s, (struct sockaddr *) &sin, sizeof(sin)) < 0) {
			(void) close(s);

			if (iamdaemon == TRUE)
				blog(BLOG_ERR, econn, host);
			else
				blog(BLOG_ERR, mess);

			return(-1);
		}

		return (s);	
#endif
	}

}
