#ifndef lint
static char rcsid[] =
	"$Id: groups.c,v 4.0 1993/09/17 00:41:39 srp Exp srp $";
#endif

/*
 *
 * Author: Scott R. Presnell
 *
 * Copyright (C) 1989, 1990, The Regents of the University of California
 *                     All Rights Reserved
 *
 * THE REGENTS OF THE UNIVERSITY OF CALIFORNIA MAKE NO REPRESENTATION OR
 * WARRANTIES WITH RESPECT TO THE CONTENTS HEREOF AND SPECIFICALLY DISCLAIM
 * ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
 * PURPOSE.
 *
 * Further, the Regents of the University of California reserve the right to
 * revise this software and/or documentation and to make changes from time
 * to time in the content hereof without obligation of the Regents of the
 * University of California to notify any person of such revision or change.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <grp.h>
#include <pwd.h>

#include <sys/types.h>
#include <sys/param.h>

#include "config.h"

#include H_STRING

#include "batch.h"

extern int	ingroup P((char *, GIDTYPE, char *));

#ifdef TEST

char *progname;
char *groupstring;

main(argc, argv)
int	argc;
char *	argv[];
{

	char *	queue;
	struct passwd *	pwp;

	progname = argv[0];

	if (argc != 3) {
		(void) fprintf(stderr, "usage %s: <user> <rg string>\n",
			progname);
		exit(1);
	}


	if ((pwp = getpwnam(argv[1])) == NULL) {
		(void) fprintf(stderr, "cannot find pwent for %s\n", argv[1]);
		exit(1);
	}
    
	groupstring = argv[2];
	
	if (check_group(pwp->pw_name, getuid(), getgid(), "test") < 0)
		(void) printf("test: not a member of appropriate group(s)\n");
	else 
		(void) printf("test: is a member of appropriate group(s)\n");


	exit(0);
}

#endif


/*
 * Check the group access for a given login to a queue.  
 * Return < 0 if access denied. = 0 is access not required
 * > 0 if access checks out OK.
 */
int
check_group(user, uid, gid, que)
char *	user;
UIDTYPE	uid;
GIDTYPE	gid;
char *	que;
{

	char *	gp;		/* Restricted group pointer */
	char *	rg;
	char *	bp;
	char	pbuf[BUFFER];
	int	mygroup;
	int	have_access;	/* Step wise access */
	int	check;		/* Final check */

	bp = pbuf;

#ifdef AND_GROUPS
	check = 1;		/* ... strangely enough */
#else
	check = -1;
#endif

	/*
	 * If there is no rg string then there is no group to check.
	 * If the user is root or a gid less than 10 no access check req'd.
	 */
#ifdef TEST
	rg = groupstring;
#else
	if ((rg = bgetstr("rg", &bp)) == NULL || uid == 0 || gid < 10)
		return 0;
#endif
	/*
	 * If its a bum rg parameter, return with access denied
	 */
	if ((gp = strtok(rg, ",")) == NULL) {
		blog(BLOG_WARNING,
		    "malformed restricted group parameter: \"%s\"", rg);
		return(-1);
	}

	/*
	 *
	 * 1) while there is a group to check
	 * 2) get the group pointer
	 * 3) Check if we are in that group right now.
	 * 4) check if our name matches a name on the list
	 *
	 */
	do {

		mygroup = -1;

		/*
		 * If we are checking for absence of group membership,
		 * invert the sense of the test. Move the pointer ahead
		 * one.
		 */
		if (*gp == '!') {
			have_access = 1;	/* Starts *with* access */
			gp++;
		} else {
			have_access = -1;	/* Starts *without* access */
		}

		if ((mygroup = ingroup(user, gid, gp)) < 0) {
			blog(BLOG_WARNING, "bad restricted group: %s", gp);
			continue;
		} else if (mygroup == 0) {
			have_access *= -1;
		}
#ifdef TEST
		printf("after %s check is %d, access is %d\n",
		       gp, check, have_access);
#endif

#ifdef AND_GROUPS
		/* Need all true for access */
		if (check == 1 && have_access > 0)
			check = 1;
		else
			check = -1;
#else
		/* Starts off denied */
		if (have_access > 0) {
			check = 1;
			break;		/* Haul out when we find first true */
		}
#endif	

	} while ((gp = strtok((char *) NULL, ",")) != NULL);

	return check;
	
}


