#ifndef lint
static char rcsid[] =
	"$Id: lock.c,v 4.0 1993/09/17 00:41:39 srp Exp srp $";
#endif

/*
 *
 * Author: Scott R. Presnell
 *
 * Copyright (C) 1989, 1990, The Regents of the University of California
 *                     All Rights Reserved
 *
 * THE REGENTS OF THE UNIVERSITY OF CALIFORNIA MAKE NO REPRESENTATION OR
 * WARRANTIES WITH RESPECT TO THE CONTENTS HEREOF AND SPECIFICALLY DISCLAIM
 * ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
 * PURPOSE.
 *
 * Further, the Regents of the University of California reserve the right to
 * revise this software and/or documentation and to make changes from time
 * to time in the content hereof without obligation of the Regents of the
 * University of California to notify any person of such revision or change.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <errno.h>
#include <fcntl.h>
#include <signal.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>

#include "config.h"

#if (defined(I_STDARG) && defined(CAN_VAPROTO)) || defined(__STDC__)
#include <stdarg.h>
#else
#include <varargs.h>
#endif

#include H_STRING

#include "batch.h"
#include "batchd.h"
#include "flags.h"

/*
 * Set the status of the lock file for a given queue.
 */
void
queue_set_lock(qp, action)
Queue *	qp;
State	action;
{

	int	fd;

	if ( !(action & INTERNAL) ) {

	errno = 0;

	if (stat(qp->q_lname, &stbuf) < 0)
		if (errno == ENOENT) {
			if ((fd = open(qp->q_lname, (O_WRONLY|O_CREAT), 00664)) < 0) {
				blog(BLOG_ERR, eopen, qp->q_lname);
				return;
			} else {
				(void) close(fd);
				if (stat(qp->q_lname, &stbuf) < 0) {
					blog(BLOG_ERR, estat, qp->q_lname);
					return;
				}
			}
		} else {
			blog(BLOG_ERR, estat, qp->q_lname);
			return;
		}
        }

	switch(action) {

	case RUNNING :
		/*
		 * If the setuid bit is set, don't mark as running. The
		 * queue has been "stopped" by bac.
		 */
		if (!(stbuf.st_mode & 04000))
			if (chmod(qp->q_lname,
				  (stbuf.st_mode & 06777) & ~00100) < 0)
			blog(BLOG_ERR, echmo, qp->q_name);
		break;

	case TIMEOUT :
	case SUSPEND :
	case INTERRUPT:
		if (chmod(qp->q_lname, (stbuf.st_mode & 06777) | 00100) < 0)
			blog(BLOG_ERR, echmo, qp->q_name);
		break;

	case QUEUING :
		/*
		 * If the setgid bit is set, don't mark as queueing. The
		 * queue has been "disabled" by bac.
		 */
		if (!(stbuf.st_mode & 02000))
			if (chmod(qp->q_lname,
				  (stbuf.st_mode & 06777) & ~00010) < 0)
			blog(BLOG_ERR, echmo, qp->q_name);
		break;

	case WAITING:
		if (chmod(qp->q_lname, (stbuf.st_mode & 06777) | 00010) < 0)
			blog(BLOG_ERR, echmo, qp->q_name);
		break;

	case DISABLED:
	case STOPPED:
	case DRAINING:
		/* Nothing to do here */
		break;
	default:
		blog(BLOG_WARNING,
		    "queue_set_lock(): bogus action (%lo)", action);
		break;

	}

	return;
	
}


/*
 * Get the status of the lock file for a given queue.
 */
int
queue_get_lock(qp, action)
Queue *	qp;
State	action;
{

	int	fd;

	errno = 0;

	if (stat(qp->q_lname, &stbuf) < 0)
		if (errno == ENOENT) {
			if ((fd = open(qp->q_lname, (O_WRONLY|O_CREAT), 00664)) < 0) {
				blog(BLOG_ERR, eopen, qp->q_lname);
				return (-1);
			} else {
				(void) close(fd);
				if (stat(qp->q_lname, &stbuf) < 0) {
					blog(BLOG_ERR, estat, qp->q_lname);
					return (-1);
				}
			}
		} else {
			blog(BLOG_ERR, estat, qp->q_lname);
			return (-1);
		}

	switch(action) {

	 case DRAINING :
		 return (!((stbuf.st_mode & 02000) &&
			   (stbuf.st_mode & 00010) &&
			   (stbuf.st_mode & 00001)));
		 break;

	case STOPPED :
		 return (!((stbuf.st_mode & 04000) && (stbuf.st_mode & 00100)));
		 break;

	 case DISABLED :
		 return (!((stbuf.st_mode & 02000) && (stbuf.st_mode & 00010)));
		 break;

	 default:
		blog(BLOG_WARNING,
		     "queue_get_lock(): bogus action (%lo)", action);
		return 1;
		break;
	}

	return 0;

}



/*
 * Grab a nonblocking lock on the lock file of the queue.  Assume that 
 * we are in the spooling directory for that queue.
 * return -1 = error in opening
 * return  0 = could not get a lock
 * return >0 = open, locked, file descriptor
 */
int
queue_grab_lock(qp)
Queue *	qp;
{

	int	lfd;

	if ((lfd = open(qp->q_lname, O_RDWR)) < 0) {
		return (-1);
	}

	/*
	 * Try to get a lock
	 */
#ifdef HAVE_LOCKF
	if (lockf(lfd, F_TLOCK, 0L) < 0) {
		if (errno == EACCES || errno == EAGAIN) {
#else
#ifdef HAVE_FLOCK
	if (flock(lfd, (LOCK_EX|LOCK_NB)) < 0) {
		if (errno == EWOULDBLOCK) {
#endif /* HAVE_FLOCK */
#endif /* HAVE_LOCKF */

			(void) close(lfd);
			return 0;

	        }
	}
	return lfd;
}


/*
 * Release the lock on the queue
 */
int
queue_release_lock(qp, lfd)
Queue *	qp;
int	lfd;
{

	return close (lfd);
}
