#ifndef lint
static char rcsid[] =
	"$Id: notify.c,v 4.0 1993/09/17 00:41:39 srp Exp srp $";
#endif

/*
 *
 * Author: Scott R. Presnell
 *
 * Copyright (C) 1989, 1990, The Regents of the University of California
 *                     All Rights Reserved
 *
 * THE REGENTS OF THE UNIVERSITY OF CALIFORNIA MAKE NO REPRESENTATION OR
 * WARRANTIES WITH RESPECT TO THE CONTENTS HEREOF AND SPECIFICALLY DISCLAIM
 * ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
 * PURPOSE.
 *
 * Further, the Regents of the University of California reserve the right to
 * revise this software and/or documentation and to make changes from time
 * to time in the content hereof without obligation of the Regents of the
 * University of California to notify any person of such revision or change.
 *
 */

/*
 * Sending mail back to the users...
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>

#include <ctype.h>

#include <limits.h>

#if defined(MALLOC_TRACE) || defined(MALLOC_LEAK)
#include "dmalloc.h"
#else
#include <malloc.h>
#endif

#include <signal.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/file.h>
#include <sys/dir.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include "config.h"

#if (defined(I_STDARG) && defined(CAN_VAPROTO)) || defined(__STDC__)
#include <stdarg.h>
#else
#include <varargs.h>
#endif

#include H_STRING

#include "batch.h"
#include "batchd.h"
#include "flags.h"
#include "pathnames.h"


#ifdef SENDMAIL					/* Sendmail path */
static char *	sendmail = SENDMAIL;
#else
static char *	sendmail = _PATH_SENDMAIL;
#endif

#ifdef BINWRITE					/* Write path */
static char *	binwrite = BINWRITE;
#else
static char *	binwrite = _PATH_WRITE;
#endif

static char	epipe[] = "cannot create pipe %s: (%m)";

void	 	send_results P((Queue *, Job *, wait_t *, Notify, char *));


static FILE *	smfopen P((char *));
static int	smfclose P((FILE *));
static void	dump_file P((FILE *, char *));

extern char *	sys_siglist[];			/* From libc.a, or faults.c */


/*
 * Send the results of a terminated job to the user that started the job.
 */
void
send_results(qp, jp, rtn, mailstat, expl)
Queue *	qp;
Job *	jp;
wait_t *	rtn;
Notify	mailstat;
char *	expl;
{

	char *	state;
	char	junk[LINE];
	int	termsig = 0;
	int	retcode = 0;
	int	cfd = 0;
	int	have = 0;
	int	want = 0;
	int	lastsig = 0;
	int	kpid, ksig, ptype;
	time_t	uptime = -1;		/* Time machine has been up */
	time_t	elaptime = 0;		/* Gross elapsed run time */
	time_t	runtime = 0;		/* Agregate run time */
	time_t	lastrun = 0;
	FILE *	mail, *	cfp;
	time_t	ktime;


	now = time((time_t *) 0);

	if (rtn != NULL) {
		if (WIFEXITED(*rtn))
			retcode = WEXITSTATUS(*rtn);
		if (WIFSIGNALED(*rtn))
			termsig = WTERMSIG(*rtn);
	}

	/* Working with output (of) file */
	jp->j_file[0] = 'o';

	/*
	 * Do we have output? Even an empty one?
	 */
	if ((have = stat(jp->j_file, &stbuf)) < 0) {
		if (!(jp->j_state & JOB_ATTACH))
			blog(BLOG_WARNING, estat, jp->j_file);
		stbuf.st_size = 0;
		stbuf.st_ctime = 0;
	}
	    
	/*
	 * If the user has not asked for the end data, and there is some,
	 * remove the file, else if there isn't any, fine, don't try to
	 * remove the file.
	 */
	if (!(jp->j_usermail & mailstat)) {
		want = -1;
		if (stbuf.st_ctime != 0) {
			if (unlink(jp->j_file) < 0)
				blog(BLOG_ERR, eunlk, jp->j_file);
		} else {
			; /* We don't have output, fine. */
		}
		jp->j_file[0] = 'c';
		return;
	} else 
		want = 1;

	/* Working with control (cf) file */
	jp->j_file[0] = 'c';

	/* Update the control file with exit status */
	if ((cfd = open(jp->j_file, (O_RDWR|O_APPEND))) < 0) {
		blog(BLOG_WARNING, eopen, jp->j_file);
	} else {
		(void) sprintf(line, "%c%s %d %d %ld\n",
			       'F', lhost, retcode, termsig, jp->j_cputime);
		(void) ewrite(cfd, line, strlen(line), "results()");
		(void) close(cfd);
	}

	/* Maybe this user doesn't want any mail. */
	if (jp->j_umaddress == NULL)
		goto done;

	if ((mail = smfopen(jp->j_umaddress)) == NULL) {
		blog(BLOG_WARNING, "smfopen: dumping data");
		mail = fopen(_PATH_ERR, "w");
	}

	(void) fprintf(mail, "Sender: root@%s\n", lhost);
	(void) fprintf(mail, "Reply-To: %s\n", qp->q_smaddress);

	/*
	 * Give some sort of hopefully useful subject.
	 */
	if (mailstat & NOTIFY_CRASH)
		state = "Crash";
	else if (termsig > 0)
		state = "Signal";
	else if (retcode != 0)
		state = "Failure";
	else
		state = "Success";

	(void) fprintf(mail, "Subject: %s%s: job %d in %s queue\n\n",
		       state, stbuf.st_size ? "+Output" : "",
		       jp->j_id, qp->q_name );


	(void) fprintf(mail, "%s %d (%s) in the %s queue",
		       (jp->j_state & JOB_ATTACH) ? "Attached job" : "Job",
		       jp->j_id, jp->j_jname ? jp->j_jname : unk, qp->q_name);

	if (mailstat & NOTIFY_CRASH) {

		/*
		 * If boottime has been set, and it's less than ten minutes,
		 * The crash was probably a machine crash.
		 *
		 * Otherwise assume it was a batchd crash.
		 */
		if (boottime != -1)
			uptime = now - boottime;
		else
			uptime = -1;

		if (uptime != -1 && uptime < (SECPERMIN * 5)) {
			(void) fputs(" has crashed due to a machine reboot.\n", mail);
			(void) fprintf(mail, "The job was %s.\n", expl);
		} else {
			(void) fputs(" is probably running unattended\n", mail);
			(void) fputs("due to a batch daemon failure.\n", mail);
		}

	} else if (termsig < 0) {

		(void) fputs(" completed.\n", mail);

	} else {

		(void) fputs(" completed", mail);

		if (termsig == 0 && retcode == 0) {
			(void) fputs(" successfully", mail);
		} else {
			if (termsig) {
				if (termsig > NSIG)
					(void) fprintf(mail,
						" with unknown status %o",
						rtn->w_status);
				else
					(void) fprintf(mail,
					       " with signal \"%s\"",
					       sys_siglist[termsig]);
				if (rtn->w_coredump)
					(void) fputs(" and a core dump", mail);
			} else {
				(void) fprintf(mail,
				       " with exit code %d", retcode);
				if (retcode == RTN_ERROR) {
	(void) fputs(".\nThis may be the result of an error in the batch daemon after the job forked.\nSee the batch queue administrator", mail);
                                }
			}
		}
		(void) fputs(".\n", mail);
	}

	/*
	 * Calculate the gross percentage of CPU time, CPU seconds over
	 * the elapsed "real" time for the job, and the aggregate percentage
	 * of CPU time, use only the time the job/queue was running).
	 * Use the 'K' records in the cf file.
	 */

	/* Working with control (cf) file */
	lastrun = jp->j_stime;
	lastsig = SIGCONT;
	if ((cfp = fopen(jp->j_file, "r")) != NULL) {
		while (fgets(line, LINE, cfp) != NULL) {

			/*
			 * Job may have remained after batch death,
			 * if so, scan in start time.
			 */
			if (line[0] == 'P' && jp->j_stime == 0) {
				sscanf(&line[1], "%s %d %d %ld",
				       junk, &kpid, &ptype, &(jp->j_stime));
				lastrun = jp->j_stime;
			}
				
			if (line[0] == 'K') {
				sscanf(&line[1], "%s %d %d %ld",
				       junk, &kpid, &ksig, &ktime);
				if (ksig == SIGSTOP) {
					lastsig = SIGSTOP;
					runtime += ktime - lastrun;
				}
				if (ksig == SIGCONT) {
					lastsig = SIGCONT;
					lastrun = ktime;
				}
			}
		}
		(void) fclose(cfp);
	}

	if (lastsig == SIGCONT) {
		runtime += now - lastrun;
	}

	/* Working with output (of) file */
	jp->j_file[0] = 'o';
	elaptime = now - jp->j_stime;

	(void) strcpy(line, ctime(&(jp->j_stime)));
	line[19] = '\0';

	(void) strcpy(line2, ctime(&now));
	line2[19] = '\0';

	(void) fprintf(mail, "Job %s %s; finished: %s\n",
		       (jp->j_state & JOB_ATTACH) ? "attached" : "started",
		       line, line2);

	(void) fprintf(mail, "%s time: %s\n",
		       (jp->j_state & JOB_ATTACH) ? "Monitored" : "Elapsed",
		       prttime(elaptime));

	if (jp->j_cputime)
		(void) fprintf(mail, "%s time: %lds\n",
		       (jp->j_state & JOB_ATTACH) ? "Monitored CPU" : "CPU",
			       jp->j_cputime);

	if (jp->j_cputime && elaptime && runtime)
		(void) fprintf(mail,
			       "Gross %%CPU %.2f; Net %%CPU %.2f\n",
			       100 * ((double) jp->j_cputime/elaptime),
			       100 * ((double) jp->j_cputime/runtime));

	/*
	 * If NOTIFY_DEBUG is on, send the user the cf and df files.
	 * This is not an advertized option.
	 */
	if (jp->j_usermail & NOTIFY_DEBUG) {

		/* Working with control (cf) file */
		jp->j_file[0] = 'c';
		dump_file(mail, jp->j_file);

		/* Working with data (df) file */
		jp->j_file[0] = 'd';
		dump_file(mail, jp->j_file);
	}

	/* Working with output (of) file */
	jp->j_file[0] = 'o';
	/*
	 * It's not necessarily an error if there is no output file.
	 * So tell them, but don't make it seem like an error.
	 *
	 * Attached jobs never generate output.
	 */
	if (jp->j_state & JOB_ATTACH) {
		;
	} else if (have == -1) {
		(void) fprintf(mail, "No output file found!?!\n\n");
	} else if (stbuf.st_size > 0) {
		(void) fprintf(mail, "Output%s follows:\n\n",
			       mailstat & NOTIFY_CRASH ? " to crash point": "");
		dump_file(mail, jp->j_file);
	}

	(void) smfclose(mail);

 done:
	/*
	 * Remove the output file if it exists - report any falures.
	 */
	if (have >= 0) {
		if (unlink(jp->j_file) < 0)
			blog(BLOG_ERR, eunlk, jp->j_file);
	}

	/* Restore to control (cf) file */
	jp->j_file[0] = 'c';

	return;
}


/*
 * Dump all data from file to the FILE pointer fp.
 */
static void
dump_file(fp, file)
FILE *	fp;
char *	file;
{

	int	n, output;
	char	buf[BUFFER];


	if ((output = open(file, 0)) == -1) {
		blog(BLOG_WARNING, eopen, file);
		(void) fprintf(fp, "***Daemon can't read the %cf file!\n",
			       file[0]);
	} else {
		if (file[0] != 'o')
			(void) fprintf(fp, "\n=== %cf file: %s ===\n",
				       file[0], file);

		while ((n = read(output, buf, sizeof(buf))) > 0)
			(void) fwrite(buf, n, 1, fp);

		(void) close(output);
	}
	return ;
}


/*
 * Fork off a process to send mail to a user.
 * Write to stdin of the "sendmail" utility program.
 */
static FILE *
smfopen(to)
char *	to;
{

	int	pid;
	int	mpipe[2];
	FILE *	mailer;


	/* Setup comminications pipe */
	if (pipe(mpipe) < 0) {
		blog(BLOG_ERR, epipe, "in smfopen()" );
		return NULL;
	}

	(void) sighold(SIGCHLD);		/* until in parent table */

	switch (pid = fork()) {
	case -1:	/* ERROR */
		blog(BLOG_ERR, efork, sendmail);
		return NULL;
		break;
	case 0:		/* CHILD -> set pipes, exec sendmail */
		(void) dup2(mpipe[0], fileno(stdin));
		(void) close(mpipe[1]);
		(void) setpriority(PRIO_PROCESS, 0, 4);	/* lower priority */
		(void) execl(sendmail, sendmail, to, (char *) 0);
		blog(BLOG_ERR, "smfopen(): execl of %s", sendmail);
		_exit(2);
		break;
	default:	/* PARENT -> add process, send info down pipe */
		(void) close(mpipe[0]);
		(void) process_add(pid, (Queue *) NULL, (Job *) NULL, sendmail);
		(void) sigrelse(SIGCHLD);
		break;
	}

	if ((mailer = fdopen(mpipe[1], "w")) == NULL) {
		(void) close(mpipe[1]);
		blog(BLOG_ERR, efdop, "smfopen");
		return NULL;
	}

	(void) fprintf(mailer, 
		       "To: %s\nFrom: batchd@%s (The Batch Daemon)\n",
		       to, lhost);

	(void) fflush(mailer);

	return mailer;
}

static int
smfclose(f)
FILE *	f;
{

	if (f != stderr && f != stdout)
		return fclose(f);
	else
		return (-1);
}


/*VARARGS*/
#if (defined(I_STDARG) && defined(CAN_VAPROTO)) || defined(__STDC__)

void
send_mail(char *to, char *sub, char *fmt, ...)

#else

void
send_mail(to, sub, fmt, va_alist)
char *	to;
char *	sub;
char *	fmt;
va_dcl

#endif /* __STDC__ */
{

	va_list	args;
	FILE *	mail;


	if (to == NULL)
		return;

	/*
	 * Fire up the variable argument mechanism.
	 */
	/*CONSTCOND*/
#if (defined(I_STDARG) && defined(CAN_VAPROTO)) || defined(__STDC__)
	va_start(args, fmt);
#else
	va_start(args);
#endif

	/* Maybe this user doesn't want any mail. */
	if (to == NULL) {
		return ;
	} else if (strchr(to, '/') != NULL) {

		if ((mail = fopen(to, "a+")) == NULL) {
			blog(BLOG_WARNING, eopen, to);
			mail = stderr;
		}

#ifdef USE_LOCALBUF
		(void) setbuf(mail, iobuf);
#endif

		now = time((time_t *) 0);
		(void) fprintf(mail, "%-15.15s ", ctime(&now) + 4);

	} else {

		if ((mail = smfopen(to)) == NULL) {
			blog(BLOG_WARNING, "smfopen: dumping data");
			mail = fopen(_PATH_ERR, "w");
		}

		(void) fprintf(mail, "Sender: root@%s\n", lhost);
		(void) fprintf(mail, "Reply-To: root@%s\n", lhost);
		(void) fprintf(mail, "Subject: %s\n\n", sub);
		
	}

	if (fmt != NULL) {
		(void) vfprintf(mail, fmt, args);
		(void) fputc('\n', mail);
	}

	(void) smfclose(mail);

	/*
	 * Shut down the variable argument mechanism.
	 */
	va_end(args);

	return;
}


/*VARARGS*/
#if (defined(I_STDARG) && defined(CAN_VAPROTO)) || defined(__STDC__)

void
send_write(char *to, char *fmt, ...)

#else

void
send_write(to, fmt, va_alist)
char *	to;
char *	fmt;
va_dcl

#endif /* __STDC__ */
{

	char *	cp;
	char *	wuser, *whost;
	char	command[LINE];
	char	waddr[LINE];
	int	w, pid;
	int	wpipe[2];
	va_list	args;
	FILE *	wp;


	if (to == NULL)
		return;

	/*
	 * Fire up the variable argument mechanism.
	 */
	/*CONSTCOND*/
#if (defined(I_STDARG) && defined(CAN_VAPROTO)) || defined(__STDC__)
	va_start(args, fmt);
#else
	va_start(args);
#endif

	/* Maybe this user doesn't want any mail. */
	if (to == NULL)
		return ;

	(void) strcpy(waddr, to);
	if ((cp = strchr(waddr, '@')) == NULL) {
		whost = lhost;
	} else {
		*cp++ = '\0';
		whost = cp;
	}

	wuser = waddr;

	dbg((BLOG_PROC, "setting up to write %s on %s", wuser, whost));

#ifdef USE_WRITED

	if ((w = getsocket(whost, "write", tcp,
			   (u_long *) 0, (u_short *) 0)) < 0)
		return;

	if ((wp = fdopen(w, "w+")) == NULL) {
		blog(BLOG_ERR, efdop, "send_write");
		return;
	}

	/* Write the user name, then the message */
	(void) fputs(wuser, wp);
	(void) fputs("\r\n", wp);

#else

	if (islocalhost(whost) != 0)
		return ;

	/* Setup comminications pipe */
	if (pipe(wpipe) < 0) {
		blog(BLOG_ERR, epipe, "in send_write()");
		return ;
	}

	(void) sighold(SIGCHLD);		/* until in parent table */

	switch (pid = fork()) {
	case -1:	/* ERROR */
		blog(BLOG_ERR, efork, binwrite);
		return ;
		break;
	case 0:		/* CHILD -> set pipes, exec binwrite */
		(void) dup2(wpipe[0], fileno(stdin));
		(void) close(wpipe[1]);
		(void) setpriority(PRIO_PROCESS, 0, 4);	/* lower priority */
		(void) execl(binwrite, binwrite, wuser, (char *) 0);
		blog(BLOG_ERR, "send_write(): execl of %s", binwrite);
		_exit(2);
		break;
	default:	/* PARENT -> add process, send info down pipe */
		(void) close(wpipe[0]);
		(void) process_add(pid, (Queue *) NULL, (Job *) NULL, binwrite);
		(void) sigrelse(SIGCHLD);
	}

	if ((wp = fdopen(wpipe[1], "w")) == NULL) {
		(void) close(wpipe[1]);
		blog(BLOG_ERR, efdop, "send_write");
		return ;
	}


#endif /* USE_WRITED */

	dbg((BLOG_PROC, "pipe open for writing %s on %s", wuser, whost));

	fflush(wp);

	sleep(2);

	if (fmt != NULL) {
		(void) fprintf(wp, "\tFrom the batch daemon on %s:", lhost);
		dbg((BLOG_PROC, "<From the batch daemon on %s:>", lhost));
		(void) fputs("\r\n", wp);
		(void) fputs("\t", wp);
		(void) vfprintf(wp, fmt, args);
		(void) fputs("\r\n", wp);
	}

	dbg((BLOG_PROC, "finished writing %s on %s", wuser, whost));

	fflush(wp);

	(void) fclose(wp);

#ifdef USE_WRITED
	(void) close(w);
#endif /* USE_WRITED */

	/*
	 * Shut down the variable argument mechanism.
	 */
	va_end(args);

	return;
}

