#ifndef lint
static char rcsid[] =
	"$Id: readcap.c,v 4.0 1993/09/17 00:41:39 srp Exp srp $";
#endif

/*
 *
 * Author: Scott R. Presnell
 *
 * Copyright (C) 1989, 1990, The Regents of the University of California
 *                     All Rights Reserved
 *
 * THE REGENTS OF THE UNIVERSITY OF CALIFORNIA MAKE NO REPRESENTATION OR
 * WARRANTIES WITH RESPECT TO THE CONTENTS HEREOF AND SPECIFICALLY DISCLAIM
 * ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
 * PURPOSE.
 *
 * Further, the Regents of the University of California reserve the right to
 * revise this software and/or documentation and to make changes from time
 * to time in the content hereof without obligation of the Regents of the
 * University of California to notify any person of such revision or change.
 *
 */

/*
 * Facilities for extracting information from the batchcap file.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>

#include <errno.h>

#if defined(MALLOC_TRACE) || defined(MALLOC_LEAK)
#include "dmalloc.h"
#else
#include <malloc.h>
#endif
#include <time.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/resource.h>

#include "config.h"

#if (defined(I_STDARG) && defined(CAN_VAPROTO)) || defined(__STDC__)
#include <stdarg.h>
#else
#include <varargs.h>
#endif

#include H_STRING

#include "batch.h"
#include "batchd.h"
#include "event.h"
#include "jobsort.h"
#include "flags.h"
#include "pathnames.h"


#define	DATESEP	','
#define	TIMESEP	'.'

static char admin[] = _PATH_ADMIN;


struct names_t namestab[10] = {
#define X_DISABLED	0
	DISABLED,	"enabled",	"disabled",
#define X_WAITING	1
	WAITING,	"queuing",	"waiting",
#define X_QUEUING	2
	QUEUING,	"queuing",	"not queuing",
#define X_STOPPED	3
	STOPPED,	"restarted",	"stopped",
#define X_TIMEOUT	4
	TIMEOUT,	"timein",	"in timeout",
#define X_SUSPEND	5
	SUSPEND,	"resumed",	"suspended",
#define X_INTERRUPT	6
	INTERRUPT,	"continued",	"interrupted",
#define X_RUNNING	7
	RUNNING,	"running",	"not running",
#define X_DRAINING	8
	DRAINING,	"continuing",	"draining",
	0, "", "",
};


static int	setcurlim P((Queue *, int, int));
static int	setmaxlim P((Queue *, int, int));
static int	days_parse P((Queue *, char *));
static int	days_assign P((Day [], int, int, int));
static int	times_extract P((char *, int, int, int));
static void	queue_free_struct P((Queue *));
static void	usr_free_struct P((Usr **));

static void	queue_read_name P((void));
static int	queue_read_cap P((Queue *, int));

/* In readcap.c but used before defined */
extern char *	statetoa P((Queue *));

#ifdef HAVE_RLIMIT
extern char *	rlimittoa P((int));
#endif

/*
 * Read the queue list and load the queue information.
 */
void
cap_reread()
{

	static Bool	 first = TRUE;
	Queue *	qp;


	blog(BLOG_INFO, "reloading capabilities database");
	/*
	 * Read the queue list (again).
	 */
	queue_read_name();

	/*
	 *  Check each potential queue entry.
	 */
	ITER(qp, Queue, &queues) {

		/*
		 * If this is an empty queue then skip.
		 */
		if (qp->q_name == NULL)
			continue;

		/*
		 * Remove pending events for a re-load.
		 */
		if (qp->q_bevent) {
			(void) event_remove(qp->q_bevent);
			qp->q_bevent = 0;
		}

		if (qp->q_pevent) {
			(void) event_remove(qp->q_pevent);
			qp->q_pevent = 0;
		}

		if (qp->q_tevent) {
			(void) event_remove(qp->q_tevent);
			qp->q_tevent = 0;
		}

		if (qp->q_uevent) {
			(void) event_remove(qp->q_uevent);
			qp->q_uevent = 0;
		}

		(void) queue_read_cap(qp, first);

	}

	/*
	 * After the first time, all other calls to the readcap routines
	 * are considered rereads.
	 */
	first = FALSE;

	return ;
}



/*
 *  Read the batchcap file, looking for all queue entries.
 */
static void
queue_read_name()
{
	
	Queue *	qp;
	char *	qname;
	
	
	/*
	 * Read all in the queues one by one, and "install" them by name.
	 */
	while ((qname = getquent()) != NULL) {
		
		dbg((BLOG_QUEUES, ques, qname, "Reading"));

		ITER(qp, Queue, &queues) {
			if (qp->q_name == NULL)
				continue;
			
			if (strcmp(qname, qp->q_name) == 0)
				break;
		}
		
		/*
		 *  Was it a new queue?
		 */
		if (qp == HEAD(Queue, &queues)) {
		
			/*
			 * Make space for a new queue.
			 */
			dbg((BLOG_ALLOC, "allocating Queue structure"));

			qp = ALLOC(Queue);

			qinfo.i_queuelist++;

			qp->q_name = estrdup(qname);
			qp->q_nexec = 0;
			qp->q_return = TRUE;
			qp->q_mtime = 0;

			qp->q_jobs.q_forw = &(qp->q_jobs);
			qp->q_jobs.q_back = &(qp->q_jobs);

			qp->q_usrs.q_forw = &(qp->q_usrs);
			qp->q_usrs.q_back = &(qp->q_usrs);

			ins_tail_que(&qp->q_queues, &queues);

			dbg((BLOG_QUEUES, ques, qp->q_name, "Added"));
		}
	}
	
	(void) endbaent();
	
	return;
}


/*
 * Read an individual queue and complete the information transfer
 * to the internal "Queue" data structure.
 */
static int
queue_read_cap(qp, first)
Queue *	qp;
int	first;
{

	char *	bp;
	char *	cp, *sp, *op;
	char	days_temp[BUFSIZ];
	char	bbuf[BUFFER];
	char	pbuf[BUFFER];
	int	rtn, i, flag;
	int	begintime, endtime;
	int	day;
	int	minute;
	Job *	jp;
	Usr *	up;
	static int	oldtype = 0;
	struct tm *	local_time;

	
	/*
	 * Get the time of day
	 */
	now = time((long *) 0);
	local_time = localtime(&now);
	day = local_time->tm_wday;
	minute = (local_time->tm_hour * MINPERHOUR) + local_time->tm_min;

	bp = pbuf;


	dbg((BLOG_QUEUES, "Read capabilities %s", qp->q_name));

	if ((rtn = bgetent(bbuf, qp->q_name)) < 0) {
		blog(BLOG_FATAL, eopen, _PATH_CAPFILE);
		return (-1);
	} else if (rtn == 0) {
		/*
		 * The queue got removed.  Free the structure, free the name,
		 * and set the name to NULL so that we won't reference the
		 * queue anymore.
		 */
 		blog(BLOG_ERR, enotq, qp->q_name);
		queue_free_struct(qp);
		efree((voidp) qp->q_name);
		qp->q_name = NULL;
		rem_que(&qp->q_queues);
		return (-1);
	}
		    

	/*
	 * Only attempt to free queue structures if this is a reread of the
	 * batchcap file.
	 */
	if (first == FALSE) {
		queue_free_struct(qp);
	}

#ifdef HAVE_RLIMIT
	/*
	 * Initialization.
	 */
	for (i = 0; i < RLIM_NLIMITS; i++) {
		rlimit_t rl;

		if (getrlimit(i, &rl) < 0)
			blog(BLOG_ERR, "%s: default getrlimit(%s) failure",
			    rlimittoa(i), qp->q_name);
		else 
			dbg((BLOG_QUEUES, "default rlimit: %-22.22s: %d %d",
			    rlimittoa(i), rl.rlim_cur, rl.rlim_max));

		qp->q_rlimit[i].rlim_cur = rl.rlim_cur;
		qp->q_rlimit[i].rlim_max = rl.rlim_max;
	}
#endif

	/*
	 * Remote Host - check for localhost as well.
	 */
	if ((cp = bgetstr("rm", &bp)) == NULL || islocalhost(cp) == 0) 
		qp->q_rhost = NULL;
	else
		qp->q_rhost = estrdup(cp);

	/*
	 * Remote Queue
	 */
	if ((cp = bgetstr("rq", &bp)) == NULL)
		qp->q_rname = NULL;
	else
		qp->q_rname = estrdup(cp);


	if (qp->q_rhost != NULL && qp->q_rname == NULL)
		blog(BLOG_WARNING, ques, qp->q_name,
		    "\"rm\" set but \"rq\" is not!");

	/*
	 * Directories and information files.
	 */
	if ((cp = bgetstr("sd", &bp)) == NULL)
		cp = _DEF_SPOOL;
	qp->q_spool = estrdup(cp);

	if ((cp = bgetstr("sf", &bp)) == NULL)
		cp = _DEF_STAT;
	qp->q_sname = estrdup(cp);

	if ((cp = bgetstr("lf", &bp)) == NULL)
		cp = _DEF_LOCK;
	qp->q_lname = estrdup(cp);

	if ((cp = bgetstr("af", &bp)) == NULL)
		cp = _DEF_ACCT;
	if (strchr(cp, '/') == NULL) {
		(void) strcpy(line, admin);
		(void) strcat(line, "/");
		(void) strcat(line, cp);
		qp->q_aname = estrdup(line);
	} else {
		qp->q_aname = estrdup(cp);
	}

	if ((cp = bgetstr("ef", &bp)) == NULL)
		cp = _DEF_ERR;
	if (strchr(cp, '/') == NULL) {
		(void) strcpy(line, admin);
		(void) strcat(line, "/");
		(void) strcat(line, cp);
		qp->q_ename = estrdup(line);
	} else {
		qp->q_ename = estrdup(cp);
	}


	/*
	 * Mail reporting queue supervisor...
	 */
	if ((cp = bgetstr("su", &bp)) == NULL)
		qp->q_smaddress = estrdup("root");
	else
		qp->q_smaddress = estrdup(cp);


	/*
	 * Mail reporting, which events to report...
	 */

	/*
	 * This doesn't need to be done here... we load the defaults
	 * in the batch process.  But we do it to see what the values are.
	 */
	if ((cp = bgetstr("ma", &bp)) == NULL) 
		qp->q_usermail = 0;
	else
		qp->q_usermail = setnotify(cp);

	if ((cp = bgetstr("ms", &bp)) == NULL)
		qp->q_supmail = 0;
	else
		qp->q_supmail = setnotify(cp);

	/*
	 * Writes to queue supervisor...
	 */
	if ((cp = bgetstr("wu", &bp)) == NULL)
		qp->q_swaddress = NULL;
	else
		qp->q_swaddress = estrdup(cp);


	/*
	 * Writes, which events to report...
	 */

	/*
	 * This doesn't need to be done here... we load the defaults
	 * in the batch process.  But we do it to see what the values are.
	 */
	if ((cp = bgetstr("wa", &bp)) == NULL) 
		qp->q_userwrite = 0;
	else
		qp->q_userwrite = setnotify(cp);

	if ((cp = bgetstr("ws", &bp)) == NULL)
		qp->q_supwrite = 0;
	else
		qp->q_supwrite = setnotify(cp);

	/*
	 * Time information for stopping/running the queues.
	 * Hours and days.
	 */
	if ((begintime = bgetnum("st")) < 0)
		begintime = 0;

	if ((endtime = bgetnum("sp")) < 0)
		endtime = 0;

	/*
	 * Initialize the days array of data.
	 */
	for (i = 0; i < 7; i++) {
		qp->q_days[i].d_start = (begintime * MINPERHOUR);
		qp->q_days[i].d_stop  = (endtime * MINPERHOUR);
	}		

	if ((cp = bgetstr("da", &bp)) != NULL) {
		(void) strcpy(days_temp, cp);
		if (days_parse(qp, days_temp) < 0)
			blog(BLOG_WARNING,
			     ques, qp->q_name, "error in \"da\"");
	}

	/*
	 * Details of operation.
	 */
	if ((qp->q_maxexec = bgetnum("mx")) < 0)
		qp->q_maxexec = 1;

	if ((qp->q_loadsched = bgetnum("sq")) < 0)
		qp->q_loadsched = 25;

	if ((qp->q_loadstop = bgetnum("sr")) < 0)
		qp->q_loadstop = 50;

	/* Nice */
	if ((i = bgetnum("ni")) < 0) {
		qp->q_priority[NICE_LOW] = 10;
		qp->q_priority[NICE_HIGH] = 10;
	} else {
		qp->q_priority[NICE_LOW] = i;
		qp->q_priority[NICE_HIGH] = i;
	}

	/* Super avanced, new and improved, nice */
	qp->q_ptype = 0;
	if ((sp = bgetstr("ns", &bp)) == NULL) {
		dbg((BLOG_QUEUES, "using \"ni\" nice"));

	} else if ((op = strtok(sp, ",")) != NULL) {

		Bool ni_first = TRUE;

		do {

			/* Consider all jobs together.	 */
			if (strncasecmp(op, "gang", 4) == 0) {
				qp->q_ptype |= NICE_GANG;
				dbg((BLOG_QUEUES, "nice doing gang"));
			}

			/* Consider jobs from each queue separately. */
			if (strncasecmp(op, "separate", 8) == 0) {
				qp->q_ptype |= NICE_SEPARATE;
				dbg((BLOG_QUEUES, "nice doing separate"));
			}

			/* Interpolate the nice values.	 */
			if (strncasecmp(op, "interp", 6) == 0) {
				qp->q_ptype |= NICE_INTERP;
				dbg((BLOG_QUEUES, "nice doing interp"));
			}

			/* Step the nice values  */
			if (strncasecmp(op, "step", 4) == 0) {
				qp->q_ptype |= NICE_STEP;
				dbg((BLOG_QUEUES, "nice doing step"));
			}

			/*
			 * Jump to the lower nice value when
			 * more than one job is running.
			 */
			if (strncasecmp(op, "jump", 4) == 0) {
				qp->q_ptype |= NICE_JUMP;
				dbg((BLOG_QUEUES, "nice doing jump"));
			}

			/*
			 * Use the load to determine the factor.
			 */
			if (strncasecmp(op, "load", 4) == 0) {
				qp->q_ptype |= NICE_LOAD;
				dbg((BLOG_QUEUES, "nice using load"));
			}

			if (isdigit((int) *op)) {
				if (ni_first) {
					qp->q_priority[NICE_HIGH] = atoi(op);
					dbg((BLOG_QUEUES, "nice high = %d",
					     qp->q_priority[NICE_HIGH]));
					ni_first = FALSE;
				} else {
					qp->q_priority[NICE_LOW] = atoi(op);
					dbg((BLOG_QUEUES, "nice low = %d",
					     qp->q_priority[NICE_LOW]));
				}
			}

		} while ((op = strtok((char *) NULL, ",")) != NULL);

		/*
		 * If more than one of interp/jump/step are set, issue
		 * a warning message.
		 */
		if (((qp->q_ptype & NICE_INTERP) &&
		     (qp->q_ptype & (NICE_STEP|NICE_JUMP))) ||
		    ((qp->q_ptype & NICE_JUMP) &&
		     (qp->q_ptype & (NICE_INTERP|NICE_STEP))) ||
		    ((qp->q_ptype & NICE_STEP) &&
		     (qp->q_ptype & (NICE_INTERP|NICE_JUMP)))) {
			blog(BLOG_WARNING, ques, qp->q_name,
			     "bad \"ns\" parameter - choose one of interp/jump/step");
			qp->q_ptype |= NICE_INTERP;
			qp->q_ptype &= ~(NICE_STEP);
			qp->q_ptype &= ~(NICE_JUMP);
		}

		if ((qp->q_ptype & NICE_GANG) && (qp->q_ptype & NICE_SEPARATE)) {
			blog(BLOG_WARNING, ques, qp->q_name,
			     "bad \"ns\" parameter - choose one of gang/separate");
			qp->q_ptype |= NICE_GANG;
			qp->q_ptype &= ~(NICE_SEPARATE);
		}

		if (qp->q_priority[NICE_LOW] != qp->q_priority[NICE_HIGH] &&
		    qp->q_ptype == 0) {
			qp->q_ptype = NICE_INTERP;
			dbg((BLOG_QUEUES, ques, qp->q_name,
			     "using \"ns\" parameter \"interp\""));
		}

	} else {
		blog(BLOG_WARNING, ques, qp->q_name,
		     "bad \"ns\" parameters - using \"ni\" values");
	}

	/* Non degrading priority */
	if ((qp->q_ndpriority = bgetnum("pi")) < 0)
		qp->q_ndpriority = -1;

#ifdef HAVE_RLIMIT
	/*
	 * Resource limits.
	 */
	(void) setcurlim(qp, RLIMIT_CPU, bgetnum("rp"));

	(void) setcurlim(qp, RLIMIT_FSIZE, (bgetnum("rf") * 1024) );

	(void) setcurlim(qp, RLIMIT_DATA, (bgetnum("rd") * 1024) );

	(void) setcurlim(qp, RLIMIT_STACK, (bgetnum("rs") * 1024) );

	(void) setcurlim(qp, RLIMIT_CORE, (bgetnum("rc") * 1024) );

	(void) setcurlim(qp, RLIMIT_RSS, (bgetnum("rr") * 1024) );

#ifdef RLIMIT_VMEM
	(void) setcurlim(qp, RLIMIT_RSS, (bgetnum("rv") * 1024) );
#endif

#ifdef RLIMIT_NOFILE
	(void) setcurlim(qp, RLIMIT_NOFILE, bgetnum("ro") );
#endif

	(void) setmaxlim(qp, RLIMIT_CPU, bgetnum("xp") );

	(void) setmaxlim(qp, RLIMIT_FSIZE, (bgetnum("xf") * 1024) );

	(void) setmaxlim(qp, RLIMIT_DATA, (bgetnum("xd") * 1024) );

	(void) setmaxlim(qp, RLIMIT_STACK, (bgetnum("xs") * 1024) );

	(void) setmaxlim(qp, RLIMIT_CORE, (bgetnum("xc") * 1024) );

	(void) setmaxlim(qp, RLIMIT_RSS, (bgetnum("xr") * 1024) );

#ifdef RLIMIT_VMEM
	(void) setmaxlim(qp, RLIMIT_RSS, (bgetnum("xv") * 1024) );
#endif

#ifdef RLIMIT_NOFILE
	(void) setmaxlim(qp, RLIMIT_NOFILE, bgetnum("xo") );
#endif
#endif /* HAVE_RLIMIT */

	if ((qp->q_idle = bgetnum("gr")) < 0)
		qp->q_idle = -1;

	if ((qp->q_mustrun = bgetnum("mr")) < 0)
		qp->q_mustrun = -1;

	/*
	 * These are bitfields in the structure, they can only be 
	 * TRUE or FALSE, so we check for a negative return (absence of
	 * flag) in a temporary var until we are sure it's not -1.
	 */
	if ((flag = bgetflag("re")) < 0)
		qp->q_restart = TRUE;			/* default is TRUE */
	else
		qp->q_restart = flag;			/* set to result */

	if ((flag = bgetflag("lb")) < 0)
		qp->q_loadb = FALSE;			/* default is FALSE */
	else
		qp->q_loadb = flag;			/* set to result */

#ifdef USE_CUMULTIME
	/* Do we use the cumulative CPU time? */
	if ((flag = bgetflag("cu")) < 0)
		qp->q_cumetime = TRUE;			/* default is TRUE */
	else
		qp->q_cumetime = flag;			/* set to result */

#else
	qp->q_cumetime = FALSE;			/* Always FALSE */
	
#endif

	/* 
	 * Check to see wether we shuld consider using the users resource
	 * file.  Some administrators may want to turn this off.
	 */
	if ((flag = bgetflag("ur")) < 0)
		qp->q_userrc = TRUE;			/* default is TRUE */
	else
		qp->q_userrc = flag;			/* set to result */

	/*
	 * Clear the sort type field.
	 */
	qp->q_ftype  = 0;
	qp->q_nfunc  = 0;
	qp->q_climit = 0;
	qp->q_ulimit = 0;

	/*
	 * If there is no sort flag, sort by time only
	 * If there is a sort flag, parse the sort for keywords.
	 * If there is an error in the parse, sort by time.
	 */
	if ((sp = bgetstr("qs", &bp)) == NULL) {
		qp->q_ftype = (SORT_TIME | SORT_UID);
		qp->q_func[qp->q_nfunc++] = job_comp_mtime;
		dbg((BLOG_QUEUES, "default sort"));
	} else if ((op = strtok(sp, ",")) != NULL) {
		do {

			/*
			 * Sort by a total CPU time used (seconds)
			 * if "usage.150" sort into two secondary groups.
			 * one less than 150 CPUsec, and one greater than 150
			 * CPUsec.
			 */
			if (strncasecmp(op, "usage", 5) == 0) {
				qp->q_ftype |= SORT_USAGE;
				qp->q_func[qp->q_nfunc++] = job_comp_usage;
				if ((cp = strchr(op, TIMESEP)) != NULL) {
					cp++;
					qp->q_ulimit = atoi(cp);
				}
				dbg((BLOG_QUEUES, "sort usage (%d)",
				     qp->q_ulimit));
			}

			/*
			 * Sort by a completed jobs
			 * if "complete.10" sort into two secondary groups.
			 * one less than 10 jobs, one greater than 10 jobs.
			 */
			if (strncasecmp(op, "complete", 6) == 0) {
				qp->q_ftype |= SORT_COMPLETE;
				qp->q_func[qp->q_nfunc++] = job_comp_complete;
				if ((cp = strchr(op, TIMESEP)) != NULL) {
					cp++;
					qp->q_climit = atoi(cp);
				}
				dbg((BLOG_QUEUES, "sort complete (%d)",
				     qp->q_climit));
			}

			/* Reset the tallies daily */
			if (strncasecmp(op, "daily", 5) == 0) {
				qp->q_ftype |= SORT_DAILY;
				dbg((BLOG_QUEUES, "clear daily"));
			}

			/* Reset the tallies weekly */
			if (strncasecmp(op, "weekly", 6) == 0) {
				qp->q_ftype |= SORT_WEEKLY;
				dbg((BLOG_QUEUES, "clear weekly"));
			}

			if (strncasecmp(op, "time", 4) == 0) {
				qp->q_ftype |= SORT_TIME;
				qp->q_func[qp->q_nfunc++] = job_comp_mtime;
				dbg((BLOG_QUEUES, "sort time"));
			}

			if ((strncasecmp(op, "round", 5) == 0) ||
			    (strncasecmp(op, "robin", 5) == 0)) {
				qp->q_ftype |= SORT_ROUND;
				qp->q_func[qp->q_nfunc++] = job_comp_queued;
				dbg((BLOG_QUEUES, "sort round robin"));
			}				

			if ((strncasecmp(op, "user", 4) == 0) ||
			    (strncasecmp(op, "uid", 3) == 0)) {
				qp->q_ftype |= SORT_UID;
				dbg((BLOG_QUEUES, "cluster by uid"));
			}

			if ((strncasecmp(op, "group", 5) == 0) ||
			    (strncasecmp(op, "gid", 3) == 0)) {
				qp->q_ftype |= SORT_GID;
				dbg((BLOG_QUEUES, "cluster by gid"));
			}

			if (qp->q_nfunc >= MAXSORTFUNC - 1)
				blog(BLOG_WARNING, ques,
				    qp->q_name, "too many sort functions");
				    
		} while ((op = strtok((char *) NULL, ",")) != NULL);

		/*
		 * If both are set, or neither are set, but 
		 * usage or complete are, then warn the user.
		 */
		if ((qp->q_ftype & SORT_DAILY) && (qp->q_ftype & SORT_WEEKLY))
		    blog(BLOG_WARNING, ques, qp->q_name,
			    "bad \"qs\" parameter - choose either daily/weekly");

		if (((qp->q_ftype & SORT_USAGE) || (qp->q_ftype & SORT_COMPLETE)) &&
		     !(qp->q_ftype & (SORT_DAILY|SORT_WEEKLY)))
		    blog(BLOG_WARNING, ques, qp->q_name,
			    "bad \"qs\" parameter - must set daily or weekly");

		if (((qp->q_ftype & SORT_UID) && (qp->q_ftype & SORT_GID)) ||
		    (!(qp->q_ftype & SORT_UID) && !(qp->q_ftype & SORT_GID)))
			blog(BLOG_WARNING, ques, qp->q_name,
			    "bad \"qs\" parameter - choose one uid/gid");

	} else {
		qp->q_ftype = (SORT_TIME | SORT_UID);
		qp->q_func[qp->q_nfunc++] = job_comp_mtime;
		blog(BLOG_WARNING, ques, qp->q_name, "bad \"qs\" parameter");
	}

	/* If time has not been specified, make it the last function to use */
	if (!(qp->q_ftype & SORT_TIME)) {
		qp->q_ftype |= SORT_TIME;
		qp->q_func[qp->q_nfunc++] = job_comp_mtime;
		dbg((BLOG_QUEUES, "%s: adding trailing time sort",
		     qp->q_name));
	}

	/*
	 * If the sorting type is changed, then we have changed from
	 * uid based data to gid based data or vice versa.  SO 
	 * free all the Usr type structures for this queue.
	 *
	 * If this happens, that means the Usr structures which jobs point
	 * to will be invalidated.  Re-assign them here.
	 */
	if ((oldtype & (SORT_UID|SORT_GID)) !=
	    (qp->q_ftype & (SORT_UID|SORT_GID))) {

		/*
		 * Iterate over the Usr pointers and free all the structures.
		 */
		ITER(up, Usr, &(qp->q_usrs)) {
			rem_que(&up->u_usrs);
			efree(up);
			up = FIRST(Usr, &(qp->q_usrs));
		}
		qinfo.i_usrlist = 0;
		oldtype = qp->q_ftype;

		ITER(jp, Job, &(qp->q_jobs)) {
			if ((up = getusrbyid(&(qp->q_usrs),
					     qp->q_ftype, jp)) == NULL) {
				up = usr_alloc(&(qp->q_usrs));
				if (qp->q_ftype & SORT_UID)
					up->u_uid = jp->j_uid;
				else if (qp->q_ftype & SORT_GID)
					up->u_gid = jp->j_gid;
				else
					blog(BLOG_ERR, "usr type error");
			} 

			/* Set pointer back to usr struct */ 
			jp->j_usr = up;
		}

	}

	if (chdir(qp->q_spool) < 0) {
		blog(BLOG_ERR, echad, qp->q_spool);
		return (-1);
	}

	/*
	 * Start from a fresh state, intialize the time,
	 * and set the status message.
	 */
	if (first == TRUE) {

		qp->q_state = CLEAR;
		queue_set_flag(qp, RESET);
		queue_set_flag(qp, QUEUING);
		queue_set_lock(qp, QUEUING);
		queue_set_flag(qp, RUNNING);
		queue_set_lock(qp, RUNNING);
		qp->q_inactive = TRUE;
		qp->q_timein = TRUE;

		/*
		 * These conditionals examine the exterior state of
		 * the queue and use this to decide whether or not to
		 * "activate" the queues.  Hence previous states of
		 * the queues as definded by the control program are
		 * "sticky" across batch daemon restarts.
		 */
		if (queue_get_lock(qp, DISABLED) == 0) {
			queue_clr_flag(qp, QUEUING);
			queue_set_flag(qp, DISABLED);
		}

		if (queue_get_lock(qp, STOPPED) == 0) {
			queue_clr_flag(qp, RUNNING);
			queue_set_flag(qp, STOPPED);
		}

		qp->q_uctl =  CLEAR;
		qp->q_uctl |= ANYRUN;
		qp->q_uctl |= ANYSUB;


		/*
		 * The time status needs to be initialized on startup.
		 * The code in runjob checks to see if we stepped across 
		 * time boundaries, but it does not do an absolute test.
		 */
		if (qp->q_days[day].d_start == qp->q_days[day].d_stop) {
			if (qp->q_days[day].d_start < 0) {
				queue_clr_flag(qp, RUNNING);
				queue_set_flag(qp, TIMEOUT);
				queue_set_lock(qp, TIMEOUT);
				qp->q_timein = FALSE;
			}
		} else if (((minute >= qp->q_days[day].d_stop &&
			   minute < qp->q_days[day].d_start) &&
			    qp->q_days[day].d_stop <
			    qp->q_days[day].d_start) ||
			   ((minute >= qp->q_days[day].d_stop ||
			   minute < qp->q_days[day].d_start) &&
			    qp->q_days[day].d_stop >
			    qp->q_days[day].d_start)) {
			queue_clr_flag(qp, RUNNING);
			queue_set_flag(qp, TIMEOUT);
			queue_set_lock(qp, TIMEOUT);
			qp->q_timein = FALSE;
		}

		/*
		 * Set the status of the queue on startup
		 */
		cp = statetoa(qp);
		queue_write_status(qp, cp);

        }

#ifdef DEBUG
	dump_queue(blog, BLOG_QUEUES, qp);
#endif

	if (chdir(home) < 0) {
		blog(BLOG_ERR, echad, home);
		return (-1);
	}

	return 0;

}

static void
queue_free_struct(qp)
Queue *	qp;
{


	efree((voidp) qp->q_rhost);
	qp->q_rhost = NULL;

	efree((voidp) qp->q_rname);
	qp->q_rname = NULL;

	efree((voidp) qp->q_spool);
	qp->q_spool = NULL;

	efree((voidp) qp->q_sname);
	qp->q_sname = NULL;

	efree((voidp) qp->q_lname);
	qp->q_lname = NULL;

	efree((voidp) qp->q_aname);
	qp->q_aname = NULL;

	efree((voidp) qp->q_ename);
	qp->q_ename = NULL;

	efree((voidp) qp->q_smaddress);
	qp->q_smaddress = NULL;

	efree((voidp) qp->q_swaddress);
	qp->q_swaddress = NULL;


}


#ifdef HAVE_RLIMIT


/*
 * Set a rlimit current limit value in a queue.
 */
static int
setcurlim(qp, res, val)
Queue *	qp;
int	res;
int	val;
{

	if (val < 0)
		return (-1);

	dbg(((BLOG_QUEUES|BLOG_RESOURCE),
	     "set %s %s cur limit to %d",
	    qp->q_name, rlimittoa(res), val));

	qp->q_rlimit[res].rlim_cur = val;
	
	return 0;
}


/*
 * Set a rlimit maximum limit value for a queue.
 */
static int
setmaxlim(qp, res, val)
Queue *	qp;
int	res;
int	val;
{

	if (val < 0)
		return (-1);

	dbg(((BLOG_QUEUES|BLOG_RESOURCE),
	     "set %s %s max limit to %d",
	     qp->q_name, rlimittoa(res), val));

	qp->q_rlimit[res].rlim_max = val;
	
	return 0;
}


/*
 * Turn resource limit definition into a string.
 */
char *
rlimittoa(i)
int	i;
{

	switch (i) {
	case RLIMIT_CPU:
		return "CPU time";
		break;

	case RLIMIT_FSIZE:
		return "open file size";
		break;

	case RLIMIT_DATA:
		return "data size";
		break;

	case RLIMIT_STACK:
		return "stack size";
		break;

	case RLIMIT_CORE:
		return "coredump size";
		break;

	case RLIMIT_RSS:
		return "resident size";
		break;

#ifdef RLIMIT_NOFILE
	case RLIMIT_NOFILE:
		return "open file desc";
		break;
#endif		

#ifdef RLIMIT_VMEM
	case RLIMIT_VMEM:
		return "mmapped size";
		break;
#endif		
	}

	blog(BLOG_ERR, "rlimittoa(%d) invalid", i );
	return (char *) NULL;
}


#endif /* HAVE_RLIMIT */


/*
 * Write out queue information using the supplied function.
 */
void
dump_queue(func, lvl, qp)
void	(*func)V((int, char *, ...));
int	lvl;
Queue *	qp;
{

	char	priv[LINE];
	char	cstring[LINE], mstring[LINE];
	char *	cp;
	char	cunits, munits;
	int	i, cdiv, mdiv, cinf, minf;


	/*
	 * Print out information to the log file indicating status of
	 * the queue as it starts up.
	 * Or to where ever, given a function to print with.
	 */
	(*func)(lvl, "Canonical name:  %s", qp->q_name);
	(*func)(lvl, "Spool directory: %s", qp->q_spool);
	(*func)(lvl, "Status file:     %s", qp->q_sname);
	(*func)(lvl, "Lock file:       %s", qp->q_lname);
	(*func)(lvl, "Accounting file: %s", qp->q_aname);
	(*func)(lvl, "Error file:      %s", qp->q_ename);
	(*func)(lvl, "Supervisor:      %s", qp->q_smaddress);

	(*func)(lvl, "Mail user :      \"%s%s%s%s%s%s\"",
		(qp->q_usermail & NOTIFY_START) ? "s" : "",
		(qp->q_usermail & NOTIFY_END) ?   "e" : "",
		(qp->q_usermail & NOTIFY_CRASH) ? "c" : "",
		(qp->q_usermail & NOTIFY_DEBUG) ? "d" : "",
		(qp->q_usermail & NOTIFY_PAUSE) ? "p" : "",
		(qp->q_usermail & NOTIFY_TRANSFER) ? "t" : "",
		(qp->q_usermail & NOTIFY_RESOURCE) ? "r" : "");
	(*func)(lvl, "Mail super:      \"%s%s%s%s%s%s\"",
		(qp->q_supmail & NOTIFY_START) ?  "s" : "",
		(qp->q_supmail & NOTIFY_END) ?    "e" : "",
		(qp->q_supmail & NOTIFY_CRASH) ?  "c" : "",
		(qp->q_supmail & NOTIFY_DEBUG) ?  "d" : "",
		(qp->q_supmail & NOTIFY_PAUSE) ?  "p" : "",
		(qp->q_supmail & NOTIFY_TRANSFER) ?  "t" : "",
		(qp->q_supmail & NOTIFY_RESOURCE) ?  "r" : "");

	(void) strcpy(priv, ctime(&(qp->q_mtime)));
	priv[19] = '\0';	/* cut off just after seconds */

	(*func)(lvl, "Last modified:   %s", priv);
	(*func)(lvl, "Current State:   %s (%o)",
		statetoa(qp), qp->q_state);

	(*func)(lvl, "Suspend queueing@%d 5min runque load", qp->q_loadsched);
	(*func)(lvl, "Stop jobs @      %d 5min runque load", qp->q_loadstop);
	(*func)(lvl, "Nice priority:   %d (high), %d (low)",
		qp->q_priority[NICE_HIGH], qp->q_priority[NICE_LOW]);
	(*func)(lvl, "NonDegrad. pri.: %d", qp->q_ndpriority);
	(*func)(lvl, "Default CPU#:    %d; Min. console idle time: %ds",
		qp->q_mustrun, qp->q_idle);
	(*func)(lvl, "Concurrent jobs: %d", qp->q_maxexec);

#ifdef HAVE_RLIMIT 

	for (i = 0; i < RLIM_NLIMITS; i++ ){
		cinf = 0;	minf = 0;
		cdiv = 1024;	mdiv = 1024;
		cunits = 'K';	munits = 'K';

		if ( i == RLIMIT_CPU) {
			if (qp->q_rlimit[i].rlim_cur == RLIM_INFINITY) {
				cinf = 1;
			} else if (qp->q_rlimit[i].rlim_cur > 3600) {
				cdiv = 3600;
				cunits = 'h';
			} else {
				cdiv = 1;
				cunits = 's';
			}
			if (qp->q_rlimit[i].rlim_max == RLIM_INFINITY) {
				minf = 1;
			} else if (qp->q_rlimit[i].rlim_max > 3600) {
				mdiv = 3600;
				munits = 'h';
			} else {
				mdiv = 1;
				munits = 's';
			}
		} else 
#ifdef RLIMIT_NOFILE
		if ( i == RLIMIT_NOFILE ) {
			cdiv = 1;
			mdiv = 1;
			cunits = ' ';
			munits = ' ';
		} else
#endif
		{

			if (qp->q_rlimit[i].rlim_cur == RLIM_INFINITY) {
				cinf = 1;
			} else if (qp->q_rlimit[i].rlim_cur > 1048576) {
				cdiv = 1048576;
				cunits = 'M';
			} else if (qp->q_rlimit[i].rlim_cur > 1024) {
				cdiv = 1024;
				cunits = 'K';
			} else {
				cdiv = 1;
				cunits = 'B';
			}
			if (qp->q_rlimit[i].rlim_max == RLIM_INFINITY) {
				minf = 1;
			} else if (qp->q_rlimit[i].rlim_max > 1048576) {
				mdiv = 1048576;
				munits = 'M';
			} else if (qp->q_rlimit[i].rlim_max > 1024) {
				mdiv = 1024;
				munits = 'K';
			} else {
				mdiv = 1;
				munits = 'B';
			}
		}

		if (cinf) {
			(void) strcpy(cstring, " infinite");
		} else {
			sprintf(cstring, "%8d%c",
				qp->q_rlimit[i].rlim_cur / cdiv, cunits);
		}

		if (minf) {
			(void) strcpy(mstring, " infinite");
		} else {
			sprintf(mstring, "%8d%c",
				qp->q_rlimit[i].rlim_max / mdiv, munits);
		}

		(*func)(lvl, "Resource limits: %-16.16s -> cur: %s max: %s",
			rlimittoa(i), cstring, mstring);
	}

#endif /* HAVE_RLIMIT */

	for (i = 0; i < 7; i++) {
		(*func)(lvl, "Time limits day: %d: start@%02dh, stop@%02dh",
			i, qp->q_days[i].d_start/60, qp->q_days[i].d_stop/60);
	}

	if (qp->q_ftype & (SORT_ROUND))
		cp = "round robin";
	if (qp->q_ftype & SORT_TIME)
		cp = "time";
	if (qp->q_ftype & (SORT_USAGE|SORT_DAILY))
		cp = "daily usage";
	if (qp->q_ftype & (SORT_USAGE|SORT_WEEKLY))
		cp = "weekly usage";

	(*func)(lvl, "Jobs queued by:  %s and clustered by %s", cp,
		(qp->q_ftype & SORT_UID) ? "user" : "group");

	(*func)(lvl, "Restart allowed: %s",
		qp->q_restart == TRUE ? yes : no);

	(*func)(lvl, "Load balance:    %s",
		qp->q_loadb == TRUE ? yes : no);

	(*func)(lvl, "CPU time acct:   %s",
		qp->q_cumetime == TRUE ? "per job" : "per process");

	return;
}


/*
 * Create a phrase for the status line.
 */
char *
statetoa(qp)
Queue *	qp;
{

	char *	substring, *	exestring;
	static char myline[LINE];


	if (queue_get_flag(qp, WAITING) == 0)
		substring = namestab[X_WAITING].n_down;
	else if (queue_get_flag(qp, DISABLED) == 0)
		substring = namestab[X_DISABLED].n_down;
	else
		substring = namestab[X_QUEUING].n_up;

	if  (queue_get_flag(qp, STOPPED) == 0)
		exestring = namestab[X_STOPPED].n_down;
	else if  (queue_get_flag(qp, TIMEOUT) == 0)
		exestring = namestab[X_TIMEOUT].n_down;
	else if  (queue_get_flag(qp, SUSPEND) == 0)
		exestring = namestab[X_SUSPEND].n_down;
	else if (queue_get_flag(qp, INTERRUPT) == 0)
		exestring = namestab[X_INTERRUPT].n_down;
	else if (queue_get_flag(qp, (DRAINING|DRAINED)) == 0) {
		substring = namestab[X_DISABLED].n_down;
		exestring = namestab[X_DRAINING].n_down;
	}
	  else
		exestring = namestab[X_RUNNING].n_up;

	(void) sprintf(myline, "%s: submissions %s; executions %s",
		       qp->q_name, substring, exestring);

	return myline;
}


/*
 * Extract from the data string, the days and the times to start and stop
 * the queue.
 * The syntax of the data is "day.start.stop,..."
 */
static int
days_parse(qp, data)
Queue *	qp;
char *	data;
{
	int	dsep = DATESEP;
	int	tsep = TIMESEP;
	char  *	dp;
	char *	tp;
	int	day;
	int	begintime;
	int	endtime;

	dp = data;


	while (*dp == dsep) 
		dp++;

	while (*dp != '\0') {

		/*
		 * Extract the day...
		 */
		if ((day = times_extract(dp, tsep, dsep, tsep)) == -2)
			return (-1);

		if ((tp = strchr(dp, tsep)) == NULL)
			return (-1);
		else
			dp = tp+1;

		/*
		 * then the starting time...
		 */
		if ((begintime = times_extract(dp, tsep, dsep, tsep)) == -2)
			return (-1);

		if ((tp = strchr(dp, tsep)) == NULL)
			return (-1);
		else
			dp = tp+1;

		/*
		 * then the stopping time.
		 */
		if ((endtime = times_extract(dp, dsep, dsep, tsep)) == -2)
			return (-1);

		if ((tp = strchr(dp, dsep)) == NULL)
			goto done;
		else
			dp = tp+1;

		while (*dp == dsep) 
			dp++;

		/*
		 * Assign the times into the day structure for this queue.
		 */
		(void) days_assign(qp->q_days, day, begintime, endtime);


	}
 done:
	(void)  days_assign(qp->q_days, day, begintime, endtime);
	return 0;

}


/*
 * Assign the time values to the structure using some simple
 * modification rules.
 */
static int
days_assign(days, day, start, stop)
Day	days[];
int	day;
int	start;
int	stop;
{


	start *= MINPERHOUR;
	stop  *= MINPERHOUR;

	dbg((BLOG_QUEUES, "orignally for day %d start: %d stop: %d",
	    day, start, stop));

	/*
	 * If the day is negative, that means turn off pocessing that day.
	 */
	if (day < 0) {
		days[abs(day)].d_start = -1;
		days[abs(day)].d_stop  = -1;
	}
	day = abs(day);

	/*
	 * If the time is not the default (-1) set the day structure to the
	 * value obtained in the parse.
	 */
	if (start != -1)
		days[day].d_start = start;

	/*
	 * If the starting time is later than the stopping time, then
	 * adjust the day value so that we stop in the the *next* day.
	 */
	if (start > stop) {
		if (day == 6)
			day = 0;
		else
			day++;
	}

	if (stop != -1)
		days[day].d_stop = stop;

	dbg((BLOG_QUEUES, "ultimately for day %d start: %d stop: %d",
	    day, days[day].d_start, days[day].d_stop));

	return 0;

}


/*
 * Extract individual values from the string.
 * Do simple syntax checking on the expecteed value of the next character.
 */
static int
times_extract(dp, sep, dsep, tsep)
char *	dp;
int	sep;
int	dsep;
int	tsep;
{

	char *	tp;
	char	temp[10];

	tp = temp;

	/*
	 * While the next character is not a string separator, copy
	 * the data into a temp buffer.
	 */
	while (*dp != dsep && *dp != tsep && *dp != '\0')
		*tp++ = *dp++;

	/* Terminate the integer string */
	*tp = '\0';

	/*
	 * Error in parsing the string, the next character is wrong.
	 */
	if (sep == TIMESEP && sep != *dp)
		return (-2);

	/*
	 * Translate the string into an integer.
	 */
	return atoi(temp);

}
