#ifndef lint
static char rcsid[] =
	"$Id: recvjob.c,v 4.0 1993/09/17 00:41:39 srp Exp srp $";
#endif

/*
 *
 * Author: Scott R. Presnell
 *
 * Copyright (C) 1989, 1990, The Regents of the University of California
 *                     All Rights Reserved
 *
 * THE REGENTS OF THE UNIVERSITY OF CALIFORNIA MAKE NO REPRESENTATION OR
 * WARRANTIES WITH RESPECT TO THE CONTENTS HEREOF AND SPECIFICALLY DISCLAIM
 * ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
 * PURPOSE.
 *
 * Further, the Regents of the University of California reserve the right to
 * revise this software and/or documentation and to make changes from time
 * to time in the content hereof without obligation of the Regents of the
 * University of California to notify any person of such revision or change.
 *
 */


/*
 * recvjob.c -  Receive batch jobs from batchds on other hosts.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <errno.h>
#include <fcntl.h>
#include <signal.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/resource.h>

#include "config.h"

#include H_STRING

#include "batch.h"
#include "batchd.h"
#include "protocol.h"
#include "flags.h"


static SIGRTN	quit P((int));
static SIGRTN	die P((int));

static void	rcleanup P((void));
static int	job_spool_file P((char *, int));

static char    tfname[LINE];		/* tmp copy of cf before linking */
static char    dfname[LINE];		/* data files */


/*
 * Handle an ipc request to receive a job (two files)
 * Then kick the batchd to queue up the new job.
 */
int
job_receive(queue, host)
char *	queue, *host;
{

	char *	cp;
	int	size;
	int	nfiles = 0;
	int	result = 0;
	Queue *	qp;


	/*
	 * Find the queue ny name
	 */
	if ((qp = getqueuebyname(queue)) == (Queue *) NULL) {
		blog(BLOG_WARNING, enotq, queue);
		nack();
		return (RTN_ERROR);
	}

	/*
	 * Move to the spooling directory for the queue
	 */
	if (chdir(qp->q_spool) < 0) {
		blog(BLOG_ERR, echad, qp->q_spool);
		return (RTN_ERROR);
	}

	/*
	 * Check this queue's current ability to accept jobs.
	 * Draining is a special case.
	 */
	if (queue_get_flag(qp, DRAINING) == 0) {
		dbg((BLOG_XFER, "queue currently draining..."));
		mack();
		return (RTN_ACCESS);
	}

	if (queue_get_flag(qp, STOPPED) == 0) {
		dbg((BLOG_XFER, "queue currently stopped..."));
		mack();
		return (RTN_ACCESS);
	}

	if (queue_get_flag(qp, QUEUING) != 0) {
		dbg((BLOG_XFER, "queue not available to receive..."));
		aack();
		return (RTN_RETRY);
	}

	if (qp->q_count >= MAXJOBS) {
		dbg((BLOG_XFER, "all queue slots are full..."));
		aack();
		return (RTN_RETRY);
	}

	(void) signal(SIGHUP, quit);
	(void) signal(SIGINT, die);
	(void) signal(SIGIOT, SIG_IGN);
	(void) signal(SIGALRM, SIG_IGN);
	(void) signal(SIGPIPE, SIG_IGN);
	(void) signal(SIGUSR1, SIG_IGN);
	(void) signal(SIGUSR2, SIG_IGN);
	(void) signal(SIGTERM, die);

	dbg((BLOG_XFER, "ready to receive job (acking)..."));

	ack();

	queue_write_status(qp, "%s: receiving job from %s",
			   qp->q_name, host);

	for ( ; ; ) {

		/*
		 * Read a command to tell us what to do
		 */
		cp = line;
		do {
			if ((size = read(1, cp, 1)) != 1) {
				rcleanup();
				blog(BLOG_ERR, elost, "receivejob()");
				return (RTN_ERROR);
			}
		} while (*cp++ != '\n');

		*--cp = '\0';
		cp = line;

		switch (*cp++) {

			/*
			 * Read df file - it always comes first.
			 */
		case QC_RECVDF:	
			size = 0;

			while (*cp >= '0' && *cp <= '9')
				size = size * 10 + (*cp++ - '0');

			if (*cp++ != ' ')
				break;

			(void) strcpy(dfname, cp);

			dbg((BLOG_XFER, "accepting df file: %s", cp));

			if ((result = job_spool_file(dfname, size)) < 0) {
				rcleanup();
				return (result);
			} else if (result > 0) {
				continue;
			}

			if (chmod(dfname, DFMOD) < 0)
				blog(BLOG_ERR, echmo, cp);

			dbg((BLOG_XFER, "receieved df file: %s", cp));

			continue;

			/*
			 * Read the cf file in as a tf file, the rename
			 * after the read is over.
			 */
		case QC_RECVCF:
			size = 0;

			while (*cp >= '0' && *cp <= '9')
				size = size * 10 + (*cp++ - '0');

			if (*cp++ != ' ')
				break;

			(void) strcpy(tfname, cp);
			tfname[0] = 't';

			dbg((BLOG_XFER, "accepting tf file: %s", tfname));

			if ((result = job_spool_file(tfname, size)) < 0) {
				rcleanup();
				return (result);
			} else if (result > 0) {
				continue;
			}

			if (link(tfname, cp) < 0) {
				rcleanup();
				blog(BLOG_ERR, eopen, tfname);
				return (RTN_ACCESS);
			}

			(void) unlink(tfname);

			/*
			 * Name has now been changed in the file system
			 */
			tfname[0] = 'c';
			nfiles++;

			if (chmod(cp, CFMOD) < 0)
				blog(BLOG_ERR, echmo, cp);

			dbg((BLOG_XFER, "receieved cf file: %s", cp));

			continue;

			/*
			 * Sent the job sucessfully.
			 */
		case QC_DONE:
			if (nfiles <= 0) {
				rcleanup();
				blog(BLOG_WARNING,
				     "got a QC_DONE but nfiles <= 0 ?");
				return (RTN_ERROR);
			}

			dbg((BLOG_XFER, "both files received - alarming"));

			queue_write_status(qp, "%s: received job from %s",
					   qp->q_name, host);

			if (kill(batchd_pid, SIGALRM) < 0) {
				(void) ewriteone(1, QR_EPERM);
				return (RTN_ERROR);
			} else {
				(void) ewriteone(1, QR_OK);
				return (RTN_OK);
			}
			break;

			/*
			 * Cleanup because data sent was bad  -
			 * there is no recvovery from an abort.
			 */
		case QC_ABORT:	
			dbg((BLOG_XFER, "oops - got an abort"));

			rcleanup();
			return (RTN_ERROR); 
			break;

			/*
			 * Didn't understand the message - send QR_ENOIMP.
			 */
		default:
			huh();
			break;

		}
	}

	/*NOTREACHED*/

}


/*
 * Read individual job files send by batchd and copy them to the
 * spooling directory.
 */
static int
job_spool_file(file, size)
char *	file;
int	size;
{

	char *	cp;
	int	i, j;
	int	fd, error;
	unsigned int amt;
	char	recvbuf[BUFSIZ];


	if ((fd = open(file, (O_WRONLY|O_CREAT), CFMOD)) < 0) {
		eack();
		blog(BLOG_ERR, eopen, file);
		return (RTN_ACCESS);
	} else {
		ack();
		error = 0;
	}

	for (i = 0; i < size; i += BUFSIZ) {
		amt = BUFSIZ;
		cp = recvbuf;
		if (i + amt > size)
			amt = size - i;
		do {
			/*
			 * If we loose the signal we just pack it in.
			 * Don't even return - just bag it.
			 */
			if ((j = read(1, cp, amt)) <= 0) {
				(void) close(fd);
				blog(BLOG_ERR, elost, "spoolfile()");
				return (RTN_ERROR);
			}

			amt -= j;
			cp += j;

		} while (amt > 0);

		amt = BUFSIZ;
		if (i + amt > size)
			amt = size - i;
		if (ewrite(fd, recvbuf, amt, "spoolfile()") != amt) {
			error++;
			break;
		}
	}

	(void) close(fd);

	if (error) {
		aack();
		(void) unlink(file);
		blog(BLOG_ERR, "write error on %s", file);
		return (RTN_RETRY);
	} else {
		ack();
		return (RTN_OK);
	}

	/*NOTREACHED*/
}


/*
 * Remove all the files associated with the current job being transfered.
 */
static void
rcleanup()
{

	if (tfname[0])
		(void) unlink(tfname);

	if (dfname[0])
		(void) unlink(dfname);

}


/*
 * Function (intended for SIGHUP) to "clear" the process if it get's stuck.
 */
static SIGRTN
quit(sig)
int	sig;
{
	blog(BLOG_WARNING, "receive aborted by signal %d", sig);
	exit(RTN_RETRY);
}


/*
 * Function (indended for SIGINT, SIGTERM) to nuke the job and the process.
 */
static SIGRTN
die(sig)
int	sig;
{
	rcleanup();
	blog(BLOG_WARNING, "receive and job aborted by signal %d", sig);
	exit(RTN_SIGNAL);
}

