#ifndef lint
static char rcsid[] =
	"$Id: request.c,v 4.0 1993/09/17 00:41:39 srp Exp srp $";
#endif

/*
 *
 * Author: Scott R. Presnell
 *
 * Copyright (C) 1989, 1990, The Regents of the University of California
 *                     All Rights Reserved
 *
 * THE REGENTS OF THE UNIVERSITY OF CALIFORNIA MAKE NO REPRESENTATION OR
 * WARRANTIES WITH RESPECT TO THE CONTENTS HEREOF AND SPECIFICALLY DISCLAIM
 * ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
 * PURPOSE.
 *
 * Further, the Regents of the University of California reserve the right to
 * revise this software and/or documentation and to make changes from time
 * to time in the content hereof without obligation of the Regents of the
 * University of California to notify any person of such revision or change.
 *
 */

/*
 * Batch service request functions (used to be bcommon.c).
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <errno.h>
#include <ctype.h>

#include <sys/types.h>
#include <sys/param.h>

#include "config.h"

#include H_STRING

#include "batch.h"
#include "protocol.h"
#include "reply.h"

struct request_t {
	int	r_ident;		/* Identifier */
	char *	r_name;			/* Name of the command */
	int   	(* r_okhandler)();	/* Function to use on OK */
	char *	r_okfmt;		/* and the format to expect */
	int	(* r_mthandler)();	/* Function to use on MT */
	char *	r_mtfmt;		/* and the format to use */
	int	(* r_enoenthandler)();	/* Function to use on ENOENT */
	char *	r_enoentfmt;		/* format */
	int	(* r_epermhandler)();	/* Function to use on EPERM error */
	char *	r_epermfmt;		/* and the format to use */
	int	(* r_ehpermhandler)();	/* Function to use on EHPERM error */
	char *	r_ehpermfmt;		/* Format */
	int	(* r_enoimphandler)();	/* Function to use on ENOIMP error */
	char *	r_enoimpfmt;		/* format */
};

typedef struct request_t Request;

Request *	 sendrequest P((int));

static char qismt[] = "%s: empty.\n";
static char nobal[] = "%s: does not load balance.\n";
static char enoimp[] = "%s: command not implemented.\n";
static char eperm[] = "%s: permission denied.\n";
static char ehperm[] = "%s: access denied to your host.\n";


Request requesttab[] = {

{ QC_KICK,	"kick",
	  reply_ok, NULL,		reply_mt, NULL,
	  reply_info, enotq,		reply_info, eperm,
	  reply_info, ehperm,		reply_info, enoimp,
  },

{ QC_DISP, 	"display",
	  reply_ok, DISP_SCAN,		reply_mt, qismt,
	  reply_info, enotq,		reply_info, eperm,
	  reply_info, ehperm,		reply_info, enoimp,
  },	  

{ QC_REM, 	"remove",
	  reply_ok, NULL,		reply_ok, NULL,
	  reply_info, enotq,		reply_ok, NULL,
	  reply_info, ehperm,		reply_info, enoimp,
  },
		  
{ QC_BID, 	"bid",
	  reply_ok, BID_FMT,		reply_info, nobal,
	  reply_info, enotq,		reply_info, eperm,
	  reply_info, ehperm,		reply_info, enoimp,
  },

{ QC_STATS, 	"stats",
	  reply_ok, STATS_FMT,		reply_info, NULL,
	  reply_info, enotq,		reply_info, eperm,
	  reply_info, ehperm,		reply_info, enoimp,
  },

{ QC_INFO, 	"info",
	  reply_ok, NULL,		reply_info, NULL,
	  reply_info, enotq,		reply_info, eperm,
	  reply_info, ehperm,		reply_info, enoimp,
  },

{ 0 },

};

extern int	s;			/* 's' is the socket we are writing */

Request *
sendrequest(command)
int	command;
{

	Request * r;


	for (r = requesttab; r != NULL; r++) {
		if ( command == r->r_ident )
			return r;
	}

 	return (Request *) NULL;
}


int
host_query(thost, command, args)
char *	thost;
int	command;
char *	args;
{


	int	n, rval = 0;
	char	msgbuf[BUFSIZ];
	char	queue[LINE];
	char *	aptr, *qptr;
	Request * r;
	FILE *	fp;


	/* Find the command structure so we can set up the callbacks. */
	if ((r = sendrequest(command)) == NULL) {
		blog(BLOG_WARNING, "cannot find structure for %o", command);
		return (-1);
	}

	/* Get a connection to the (possibly remote) host. */
	if ((s = getsocket(thost, service, tcp,
			   (u_long *) 0, (u_short *) 0)) < 0)
		return (-1);

	/*
	 * Copy the first argument of "args" into the string queue: by
	 * definition the first argument *is* always the queue of interest
	 */
	aptr = args;
	qptr = queue;

	while (*aptr && !isspace((int) *aptr))
	       *qptr++ = *aptr++;

	*qptr = '\0';	/* Terminate our string */
	
	(void) sprintf(msgbuf, "%c%s\n", command, args);

	n = strlen(msgbuf);

	if (write(s, msgbuf, n) != n) {
		(void) close(s);
		blog(BLOG_ERR, elost, "host_query(): write");
		return (-1);
	}

	/*
	 * Set up for a buffered read
	 */
	if ((fp = fdopen(s, "r")) == NULL) {
		(void) close(s);
		blog(BLOG_ERR, efdop, "host_query");
		return (-1);
	}

	if (read(s, msgbuf, 1) == 1) {
		switch ((int) msgbuf[0]) {

		case QR_OK:
			(*r->r_okhandler)(fp, r->r_okfmt, queue);
			break;

		case QR_MT:
			(*r->r_mthandler)(fp, r->r_mtfmt, queue);
			break;

		case QR_ENOENT:
			(*r->r_enoenthandler)(fp, r->r_enoentfmt, queue);
			rval = 1;
			break;

		case QR_EPERM:
			(*r->r_epermhandler)(fp, r->r_epermfmt, queue);
			rval = 1;
			break;
			
		case QR_EHPERM:
			(*r->r_ehpermhandler)(fp, r->r_ehpermfmt, queue);
			rval = 1;
			break;
			
		case QR_ENOIMP:
			(*r->r_enoimphandler)(fp, r->r_enoimpfmt, queue);
			rval = 1;
			break;
			
		default:
			(void) close(s);
			blog(BLOG_ERR, erecv, (int) msgbuf[0]);
			return (-1);
			break;
		}

	} else {
		(void) close(s);
		blog(BLOG_ERR, elost, "host_query(): read");
		return (-1);
	}

	(void) close(s);
	return rval;
}

