#ifndef lint
static char rcsid[] =
	"$Id: resource.c,v 4.0 1993/09/17 00:41:39 srp Exp srp $";
#endif

/*
 *
 * Author: Scott R. Presnell
 *
 * Copyright (C) 1989, 1990, The Regents of the University of California
 *                     All Rights Reserved
 *
 * THE REGENTS OF THE UNIVERSITY OF CALIFORNIA MAKE NO REPRESENTATION OR
 * WARRANTIES WITH RESPECT TO THE CONTENTS HEREOF AND SPECIFICALLY DISCLAIM
 * ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
 * PURPOSE.
 *
 * Further, the Regents of the University of California reserve the right to
 * revise this software and/or documentation and to make changes from time
 * to time in the content hereof without obligation of the Regents of the
 * University of California to notify any person of such revision or change.
 *
 */

/*
 * resource.c - determine the load on the machine, check the process
 *	variables and keep track of the time a process uses.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <errno.h>
#include <signal.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>

#include <sys/file.h>

#include "config.h"
#include "batch.h"
#include "batchd.h"
#include "protocol.h"


extern char *	rlimittoa P((int));			/* in readcap.c */


/*
 * Check the resource limits of a particular job.  Kill the job if overrun.
 */
int
job_check_usage(jp)
Job *	jp;
{

	Queue *	qp = jp->j_queue;
	static char *	mess =
		"Job %d in queue %s was terminated because of a\n%s resource overrun: %d (maxvalue: %d).\n";
	int	pid;
	int	limit;
	char *	type;


#if defined(HAVE_RLIMIT) && defined(USE_CUMULTIME)

#define SIG_BUG

#ifdef SIG_BUG

	/*
	 * For some reason, if we do a pgrp kill (using killjob),
	 * *this* process (the batchd process) also gets killed
	 * and since our final kill is SIGKILL, the batchd can't
	 * ignore it, and we die.
	 *
	 * However, if we fork off, set to a new process group,
	 * then kill the process, the batchd process is insulated.
	 * What's the problem?
	 *
	 * It means another fork, but it works.
	 */

	(void) sighold(SIGCHLD);		/* until in parent table */

	switch (pid = fork()) {
	case -1:	/* ERROR */
		(void) sigrelse(SIGCHLD);
		blog(BLOG_ERR, efork, "resource check");
		return (RTN_RETRY);
		break;
	case 0:		/* CHILD -> setpgrp(), check/kill cpu hogs */
		/* for symmetry and completeness */
		(void) sigrelse(SIGCHLD);
		break;
	default:	/* PARENT -> return */
		/* Add the process to the table	 */
		process_add(pid, (Queue *) NULL, (Job *) NULL, "resource check");

		/* Job is safely in table now */
		(void) sigrelse(SIGCHLD);

		/* We're the parent, we go back to our usual business. */
		return (0);
		break;
	}	
	
#ifdef HAVE_SETSID
	(void) setsid();
#else
#ifdef HAVE_BSDPGRP
	(void) setpgrp(0, 0);
#else
	(void) setpgrp();
#endif
#endif

#endif /* SIG_BUG */

	/*
	 * Make a subject line
	 */
	(void) sprintf(line, quej, qp->q_name, jp->j_id, "resource overrun");

	/*
	 * From here on out we check to see if this process has
	 * overrun a set limit for the queue.  If so, kill the process
	 * and possibly send mail about the death.  Then return.
	 */
	limit = MIN(jp->j_rlimit[RLIMIT_CPU].rlim_cur, 
		    jp->j_rlimit[RLIMIT_CPU].rlim_max);
	type = rlimittoa(RLIMIT_CPU);

	if (qp->q_cumetime == TRUE && jp->j_cputime > limit) {
#ifdef SIGXCPU
		job_kill(jp, SIGXCPU);
#else
		job_kill(jp, SIGINT);
#endif

		if (jp->j_usermail & NOTIFY_RESOURCE)
			send_mail(jp->j_umaddress, line,
				  mess, jp->j_id, qp->q_name,
				  "CPU time", jp->j_cputime, limit);

		if (qp->q_supmail & NOTIFY_RESOURCE)
			send_mail(qp->q_smaddress, line,
				  mess, jp->j_id, qp->q_name,
				  type, jp->j_cputime, limit);

		if (jp->j_userwrite & NOTIFY_RESOURCE)
			send_write(jp->j_uwaddress,
				   mess, jp->j_id, qp->q_name,
				   "CPU time", jp->j_cputime, limit);

		if (qp->q_supwrite & NOTIFY_RESOURCE)
			send_write(qp->q_swaddress,
				   mess, jp->j_id, qp->q_name,
				   type, jp->j_cputime, limit);


#ifdef SIG_BUG
		exit (RTN_REPORT);
#else
		return (RTN_REPORT);
#endif
	}

#ifdef SIG_BUG
	exit (RTN_OK);
#else
	return (RTN_OK);
#endif

#else  /* USE_CUMULTIME */

	return (RTN_OK);

#endif /* HAVE_RLIMIT && USE_CUMULTIME */

}



