#ifndef lint
static char rcsid[] =
	"$Id: rmjob.c,v 4.0 1993/09/17 00:41:39 srp Exp srp $";
#endif

/*
 *
 * Author: Scott R. Presnell
 *
 * Copyright (C) 1989, 1990, The Regents of the University of California
 *                     All Rights Reserved
 *
 * THE REGENTS OF THE UNIVERSITY OF CALIFORNIA MAKE NO REPRESENTATION OR
 * WARRANTIES WITH RESPECT TO THE CONTENTS HEREOF AND SPECIFICALLY DISCLAIM
 * ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
 * PURPOSE.
 *
 * Further, the Regents of the University of California reserve the right to
 * revise this software and/or documentation and to make changes from time
 * to time in the content hereof without obligation of the Regents of the
 * University of California to notify any person of such revision or change.
 *
 */

/*
 * rmjob.c - remove and/or kill active jobs.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <ctype.h>

#include <pwd.h>
#include <grp.h>

#include <signal.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/resource.h>

#include "config.h"

#include H_STRING

#include "batch.h"
#include "batchd.h"
#include "protocol.h"
#include "flags.h"


#define REMOVE_USER		1
#define REMOVE_JOB		2
#define REMOVE_ALL		4

static int job_remove_one P((Queue *, char *, int, char *, int));

/*
 * Parse the string for removing jobs
 */
int
job_remove(que, cp)
char *	que;
char *	cp;
{


	Queue *	qp;
	char *	person, *job;
	int	lfd = 0, uid = 0, jid = 0, errs = 0;
	int	i;
	Bool	done = FALSE;
	int	retries = 3;
	struct passwd *	pwent;

	person = cp;
	while (*cp && *cp != ' ' && *cp != '\n')
		cp++;
	*cp++ = '\0';

#ifdef ndef	
	/* There *is* a job isn't there? */
	if (cp == NULL) {
		(void) sprintf(line, "%c%s (%s)\n",
			       QR_EPERM, ": no jobs specified", cp);
		(void) ewrite(1, line, strlen(line), "rmjob()");
		return (-1);
	}
#endif

	/*
	 * This person must be recognized in the passwd file
	 * on the host machine running the job.
	 */
	if ((pwent = getpwnam(person)) != NULL) {
		uid = pwent->pw_uid;
	} else { 
		(void) sprintf(line, "%c%s\n", QR_EPERM, ": who are you?");
		(void) ewrite(1, line, strlen(line), "rmjob()");
		return(-1);
	}

	if ((qp = getqueuebyname(que)) == (Queue *) NULL) {
		nack();
		return (RTN_ERROR);
	}

	if (chdir(qp->q_spool) < 0) {
		(void) sprintf(line, "%c%s\n", QR_EPERM, ": cannot chdir()");
		(void) ewrite(1, line, strlen(line), "rmjob()");
		blog(BLOG_ERR, echad, qp->q_spool);
		return(RTN_ERROR);
	}
	
	/*
	 * While there are jobs, try to remove them.
	 */

	/* Grab lock */
	i = 1;
 lock_again:
	switch (lfd = queue_grab_lock(qp)) {
	case -1: /* error in opening the file */
		(void) sprintf(line, "%c%s\n", QR_EPERM, "cannot open lock file?");
		(void) ewrite(1, line, strlen(line), "rmjob()");
		blog(BLOG_ERR, elock, qp->q_lname);
		return(RTN_ERROR);
		break;

	case 0:	/* file is already locked */
		if (i < retries) {
			i++;
			(void) sleep((unsigned) i*i);
			goto lock_again;
		} else {
			(void) sprintf(line, "%c%s\n", QR_EAGAIN, "queue is busy");
			(void) ewrite(1, line, strlen(line), "rmjob()");
			return(RTN_RETRY);
		}
		break;

	default:
		/* All is cool, we've locked the file */
		break;
	}

	blog(BLOG_TRACE, "deleting jobs \"%s\"", cp);

	do {
		job = cp;
		while (*cp && *cp != ' ' && *cp != '\n') {
			cp++;
		}

		/*
		 * If the separator is a NULL,  we've hit the end of the
		 * list.  Don't try to parse any more jobs. Otherwise,
		 * turn the separator into a NULL, and walk past.
		 */
		if (*cp == '\0')
			done = TRUE;
		else
			*cp++ = '\0';

		if ((pwent = getpwnam(job)) != NULL)
			jid = pwent->pw_uid;
		else
			jid = -1;

		errs += job_remove_one(qp, person, uid, job, jid);

	} while (!done && *cp);
	
	/* Release the lock when were done. */
	if (queue_release_lock(qp, lfd) < 0) {
		blog(BLOG_ERR, elock, qp->q_lname);
		return(RTN_ERROR);
	}

	/*
	 * Signal remake of queue after all jobs have been removed
	 */
	(void) kill(batchd_pid, SIGALRM);

	return errs;
}


/*
 * Remove a job or all jobs for a person from a queue
 */
static int
job_remove_one(qp, person, uid, job, jid)
Queue *	qp;
char *	person;
int	uid;
char *	job;
int	jid;
{

	char	job_desc[LINE];
	char *	funcname = "remv";
	char *	tag =  "%s/%s: removing %s";
	int	intjob = -1;
	int	npass = 0, class = 0;
	Job *	jp;


#ifdef ndef
	/*
	 * Gave us the entire name.
	 */
	if ((job[0] == 'c' || job[0] == 'd') &&
	    job[1] == 'f' && isdigit((int) job[2])) {

		(void) sscanf(job, "%*[a-zA-Z]%d%*[a-zA-Z.:\\]", &intjob);

		(void) sprintf(job_desc, "job %d", intjob);

		blog(BLOG_TRACE, tag, funcname, "file", job_desc);

		class = REMOVE_JOB;

	} else
#endif
	/*
	 * Gave us the job number
	 */
	if (isdigit((int)job[0])) {

		intjob = atoi(job);
		(void) sprintf(job_desc, "job %d", intjob);

		blog(BLOG_TRACE, tag, funcname, "jobid", job_desc);

		class = REMOVE_JOB;

	/*
	 * Assume we have user id
	 */
	} else if (strcmp(job, all) == 0) {

		(void) strcpy(job_desc, "jobs for all users");

		blog(BLOG_TRACE, tag, funcname, "all", job_desc);

		class = REMOVE_ALL;

	} else {

		(void) sprintf(job_desc, "jobs for user %s", job);

		blog(BLOG_TRACE, tag, funcname, "user", job_desc);

		class = REMOVE_USER;

	}

	if ((class == REMOVE_USER &&
	     (strcmp(person, job) && strcmp(person, "root")) ) ||

	    (class == REMOVE_ALL && strcmp(person, "root")) ) {
		(void) sprintf(line, "%c%s: %s %s.\n",
			       QR_EPERM, qp->q_name,
			       "permission denied removing",
			       job_desc);
		(void) ewrite(1, line, strlen(line), "rmjob()");
		return(RTN_RETRY);
	}

	/*
	 * For all jobs in that queue.
	 */
	ITER(jp, Job, &(qp->q_jobs)) {

		if (jp->j_file == NULL)
			continue;

		if (class == REMOVE_JOB) {

			if (intjob != jp->j_id)
				continue;

			/*
			 * We've found the job, now if the uid's
			 * don't match up, it's a permission error,
			 * not just a skip.
			 */
			if (uid && jp->j_uid != uid) {
				(void) sprintf(line, "%c%s: %s %s.\n",
					       QR_EPERM, qp->q_name,
					       "permission denied removing", job);
				(void) ewrite(1, line, strlen(line), (char *) NULL);
				return(RTN_RETRY);
			}

		} else if (class == REMOVE_USER) {

			/*
			 * If the uid's don't match up skip this job.
			 */
			if ((uid && jp->j_uid != uid) ||
			    (uid == 0 && jp->j_uid != jid))
				continue;

		} else if (class == REMOVE_ALL) {

			/*
			 * If uid is other than root, skip all jobs.
			 */
			if (uid)
				continue;
		}

		/*
		 * We've passed all the tests - 
		 * 	if the job is running, kill it.
		 *	and let the terminator clean up the files and entry.
		 * or
		 * 	if the job is queued, or if it is being
		 * 	transferred, remove the files.
		 *	and let the jobqueue clean up the entry.
		 *
		 *
		 * We have to guarantee that the files are removed
		 * before the queue is remade again. so we do it
		 * here immediately (albeit redundantly).
		 *
		 * In either case remake the queues (see below).
		 */

		if (jp->j_pid > 0) {
			job_kill(jp, SIGINT);

			/*
			 * If the job is attached (and alrady has a pid),
			 * but is not yet running, then also remove the
			 * queued files.
			 */
			if ((jp->j_state & JOB_ATTACHED) &&
			    !(jp->j_state & JOB_RUNNING))
				job_unlink(jp);
		} else {
			job_unlink(jp);
		}

		npass++;

		if (npass == 1)
			ack();
		
		/*  Indicate the job was removed */
		(void) sprintf(line, quej, qp->q_name, jp->j_id, "removed");

		/* Write a message to the syslog */
		blog(BLOG_STATS, line);

		/* Write a message to the status file */
		queue_write_status(qp, line);


		if (npass) {
			(void) ewrite(1, line, strlen(line), (char *) NULL);
			(void) ewrite(1, ".\n", 2, (char *) NULL);
		}
	}


	if (!npass) {

		/*
		 * Empty queue
		 */
		(void) sprintf(line, "%c%s: %s %s.\n", QR_MT,
			       qp->q_name, "could not find", job_desc);
		(void) ewrite(1, line, strlen(line), (char *) NULL);
		return (RTN_RETRY);

	} else  {
		return (RTN_OK);
	}



}


/*
 * Send signals to kill a particular job.
 */
void
job_kill(jp, sig)
Job *	jp;
int	sig;
{

	errno = 0;

#ifdef HAVE_PIDCHECK
	if (log_lvl & BLOG_TRACE)
		if (kill(jp->j_pid, 0) < 0)
			blog(BLOG_ERR, ekill, jp->j_queue->q_name, jp->j_pid, 0);
#endif

	/*
	 * First send a sig (SIGINT|SIGXCPU), then a SIGTERM, finally a
	 * SIGKILL.  Internal jobs for sending processes ignore SIGINT.
	 */
	if (job_killpg(jp, sig) > 0)
		blog(BLOG_WARNING, ekill, jp->j_queue->q_name, jp->j_pid, sig);

	(void) sleep(2);

	if (job_killpg(jp, SIGTERM) > 0)
		return ;

	(void) sleep(2);

	(void) job_killpg(jp, SIGKILL);

	(void) sleep(2);

	if (kill(jp->j_pid, 0) >= 0)
		blog(BLOG_ERR, "%s: job: %d (pid:%d) exists after SIGKILL!?",
		     jp->j_queue->q_name, jp->j_id, jp->j_pid);

	return ;
}


/*
 * Unlink the control and data files for a job.
 */
void
job_unlink(jp)
Job *	jp;
{

	char kffile[MAXPATHLEN];


	dbg((BLOG_JOBS, quej, jp->j_queue->q_name, jp->j_id,
	     "unlinking files"));

	jp->j_file[0] = 'c';
	if (jp->j_state & JOB_KEEP) {
		/*
		 * Keep control file - report failure if not ENOENT.
		 */
		(void) strcpy(kffile, jp->j_file);
		kffile[0] = 'k';
		errno = 0;
		if (rename(jp->j_file, kffile) < 0 && errno != ENOENT)
			blog(BLOG_ERR, eunlk, kffile);
	} else {
		/*
		 * Unlink control file - report failure if not ENOENT.
		 */
		errno = 0;
		if (unlink(jp->j_file) < 0 && errno != ENOENT)
			blog(BLOG_ERR, eunlk, jp->j_file);
	}

	/*
	 * Unlink data file - report failure if not ENOENT.
	 */
	errno = 0;
	jp->j_file[0] = 'd';

	if (unlink(jp->j_file) < 0 && errno != ENOENT)
		blog(BLOG_ERR, eunlk, jp->j_file);

	/*
	 * Change the name back
	 */
	jp->j_file[0] = 'c';

	/*
	 * Leave the output file behind for the terminator to scan
	 */
	return ;

}


