#ifndef lint
static char rcsid[] =
	"$Id: util.c,v 4.0 1993/09/17 00:42:25 srp Exp srp $";
#endif

/*
 *
 * Author: Scott R. Presnell
 *
 * Copyright (C) 1989, 1990, The Regents of the University of California
 *                     All Rights Reserved
 *
 * THE REGENTS OF THE UNIVERSITY OF CALIFORNIA MAKE NO REPRESENTATION OR
 * WARRANTIES WITH RESPECT TO THE CONTENTS HEREOF AND SPECIFICALLY DISCLAIM
 * ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
 * PURPOSE.
 *
 * Further, the Regents of the University of California reserve the right to
 * revise this software and/or documentation and to make changes from time
 * to time in the content hereof without obligation of the Regents of the
 * University of California to notify any person of such revision or change.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <ctype.h>

#include <time.h>

#include <grp.h>
#include <pwd.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>

#include "config.h"

#include H_STRING

#include "batch.h"
#include "pathnames.h"
#include "patchlevel.h"

/* Copyright information */
static char copyright_srp[] =
	"Copyright (C) 1989, Scott R. Presnell";
static char copyright_uc[] =
	"Copyright (C) 1989, The Regents of the University of California";


static FILE *	bhf;			/* Batch hosts file ptr */
static FILE *	rcf;			/* rc file input */

static char	myrcbuf[LINE];

static char *	getrcline P((FILE **, char *));
char *		scanrcline P((char *, char *, char *, char *));


/*
 * Get a host name, one line at a time, from the file specified by
 * BATCHHOSTS.  Return each host name as it is secured.
 * This function returns the host names into a static area.  Copy the 
 * returned value to retain that information across getbhent() calls.
 */
char *
getbhent()
{
	char *cp, *sp;
	cp = getrcline(&bhf, batchhosts);

	/* 
	 * Truncate string at first whitespace character.
	 */
	if (cp && (sp = strchr(cp, ' ')) != NULL)		/* space */
		*sp = '\0';
	else if (cp && (sp = strchr(cp, '	')) != NULL)	/* tab */
		*sp = '\0';
	
	return cp;

}


void
endbhent()
{

	if (bhf != NULL) {
		(void) fclose(bhf);
		bhf = NULL;
	}

	return;
}


/*
 * Scan the rcfile for settings.
 */
char *
scanrcline(filename, mp, vp, sp)
char *	filename;
char *	mp, *vp, *sp;
{


	char *	cp, *lp;


	if ((lp = getrcent(filename)) != NULL) {

		cp = lp;

		/* Skip leading white space */
		while (cp && isspace((int) *cp))
			cp++;
			
		/* get first string  - should be "set" */
		while (cp && !isspace((int) *cp))
			*mp++ = *cp++;

		*mp = '\0';

		/* Skip intermediate white space */
		while (cp && isspace((int) *cp))
			cp++;

		/* get second string  - the variable */
		while (cp && (!isspace((int) *cp) && *cp != '='))
			*vp++ = *cp++;

		*vp = '\0';

		/* Skip intermediate white space or equals. */
		while (cp && (isspace((int) *cp) || *cp == '='))
			cp++;

		/* get third string */
		while ((*sp++ = *cp++) != '\0')
			;

		*sp = '\0';

		return lp;
	}

	return NULL;

}


char *
getrcent(filename)
char *	filename;
{
	char *cp, *sp;
	cp = getrcline(&rcf, filename);

#ifdef ndef
	/* 
	 * Truncate string at first whitespace character.
	 */
	if (cp && (sp = strchr(cp, ' ')) != NULL)		/* space */
		*sp = '\0';
	else if (cp && (sp = strchr(cp, '	')) != NULL)	/* tab */
		*sp = '\0';
#endif
	
	return cp;

}


/*
 * Close the batch hosts file ptr if it is open.
 */
void
endrcent()
{

	if (rcf != NULL) {
		(void) fclose(rcf);
		rcf = NULL;
	}

	return;
}


static char *
getrcline(fp, filename)
FILE **	fp;
char *	filename;
{
	
	char *ptr;

	if (*fp == NULL) {
		if ((*fp = fopen(filename, "r")) == NULL) {
			blog(BLOG_WARNING, eopen, filename);
			return (char *) NULL;
#ifdef USE_LOCALBUF
		} else {
			(void) setbuf(*fp, rciobuf);
#endif
		}
	}

 again:
	if (fgets(myrcbuf, sizeof(myrcbuf), *fp) != NULL) {

		/*
		 * If the first character is a comment, skip the line
		 */
		if (myrcbuf[0] == '#' || myrcbuf[0] == '\n')
			goto again;

		/*
		 * Chop off the newline character
		 */
		myrcbuf[strlen(myrcbuf) - 1] = '\0';

		/*
		 * Chop off everything beyond the first comment in the line
		 */
		if ((ptr = strchr(myrcbuf,'#')) != NULL)
			*ptr = '\0';

		return myrcbuf;

	} else {
		(void) fclose(*fp);
		return (char *) NULL;
	}
}


/*
 * Get the next queue entry.
 */
char *
getquent()
{

	char	bbuf[BUFFER];
	int	status;


	if ((status = getbaent(bbuf)) < 0) {
		blog(BLOG_FATAL, eopen, _PATH_CAPFILE);
	} else if (status == 0) {
		(void) endbaent();
	} else {

		/*
		 * Return the canonical name for that queue
		 */
		return bnamget(bbuf);
	}

	return (char *) NULL;
}

void
endquent()
{
	endbaent();
}

/*
 * islocalhost(char *)
 * 
 * Determine if the given host is the local host.
 * return -1 if localhost not set.
 * return  0 localhost.
 * return  1 not localhost.
 */
int
islocalhost(host)
char *	host;
{

	if (!lhost[0])
		return (-1);
	else if (strcasecmp(host, localhost) == 0)
		return 0;
	else if (strncasecmp(host, lhost, strlen(host)) == 0)
		return 0;
	else
		return 1;

}


/*
 * prttime prints a time in hours and minutes or minutes and seconds.
 * returns a pointer to a static buffer holding the string.
 */
char *
prttime(tim)
time_t	tim;
{
	
	char *	mptr;
	char *	comma = ",";
	char *	plural = "s";
	int days = 0, hrs = 0, mins = 0, secs = 0;
	static char mybuf[80];


	(void) memset((voidp) mybuf, 0, 80);

	days = tim / SECPERDAY;
	tim %= SECPERDAY;
	hrs = tim / SECPERHOUR;
	tim %= SECPERHOUR;
	mins = tim / SECPERMIN;
	tim %= SECPERMIN;
	secs = tim;

	mptr = mybuf;

	if (days > 0) {
		(void) sprintf(mptr, "%3d day%s%s",
			       days, days > 1 ? plural : "",
			       (hrs + mins + secs) > 0 ? comma : "");
		mptr += 7 + (days > 1) + ((hrs + mins + secs) > 0);
	}

	if (hrs > 0) {
		(void) sprintf(mptr, " %2d hour%s%s",
			       hrs, hrs > 1 ? plural : "",
			       (mins + secs) > 0 ? comma : "");
		mptr += 8 + (hrs > 1) + ((mins + secs) > 0);
	}

	if (mins > 0) {
		(void) sprintf(mptr, " %2d minute%s%s",
			       mins, mins > 1 ? plural : "",
			       secs > 0 ? comma : "");
		mptr += 10 + (mins > 1) + (secs > 0);
	}

/*	if (secs > 0)*/
	(void) sprintf(mptr, " %2d second%s",
		       secs, secs > 1 ? plural : "");

	return mybuf;
}


/*
 *  Return number of seconds since device file has been modified.
 */
time_t
getmtime(device)
char *	device;
{

	if (stat(device, &stbuf) < 0) {
		blog(BLOG_ERR, estat, device);
		return -1;
	}

	/* Modification time is what we want */
	return stbuf.st_mtime;
}


#ifdef HAVE_H_ERRLIST

extern int	h_nerr;
extern char *	h_errlist[];

#else

char    *h_errlist[] = {
        "Error 0",				/* 0 ERROR_NOT_FOUND */
        "Unknown host",                         /* 1 HOST_NOT_FOUND */
        "Host name lookup failure",             /* 2 TRY_AGAIN */
        "Unknown server error",                 /* 3 NO_RECOVERY */
        "No address associated with name",      /* 4 NO_ADDRESS */
};
int     h_nerr = { sizeof(h_errlist)/sizeof(h_errlist[0]) };

#endif	/* HAVE_H_ERRLIST */

extern int	sys_nerr;
extern char *	sys_errlist[];


/*
 * Take a log format string and expand occurences of %m
 * with the current error code take from errno.
 */
void
expand_error(f, e)
char *	f;
char *	e;
{

	char *	p;
	int	error = errno;

	for (p = f; *e = *p; e++, p++) {
		if (p[0] == '%' && p[1] == 'm') {
			char *errstr;
			if (error < 0 || error >= sys_nerr)
				errstr = 0;
			else
				errstr = sys_errlist[error];
			if (errstr)
				(void) strcpy(e, errstr);
			else
				(void) sprintf(e, "Error %d", error);
			e += strlen(e) - 1;
			p++;
		}

		if (p[0] == '%' && p[1] == 'h') {
			char *errstr;
			if (h_errno < 0 || h_errno >= h_nerr)
				errstr = 0;
			else
				errstr = h_errlist[h_errno];
			if (errstr)
				(void) strcpy(e, errstr);
			else
				(void) sprintf(e, "Host error %d", h_errno);
			e += strlen(e) - 1;
			p++;
		}

#ifdef ndef
		if (p[0] == '%' && p[1] == 'b') {
			char *errstr;
			if (b_error < 0 || b_error >= b_nerr)
				errstr = 0;
			else
				errstr = b_errlist[b_error];
			if (errstr)
				(void) strcpy(e, errstr);
			else
				(void) sprintf(e, "Batch error %d", b_error);
			e += strlen(e) - 1;
			p++;
		}
#endif
	}
}


/*
 * Atov - return a value as modified by the last character
 */
int
atov(cp)
char *	cp;
{
	int mult = 0, val = 0;


	mult = *(cp + strlen(cp) - 1);

	if (isalpha(mult)) {
		*(cp + strlen(cp) - 1) = '\0';
	}

	val = atoi(cp);

	if (isalpha(mult)) {
		switch(mult) {

		case 'm':
		case 'M':
			val *= 1024 * 1024;
			break;

		case 'k':
		case 'K':
			val *= 1024;
			break;

		case 'b':
		case 'B':
			break;
		default:
			blog(BLOG_ERR, "bad modifier in \"%s%c\"", cp, mult);
		}
	}

	return val;
}


#ifdef HAVE_RLIMIT

/*
 * Turn a string of  '<char><int>' into a pointer for a limit type,
 * and a value and set them in the passed array.
 */
int
atorlimit(lp, cp)
struct rlimit lp[];
char *	cp;
{
	int	val;

	val = *cp;
	cp++;

	switch(val) {
	case 'p':
		lp[RLIMIT_CPU].rlim_cur = atov(cp);
		break;
	case 'f':
		lp[RLIMIT_FSIZE].rlim_cur = atov(cp);
		break;
	case 'd':
		lp[RLIMIT_DATA].rlim_cur = atov(cp);
		break;
	case 's':
		lp[RLIMIT_STACK].rlim_cur = atov(cp);
		break;
	case 'c':
		lp[RLIMIT_CORE].rlim_cur = atov(cp);
		break;
	case 'r':
		lp[RLIMIT_RSS].rlim_cur = atov(cp);
		break;
#ifdef RLIMIT_NOFILE
	case 'o':
		lp[RLIMIT_NOFILE].rlim_cur = atov(cp);
		break;
#endif
#ifdef RLIMIT_VMEM
	case 'm':
		lp[RLIMIT_VMEM].rlim_cur = atov(cp);
		break;
#endif
	default:
		blog(BLOG_ERR, "error in resource spec: \"%s\"", cp);
		return (-1);
		break;
	}

	return (0);
}

/*
 * Turn a an int into a unique rlimit char.
 */
int
rlimittoc(val)
int	val;
{

	int	lim = 'x';

	switch(val) {
	case RLIMIT_CPU:
		lim = 'p';
		break;
	case RLIMIT_FSIZE:
		lim = 'f';
		break;
	case RLIMIT_DATA:
		lim = 'd';
		break;
	case RLIMIT_STACK:
		lim = 's';
		break;
	case RLIMIT_CORE:
		lim = 'c';
		break;
	case RLIMIT_RSS:
		lim = 'r';
		break;
#ifdef RLIMIT_NOFILE
	case RLIMIT_NOFILE:
		lim = 'o';
		break;
#endif
#ifdef RLIMIT_VMEM
	case RLIMIT_VMEM:
		lim = 'm';
		break;
#endif
	default:
		blog(BLOG_ERR, "error in resource spec: \"%c\"", val);
		lim = 'x';
		break;
	}

	return lim;
}

#endif /* HAVE_RLIMIT */

/*
 * Turn a string with significant characters into a bitstring for
 * determining under which conditions to notify a user
 */
Notify
setnotify(str)
char *	str;
{

	Notify val = NOTIFY_NONE;

	if (str == NULL)
		return val;

	if (strcmp(str, "none") == 0) {
		val |= NOTIFY_NONE;
		return val;
	}
	
	if (strcmp(str, all) == 0) {
		val = ~(NOTIFY_NONE);
		return val;
	}

	if (strchr(str, 's'))
		val |= NOTIFY_START;
	if (strchr(str, 'e'))
		val |= NOTIFY_END;
	if (strchr(str, 'c'))
		val |= NOTIFY_CRASH;
	if (strchr(str, 'd'))
		val |= NOTIFY_DEBUG;
	if (strchr(str, 't'))
		val |= NOTIFY_TRANSFER;
	if (strchr(str, 'p'))
		val |= NOTIFY_PAUSE;
	if (strchr(str, 'r'))
		val |= NOTIFY_RESOURCE;
	if (strchr(str, 'n'))
		val |= NOTIFY_NONE;

	return val;
}


/*
 * Return a random number in the range given
 */
int
getrand(low, high)
int	low, high;
{


	double percent;

	/*
	 * Get a random number and calculate it's value as a percentage
	 * of the possible range;
	 */
#ifndef RAND_MAX
#define RAND_MAX 32767
#endif
	percent = rand() / (double) RAND_MAX;

	/*
	 * Our range...
	 */
	return ((int) (low + (percent * (double) (high - low)) + 0.5) );
}


/*
 * Dump out the version information using a supplied function.
 */
void
dump_version(func, lvl)
void	(*func)V((int, char *, ...));
int	lvl;
{



	if (!lvl) {
		(*func)(lvl, copyright_srp);
		(*func)(lvl, copyright_uc);
	}

	(*func)(lvl, "batch $Revision: 4.0 $, patchlevel %d", PATCHLEVEL);

	if (!lvl) {
		(*func)(lvl, "built for a %s running %s", ARCH, OS);
	}
	
}


/*
 * Return non-zero if the user is a member of the given group
 */
int
ingroup(username, gid, group)
char *	username;
GIDTYPE	gid;
char *	group;
{

#ifdef HAVE_GETGROUPS
	int	i, ngrps;
	GRPTYPE	groups[NGROUPS];
#endif
	struct group  *gr;


	/*
	 * Get the group entry for the group we are checking against
	 */
	if ((gr = getgrnam(group)) == NULL) {
		return (-1);
	}

	/*
	 * If the effective group id is the same as the named group then OK
	 */
	if (gid == gr->gr_gid)
		return (0);

#ifdef HAVE_GETGROUPS

	/*
	 * Get the groups for this user
	 */
	if ((ngrps = getgroups(NGROUPS, groups)) < 0) {
		blog(BLOG_ERR, "cannot getgroups for %s (%m)", username);
		return(-1);
	}

	/*
	 * Check each of the gotten groups against the given group
	 * Do the comparason in the native type that getgroups returns.
	 */
	for (i = 0; i < ngrps; i++)
		if (groups[i] == (GRPTYPE) gr->gr_gid)
			return (0);
#else
	/*
	 * Check the names listed in the group entry against the user's name
	 */
	while (*(gr->gr_mem) != NULL) {
		if (strcmp(*(gr->gr_mem), username) == 0)
			return (0);
		gr->gr_mem++;
	}
#endif

	/*
	 * If all that fails - the user is not from the correct group
	 */
	return (1);
}


/*
 * Error wrappers around write
 */
int
ewrite(fd, buf, bytes, msg)
int	fd;
voidp	buf;
size_t	bytes;
char *	msg;
{
	int	w;

	if (msg == NULL)
		msg = "<null>";

	if ((w = write(fd, buf, bytes)) != bytes)
		blog(BLOG_ERR, ewrit, fd, msg);

	return w;
}

int
ewriteone(fd, buf)
int	fd;
int	buf;
{
	char	let = (char) buf;
	int	w;


	if ((w = write(fd, (voidp) &let, 1)) != 1)
		blog(BLOG_ERR, elost, fd);

	return w;
}
