#ifndef lint
static char rcsid[] =
	"$Id: xlog.c,v 4.0 1993/09/17 00:42:25 srp Exp srp $";
#endif

/*
 *
 * Author: Scott R. Presnell
 *
 * Copyright (C) 1989, 1990, The Regents of the University of California
 *                     All Rights Reserved
 *
 * THE REGENTS OF THE UNIVERSITY OF CALIFORNIA MAKE NO REPRESENTATION OR
 * WARRANTIES WITH RESPECT TO THE CONTENTS HEREOF AND SPECIFICALLY DISCLAIM
 * ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
 * PURPOSE.
 *
 * Further, the Regents of the University of California reserve the right to
 * revise this software and/or documentation and to make changes from time
 * to time in the content hereof without obligation of the Regents of the
 * University of California to notify any person of such revision or change.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <errno.h>
#include <fcntl.h>

#include <time.h>
#include <syslog.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/resource.h>

#include "config.h"

#if (defined(I_STDARG) && defined(CAN_VAPROTO)) || defined(__STDC__)
#include <stdarg.h>
#else
#include <varargs.h>
#endif

#include H_STRING

#include "batch.h"
#include "batchd.h"
#include "bid.h"
#include "event.h"
#include "pathnames.h"


#ifdef DBFILE					/* db dump file */
char *	dbfile = DBFILE;
#else
char *	dbfile = _PATH_DB;
#endif

#ifdef STATSFILE				/* stats dump file */
char *	statsfile = STATSFILE;
#else
char *	statsfile = _PATH_STATS;
#endif


#ifdef USE_LOCALBUF
static char logiobuf[BUFSIZ];			/* Logging buffer */
#endif

FILE *		logfp = stderr;
FILE *		sfp = stderr;			/* Stats logging descriptor */

extern Bool	logging, debugging, dbging;
extern char *	logfile;

static int	syslogging = 0;

extern void	expand_error P((char *, char *));
int		switch_to_logfile P((char *));
static void	fakelog P((int));

/*
 * Dump system statistics into a file.
 */
void
dump_stats()
{
	Queue *	qp;
	char	priv[LINE];

#ifdef USE_LIBMALLOC
	struct mallinfo mi;
#endif

	blog(BLOG_INFO, "dumping daemon statistics");

	if (!(log_lvl & BLOG_NOFORK)) {
		if ((sfp = fopen(statsfile, "a+")) == NULL) {
			blog(BLOG_WARNING, eopen, statsfile);
			return ;
		}
#ifdef USE_LOCALBUF
		(void) setbuf(sfp, (char *) NULL);
#endif
	} else {
		sfp = stderr;
	}

	now = time((time_t *) 0);
	(void) strcpy(priv, ctime(&now));
	priv[strlen(priv) - 1] = '\0';
	(void) fprintf(sfp, "Batchd statistics for %s:\n", priv);

	(void) strcpy(priv, ctime(&qinfo.i_start));
	priv[strlen(priv) - 1] = '\0';
	(void) fprintf(sfp, "Daemon started: %s\n", priv);
	(void) fprintf(sfp, "Logging to \"%s\" is %s, debugging is %s\n",
		       logfile, logging ? "on" : "off",
		       debugging ? "on" : "off");

	(void) fprintf(sfp, "\nCycles:\n");
	(void) fprintf(sfp, "forks: %-4d waits: %-4d cycles: %-4d\n",
		       qinfo.i_fork, qinfo.i_wait, qinfo.i_main);

	(void) fprintf(sfp, "\nConnections:\n");
	(void) fprintf(sfp, "local: %-4d inet:  %-4d udp:    %-4d\n",
		       qinfo.i_loc, qinfo.i_inet, qinfo.i_udp);

	(void) fprintf(sfp, "\nLinked list allocated:\n");
	(void) fprintf(sfp, "Run: %-4d Queue: %-4d Job: %-4d Host: %-4d Bid: %-4d Event: %-4d\n",
		       qinfo.i_runlist, qinfo.i_queuelist, qinfo.i_joblist,
		       qinfo.i_hostlist, qinfo.i_bidlist, qinfo.i_eventlist);

	(void)	fprintf(sfp, "Free Jobs: %-4d Free Events: %-4d\n",
			qinfo.i_fjoblist, qinfo.i_feventlist);

#ifdef USE_LIBMALLOC
	mi = mallinfo();

	(void) fprintf(sfp, "\nMemory units allocated (libmalloc) :\n");
	(void) fprintf(sfp, "Ordinary: %d Small: %d\n", mi.ordblks, mi.smblks);
	(void) fprintf(sfp, "Total: %d Oused: %d Ofree: %d Sused: %d Sfree: %d\n",
		      mi.arena, mi.uordblks, mi.fordblks, mi.usmblks, mi.fsmblks);
#endif /* USE_LIBMALLOC */

	ITER(qp, Queue, &queues) {
		(void) strcpy(priv, ctime(&qp->q_mtime));
		priv[strlen(priv) - 1] = '\0';
		(void) fprintf(sfp, "\nQueue %s; last action: %s \n\t%s ago\n",
			       qp->q_name, priv, prttime(now - qp->q_mtime));
		(void) fprintf(sfp, "Jobs:\n");
		(void) fprintf(sfp, "\tstarted:    %-4d returned: %-4d\n",
			       qp->q_stats.s_started,
			       qp->q_stats.s_returned);
		(void) fprintf(sfp, "\ttransfered: %-4d received: %-4d ",
			       qp->q_stats.s_transfered,
			       qp->q_stats.s_received);
		(void) fprintf(sfp, "signaled: %-4d\n",
			       qp->q_stats.s_signaled);
	}

	(void) fprintf(sfp, "\n==========\n");

	if (!(log_lvl & BLOG_NOFORK))
		(void) fclose(sfp);
	else
		sfp = NULL;

	return ;
}

/*
 * Dump system db into a file.
 */
void
dump_db()
{

	char	priv[LINE];
	Queue *	qp;
	Hostinfo *	hip;
	Bidinfo	 *	bip;


	blog(BLOG_INFO, "dumping capabilities database");

	if (!(log_lvl & BLOG_NOFORK)) {
		if ((sfp = fopen(dbfile, "a+")) == NULL) {
			blog(BLOG_WARNING, eopen, dbfile);
			return ;
		}
#ifdef USE_LOCALBUF
		(void) setbuf(sfp, (char *) NULL);
#endif
	} else {
		sfp = stderr;
	}

	now = time((time_t *) 0);
	(void) strcpy(priv, ctime(&now));
	priv[strlen(priv) - 1] = '\0';
	(void) fprintf(sfp, "Batchd database for %s:\n", priv);

	(void) strcpy(priv, ctime(&qinfo.i_start));
	priv[strlen(priv) - 1] = '\0';
	(void) fprintf(sfp, "Daemon started: %s\n", priv);

	(void) fprintf(sfp, "Load averages: %.2lf %.2lf %.2lf\n",
		       loadave[0], loadave[1], loadave[2]);

	(void) fprintf(sfp, "Child processes: %d; Jobs running: %d\n",
		       njobs, nkids);

	ITER(qp, Queue, &queues) {
		(void) fprintf(sfp, "\nQueue %s Data:\n", qp->q_name);
		dump_queue(printsf, 0, qp);

		(void) fprintf(sfp, "\nUser Data:\n");
		dump_usr(printsf, 0, qp);
	}

	(void) fprintf(sfp, "\nBid Data:\n");
	ITER(hip, Hostinfo, &hosts) {
		(void) fprintf(sfp, "\nHost: %s, load [%.2lf, %.2lf, %.2lf]\n",
			       hip->hi_name, hip->hi_loadave[0], hip->hi_loadave[1], hip->hi_loadave[2]);
			
		ITER(bip, Bidinfo, &hip->hi_bids) {
			(void) fprintf(sfp,"\tQueue: %-8.8s bid: %.2lf - %s ago\n",
				       bip->bi_name, bip->bi_bid,
				       prttime(now - bip->bi_stamp));
		}
	}

	(void) fprintf(sfp, "\nPending Events:\n");
	event_dump(printsf, 0);

	(void) fprintf(sfp, "\n==========\n");

	if (!(log_lvl & BLOG_NOFORK))
		(void) fclose(sfp);
	else
		sfp = NULL;

	return ;
}

/*
 * Report a fatal error
 */
/*VARARGS*/
#if (defined(I_STDARG) && defined(CAN_VAPROTO)) || defined(__STDC__)

void
blog(int lvl, char *fmt, ...)

#else

void
blog(lvl, fmt, va_alist)
int	lvl;
char *	fmt;
va_dcl

#endif /* __STDC__ */
{
	va_list	args;
	char	priv[LINE];
	char	nfmt[LINE];
	char *	ptr;
	int	level;		/* real syslog level */	


	ptr = priv;

	/*
	 * Fire up the variable argument mechanism.
	 */
	/*CONSTCOND*/
#if (defined(I_STDARG) && defined(CAN_VAPROTO)) || defined(__STDC__)
	va_start(args, fmt);
#else
	va_start(args);
#endif

	/* If we're not logging that level - return */
	if ( !(lvl & log_lvl) ||
	    ((lvl & BLOG_ALL) && logging == FALSE) ||
	    ((lvl & BLOG_FULL) && debugging == FALSE))
		return ;

	switch (lvl) {
	case BLOG_FATAL:
		level = LOG_CRIT; break;
	case BLOG_ERR:
		level = LOG_ERR; break;
	case BLOG_WARNING:
		level = LOG_WARNING; break;
	case BLOG_USER:
		level = LOG_WARNING; break;
	case BLOG_INFO:
		level = LOG_INFO; break;
	case BLOG_STATS:
		level = LOG_INFO; break;
	case BLOG_TRACE:
		level = LOG_DEBUG; break;
	default:
		level = LOG_ERR; break;
	}

	if (lvl & BLOG_FULL) {
		lvl |= BLOG_DEBUG;
		level = LOG_DEBUG;
	}

	expand_error(fmt, nfmt);

	(void) vsprintf(priv, nfmt, args);

	ptr += strlen(ptr);
	
	if (syslogging) {
		if (ptr[-1] == '\n')
			*--ptr  = '\0';

		syslog(level, priv);
	} else {
		if (ptr[-1] != '\n') {
			*ptr++ = '\n';
			*ptr = '\0';
		}

		fakelog(lvl);
		(void) fwrite(priv, ptr - priv, 1, logfp);
		(void) fflush(logfp);
	}
	
	/*
	 * Shut down the variable argument mechanism.
	 */
	va_end(args);

	/*
	 * Make a core dump so we can track the problem.
	 */
	if ( (log_lvl & BLOG_NOFORK) && (level < LOG_ERR))
		(void) abort();

	return ;

}



/*
 * Output the time of day and hostname to the logfile
 */
static void
fakelog(lvl)
int	lvl;
{

	char *	sev;
	time_t	t = time((time_t *) 0);
	static char	last_ctime[LINE];
	static time_t	last_t = 0;


	if (t != last_t) {
		(void) strcpy(last_ctime, ctime(&t));
		last_t = t;
	}

	switch (lvl) {
	case BLOG_FATAL:
		sev = "fatal:"; break;
	case BLOG_ERR:
 		sev = "error:"; break;
	case BLOG_USER:
		sev = "user: "; break;
	case BLOG_WARNING:
		sev = "warn: "; break;
	case BLOG_INFO:
		sev = "info: "; break;
	case BLOG_STATS:
		sev = "stats:"; break;
	case BLOG_TRACE:
		sev = "trace:"; break;
	default:
		sev = "unk:  "; break;
	}

	if (lvl & BLOG_DEBUG)
		sev = "debug:";

	(void) fprintf(logfp, "%15.15s %s %s[%d]/%s ",
		       last_ctime+4, shost, progname, getpid(), sev);
}

/*
 * Change current logfile
 */
int
switch_to_logfile(newlogfile)
char *	newlogfile;
{

	FILE *	new_logfp = stderr;


	if (newlogfile) {
		syslogging = 0;
		if (strcmp(newlogfile, "/dev/stderr") == 0)
			new_logfp = stderr;
		if (strcmp(newlogfile, "stderr") == 0)
			new_logfp = stderr;
		else if (strcmp(newlogfile, "syslog") == 0) {
			syslogging = 1;
			new_logfp = stderr;
#ifdef LOG_DAEMON
			/*
			 * Existence of LOG_DAEMON indicates
			 * three parameter openlog().
			 */
			openlog(progname, LOG_PID, LOG_DAEMON);
#else
			openlog(progname, LOG_PID);
#endif /* LOG_DAEMON */

		} else {
			new_logfp = fopen(newlogfile, "a");
#ifdef USE_LOCALBUF
			(void) setbuf(new_logfp, logiobuf);
#endif
		}
	}

	/*
	 * If we couldn't open a new file, then continue using the old.
	 */
	if (!new_logfp && newlogfile) {
		blog(BLOG_USER, "%s: Can't open logfile: %m", newlogfile);
		return (-1);
	}

	/*
	 * Close the previous file
	 */
	if (logfp && logfp != stderr)
		(void) fclose(logfp);
	logfp = new_logfp;
	return 0;
}


/*VARARGS*/
#if (defined(I_STDARG) && defined(CAN_VAPROTO)) || defined(__STDC__)

void
printsf(int lvl, char *fmt, ...)

#else

void
printsf(lvl, fmt, va_alist)
int	lvl;
char *	fmt;
va_dcl

#endif /* __STDC__ */
{
	va_list	args;


	if (sfp == NULL)
		return;

	/*
	 * Fire up the variable argument mechanism.
	 */
	/*CONSTCOND*/
#if (defined(I_STDARG) && defined(CAN_VAPROTO)) || defined(__STDC__)
	va_start(args, fmt);
#else
	va_start(args);
#endif

	/*
	 * Print using the varaible argument fprintf function
	 */
	(void) vfprintf(sfp, fmt, args);

	/*
	 * Terminate with a newline.
	 */
	(void) fputc('\n', sfp);

	(void) fflush(sfp);

	/*
	 * Shut down the variable argument mechanism.
	 */
	va_end(args);

	return;
}


State
decode_loglvl(opts)
char *	opts;
{

	State val = 0;

	if (strstr(opts, "All") != NULL)
		val |= BLOG_ALL;

	if (strstr(opts, "nofatal") != NULL)
		val &= ~BLOG_FATAL;
	else if (strstr(opts, "fatal") != NULL)
		val |= BLOG_FATAL;

	if (strstr(opts, "noerr") != NULL)
		val &= ~BLOG_ERR;
	else if (strstr(opts, "err") != NULL)
		val |= BLOG_ERR;

	if (strstr(opts, "nowarning") != NULL)
		val &= ~BLOG_WARNING;
	else if (strstr(opts, "warning") != NULL)
		val |= BLOG_WARNING;

	if (strstr(opts, "nouser") != NULL)
		val &= ~BLOG_USER;
	else if (strstr(opts, "user") != NULL)
		val |= BLOG_USER;

	if (strstr(opts, "noinfo") != NULL)
		val &= ~BLOG_INFO;
	else if (strstr(opts, "info") != NULL)
		val |= BLOG_INFO;

	if (strstr(opts, "notrace") != NULL)
		val &= ~(BLOG_TRACE);
	else if (strstr(opts, "trace") != NULL)
		val |= BLOG_TRACE;

	if (strstr(opts, "nostats") != NULL)
		val &= ~BLOG_STATS;
	else if (strstr(opts, "stats") != NULL)
		val |= BLOG_STATS;

	return val;

}


State
decode_dbglvl(opts)
char *	opts;
{

	State val = 0;

	if (strstr(opts, "Full") != NULL)
		val |= BLOG_FULL;

	if (strstr(opts, "alloc") != NULL)
		val |= BLOG_ALLOC;

	if (strstr(opts, "nofork") != NULL)
		val |= BLOG_NOFORK;

	if (strstr(opts, "event") != NULL)
		val |= BLOG_EVENT;

	if (strstr(opts, "load") != NULL)
		val |= BLOG_LOAD;

	if (strstr(opts, "xfer") != NULL)
		val |= BLOG_XFER;

	if (strstr(opts, "proc") != NULL)
		val |= BLOG_PROC;

	if (strstr(opts, "job") != NULL)
		val |= BLOG_JOBS;

	if (strstr(opts, "queue") != NULL)
		val |= BLOG_QUEUES;

	if (strstr(opts, "resource") != NULL)
		val |= BLOG_RESOURCE;

	return val;
}

